require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function steerToPoint(px, py)
    local me, player = getMe(), getPlayer()

    -- Our current rotation; degrees, +-180
    local myAngle = me.ArtRotation  -- (-180, +180)
    myAngle = math.rad(myAngle)

    -- Vector for me, the target
    local vMe = vec_pos(me)

    local vTarget = vec_pos(player)
    if px and py
    then
      vTarget.x = px
      vTarget.y = py
    end

    -- Vector for where I am facing
    local vMyFacing = { x = math.cos(myAngle), y = math.sin(myAngle) }
    vMyFacing = vec_norm(vMyFacing)

    -- |Me| - |Pl| 
    local vDiffWS = { x = vTarget.x - vMe.x, y = vTarget.y - vMe.y }
    local vDiffLS = { x = vDiffWS.x * me.FacingDir, y = vDiffWS.y }

    -- The actual angle to the target
    local playerAngle = math.atan2(vDiffLS.y, vDiffLS.x)
    trace( "%15s: %+07.2f" % { "playerAngle before anything", math.deg(playerAngle) })

    -- Wrap around the angle, but also put it in degrees
    playerAngle = math.deg(playerAngle)
    if playerAngle > 180 then
      playerAngle = playerAngle - 360
    elseif playerAngle < -180 then
      playerAngle = playerAngle + 360
    end
    -- put it back --
    playerAngle = math.rad(playerAngle)

    -- Get the difference of the two angles, but also put it in degrees
    local angleDiff = playerAngle - myAngle
    angleDiff = math.deg(angleDiff)
    if angleDiff > 180 then
      trace(">180")
      angleDiff = angleDiff - 360
    elseif angleDiff < -180 then
      trace("<180")
      angleDiff = angleDiff + 360
    end
    -- put it back 
    angleDiff = math.rad(angleDiff)

    --trace( "%15s: " % { })
    trace( "%15s: %+07.2f" % { "artRot", math.deg(myAngle) })
    trace( "%15s: %+07.2f" % { "playerAngle", math.deg(playerAngle) })
    trace( "%15s: %+07.2f" % { "angleDiff After", math.deg(angleDiff) })
    
    draw_point( vMe.x, vMe.y , { 255, 255, 255 }) -- White
    draw_point( vTarget.x, vTarget.y, { 255, 255, 255 }) -- White

    draw_point(
      vMe.x + math.cos(myAngle) * 320, 
      vMe.y + math.sin(myAngle) * 320,
    {255, 0, 0})

    draw_point(
      vMe.x + math.cos(playerAngle) * 320, 
      vMe.y + math.sin(playerAngle) * 320,
    {0, 255, 0})

    draw_point(
      vMe.x + math.cos(angleDiff) * 480, 
      vMe.y + math.sin(angleDiff) * 480,
    {0, 0, 255})

    if angleDiff > 0
    then
      addInput(dirs.UB)
    elseif angleDiff < 0
    then
      addInput(dirs.UF)
    else
      addInput(dirs.F)
    end
end

-- Experiment
local function steerToPoint_manual( rotSpeed )
    local me, player = getMe(), getPlayer()

    local _rotation = rotSpeed or 30

    -- Our current rotation; degrees, +-180
    local myAngle = me.Slot6  -- (-180, +180)
    myAngle = math.rad(myAngle)

    -- Vector for me, the target
    local vMe = vec_pos(me)
    local vPl = vec_pos(player)

    -- Vector for where I am facing
    local vMyFacing = { x = math.cos(myAngle), y = math.sin(myAngle) }
    vMyFacing = vec_norm(vMyFacing)

    -- |Me| - |Pl| 
    local vDiffWS = { x = vPl.x - vMe.x, y = vPl.y - vMe.y }
    local vDiffLS = { x = vDiffWS.x * me.FacingDir, y = vDiffWS.y }

    -- The actual angle to the target
    local playerAngle = math.atan2(vDiffLS.y, vDiffLS.x)
    trace( "%15s: %+07.2f" % { "playerAngle before anything", math.deg(playerAngle) })

    -- Wrap around the angle, but also put it in degrees
    playerAngle = math.deg(playerAngle)
    if playerAngle > 180 then
      playerAngle = playerAngle - 360
    elseif playerAngle < -180 then
      playerAngle = playerAngle + 360
    end
    -- put it back --
    playerAngle = math.rad(playerAngle)

    -- Get the difference of the two angles, but also put it in degrees
    local angleDiff = playerAngle - myAngle
    angleDiff = math.deg(angleDiff)
    if angleDiff > 180 then
      trace(">180")
      angleDiff = angleDiff - 360
    elseif angleDiff < -180 then
      trace("<180")
      angleDiff = angleDiff + 360
    end
    -- put it back 
    angleDiff = math.rad(angleDiff)

    --trace( "%15s: " % { })
    trace( "%15s: %+07.2f" % { "artRot", math.deg(myAngle) })
    trace( "%15s: %+07.2f" % { "playerAngle", math.deg(playerAngle) })
    trace( "%15s: %+07.2f" % { "angleDiff After", math.deg(angleDiff) })
    
    draw_point( vMe.x, vMe.y , { 255, 255, 255 }) -- White
    draw_point( vPl.x, vPl.y, { 255, 255, 255 }) -- White

    draw_point(
      vMe.x + math.cos(myAngle) * 320, 
      vMe.y + math.sin(myAngle) * 320,
    {255, 0, 0})

    draw_point(
      vMe.x + math.cos(playerAngle) * 320, 
      vMe.y + math.sin(playerAngle) * 320,
    {0, 255, 0})

    draw_point(
      vMe.x + math.cos(angleDiff) * 480, 
      vMe.y + math.sin(angleDiff) * 480,
    {0, 0, 255})

    local absDiff = math.abs(angleDiff)
    trace(math.deg(absDiff))
    if math.deg(angleDiff) > _rotation
    then
      trace("Greater")
      me.Slot6 = me.Slot6 - _rotation
    elseif math.deg(angleDiff) < -_rotation
    then
      trace("LessThan")
      me.Slot6 = me.Slot6 + _rotation
    else
      --me.Slot6 = math.deg(playerAngle)
    end
    -- go forward --
    addInput(dirs.U)
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  param.time = 0 
  param.angle = random(360)

  --me.Slot6 = param.angle
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  while true do
    me, player = getMe(), getPlayer()

    steerToPoint(player.XPos, player.YPos + 600)
    --steerToPoint_manual(350)

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  return toReturn
end --OnBattleAttack()