require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("movement")
require("vec2")

-- Declare these at the global scope of this file, so we can access them
-- anywhere
local me, player = nil

local param = {
  origin = { x = 0, y = 0 },
  time = 0,
  turnWaitTime = 0,
  --
  maxDistFromOrigin = 160 * 6,
  destPoint = { x = 0, y = 0 },
  lastPlayerPos = { x = 0, y = 0 },
  --
  sightXRange = 160 * 4,
  sightXRangeAlerted = 160 * 7,
  sightY = {
      top = 160 * 4,
      bottom = 160 * -1.5,
  },
  --
  searchingForPlayer = false,
  foundPlayer = false,
  usedGun = false,
  canUseGun = false, 
  --
  debug_startRoutine = false
}

local function dist_origin()
    return dist_point(param.origin.x, param.origin.y)
end

local function vec_dist_lastPlayerPos()
  local me, player = getMe()
  local vec_me = { x = me.XPos, y = me.YPos }
  return vec_dist( vec_me, param.lastPlayerPos)
end

local function dist_lastPlayerPos()
  local me, player = getMe(), getPlayer()
  local x_dist = param.lastPlayerPos.x - me.XPos
  local y_dist = param.lastPlayerPos.y - me.YPos
  --
  return x_dist, y_dist
end



local function isInSight()
  local me, player = getMe(), getPlayer()
  --
  local diff = { x = player.XPos - me.XPos ,y = player.YPos - me.YPos }
  if diff.x >= -param.sightXRange and diff.x <= param.sightXRange
  and diff.y >= param.sightY.bottom and diff.y <= param.sightY.top
  then
      return true 
  end
  return false
end

local function isInSightAlerted()
  local me, player = getMe(), getPlayer()
  --
  local diff = { x = player.XPos - me.XPos ,y = player.YPos - me.YPos }
  if diff.x >= -param.sightXRangeAlerted and diff.x <= param.sightXRangeAlerted
  and diff.y >= param.sightY.bottom and diff.y <= param.sightY.top
  then
      return true 
  end
  return false
end

local function attack_kick( data )
  if data ~= nil then 
    data.dir = dirs.D
    --data.dir = dirs.N
    data.pressAction = false
  end
  doMotion("D+Attack")
  --doMotion("X+Attack")
end

local function attack_salute( data )
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
  end
  doMotion("U+Attack")
end

local function attack_gunshot( data )
  if data ~= nil then 
    data.dir = dirs.N
    data.pressAction = false
  end
  doMotion("X+Attack")
end

-- the initialization function for this AI
function OnInit()
  -- Accessible anywhere in this file after Initialization
  me, player = getMe(), getPlayer()
  --
  param.origin.x, param.origin.y = me.XPos, me.YPos
  --
  param.destPoint.x, param.destPoint.y = param.origin.x, param.origin.y
  param.destPoint.x = param.destPoint.x + (param.maxDistFromOrigin * (me.FacingDir * -1))
  --
  param.time = 0 
  param.turnWaitTime = 0 
  param.lastPlayerPos.x, param.lastPlayerPos.y = 0, 0
  --
  param.searchingForPlayer = false
  param.foundPlayer = false
  param.usedGun = false
  --
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
end

-- the main update function for this AI
function OnUpdate()
  while true do
    if aiDebug() then
      draw_point(param.destPoint.x, param.destPoint.y)
      draw_point(param.lastPlayerPos.x, param.lastPlayerPos.y)
      draw_bar(param.time, 0, 120)
      draw_bar(param.turnWaitTime, 0, 120, 1)
      local width = param.searchingForPlayer and param.sightXRangeAlerted or param.sightXRange
      local c1 = param.searchingForPlayer and {255, 0, 255} or {255, 0, 128}
      local c2 = param.searchingForPlayer and {0, 255, 255} or {255, 128, 0}
      drawBox(me.XPos - width, me.XPos + width,
          me.YPos + param.sightY.top, me.YPos,
          1, -1, 
          unpack(c1)
      )
      drawBox(me.XPos - width, me.XPos + width,
          me.YPos, me.YPos + param.sightY.bottom,
          1, -1, 
          unpack(c2)
      )
    end
    --
    if true then
      param.turnWaitTime = math.max(0, param.turnWaitTime - 1) 
    end
    --
    local _dist = dist_point(param.destPoint.x, param.destPoint.y)
    if me.Alerted >= 30 or me.Hitstun > 0 then
      param.turnWaitTime = 0 
      param.lastPlayerPos = { x = player.XPos, y = player.YPos}
      param.searchingForPlayer = true
    end

    debugPalette()
    if me.Lifebar <= 0 then
      -- let's just do NOTHING
      debugPalette("Debug_Dark")
    elseif ( param.searchingForPlayer and isInSightAlerted() or isInSight() ) and player_vulnerable() then
      -- let's try this..
      debugPalette("Debug_Blue")
      wait()
      -- Wait a bit
      param.time = 0
      param.lastPlayerPos = { x = player.XPos, y = player.YPos} 
      param.searchingForPlayer = true
      param.foundPlayer = false
      local vec_dist = vec_dist_lastPlayerPos()
      wait(4)
      while param.time <= 100 or isInSightAlerted() do
        debugPalette("Debug_Yellow")
        draw_point(param.lastPlayerPos.x, param.lastPlayerPos.y)
        draw_bar(param.time, 0, 100)
        
        approach(param.lastPlayerPos.x)
        --
        param.lastPlayerPos = { x = player.XPos, y = player.YPos} 
        if not isInSightAlerted() then
          param.time = param.time + 1
        end
        local dist = dist_lastPlayerPos()
        -- the smaller range
        if isInSight() then
          param.foundPlayer = true
          break 
        end
        --
        wait()
      end
      --
      wait()
      if param.foundPlayer and not me:hasTag("Attack") then
        local dist = dist_lastPlayerPos()
        if math.abs(dist) > 380 and not param.usedGun and param.canUseGun then
          attack_gunshot()
          param.usedGun = true
        else
          attack_kick()
        end
        if me:hasTag("Attack") then
          debugPalette("Debug_Red")
        end
        param.time = 0 
        while param.time <= 80 and not me:hasTag("Attack") do
          param.time = param.time + 1
          if isInSightAlerted() or got_hurt() then
            break
          end
          wait()
        end
      end
      debugPalette("Debug_Green")
      --
      param.time = 0
      param.turnWaitTime = 90
    elseif param.searchingForPlayer then
      -- Let's just check the point we saw the player last
      if math.abs( dist_player() ) <= 100 then
        param.searchingForPlayer = false
      elseif obstacle_in_front() 
      or (
        math.abs( dist_lastPlayerPos() ) <= 100
      and not isInSightAlerted() 
      )
      then
        -- Hand off
        param.searchingForPlayer = false
        param.turnWaitTime = 90
      end
      debugPalette("Debug_Blink")
      if player_vulnerable() then
        approach(param.lastPlayerPos.x)
      end
    elseif param.turnWaitTime > 0 then
      debugPalette("Debug_Cyan")
    elseif math.abs( _dist ) <= me.WorldCollisionWidth then
      local side = param.origin.x - me.XPos
      if side > 0 then side = 1 else side = -1 end 
      param.destPoint.x = param.origin.x + (param.maxDistFromOrigin * (side))
      param.turnWaitTime = 60 * 1
    else
      approach(param.destPoint.x)
    end
    --
    wait() -- try to keep this at the end
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  attack_swipe(toReturn)

  return toReturn
end --OnBattleAttack()