require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function action_perch(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Heruka")
end

local function action_perchNotice(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_diveBite(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
  param.attackCooldown = 35
end

local function attack_whirlwind(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 17
end

local function attack_blow(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 40
end

local function vec_dist_player(target)
  local me, player = getMe(), (target or getPlayer())
  local vm = { x = me.XPos, y = me.YPos }
  local vp = { x = player.XPos, y = player.YPos }
  
  local result = vec_dist( vm, vp )
  
  return result
end

local function dist_origin()
  return dist_point(param.origin.x, param.origin.y)
end

local function vec_dist_origin()
  local me = getMe()
  local vm = { x = me.XPos, y = me.YPos }
  local vo = { x = param.origin.x, y = param.origin.y }

  local result = vec_dist( vm, vo )

  return result
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)
  
  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  if aiDebug() then
    drawBox(
      me.XPos - w/2,
      me.XPos + w/2,
      me.YPos + h.top,
      me.YPos + h.bottom,
      1, -1,
      255, 255, 255
    )
  end

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)
  
  local w = param.attackRangeX
  local h = param.attackRangeY
  if param.bossBurstedMode then
    w = param.attackRangeRageX
    h = param.attackRangeRageY
  end
  
  if aiDebug() then
    drawBox(
      me.XPos - w/2,
      me.XPos + w/2,
      me.YPos + h.top,
      me.YPos + h.bottom,
      1, -1,
      255, 255, 255
    )
  end
  
  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function AttackOK()
  if param.attackCooldown <= 0
  and canAttack(me)
  then
    return true
  end
  return false
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.time = 0
  param.perchWaitTime = 0
  param.moveDecay = 0 
  param.attackCooldown = 0 
  param.seekYOffset = player.WorldCollisionHeight * (-5/4)
  --

  param.origin = { x = 0, y = 0 }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  param.sightRangeX = 160 * 8
  param.sightRangeY = { top = 160 * 3, bottom = 160 * -7 }
  param.attackRangeX = 160 * 7
  param.attackRangeY = { top = 160 * 3, bottom = 160 * -4 }
  -- co-op
  param.sightRangeRageX = 160 * 15 
  param.sightRangeRageY = { top = 160 * 6, bottom = 160 * -7 }
  param.attackRangeRageX = 160 * 14
  param.attackRangeRageY = { top = 160 * 6, bottom = 160 * -6 }
  --
  param.stationary = false
  if has_spawn_command("Stationary") then
    param.stationary = true
  end
  param.startPerched = false
  param.canPerch = true
  -- hash("WaitExtended")
  -- hash("Stationary")
  param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()

    if aiDebug() then
      local show_ranges = true
      draw_bar(param.time, 0, 60 * 60, 2)
      draw_bar(param.moveDecay, 0, 60, 1, {255, 0, 255})
      draw_bar(param.attackCooldown, 0, 60)
      --
      draw_bar(me.TotalStun, 0, _BOSS_BURST_STUN_MAX, 3, _color.stun)
      draw_bar(me.Alerted, 0, _ALERTED_MAX, 4, _color.alerted)
    end

    param.time = param.time + 1
    ----------------------------------------------------------------------------
    -- The start
    ----------------------------------------------------------------------------
    if bossBursted() 
    then
      param.turnWaitTime = 0
      param.perchWaitTime = 0 
      param.moveDecay = 0 
      param.attackCooldown = 0 
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 3 
    end
    
    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    
    debugPalette()
    if not isTangled(me) then
      debugPalette("Debug_Green")
      --------------------------------------------------------------------------
      -- Ticks
      --------------------------------------------------------------------------
      if not isInSightRange(target) then
        --param.turnWaitTime = math.max(0, param.turnWaitTime - 1)
      end

      if not me:hasTag("Attack") then
        param.attackCooldown = math.max(0, param.attackCooldown - 1)
        param.moveDecay = math.max(0, param.moveDecay - 1)
        param.perchWaitTime = math.max(0, param.perchWaitTime - 1)
      end
      -- Perched!
      if me:hasTag("Perch") then
        debugPalette("Debug_Magenta")
        --
        if isInSightRange(target)
        and sprite_vulnerable(target)
        and not me:hasTag("FlipToIdle")
        then
          action_perchNotice()
          param.perchWaitTime = 60 * 2
        end
      elseif param.stationary
      then
        if vec_dist_origin() > 80
        then
          fly_to(param.origin.x, param.origin.y)
        else
          -- We're chillin'
        end
      -- I see you!
      elseif (
        param.bossBurstedMode
        and isInSightRange(target)
      )
      or ( 
        isInSightRange(target)
        and sprite_vulnerable(target)
      )
      
      then
        debugPalette("Debug_Yellow")
        -- Move to Player
        if vec_dist_player(target) > 160 * 2.2
        then
          fly_to(target.XPos, target.YPos + param.seekYOffset)
        end
        -- Attack
        if isInAttackRange(target)
        and AttackOK()
        and not me:hasTag("Attack")
        then
          debugPalette("Debug_Red")
          local yDist = select(2, dist_player())
          -- trace(yDist)
          if yDist >= 160 * -2
          and yDist <= 160 * 1
          then
            if dist_sprite(target) > 0
            then
              attack_diveBite()
              me:setEnemyLeaderTo(target)
            else
              addInput(dirs.B)
            end
          end
          param.moveDecay = 30
        end
      elseif vec_dist_origin() > 300
      and param.moveDecay <= 0 
      then
        debugPalette("Debug_Blue")
        -- Let's go home
        fly_to(param.origin.x, param.origin.y)
      elseif param.canPerch
      and me.CeilingDist <= 150
      and param.moveDecay <= 0 
      and param.perchWaitTime <= 0
      and not me:hasTag("Perch")
      then
        debugPalette("Debug_Orange")
        action_perch()
      else
        debugPalette("Debug_Cyan")
      end
    else
      debugPalette("Debug_Blink")
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  -- Disallow [U,D]+Herukas in battle for now!
  if toReturn.dir == dirs.U
  or toReturn.dir == dirs.D
  then
    toReturn.pressAction = false
  end

  return toReturn
end --OnBattleAttack()