require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function attack_drain(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.meterBurn = false
  end
  doMotion("X+Heruka")
  param.attackCooldown = 50
end

local function attack_stab(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 32
end

local function attack_drill(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
  param.attackCooldown = 44
end

local function attack_spit(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 27
end

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.attackRangeX
  local h = param.attackRangeY
  if param.bossBurstedMode then
    w = param.attackRangeRageX
    h = param.attackRangeRageY
  end

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInGrabRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  if dx >= -param.grabRangeX/2
  and dx <= param.grabRangeX/2
  and dy <= param.attackRangeY.top
  and dy >= param.attackRangeY.bottom
  then
    return true
  end
  return false
end

local function AttackOK()
  if param.attackCooldown <= 0
  and canAttack(me)
  then
    return true
  end
  return false
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.time = 0
  param.turnWaitTime = 0
  param.attackCooldown = 0 
  --
  param.origin = { x = 0, y = 0 }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  param.roamRangeX = 160 * 4
  param.roamRangeFudgeDist = 75
  param.sightRangeX = 160 * 10
  param.sightRangeY = { top = 160 * 3.3, bottom = 160 * -1.5 }
  param.attackRangeX = 160 * 4
  param.attackRangeY = { top = 160 * 3.3, bottom = 160 * -1.5 }
  param.grabRangeX = 160 * 3
  --
  param.sightRangeRageX = 160 * 23
  param.sightRangeRageY = { top = 160 * 6, bottom = 160 * -1.5 }
  param.attackRangeRageX = 160 * 4 -- shoot
  param.attackRangeRageY = { top = 160 * 5, bottom = 160 * -1.5 }
  --
  param.goalX = param.origin.x
  param.goalX = param.goalX + ((param.roamRangeX/2) * (me.FacingDir * -1))
  --
  param.useGrabAttack = false
  -- battle
  param.battleGrabCount = 0 
  --
  if me.ColorNum == 2 
  or me.ColorNum == 6 
  then
    param.useGrabAttack = true 
  end
  
  -- co-op
  param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()

    if aiDebug() then
      draw_bar(param.attackCooldown, 0, 60)
    end

    -- co-op
    if bossBursted() 
    then
      param.turnWaitTime = 0
      param.attackCooldown = 25 -- we actually don't wanna hurt someone IMMEDIATELY!
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 3 
    end
    
    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
      
    ----------------------------------------------------------------------------
    -- The start
    ----------------------------------------------------------------------------
    debugPalette()
    if not isTangled(me) then
      debugPalette("Debug_Green")
      --------------------------------------------------------------------------
      -- Ticks
      --------------------------------------------------------------------------
      if not isInSightRange(target) then
        param.turnWaitTime = math.max(0, param.turnWaitTime - 1)
      end

      if not me:hasTag("Attack") then
        param.attackCooldown = math.max(0, param.attackCooldown - 1)
      end
      
      
      --
      local dist_goalPoint = dist_point(param.goalX, me.YPos)
      if math.abs( dist_goalPoint ) <= me.WorldCollisionWidth/2 then
        local side = param.origin.x - me.XPos
        if side >= 0 then side = 1 else side = -1 end 
        param.goalX = param.origin.x + (param.roamRangeX/2 * (side))
        param.turnWaitTime = 180
      end
      -- I see you!
      if ( isInSightRange(target)
        and param.bossBurstedMode
      )
      or ( isInSightRange(target)
        and sprite_vulnerable(target)
      )
      then
        debugPalette("Debug_Yellow")
        -- Move to Player
        approach_withTurn(target.XPos, 8)
        -- Attack
        if isInAttackRange(target)
        and AttackOK()
        and not me:hasTag("Attack")
        then
          debugPalette("Debug_Red")
          if select(2, dist_sprite(target)) >= 160 * 1 then
            attack_drill()
          elseif param.useGrabAttack then
            attack_drain()
          else
            attack_stab()
          end
        end
      elseif param.turnWaitTime > 0 then
        debugPalette("Debug_Blue")
        -- we're just waitin'
        if param.turnWaitTime == 90 then
          turn_around(8)
        end
      else
        debugPalette("Debug_Cyan")
        approach_withTurn(param.goalX, 8)
      end
    else
      debugPalette("Debug_Blink")
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  
  local actionChance = random(1000)

  param.battleGrabCount = param.battleGrabCount + 1

  if param.useGrabAttack then
    attack_drill(toReturn)

    if actionChance > 500
    and param.battleGrabCount >= 3 
    then
      attack_drain(toReturn)
      param.battleGrabCount = 0 
    end
  else
    attack_stab(toReturn)
    if actionChance > 500
    then
      attack_spit(toReturn)
    end
  end

  if aiDebug()
  then
    attack_drain(toReturn)
    attack_spit(toReturn)
  end

  return toReturn
end --OnBattleAttack()