require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

-- Attacks --
local function attack_attack(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_uatk(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_datk(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_natk(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_super(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = true
  end
  doMotion("X+Attack+MeterBurn")
end

local function attack_super_lv1(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = true
  end
  doMotion("X+Dash")
end

local function attack_super_bowScream(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = true
  end
  doMotion("X+Attack+Heruka+MeterBurn")
end

-- sequence keys
local sequence = {
  none = 0,
  default = 0,
  -- default
  triple = 1,
  -- supers --
  super = 10
}


-- the initialization function for this AI
function OnInit()
  local me = getMe()

  param.time = 0 
  --
  param.battleSeq = 0  
  param.battleSubSeq = 0
  param.battleMaxTimeout = 60 * 5 -- just in case
  param.battleTimeout = 60 * 1 -- default for our own sequence

  --
  param.isBattleChallenge = false
  if has_spawn_command("BattleChallenge")
  then
    param.isBattleChallenge = true
  end
  param.hasChallengedPlayer = false

  -- dangerous --
  if param.isBattleChallenge
  then
    me.MaxLife = 1000000
    me.Lifebar = me.MaxLife
  end
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()

    if param.isBattleChallenge
    then
      if isInRange(player, me.XPos, me.YPos, 160 * 8, 160 * 8)
      then
        trace("ATTACKING THE PLAYER")
        param.hasChallengedPlayer = me:startBattle(player)
      end
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  -- draw it out
  if true
  then
    draw_bar(param.time, 0, param.battleMaxTimeout, 0)
    draw_bar(param.time, 0, param.battleTimeout, 1, {255, 0, 0})
  end

  -- sequence picker
  param.battleSeq = 0
  param.battleSubSeq = 0 
  
  param.time = 0

  -- reset commands
  toReturn.meterBurn = false
  toReturn.pressAction = false
  toReturn.pressDash = false

  -- sequence assignment
  param.battleSeq = sequence.triple
  attack_attack(toReturn)
  param.battleSeq = sequence.super
  attack_super_lv1(toReturn)

  return toReturn
end --OnBattleAttack()

function OnBattleTurnUpdate(battleData)
  local me = getMe()

  if aiDebug() 
  then
    trace("%04i:%01i:%01i" % { param.time, param.battleSeq, param.battleSubSeq})
    draw_bar(param.time, 0, param.battleMaxTimeout, 0)
    draw_bar(param.time, 0, param.battleTimeout, 1, {255, 0, 255})
  end

  -- the loop --
  if param.time <= param.battleMaxTimeout
  then
    if param.battleSeq == sequence.default
    then
      if param.time >= param.battleTimeout
      then
        debugPalette("Debug_Black")
      else
        debugPalette("Debug_White")
        addInput(dirs.N)
      end
    elseif param.battleSeq == sequence.super
    then
      if param.time >= param.battleTimeout
      then
        debugPalette("Debug_Black")
      elseif me:hasTag("Attack")
      then
        -- don't do anything actually? --
      else
        debugPalette("Debug_White")
        addInput(dirs.N)
      end
    -- Triple NAtk --
    elseif param.battleSeq == sequence.triple
    then
      if param.time >= param.battleTimeout
      then
        debugPalette("Debug_Black")
      elseif param.battleSubSeq == 0 
      then
        debugPalette("Debug_Orange")
        if me.NumTouches > 0 
        and param.time >= 0 
        then
          attack_natk(toReturn)
          debugPalette("Debug_LightOrange")
          param.battleSubSeq = 1
          param.time = -1
        else
          addInput(dirs.N)
        end
      elseif param.battleSubSeq == 1
      then
        debugPalette("Debug_Orange")
        if me.NumTouches > 0
        and param.time >= 0 
        then
          attack_natk(toReturn)
          debugPalette("Debug_LightOrange")
          param.battleSubSeq = 2
          param.time = -1
        else
          addInput(dirs.N)
        end
      elseif param.battleSubSeq == 2
      then
        debugPalette("Debug_Orange")
        if me.NumTouches > 0 
        then
          attack_datk(toReturn)
          debugPalette("Debug_LightOrange")
          param.battleSubSeq = 3
          param.time = -1
        else
          addInput(dirs.N)
        end
      elseif param.battleSubSeq == 3
      then
        debugPalette("Debug_Orange")
        if me.NumTouches > 0 
        and param.time >= 0 
        then
          attack_datk(toReturn)
          debugPalette("Debug_LightOrange")
          param.time = param.battleMaxTimeout
        else
          addInput(dirs.N)
        end
      else
        debugPalette("Debug_White")
        addInput(dirs.N)
      end
    end
  end
  
  -- update only if not hitstop
  if update_ok()
  then
    param.time = param.time + 1
  end
end