require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("math_ext")
require("movement")
require("util")

local function dist_player()
  local me, player = getMe(), getPlayer()
  local dx = player.XPos - me.XPos
  dx = dx * me.FacingDir
  local dy = player.YPos - me.YPos
  return dx, dy
end

local function vec_dist_player()
  local me, player = getMe(), getPlayer()
  local vm = { x = me.XPos, y = me.YPos }
  local vp = { x = player.XPos, y = player.YPos }
  
  local result = vec_dist( vm, vp )
  
  return result
end

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  --dx = dx * me.FacingDir
  return dx, dy 
end

local function vec_dist_origin()
  local me = getMe()
  local vm = { x = me.XPos, y = me.YPos }
  local vo = { x = param.origin.x, y = param.origin.y }

  local result = vec_dist( vm, vo )

  return result
end

local function isInSightRange()
  local me, player = getMe(), getPlayer()

  if dist_player() > -param.sightRangeX/2
  and dist_player() < param.sightRangeX/2
  and select(2,dist_player()) < param.sightRangeY.top
  and select(2,dist_player()) > param.sightRangeY.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange()
  local me, player = getMe(), getPlayer()

  if dist_player() > -param.attackRangeX/2
  and dist_player() < param.attackRangeX/2
  and select(2,dist_player()) < param.attackRangeY.top
  and select(2,dist_player()) > param.attackRangeY.bottom
  then
    return true
  end
  return false
end

local function attack_gun(data)
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 70
  param.attackPeriod = 40
  param.movementCooldown = 30
end

-- the initialization function for this AI
function OnInit()
    -- Accessible anywhere in this file after Initialization
    local me, player = getMe(), getPlayer()
    param.time = 0 
    param.attackCooldown = 0
    param.movementCooldown = 0 
    param.attackPeriod = 0 
    -- Set origin
    param.origin = {
        x = me.XPos,
        y = me.YPos,
        z = me.ZPos
    }
    if me:getSpawnpointPos() ~= nil then
        param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
    end
    --
    param.sightRangeX = 160 * 8
    param.sightRangeY = { top = 160 * 3, bottom = 160 * -9 }
    param.attackRangeX = 160 * 7
    param.attackRangeY = { top = 160 * 3, bottom = 160 * -4.25 }

    --
    param.isSaw = false
    if has_spawn_command("Saw") then
      param.isSaw = true
    end
    param.keepHoldingSaw = false

    --
    param.isSentryMode = false
    if has_spawn_command("Sentry")
    or has_spawn_command("Stationary")
    then
      param.isSentryMode = true
    end
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end

  while true do
    me, player = getMe(), getPlayer()
    --
    if aiDebug() then
      draw_bar(param.movementCooldown, 0, 120, 3, {255, 255, 255})
      draw_bar(param.attackPeriod, 0, 90, 2, {0, 255, 255})
      draw_bar(param.time, 0, 90, 1, {0, 128, 128})
      draw_bar(param.attackCooldown, 0, 90, 0, {255, 0, 0})
      draw_point(player.XPos, player.YPos + player.WorldCollisionHeight)
    end

    if tick_ok() and not is_hurt(me) and not me:hasTag("Attack") then
      param.movementCooldown = math.max(0, param.movementCooldown - 1)
      param.attackCooldown = math.max(0, param.attackCooldown - 1)
    end

    -- 
    if isInAttackRange() 
    and not param.keepHoldingSaw
    then
      param.keepHoldingSaw = true
    elseif not isInSightRange() 
    and param.keepHoldingSaw 
    then
      param.keepHoldingSaw = false
    end

    --if has_spawn_command("Stationary") then
    --  if vec_dist_origin() > 80
    --  then
    --    fly_to(param.origin.x, param.origin.y)
    --  else
    --    -- We're chillin'
    --  end
    --end
--
    if param.keepHoldingSaw then
      addInput(dirs.N, btns.Heruka)
      --doMotion("X+Heruka")
    end

    if isInSightRange() 
    and not param.isSentryMode
    then
      if vec_dist_player() > 500 
      and param.movementCooldown <= 0
      and not param.isSaw
      then
        fly_to(player.XPos, player.YPos + player.WorldCollisionHeight * 1.5,
        { min = 5, max = 22, range = 160 * 3})
      end
    elseif vec_dist_origin() > 30 
    and param.movementCooldown <= 0 
    then
      fly_to(param.origin.x, param.origin.y, { min = 3, max = 30, range = 160 * 4 })
    else
      -- don't do anything
    end
    --
    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
    local toReturn = DefaultOnBattleAttack(battleData)
    local me = getMe()
    return toReturn
end --OnBattleAttack()