require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function block()
  doMotion("X+WorldGuard")
end

local function attack_grabSlash(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Heruka")
end

local function attack_doubleCut(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function attack_steal(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 44
end

local function attack_summon(data)
  param.battleSummonedOnce = true

  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_feignDeath_wakeup(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.attackRangeX
  local h = param.attackRangeY
  if param.bossBurstedMode then
    w = param.attackRangeRageX
    h = param.attackRangeRageY
  end

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function bulletDist()
  local _x, _y = nil, nil
  local _ret = nil
  if select(1,dist_blang()) ~= nil then
    return dist_blang()
  elseif select(1,dist_spear()) ~= nil then
    return dist_spear()
  elseif  select(1,dist_arrow()) ~= nil then
    return dist_arrow()
  end
  return -256^2, -256^2
end

local function summonOK()
  local me = getMe()

  if me.ColorNum > 1
  then
    return true
  else
    return false
  end

  return false
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.origin = {
    x = me.XPos,
    y = me.YPos
  }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  param.time = 0
  param.attackCooldown = 0 
  param.searchingForPlayer = false
  --
  param.sightRangeX = 160 * 9
  param.sightRangeY = { top = 160 * 1.5, bottom = 160 * -1 }
  param.attackRangeX = 160 * 6
  param.attackRangeY = { top = 160 * 2.5, bottom = 160 * -1.5 }
  --
  param.sightRangeRageX = 160 * 19 
  param.sightRangeRageY = { top = 160 * 6, bottom = 160 * -7 }
  param.attackRangeRageX = 160 * 7
  param.attackRangeRageY = { top = 160 * 6, bottom = 160 * -6 }
  --
  param.battleSummonedOnce = false
  param.battleSummonHealthPct = 33 / 100
  --
  if has_spawn_command("Summoned")
  then
    param.battleSummonedOnce = true
  end
  
  param.bossBurstedMode = false
end

-- The world update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()
  
  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    
    if aiDebug() then
      draw_bar(param.attackCooldown, 0, 90, 0)
    end

    if tick_ok() and not is_hurt(me) then
      param.attackCooldown = math.max(0, param.attackCooldown - 1)
    end

    if tick_ok() then
      param.time = param.time + 1
    end
    
    -- co-op 
    if bossBursted() 
    then
      param.time = 0 
      param.attackCooldown = 0 
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 5 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    
    if me:hasTag("FeignDeath") 
    then
      -- We're.. very.. sleepy..
      if isInSightRange()
      -- Either in front and not facing us
      -- OR Behind us and far enough in the deadzone
      and 
      (
      (me.FacingDir == player.FacingDir and dist_player() > param.attackRangeX/2 and dist_player() < param.sightRangeX/2)
      or (me.FacingDir ~= player.FacingDir and dist_player() < -param.attackRangeX/2 )
      )
      and player_vulnerable()
      and not me:hasTag("FeignDeath_Wakeup")
      then
        addInput(dirs.N, btns.Attack)
      end
    else
      if isInSightRange(target) then
        approach_withTurn(target.XPos, 10)
        if isInAttackRange(target) and dist_sprite(target) > 0 
        and param.attackCooldown <= 0
        and (
          ( sprite_vulnerable(target)
            and canAttack()
          ) or ( param.bossBurstedMode 
            and canAttack()
          )
        )
        then
          if canAttack() then
            attack_steal()
          end
        end
      else
        if math.abs( dist_origin() ) >= 160 * 0.25
        and math.abs( select(2,dist_origin()) ) <= 160 * 4
        and tick_ok()
        and param.attackCooldown <= 0 
        and me.OnGround > 0 
        then
          approach_withTurn(param.origin.x, 10)
        end
      end
    end

    wait()
  end -- the while loop
end -- the function

-- The battle action function for this AI 
function OnBattleAttack(battleData)
  local me = getMe()
  local toReturn = DefaultOnBattleAttack(battleData)

  local dirChance = random(1000)
  local actionChance = random(1000)

  if has_spawn_command("Summoned")
  then
    param.battleSummonedOnce = true
  end
  
  -- Default
  attack_steal(toReturn)

  -- Special
  
  -- Actions
  if actionChance > 500
  and dirChance > 500 then
    attack_doubleCut()
  end

  -- Chances
  if dirChance > 667 then
    attack_steal(toReturn)
  end

  if not param.battleSummonedOnce 
  and me.HealthPct <= param.battleSummonHealthPct
  and summonOK()
  then
    attack_summon(toReturn)
    param.battleSummonedOnce = true
  end
  
  return toReturn
end --OnBattleAttack()