require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function attack_bellyFlop(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.meterBurn = false
  end
  doMotion("U+Heruka")
  param.attackCooldown = 33
end

local function attack_daggerToss(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.meterBurn = false
  end
  doMotion("X+Heruka")
  param.attackCooldown = 40
end

local function attack_unblockable(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.meterBurn = false
  end
  doMotion("D+Heruka")
  param.attackCooldown = 40
end

local function attack_charge(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
  param.attackCooldown = 27
end

local function attack_roll(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 30
end

local function attack_slash(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 27
end

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInFrontSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  if dx >= 0
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.attackRangeX
  local h = param.attackRangeY
  if param.bossBurstedMode then
    w = param.attackRangeRageX
    h = param.attackRangeRageY
  end

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInGrabRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  if dx >= -param.grabRangeX/2
  and dx <= param.grabRangeX/2
  and dy <= param.attackRangeY.top
  and dy >= param.attackRangeY.bottom
  then
    return true
  end
  return false
end

local function AttackOK()
  if param.attackCooldown <= 0
  and canAttack(me)
  then
    return true
  end
  return false
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()
  --
  local players = getPlayers()
  -- Set origin
  param.time = 0
  param.turnWaitTime = 0
  param.attackCooldown = 0 
  --
  param.origin = { x = 0, y = 0 }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  param.roamRangeX = 160 * 10
  param.roamRangeFudgeDist = 80
  param.sightRangeX = 160 * 12
  param.sightRangeY = { top = 160 * 3.3, bottom = 160 * -1.5 }
  param.attackRangeX = 160 * 8
  param.attackRangeY = { top = 160 * 3.3, bottom = 160 * -1.5 }
  --
  param.sightRangeRageX = 160 * 26
  param.sightRangeRageY = { top = 160 * 6, bottom = 160 * -1.5 }
  param.attackRangeRageX = 160 * 9
  param.attackRangeRageY = { top = 160 * 6, bottom = 160 * -1.5 }
  --
  param.turnAroundWhenPlayerBehind = true
  param.turnedAround = false
  -- Hashes
  param.isMiniboss = has_spawn_command("ForceMiniboss")
  param.doWideAttacks = has_spawn_command("DoWideAtk")
  param.isReviveTutorial = has_spawn_command("ReviveTutorial")
  param.isAllGuardTutorial = has_spawn_command("AllGuardTutorial")
  param.isCleanBlockTutorial = false
  if has_spawn_command("EasyFlop_Tutorial")
  then
    param.isCleanBlockTutorial = true
  end

  -- Green Belu don't roam..
  if me.ColorNum == 1
  then
    param.roamRangeX = 160 * 2
    param.roamRangeFudgeDist = 40
  end
  -- Tutorial Revive buddy doesn't roam.
  if param.isReviveTutorial
  then
    param.roamRangeX = 160 * 2
    param.roamRangeFudgeDist = 40
  end
  -- Blue (Miniboss) turns around
  if me.ColorNum == 3
  then
    param.turnAroundWhenPlayerBehind = true
    param.isMiniboss = true
  end
  -- Ravannavar Buddy
  if me.ColorNum == 4
  then
    param.sightRangeX = 160 * 14
    param.roamRangeX = 160 * 2
    param.roamRangeFudgeDist = 40
    param.attackRangeX = 160 * 4.7 * 2
  end
  -- Port back this behavior?
  if param.isMiniboss then
    me.IsBoss = hash("Miniboss")
    me.NumBossLifebars = 1
  end
  -- Setup goal here, or else we'll get stuck trying to turn from Left
  param.goalX = param.origin.x
  param.goalX = param.goalX + ((param.roamRangeX/2) * (me.FacingDir * -1))
  
  param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    if aiDebug() then
      draw_bar(param.turnedAround and 100 or 1, 0, 100, 2, {255, 255, 255})
      draw_bar(param.turnWaitTime, 0, 300, 1)
      draw_bar(param.attackCooldown, 0, 60)
    end

    -- co-op 
    if bossBursted() 
    then
      param.turnWaitTime = 0
      param.perchWaitTime = 0 
      param.attackCooldown = 0 
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 3 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    ----------------------------------------------------------------------------
    -- The start
    ----------------------------------------------------------------------------
    debugPalette()
    if not isTangled(me) then
      debugPalette("Debug_Green")
      --------------------------------------------------------------------------
      -- Ticks
      --------------------------------------------------------------------------
      -- We calculate our sightDistance manually here because 
      if not isInFrontSightRange(target)
      then
        param.turnWaitTime = math.max(0, param.turnWaitTime - 1)
      end

      if not me:hasTag("Attack") then
        param.attackCooldown = math.max(0, param.attackCooldown - 1)
      end
      --
      local dist_goalPoint = dist_point(param.goalX, me.YPos)
      if math.abs( dist_goalPoint ) <= me.WorldCollisionWidth/2 then
        local side = param.origin.x - me.XPos
        if side >= 0
        then 
          side = 1
        else 
          side = -1
        end 
        param.goalX = param.origin.x + (param.roamRangeX/2 * (side))
        param.turnWaitTime = 180
        trace("Ding")
      end
      -- I see you!
      -- But we're a belu and dumb so only chase if I see you in front of me
      if isInFrontSightRange(target)
      and ( param.bossBurstedMode or sprite_vulnerable(target))
      then
        debugPalette("Debug_Yellow")
        --param.turnedAround = false
        -- Move to Player
        local check_bonus = me.XVel
        if me.XVel <= 8 then
          check_bonus = 0 
        end
        check_bonus = check_bonus * 4
        trace(check_bonus)
        if math.abs( dist_sprite(target) ) >= (160 * 3) + check_bonus
        then
          debugPalette("Debug_DarkYellow")
          approach_withTurn(target.XPos, 8)
        end
        -- Attack
        if isInAttackRange(target)
        and AttackOK()
        and not me:hasTag("Attack")
        then
          debugPalette("Debug_Red")
          param.turnedAround = false
          -- Anti-air
          if select(2, dist_sprite(target)) >= 160 * 0.75 then
            debugPalette("Debug_Orange")
            -- charge up!
            param.time = param.time + 1
            local waitTime = param.bossBurstedMode and 2 or 7
            if param.time > 7 then
              attack_roll()
            end
          else
            param.time = 0 
            attack_slash()
          end
        end
      elseif isInSightRange(target)
      and ( param.bossBurstedMode or sprite_vulnerable(target) )
      and ( param.bossBurstedMode or param.turnAroundWhenPlayerBehind )
      and dist_sprite(target) < 0
      then
        debugPalette("Debug_Magenta")
        param.turnedAround = true
        turn_around(13)
      elseif param.turnWaitTime > 0 then
        debugPalette("Debug_Blue")
        -- we're just waitin'
        if param.turnWaitTime == 90
        and not param.turnedAround
        then
          turn_around(13)
          param.turnedAround = true 
        end
      else
        debugPalette("Debug_Cyan")
        param.turnedAround = false
        approach_withTurn(param.goalX, 13)
      end
    else
      debugPalette("Debug_Blink")
      param.turnedAround = false
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  
  local dirChance = random(1000)
  local actionChance = random(1000)

  attack_slash(toReturn)

  if dirChance > 667
  then
    attack_roll(toReturn)
  elseif dirChance < 334
  then
    attack_charge(toReturn)
  end

  --[[
  if actionChance > 500
  and me.ColorNum >= 3 and me.ColorNum <= 5 
  then
    attack_bellyFlop(toReturn)
  end
  ]]

  if me.SpawnCommand == hash("HeadbuttOnly") then
    attack_charge(toReturn)
  end

  if param.doWideAttacks
  then
    attack_slash(toReturn)
  end

  -- If we're a tutorial Belu, we need to do this thing IMMEDIATELY tho --
  if param.isReviveTutorial
  and battleData.turnNum == 1
  and false -- DISABLED :)
  then
    attack_unblockable(toReturn)
  end

  -- If we're the AllGuard tutorial, we always do roll --
  if param.isAllGuardTutorial
  then
    attack_roll(toReturn)
  end

  if param.isCleanBlockTutorial
  then
    trace("Clean Block Action")
    attack_bellyFlop(toReturn)
  end

  return toReturn
end --OnBattleAttack()