require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("movement")
require("vec2")

require("behavior")

local function dist_origin()
    return dist_point(param.origin.x, param.origin.y)
end

local function dist_player(target)
    local me, player = getMe(), (target or getPlayer())
    -- 
    local vec_me = { x = me.XPos, y = me.YPos }
    local vec_pl = { x = player.XPos, y = player.YPos }

    return vec_dist( vec_me, vec_pl)
end

local function dist_lastPlayerPos()
    local me = getMe()
    local vec_me = { x = me.XPos, y = me.YPos }
    return vec_dist( vec_me, param.lastPlayerPos)
end

local function isInSight(target)
    local me, player = getMe(), (target or getPlayer())
    
    local w = param.sightXRange
    local h = param.sightY
    if param.bossBurstedMode then
      w = param.sightXRageRange
      h = param.sightYRageRange
    end
    --
    local diff = { x = player.XPos - me.XPos ,y = player.YPos - me.YPos }
    if diff.x >= -w and diff.x <= w
    and diff.y >= h.bottom and diff.y <= h.top
    then
        return true 
    end
    return false
end

local function isInSightAlerted(target)
    local me, player = getMe(), (target or getPlayer())
    
    local w = param.sightXRangeAlerted
    local h = param.sightY
    if param.bossBurstedMode then
      w = param.sightXRageRange
      h = param.sightYRageRange
    end
    --
    local diff = { x = player.XPos - me.XPos ,y = player.YPos - me.YPos }
    if diff.x >= -w and diff.x <= w
    and diff.y >= h.bottom and diff.y <= h.top
    then
        return true 
    end
    return false
end

local function attack_swipe(data)
    if data ~= nil then
        data.dir = dirs.N
        data.pressAction = false
    end
    doMotion("X+Attack")
end

local function attack_dive(data)
    if data ~= nil then
        data.dir = dirs.U
        data.pressAction = false
    end
    doMotion("U+Attack")
end

local function attack_headbutt(data)
    if data ~= nil then
        data.dir = dirs.D
        data.pressAction = false
    end
    doMotion("D+Attack")
end

local function attack_kidnap(data)
    if data ~= nil then
        data.dir = dirs.N
        data.pressAction = true
    end
    doMotion("X+Heruka")
end

-- the initialization function for this AI
function OnInit()
    -- Accessible anywhere in this file after Initialization
    local me, player = getMe(), getPlayer()
    --
    param.origin = {
        x = me.XPos,
        y = me.YPos
    }
    if me:getSpawnpointPos() ~= nil then
      param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
    end
    --
    param.destPoint = {
        x = param.origin.x,
        y = param.origin.y
    }
    param.maxDistFromOrigin = 160 * 4
    param.destPoint.x = param.destPoint.x + (param.maxDistFromOrigin * (me.FacingDir * -1))
    --
    --
    param.sightXRange = 160 * 2
    param.sightXRangeAlerted = 160 * 3.5
    param.sightY = {
        top = 160 * 3,
        bottom = 160 * -4
    }
    --
    param.sightXRageRange = 160 * 8
    param.sightYRageRange = {
        top = 160 * 6,
        bottom = 160 * -8
    }
    --
    param.time = 0 
    param.turnWaitTime = 0 
    param.lastPlayerPos = { x = 0, y = 0 }
    --
    param.searchingForPlayer = false
    --
    param.debug_startRoutine = false
    -- COOP
    param.COOP_newTargetOK = true
    param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()
  
  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()
    if param.COOP_newTargetOK then
      target = getClosestPartyMember()
    end
    --
    if aiDebug() then
      draw_point(param.destPoint.x, param.destPoint.y)
      draw_point(param.lastPlayerPos.x, param.lastPlayerPos.y)
      draw_bar(param.turnWaitTime, 0, 60, 1, {255, 0, 255})
      draw_bar(param.time, 0, 120)
    end
    --

    if me:hasTag("Dive_Start")
    then
      trace("WARNING: SETTING ENEMY MANUALLY")
      me:setEnemyLeaderTo(target)
      param.COOP_newTargetOK = false
    else
      param.COOP_newTargetOK = true
    end 
    
    -- co-op 
    if bossBursted() 
    then
      param.turnWaitTime = 0
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 3 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end

    if tick_ok() then
      param.turnWaitTime = math.max(0, param.turnWaitTime - 1) 
      --
      local _dist = dist_point(param.destPoint.x, param.destPoint.y)
      if me.Alerted >= 30 or me.Hitstun > 0 
      and me ~= nil then
        param.lastPlayerPos = { x = target.XPos, y = target.YPos + 70 }
        param.searchingForPlayer = true
      end
      if me.Lifebar <= 0 then
        -- let's just do NOTHING
        debugPalette("Debug_Dark")
        wait(30)
      elseif has_spawn_command("Stationary") then
        bh_flying()
      elseif ( param.searchingForPlayer and isInSightAlerted(target) or isInSight(target) ) and ( param.bossBurstedMode or sprite_vulnerable(target) ) then
        -- let's try this..
        debugPalette("Debug_Blue")
        -- Wait a bit
        param.time = 0
        wait()
        param.lastPlayerPos = { x = target.XPos, y = target.YPos + 40 } 
        local vec_dist = dist_lastPlayerPos()
        param.searchingForPlayer = true
        wait(10)
        while param.time <= 100 do
          debugPalette("Debug_Yellow")
          draw_point(param.lastPlayerPos.x, param.lastPlayerPos.y)
          draw_point(param.lastPlayerPos.x, param.lastPlayerPos.y + 70)
          draw_bar(param.time, 0, 100)

          -- Boost to catch the player
          if vec_dist > 160 * 2 then 
              me.Slot1 = 24
          end
          
          fly_to(param.lastPlayerPos.x, param.lastPlayerPos.y)
          --
          vec_dist = dist_lastPlayerPos()
          param.lastPlayerPos = { x = target.XPos, y = target.YPos + 70 } 
          param.time = param.time + 1
          --
          if vec_dist <= 140 then 
            param.time = 1024 
            break 
          end
          --
          wait()
        end
        --
        debugPalette("Debug_Red")
        if sprite_vulnerable(target) and param.time >= 1024 then
          faceSprite(target)
          attack_dive()
          wait(10)
        end
        debugPalette("Debug_Green")
        --
        param.time = 0
      elseif param.searchingForPlayer then
        -- Let's just check the point we saw the player last
        if dist_sprite(target) <= 100 then
          param.searchingForPlayer = false
        elseif dist_lastPlayerPos() <= 100 then
          -- Hand off to the insight thing
          param.searchingForPlayer = false
          wait(30)
        end
        fly_to(param.lastPlayerPos.x, param.lastPlayerPos.y)
      elseif param.turnWaitTime > 0 then
      elseif math.abs( _dist ) <= me.WorldCollisionWidth / 2 then
        param.destPoint.x = param.origin.x + (param.maxDistFromOrigin * (me.FacingDir * -1))
        param.turnWaitTime = 60 * 1
      else
        fly_to(param.destPoint.x, param.destPoint.y)
      end
    end

    --
    wait() -- try to keep this at the end
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  local dirChance = random(1000)
  local actionChance = random(1000)

  if dirChance > 500
  then
    attack_dive(toReturn)
  else
    attack_swipe(toReturn)
  end

  if aiDebug()
  then
    --attack_headbutt(toReturn)
    attack_swipe(toReturn)
  end

  return toReturn
end --OnBattleAttack()