require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function isInSightRange(_spr)
  local me = getMe()
  local spr = _spr or getPlayer()
  
  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  if isInRange(spr, me.XPos, me.YPos, w, h )
  then
    return true
  end
  return false
end

local function isInAttackRange(_spr)
  local me = getMe()
  local spr = _spr or getPlayer()
  
  local w = param.attackRangeX
  local h = param.attackRangeY
  if param.bossBurstedMode then
    w = param.attackRangeRageX
    h = param.attackRangeRageY
  end
  
  if isInRange(spr, me.XPos, me.YPos, w, h )
  then
    return true
  end
  return false
end

local function attack_phono(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.meterBurn = false
  end
  doMotion("U+Heruka")
end

local function attack_beam(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.meterBurn = false
  end
  doMotion("X+Heruka")
  param.attackCooldown = 40
end

local function attack_pillars(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 120
end

local function attack_whip(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 70
end

local function debugStuff()
  draw_bar(param.attackCooldown, 0, 100)
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()
  -- Set origin
  param.time = 0
  param.attackCooldown = 0 
  param.origin = {
    x = me.XPos,
    y = me.YPos
  }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  param.sightRangeX = 160 * 18
  param.sightRangeY = 160 * 15
  param.attackRangeX = 160 * 9
  param.attackRangeY = 160 * 3.5
  --
  param.sightRangeRageX = 160 * 38
  param.sightRangeRageY = 160 * 20
  param.attackRangeRageX = 160 * 19
  param.attackRangeRageY = 160 * 6
  
  param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    if aiDebug() then
      debugStuff()
    end

    if tick_ok() then
      if not me:hasTag("Attack") then
        param.attackCooldown = math.max(0, param.attackCooldown - 1)
      end
      param.time = param.time + 1
    end

    -- co-op 
    if bossBursted() 
    then
      param.attackCooldown = 20
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 3 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    
    if update_ok()
    then
      if sprite_vulnerable(target)
      or param.bossBurstedMode then
        if param.attackCooldown <= 0
        and isInAttackRange(target)
        and dist_point(target.XPos, target.YPos) >= 0 
        then
          --debugPalette("Debug_Blue")
          attack_whip()
          if param.bossBurstedMode then
            attack_beam()
          end
        elseif isInSightRange(target)
        and dist_sprite() < 0
        and not me:hasTag("Attack")
        and not me:hasTag("TurnStanding")
        and not is_hurt()
        then
          --debugPalette("Debug_Red")
          -- turn_around() doesn't seem to work! Let's investigate why later
          wait(1)
          addInput(dirs.B)
          wait(1)
          addInput(dirs.B)
        end
      end
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  
  local dirChance = random(1000)
  local actionChance = random(1000)

  --attack_beam(toReturn)
  --attack_whip(toReturn)
  --attack_pillars(toReturn)

  return toReturn
end --OnBattleAttack()