require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local me, player = nil
local param = {
  origin = { x = 0, y = 0 },
  time = 0
}

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function isInBattleStartRange()
  local me, player = getMe(), getPlayer()
  local _w, _h = 160 * 8, 160 * 5

  return isInRange(player, me.XPos, me.YPos, 
    _w, _h
  )
end

-- Battle Attacks --
local function attack_slash(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 30
end

local function attack_moonslash(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
  param.attackCooldown = 30
end

local function attack_chargePillar(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 30
end

-- For Lhan Fight --
local function action_pillar(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function action_getOutOfBattle(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function action_superPillar(data)
  if data ~= nil then
    data.dir = dirs.D  -- Lv1
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("N+Hrk")
end

local function numCharges()
  local me = getMe()

  if me.LightR ~= nil
  then
    return me.LightR
  else
    return 0 
  end

  return 0 
end

-- the initialization function for this AI
function OnInit()
  -- Accessible anywhere in this file after Initialization
  me, player = getMe(), getPlayer()
  local players = getPlayers()

  param.attackCooldown = 0 
  param.time = 0

  param.isAshwatMiniboss = false
  if has_spawn_command("AshwatMiniboss")
  then
    param.isAshwatMiniboss = true
  end

  param.isShadowBoss = false
  if has_spawn_command("ShadowBoss")
  then
    param.isShadowBoss = true
  end

  -- Make sure!
  param.hasSpawnedPillar = false

  -- BATTLE 
  param.numTurnsSinceCharge = 0
  param.maxTurnsUntilCharge = 1
end

-- the main update function for this AI
function OnUpdate()
  me, player = getMe(), getPlayer()
  --
  while true do
    me, player = getMe(), getPlayer()
    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  
  if battleData.numChargesDone == nil
  then
    battleData.numChargesDone = 0 
  end
  battleData.numChargesDone = 0 
  
  if battleData.turnsSinceSuper == nil
  then
    battleData.turnsSinceSuper = 0 
  end
  
  if battleData.attackAfterCharge == nil
  then
    battleData.attackAfterCharge = false
  end

  local dirChance = random(1000)
  local actionChance = random(1000)
  local superOK = false --me.LifePct <= 0.55
  battleData.attackAfterCharge = false

  trace( numCharges() )

  -- Act 1 Ashwat Miniboss --
  if param.isAshwatMiniboss
  or has_spawn_command("ShadowBoss")
  then
    if battleData.turnsSinceSuper > 4
    and superOK
    and actionChance > 650
    then
      action_superPillar(toReturn)
      battleData.turnsSinceSuper = 0
    elseif param.numTurnsSinceCharge >= param.maxTurnsUntilCharge
    then
      attack_chargePillar(toReturn)
      battleData.attackAfterCharge = true
      --me.BattleSpeedMult = 1000
      --me.BaseBattleSpeed = 1000
      param.numTurnsSinceCharge = 0 
      battleData.turnsSinceSuper = battleData.turnsSinceSuper + 1
    else
      if dirChance > 550
      and numCharges() == 1
      then
        attack_moonslash(toReturn)
      else
        attack_slash(toReturn)
      end
      --
      param.numTurnsSinceCharge = param.numTurnsSinceCharge + 1
      battleData.turnsSinceSuper = battleData.turnsSinceSuper + 1
    end
  -- Act 3 Shadow Boss -- 
  elseif param.isShadowBoss
  then
    attack_slash(toReturn)
  -- NO IDEA WHAT THIS FOR
  else
    attack_slash(toReturn)
  end

  return toReturn
end --OnBattleAttack()

function OnBattleTurnUpdate(battleData)
  local me = getMe()

  local charges = numCharges()
  local secondSlashOK = me.LifePct <= 0.55
  local thirdSlashOK = (me.LifePct <= 0.25) and (charges <= 0) 
  local secondChargeOK = me.LifePct <= 0.65
  local thirdChargeOK = false -- me.LifePct <= 0.33
  if has_spawn_command("ShadowBoss")
  then
    thirdChargeOK = true
  end

  param.attackCooldown = math.max(0, param.attackCooldown - 1)

  -- [[ SLASHES ]]
  if me:hasTag("Attack")
  and me:hasTag("ThreeSlashes")
  then
    param.time = param.time + 1
    
    if me:hasTag("ThreeSlashes_First")
    and secondSlashOK
    and param.time >= 22
    and param.attackCooldown <= 0 
    then
      attack_slash()
      param.time = 0 
    elseif me:hasTag("ThreeSlashes_Second")
    and thirdSlashOK
    and param.time >= 22
    and param.attackCooldown <= 0 
    then
      attack_slash()
      param.time = 0 
    end
  -- [[ CHARGES ]]
  elseif me:hasTag("Attack")
  and me:hasTag("StoneCharge")
  then
    param.time = param.time + 1

    if secondChargeOK
    and param.time >= 52
    and param.attackCooldown <= 0
    and battleData.numChargesDone <= 0
    then
      attack_chargePillar()
      param.attackCooldown = 3
      battleData.numChargesDone = battleData.numChargesDone + 1
    elseif thirdChargeOK
    and param.time >= 82
    and param.attackCooldown <= 0
    and battleData.numChargesDone <= 1
    then
      attack_chargePillar()
      param.attackCooldown = 1000
      battleData.numChargesDone = battleData.numChargesDone + 1
    end
  --[[ COMBO SLASH ]]
  --[[
  elseif me:hasTag("Standing")
  and secondChargeOK
  and param.time >= 82
  and param.attackCooldown <= 0
  then
    attack_slash(toReturn)
  ]]
  else
    param.time = 0 
    -- Send No Inputs
  end
  
  --trace("time: %i" % { param.time })
  --trace("atkcd: %i" % { param.attackCooldown })
  --trace("numCh: %i" % { battleData.numChargesDone })
end