require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function block()
  doMotion("X+WorldGuard")
end

local function attack_tongueSlam(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
  end
  doMotion("U+Attack")
  param.attackCooldown = 30
end

local function attack_bonk(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 7
end

local function attack_roll(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 14
end

local function attack_shower(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
  end
  doMotion("U+Heruka")
  param.attackCooldown = 14
end

local function attack_shellFlip(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
  end
  doMotion("D+Heruka")
  param.attackCooldown = 14
end

local function attack_shellFlip_coinFlip(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
  end
  doMotion("X+Attack")
end

local function attack_shellFlip_shower(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
  end
  doMotion("U+Attack")
end

local function attack_shellFlip_exit(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
  end
  doMotion("X+Heruka")
end

local function attack_coinFlip(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
  end
  doMotion("X+Heruka")
  param.attackCooldown = 14
end

local function shellFlipOK()
  local me = getMe()

  if me.ColorNum > 1
  then
    return true 
  else
    return false
  end

  return false
end

local function isInAttackRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.attackRangeX
  local h = param.attackRangeY
  if param.bossBurstedMode then
    w = param.attackRangeRageX
    h = param.attackRangeRageY
  end
  
  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function bulletDist()
  local _x, _y = nil, nil
  local _ret = nil
  if select(1,dist_blang()) ~= nil then
    return dist_blang()
  elseif select(1,dist_spear()) ~= nil then
    return dist_spear()
  elseif  select(1,dist_arrow()) ~= nil then
    return dist_arrow()
  end
  return -256^2, -256^2
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  param.time = 0
  param.roamTimeLimit = 100 + random(180)
  param.attackCooldown = 0 
  param.blockTime = 0
  param.roamPoint = { x = 0, y = 0 }
  -- 
  param.attackRangeX = 160 * 8
  param.attackRangeY = { top = 160 * 2.25, bottom = 160 * -1.5 }
  param.sightRangeX = 160 * 14
  param.sightRangeY = { top = 160 * 2.5, bottom = 160 * -1.5 }
  --
  param.attackRangeRageX = 160 * 16
  param.attackRangeRageY = { top = 160 * 5, bottom = 160 * -1.5 }
  param.sightRangeRageX = 160 * 22
  param.sightRangeRageY = { top = 160 * 5, bottom = 160 * -1.5 }
  --
  param.closeAttackRangeX = 160 * 2 -- so we can BONK
  param.closeRangeX = 160 * 4 -- so we can STOP moving
  param.blockRangeX = 160 * 8

  param.isChallengePoison = false
  if me.WeaponType == hash("Challenge_Poison")
  then
    param.isChallengePoison = true
  end

  param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()
  
  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()

    --
    if aiDebug() then
      draw_bar(param.attackCooldown, 0, 120, 2, {255, 0, 0})
      draw_bar(param.blockTime, 0, 60)
      draw_bar(param.time, -100, 240, 1, {255,255,255})
    end

    if tick_ok() then
      param.time = param.time + 1
      param.blockTime = math.max(0, param.blockTime - 1)

      if not me:hasTag("Attack")
      then
        param.attackCooldown = math.max(0, param.attackCooldown - 1)
      end
    end

    -- co-op 
    if bossBursted() 
    then
      param.attackCooldown = 0 
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 3 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    
    -- Block priority
    if update_ok()
    then
      if isInAttackRange(target)
      and dist_sprite(target) >= 0
      and ( param.bossBurstedMode or sprite_vulnerable(target) )
      and not me:hasTag("Attack")
      and param.attackCooldown <= 0 
      then
        if math.abs(dist_sprite(target)) <= param.closeAttackRangeX 
        then
          attack_bonk()
          if param.bossBurstedMode then
            attack_tongueSlam()
          end
        else
          attack_tongueSlam()
          if param.bossBurstedMode then
            attack_roll()
          end
        end
      elseif isInSightRange(target)
      and dist_sprite(target) >= 0
      and ( param.bossBurstedMode or sprite_vulnerable(target) )
      and not obstacle_in_front(80)
      then
        if math.abs(dist_sprite(target)) >= param.closeRangeX
        then
          approach( target.XPos )
        end
      elseif isInSightRange(target)
      and dist_sprite(target) < -2
      and ( param.bossBurstedMode or sprite_vulnerable(target) )
      and not me:hasTag("Attack")
      and not obstacle_behind(80)
      then
        turn_around(10)
      end
    end
    
    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  local dirChance = random(1000)
  local actionChance = random(1000)

  -- reset vars
  if battleData.shellTurns == nil
  then
    battleData.shellTurns = 0 
  end

  if battleData.turnsSinceShell == nil
  then
    battleData.turnsSinceShell = 0 
  end

  -- battle counting
  if me:hasTag("ShellFlip")
  then
    battleData.shellTurns = battleData.shellTurns + 1
    battleData.turnsSinceShell = 0
  else
    battleData.shellTurns = 0 
    battleData.turnsSinceShell = battleData.turnsSinceShell + 1
  end
  trace("%02i/%02i" % { battleData.shellTurns, battleData.turnsSinceShell })
  
  -- shell
  if me:hasTag("ShellFlip")
  then
    if aiDebug()
    then
      attack_shellFlip_coinFlip(toReturn)
    elseif battleData.shellTurns > 3
    then
      trace("ShellFlip Exit")
      attack_shellFlip_exit(toReturn)
    elseif dirChance > 500
    then
      trace("ShellFlip Shower")
      attack_shellFlip_shower(toReturn)
    else
      trace("ShellFlip CoinFlip")
      attack_shellFlip_coinFlip(toReturn)
    end
  -- normal
  elseif param.isChallengePoison
  then
    if dirChance > 667
    then  
      attack_tongueSlam(toReturn)
    elseif dirChance < 334
    then
      attack_roll(toReturn)
    else
      attack_bonk(toReturn)
    end
  else
    if aiDebug()
    then
      attack_shellFlip(toReturn)
    elseif battleData.shellTurns <= 0 
    and shellFlipOK()
    and battleData.turnsSinceShell >= 3
    then
      trace("Normal ShellFlip")
      attack_shellFlip(toReturn)
    elseif actionChance > 500 then
      trace("Normal Roll")
      attack_roll(toReturn)
    elseif dirChance > 500 then
      trace("Normal Tongue")
      attack_tongueSlam(toReturn)
    else
      trace("Normal Bonk")
      attack_bonk(toReturn) 
    end
  end

  return toReturn
end --OnBattleAttack()