require("LuaLib")
require("MonsterLib")

require("util")
require("debug_ext")
require("vec2")
require("math_ext")
require("movement")

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function vec_dist_origin()
  local me = getMe()
  local vm = { x = me.XPos, y = me.YPos }
  local vo = { x = param.origin.x, y = param.origin.y }

  local result = vec_dist( vm, vo )

  return result
end

local function attack_rudeCombo(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_flight(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function attack_exitFlight(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Heruka")
end

local function attack_machineGun(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_getOuttaHere(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Heruka")
end

local function attack_swoop(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function attack_bombDrop(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_offscreenIdleEnter(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("U+Dash")
end

local function attack_offscreenIdleExit(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("D+Dash")
end

local function action_teleport()
  local me = getMe()
  doMotion("X+WorldGuard")
end

local function setTeleportInfo(_x, _y)
  local me = getMe()
  local x = _x or me.XPos
  local y = _y or me.YPos

  me.WeaponType = x
  me.DHCNum = y
end

local function setSwoopInfo(x, y, facing)
  local me = getMe()

  local _x, _y = me.XPos, me.YPos

  if x ~= nil then _x = x end
  if y ~= nil then _y = y end

  me.Slot5 = x
  me.Slot6 = y
end

local function isInFlightMode()
  local me = getMe()

  if me:hasTag("FlightMode")
  then
    trace("(!) I was in flight mode")
    return true
  end

  return false
end

local function isInBombDrop()
  local me = getMe()

  if me:hasTag("SwoopBombdrop")
  then
    return true
  else
    return false
  end

  return false
end

local function isInOffscreenIdle()
  local me = getMe()

  if me:hasTag("Offscreen_Idle")
  then
    trace("(!) I was in offscreen idle")
    return true
  end

  return false
end

local function gotHit_loopCheck()
  if got_hit(me)
  then
    trace("[GotHit] Got Hit loop check passed")
    return true
  elseif bossBursted()
  then
    trace("[GotHit] BOSS BURST")
    param.time = 300 
    me.InvinFlashTime = 400  -- longer for us
  else
    return false
  end
end

local function GetPoint( s )
  if s == nil then return nil end
  local _x, _y, _z = getObjectPos(s, false)
  if _x ~= nil then
    return {x = _x, y = _y, z = _z}
  end
end

local Mode = {
  Debug           = -100,
  GoToSpawn       = -10,
  --
  Idle            = -50,
  Swoop           = 100,
  MachineGuns     = 200,
  --
  GotHit          = 1000
}

local function PickRandomAttackPattern()
  param.swoopSidePick = 0
  param.swoopRepeats = 0  
  param.time = 0 

  local r = random(1000)
  
  if r > 500 then
    trace("Picked MachineGuns")
    param.curMode = Mode.MachineGuns
  else
    if param.isGarudaCruel
    then
      trace("Picked Grab")
      param.curMode = Mode.SwoopGrab
    else
      trace("Picked Swoop")
      param.curMode = Mode.Swoop
    end
  end

  -- Remember to add your crazy overrides!
  if aiDebug()
  then
    param.curMode = Mode.Swoop
    trace("HI THIS IS A DEBUG OVERRIDE to Swoop")
  end
  --param.curMode = Mode.SwoopGrab
end

-- the initialization function for this AI
function OnInit()
  -- Accessible anywhere in this file after Initialization
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.time = 0 
  param.origin = { x = me.XPos, y = me.YPos }
  --
  if me:getSpawnpointPos() ~= nil then
      param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  -- Assemble points
  param.Points = {
    Swoop_Right = GetPoint("Spawn_GarudaBoss_RightSwoop"),
    Swoop_Left = GetPoint("Spawn_GarudaBoss_LeftSwoop"),
    HighSwoop_Hint = GetPoint("Spawn_BossHint_Garuda_HighSwoopHint")
  }

  -- 
  param.curMode = Mode.Idle
  param.prevMode = nil

  param.modeModifier = nil
  param.swoopRepeats = 0 

  param.swoopSidePick = 0
  param.swoopGoHigh = 0 
  param.swoopHighYOffset = 160 * 3
  param.swoopShouldDropBombs = false
  param.swoopGrabbedPlayer = false
  param.swoopHighYPos = 0
  if param.Points.HighSwoop_Hint ~= nil
  then
    param.swoopHighYPos = param.Points.HighSwoop_Hint.y
    --trace( param.swoopHighYPos ) 
  end

  param.gameEventOK = true

  param.forceExitThisMode = false
  param.forceExitAnyLoopToNeutral = false
  param.forceTeleport = false

  param.speedBonus_leaving =
  { min = 40, max = 140, range = 160 * 10 }
  
  param.speedBonus_goToPoint =
  { min = 20, max = 200, range = 160 * 1 }

  param.speedBonus_machineGuns =
  { min = 6, max = 50, range = 160 * 8 }

  param.speedBonus_comingBack =
  --{ min = 40, max = 140, range = 160 * 10 }
  { min = 30, max = 80, range = 160 * 34 }

  -- 

  -- Set us up for boss modes
  param.isBoss = false
  if has_spawn_command("Boss")
  then
    param.isBoss = true
  end

  param.isGarudaCruel = false
  if has_spawn_command("A3Boss")
  then
    param.isBoss = true
    param.isGarudaCruel = true
  end

  param.quickLeaveFromFirstIdle = false
  if has_spawn_command("QuickRise")
  or has_spawn_command("Quickrise")
  then
    param.quickLeaveFromFirstIdle = true
  end

  param.isChallengeMode = false
  if me.LightX == hash("Challenge_NagaRiderTeam")
  then
    trace("Garuda Challenge Mode")
    param.isChallengeMode = true
  end
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  local players = getPlayers()
  
  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()
    --
    if aiDebug()
    then
      if param.curMode ~= param.prevMode
      then
        for k, v in pairs(Mode) do
          if param.curMode == v then
            trace(k)
          end
        end
      end

      if me:hasTag("Offscreen_Idle")
      then
        trace("Offscreen!")
      end
    end

    --
    if param.isBoss
    then
      if update_ok()
      then
        -- Just in case bailout
        if got_hit(me) then
          trace("[IsBoss] Got Hit, transitioning to GotHit")
          param.curMode = Mode.GotHit
        end
          
        if bossBursted() then
          trace("[IsBoss] BOSS BURST Got Hit, transitioning to GotHit")
          param.curMode = Mode.GotHit
          me.InvinFlashTime = 400  -- longer for us
        end

        -- Mark which mode you had
        param.prevMode = param.curMode

        --[[ Idle ]]
        if param.curMode == Mode.Idle
        then
          debugPalette("Debug_White")
          param.time = param.time + 1
          trace("[Idle] Idle Time is %i" % { param.time })
          --
          if param.quickLeaveFromFirstIdle 
          then
            param.time = 80
            param.quickLeaveFromFirstIdle = false
          end
          if param.time >= 100
          then
            param.time = 0 
            PickRandomAttackPattern()
          end
        --[[ GotHit ]]
        elseif param.curMode == Mode.GotHit
        then
          param.time = 0 
          trace("[GotHit] Entry")
          local landedBeforeTime = false

          debugPalette("Debug_Magenta")
          -- we want to use Time for a timeout, before leaving.. if we didn't land, dang!
          repeat
            debugPalette("Debug_Magenta")
            -- Breakout --
            if gotHit_loopCheck() then
              param.curMode = Mode.GotHit
              break 
            end

            param.time = param.time + 1
            trace("Time %i" % { param.time })

            -- here --
            if me.OnGround > 0 then
              landedBeforeTime = true
              break
            end
          
            debugPalette("Debug_Magenta")
            wait()
          until param.time >= 60 * 2

          -- Leave early if we didn't land...
          if landedBeforeTime then
            debugPalette("Debug_White")
            trace("Didn't land early..")
            PickRandomAttackPattern()
            --param.curMode = Mode.MachineGuns
          end

          -- We want to float upscreen then do something else, let's try it!
          repeat
            -- Can we wait until we're done with the original timer??
            if param.time <= 60 * 2
            then
              param.time = param.time + 1
              
              debugPalette("Debug_LightMagenta")
              wait()
            else
              addInput(dirs.U)
              --
              local _speed = me.Slot6 
              _speed = _speed + 1 -- faster
              me.Slot6 = _speed

              me.Slot1 = 50
              
              debugPalette("Debug_DarkMagenta")
              wait()
            end
          until me.TouchScreenUD > 1 or me.TouchScreenUD < -1 or param.time >= 60 * 2

          -- Now we choose something
          trace("[GotHit] Exit")
          PickRandomAttackPattern()
        --[[ GoToSpawn ]]
        elseif param.curMode == Mode.GoToSpawn
        then
          local distToSpawn = vec_dist_origin()
          local target = (param.swoopSidePick == 1) and param.Points.Swoop_Right or param.Points.Swoop_Left
          param.forceExitThisMode = false
          param.swoopRepeats = 0 

          trace("[GoToSpawn] Entry")

          if param.forceTeleport
          then
            if isInOffscreenIdle()
            then
              action_teleport()
              setTeleportInfo(target.x, target.y)
              wait(3)
              wait(5)
              attack_offscreenIdleExit()
            else
              attack_offscreenIdleEnter()
              wait(5)
              action_teleport()
              setTeleportInfo(target.x, target.y)
              wait(3)
              wait(5)
              attack_offscreenIdleExit()
            end
            wait(3)
            param.forceTeleport = false
          end

          fly_to( param.origin.x, param.origin.y,
            param.speedBonus_comingBack
          ) 
          if distToSpawn <= 80
          then
            param.curMode = Mode.Idle
          end

          debugPalette("Debug_Dark")
        --[[ MachineGuns ]]
        elseif param.curMode == Mode.MachineGuns
        then
          -- LET'S GO UPPPPPPPPPPP --
          trace("[MachineGuns] Going up!")
          repeat 
            -- Breakout --
            if gotHit_loopCheck() then
              param.curMode = Mode.GotHit
              break 
            end

            addInput(dirs.U)
            me.Slot1 = 80 -- Set maximum flying speed

            debugPalette("Debug_DarkCyan")
            wait()
          until me.TouchScreenUD > 1 or me.TouchScreenUD < -1
        
          -- Wait a bit --
          trace("[MachineGuns] Wait!")
          wait(20)

          -- Ensure we are no longer idle offscreen
          trace("[MachineGuns] Offscreen Idle Check")
          repeat
            trace("[MachineGuns] Attempting to not be Offscreen Idle!")
            attack_offscreenIdleExit()
            wait(2)
          until not isInOffscreenIdle()

          if not me:hasTag("FaceFront_MachineGunHands_Chase_Start")
          then
            attack_machineGun()
            wait(6) -- To prevent any extra Valid
          end

          -- Come on DOWN!!!!! --
          local function isNearPlayer()
            return isInRange( player,
              me.XPos, me.YPos,
              160 * 8, 160 * 8
            )
          end

          local _timeSeen = 0 -- Max should be 3?
          local _timeout = 60 * 6

          trace("[MachineGuns] Tracking Player")
          repeat
            -- Breakout --
            if gotHit_loopCheck() then
              param.curMode = Mode.GotHit
              break 
            end

            -- Lets use the dist instead
            local vecMe = { x = me.XPos, y = me.YPos } 
            local vecPlayer = { x = player.XPos, y = me.YPos } 
            local _dist = vec_dist( vecMe, vecPlayer ) 


            if isNearPlayer() then
              debugPalette("Debug_DarkCyan")
              _timeSeen = _timeSeen + 3 -- 300/2 = 100f
            else
              debugPalette("Debug_Cyan")
              _timeSeen = math.max(0, _timeSeen - 1)
            end

            -- timeout too
            _timeout = math.max(0, _timeout - 1)

            -- we fly to you, but let's change the offset + the range
            -- based on the time..?

            local rangeOffsetMult = 0.5

            --[[
            if _timeout <= 80
            or _timeSeen >= 60 * 2
            then
              rangeOffsetMult = 1 
            end
            ]]

            if _dist >= (80 * rangeOffsetMult) then
              fly_to( player.XPos, player.YPos + 60,
              param.speedBonus_machineGuns
              )
            end

            debugPalette("Debug_Cyan")
            wait()
          until _timeout <= 0 or _timeSeen >= 60 * 3

          trace("[MachineGuns] Firing")
          attack_machineGun()
          wait()

          repeat
            -- Breakout --
            if gotHit_loopCheck() then
              param.curMode = Mode.GotHit
              break 
            end

            debugPalette("Debug_Cyan")
            wait()
          until not me:hasTag("FaceFront_MachineGunHands")

          debugPalette("Debug_Cyan")

          trace("[MachineGuns] End")
          -- Exit State --
          param.curMode = Mode.Idle
        --[[ Swoop GRAB ]]
        elseif param.curMode == Mode.SwoopGrab
        then
          -- We SHOULD be allowed to do this
          param.gameEventOK = true

          -- Just in case we ever get here
          if gotHit_loopCheck()
          then
            param.curMode = Mode.GotHit
          -- We probably have not PICKED a side so lets
          elseif param.swoopSidePick == 0
          then
            -- We'll use this block to pick a side and GO to it! --
            local r = random(1000)
            param.swoopSidePick = (r > 500) and 1 or -1
            param.swoopGoHigh = 0 

            local target = (param.swoopSidePick == 1) and param.Points.Swoop_Right or param.Points.Swoop_Left
            local yOffset = 160 * 1.5
            local _dist = vec_dist(
              { x = me.XPos, y = me.YPos }, 
              { x = target.x, y = target.y + yOffset }
            )

            trace("Seeking a starting point")
            repeat
              -- Breakout --
              if gotHit_loopCheck() then
                param.curMode = Mode.GotHit
                param.forceExitThisMode = true
                break 
              end

              _dist = vec_dist( 
                { x = me.XPos, y = me.YPos }, 
                { x = target.x, y = target.y + yOffset }
              )

              fly_to( target.x, target.y + yOffset,
                param.speedBonus_comingBack
              )

              if isInOffscreenIdle() then
                trace("Going faster the plz")
                if not isOnScreen(me)
                then
                  action_teleport()
                  setTeleportInfo(target.x, target.y + yOffset) 
                  wait(3)
                end
              end

              debugPalette("Debug_DarkRed")
              wait()
            until _dist <= 50

            -- standing from the bottom in training mode version is
            -- 1291.42px
            param.swoopGoHigh = 0 
            if player.YPos >= param.swoopHighYPos - (player.WorldCollisionHeight/2)
            then
              param.swoopGoHigh = 1
            end

            param.swoopShouldDropBombs = me.LifePct <= 0.4
            -- Now let's face AWAY from the sideswoop and get on!
            if not param.forceExitThisMode then
              me.FacingDir = param.swoopSidePick * 1

              trace("wait 7f before doing ANY grabs")
              wait(7)
            end
          elseif param.swoopRepeats >= 1
          then
            if isInOffscreenIdle() then
              trace("lets get out of offscreen!")
              trace("wait for us to get to offscreen_idle for real?!")

              repeat
                wait()
              until me:hasTag("Offscreen_Idle") and not me:hasTag("SwoopBasic_Loop")

              trace("ok we're gonna offscreen idle BUTTON")

              trace("wait 1f")
              wait()
            end
            
            trace("going to GoToSpawn")
            param.swoopRepeats = 0
            param.curMode = Mode.GoToSpawn
            param.forceTeleport = true
            -- Left to Right
          elseif param.swoopSidePick == -1
          then
            trace("Swoop LTR")
            local _facingDir = me.FacingDir
            
            -- MAKE SURE WE'RE NOT IN OFFSCREEN IDLE
            repeat
              trace("LTR WAIT FOR OFFSCREEN")
              attack_offscreenIdleExit()
              wait()
            until not isInOffscreenIdle()

            trace("LTR WAITING FOR SWOOPZONE")
            
            -- EVENT
            if param.gameEventOK
            then
              doGameEvent("Event_Garuda_Swoop_FromLeft")
              param.gameEventOK = false
              wait()
            end

            trace("wait 25f before a grab")
            wait(25)

            -- Swoopzone
            me.FacingDir = 1
            -- Manually calc offset
            local _y = param.Points['Swoop_Left'].y
            if param.swoopGoHigh > 0
            then
              _y = _y + param.swoopHighYOffset
            end
            setSwoopInfo(param.Points['Swoop_Left'].x, _y)
            --
            me.LightY = 0 
            if param.swoopShouldDropBombs
            then
              me.LightY = hash("Bomb")
            elseif param.isGarudaCruel
            then
              me.LightY = hash("Grab")
            end
            attack_swoop()
            wait()

            trace("LTR DONE SWOOPZONE")
            param.time = 0 
            repeat
              param.time = param.time + 1
              wait()

              trace("AAAAAA")

              if me:hasTag("SwoopGrab_Success")
              or me:hasTag("SwoopGrab_Success_Hold_Loop")
              then
                param.swoopGrabbedPlayer = true
              end
            until me:hasTag("SwoopGrab_Reposition")
            or isInOffscreenIdle()
            or me:hasTag("Stand")
            or ((me.TouchScreenLR >= 2 or me.TouchScreenLR <= -2) and param.time >= 30)
            or ((me.TouchScreenUD >= 2 or me.TouchScreenUD <= -2) and param.time >= 7)
            or me.InBattle > 0 

            trace("SWOOPZONE BETWEEN END")

            while me:hasTag("SwoopGrab_Reposition")
            or me:hasTag("Swoop_Grab_Throw_To_Ground")
            or me:hasTag("SwoopGrab_Recovery")
            do
              trace("BBBBBBBBBBBBB")
              debugPalette("Debug_BlinkRed")
              wait()
            end

            param.swoopRepeats = param.swoopRepeats + 1
            param.swoopSidePick = param.swoopSidePick * -1
            --me.FacingDir = param.swoopSidePick * -1


            trace("LTR THIS IS THE END OF THE BLOCK WHAT THE HECK")
          elseif param.swoopSidePick == 1
          then
            trace("Swoop RTL")
            local _facingDir = me.FacingDir
            
            -- MAKE SURE WE'RE NOT IN OFFSCREEN IDLE
            repeat
              trace("RTL WAIT FOR OFFSCREEN")
              attack_offscreenIdleExit()
              wait()
            until not isInOffscreenIdle()
            
            trace("RTL WAITING FOR SWOOPZONE")

            -- EVENT
            if param.gameEventOK
            then
              doGameEvent("Event_Garuda_Swoop_FromLeft")
              param.gameEventOK = false
              wait()
            end

            trace("wait 25f before a grab")
            wait(25)

            -- Swoopzone
            me.FacingDir = -1
            -- Manually calc offset
            local _y = param.Points['Swoop_Right'].y
            if param.swoopGoHigh > 0
            then
              _y = _y + param.swoopHighYOffset
            end
            setSwoopInfo(param.Points['Swoop_Right'].x, _y)
            --
            if param.swoopShouldDropBombs
            then
              me.LightY = hash("Bomb")
            elseif param.isGarudaCruel
            then
              me.LightY = hash("Grab")
            end
            attack_swoop()
            wait()
            
            trace("RTL DONE SWOOPZONE")

            param.time = 0 
            repeat
              param.time = param.time + 1
              wait()

              trace("AAAAAA")

              if me:hasTag("SwoopGrab_Success")
              or me:hasTag("SwoopGrab_Success_Hold_Loop")
              then
                param.swoopGrabbedPlayer = true
              end
            until me:hasTag("SwoopGrab_Reposition")
            or isInOffscreenIdle()
            or me:hasTag("Stand")
            or me.InBattle > 0
            or ((me.TouchScreenLR >= 2 or me.TouchScreenLR <= -2) and param.time >= 30)
            or ((me.TouchScreenUD >= 2 or me.TouchScreenUD <= -2) and param.time >= 7)

            trace("SWOOPZONE BETWEEN END")

            while me:hasTag("SwoopGrab_Reposition")
            or me:hasTag("Swoop_Grab_Throw_To_Ground")
            or me:hasTag("SwoopGrab_Recovery")
            do
              trace("BBBBBBBBBBBBB")
              debugPalette("Debug_BlinkRed")
              wait()
            end

            trace("RTL I GUESS WE ARE DONE HERE")

            param.swoopRepeats = param.swoopRepeats + 1
            param.swoopSidePick = param.swoopSidePick * -1
            --me.FacingDir = param.swoopSidePick * -1

            trace("RTL THIS IS THE END OF THE BLOCK WHAT THE HECK")
          end
          debugPalette("Debug_Red")
        --[[ Swoop ]]
        elseif param.curMode == Mode.Swoop
        then
          -- We SHOULD be allowed to do this
          param.gameEventOK = true

          -- Just in case we ever get here
          if gotHit_loopCheck()
          then
            param.curMode = Mode.GotHit
          -- We probably have not PICKED a side so lets
          elseif param.swoopSidePick == 0
          then
            -- We'll use this block to pick a side and GO to it! --
            local r = random(1000)
            param.swoopSidePick = (r > 500) and 1 or -1
            param.swoopGoHigh = 0 

            local target = (param.swoopSidePick == 1) and param.Points.Swoop_Right or param.Points.Swoop_Left
            local _dist = vec_dist(
              { x = me.XPos, y = me.YPos }, 
              { x = target.x, y = target.y }
            )

            trace("Seeking a starting point")
            repeat
              -- Breakout --
              if gotHit_loopCheck() then
                param.curMode = Mode.GotHit
                param.forceExitThisMode = true
                break 
              end

              _dist = vec_dist( 
                { x = me.XPos, y = me.YPos }, 
                { x = target.x, y = target.y }
              )

              fly_to( target.x, target.y,
                param.speedBonus_comingBack
              )

              if isInOffscreenIdle() then
                trace("Going faster the plz")
                if not isOnScreen(me)
                then
                  action_teleport()
                  setTeleportInfo(target.x, target.y)
                  wait(3)
                end
              end

              debugPalette("Debug_DarkBlue")
              wait()
            until _dist <= 50

            -- We'll pick being high based on her ydist from US
            local ydist = select(2, dist_player())
            trace(ydist)

            -- standing from the bottom in training mode version is
            -- 1291.42px
            param.swoopGoHigh = 0 
            if player.YPos >= param.swoopHighYPos - (player.WorldCollisionHeight/2)
            then
              param.swoopGoHigh = 1
            end

            param.swoopShouldDropBombs = me.LifePct <= 0.4
            -- Now let's face AWAY from the sideswoop and get on!
            if not param.forceExitThisMode then
              me.FacingDir = param.swoopSidePick * 1

              trace("wait 7f before doing ANY sweeps")
              wait(7)
            end
          -- Leave before we swoop ANYMORE
          elseif param.swoopRepeats >= 3
          then
            if isInOffscreenIdle() then
              trace("lets get out of offscreen!")
              trace("wait for us to get to offscreen_idle for real?!")

              repeat
                wait()
              until me:hasTag("Offscreen_Idle") and not me:hasTag("SwoopBasic_Loop")

              trace("ok we're gonna offscreen idle BUTTON")
              --attack_offscreenIdleExit()

            end
            
            trace("going to GoToSpawn")
            param.swoopRepeats = 0
            param.curMode = Mode.GoToSpawn
            param.forceTeleport = true
            -- Left to Right
          elseif param.swoopSidePick == -1
          then
            trace("Swoop LTR")
            local _facingDir = me.FacingDir
            
            -- MAKE SURE WE'RE NOT IN OFFSCREEN IDLE
            repeat
              trace("LTR WAIT FOR OFFSCREEN")
              attack_offscreenIdleExit()
              wait()
            until not isInOffscreenIdle()

            trace("LTR WAITING FOR SWOOPZONE")
            -- EVENT
            if param.gameEventOK
            then
              doGameEvent("Event_Garuda_Swoop_FromLeft")
              param.gameEventOK = false
              wait()
            end

            trace("wait 25f before a swoop")
            wait(25)

            -- Swoopzone
            me.FacingDir = 1
            -- Manually calc offset
            local _y = param.Points['Swoop_Left'].y
            if param.swoopGoHigh > 0
            then
              _y = _y + param.swoopHighYOffset
            end
            setSwoopInfo(param.Points['Swoop_Left'].x, _y)
            --
            me.LightY = 0 
            if param.swoopShouldDropBombs
            then
              me.LightY = hash("Bomb")
            end
            attack_swoop()
            wait()

            trace("LTR DONE SWOOPZONE")

            param.time = 0 
            repeat
              --trace("LTR WE ARE GOING TO WAIT TO BE OFFSCREEN")
              --
              -- Breakout --
              if gotHit_loopCheck() then
                param.curMode = Mode.GotHit
                param.forceExitThisMode = true
                break 
              end
              param.time = param.time + 1

              wait()
            until (me.TouchScreenLR >= 2 or me.TouchScreenLR <= -2)
            --or (isInOffscreenIdle() and param.time >= 30)
            --or ((me.TouchScreenUD >= 2 or me.TouchScreenUD <= -2) and param.time >= 30)
            --or me.InBattle > 0 
            and isInOffscreenIdle()

            trace("LTR I GUESS WE ARE DONE HERE")

            param.swoopRepeats = param.swoopRepeats + 1

            param.swoopSidePick = param.swoopSidePick * -1
            me.FacingDir = param.swoopSidePick * -1

            trace("wait 7f after a swoop")
            wait(7)

            trace("LTR THIS IS THE END OF THE BLOCK WHAT THE HECK")
          elseif param.swoopSidePick == 1
          then
            trace("Swoop RTL")
            local _facingDir = me.FacingDir
            
            -- MAKE SURE WE'RE NOT IN OFFSCREEN IDLE
            repeat
              trace("RTL WAIT FOR OFFSCREEN")
              attack_offscreenIdleExit()
              wait()
            until not isInOffscreenIdle()
            
            trace("RTL WAITING FOR SWOOPZONE")

            -- EVENT
            if param.gameEventOK
            then
              doGameEvent("Event_Garuda_Swoop_FromRight")
              param.gameEventOK = false
              wait()
            end

            trace("wait 25f before a swoop")
            wait(25)

            -- Swoopzone
            me.FacingDir = -1
            -- Manually calc offset
            local _y = param.Points['Swoop_Right'].y
            if param.swoopGoHigh > 0
            then
              _y = _y + param.swoopHighYOffset
            end
            setSwoopInfo(param.Points['Swoop_Right'].x, _y)
            --
            if param.swoopShouldDropBombs
            then
              me.LightY = hash("Bomb")
            end
            attack_swoop()
            wait()
            
            trace("RTL DONE SWOOPZONE")

            param.time = 0
            repeat
              --trace("RTL WE ARE GOING TO WAIT TO BE OFFSCREEN")
              --
              -- Breakout --
              if gotHit_loopCheck() then
                param.curMode = Mode.GotHit
                param.forceExitThisMode = true
                break 
              end

              param.time = param.time + 1

              wait()
            until (me.TouchScreenLR >= 2 or me.TouchScreenLR <= -2)
            --or (isInOffscreenIdle() and param.time >= 30)
            --or ((me.TouchScreenUD >= 2 or me.TouchScreenUD <= -2) and param.time >= 30)
            --or me.InBattle > 0 
            and isInOffscreenIdle()

            trace("RTL I GUESS WE ARE DONE HERE")

            param.swoopRepeats = param.swoopRepeats + 1

            param.swoopSidePick = param.swoopSidePick * -1

            trace("wait 7f after a swoop")
            wait(7)

            trace("RTL THIS IS THE END OF THE BLOCK WHAT THE HECK")
          end
          debugPalette("Debug_Blue")
        --[[ No Mode ]]
        else
          debugPalette("Debug_Blink")
        end
      end
    -- Not a boss --
    else
      debugPalette("Debug_Blink")
    end

    if param.isGarudaCruel
    then
      debugPalette("Debug_LightOrange")
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  local dirChance = random(1000)
  local actionChance = random(1000)

  if param.isChallengeMode then
    trace("Garuda Challenge Mode!")
    if dirChance > 500
    then
      attack_rudeCombo(toReturn)
      --attack_machineGun(toReturn)
    else
      attack_rudeCombo(toReturn)
    end
  else
    attack_getOuttaHere(toReturn)

    -- WE ALWAYS want to go to LEAVING!!! --
    param.curMode = Mode.GotHit
    param.forceExitAnyLoopToNeutral = true
  end
  return toReturn
end --OnBattleAttack()