require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("movement")
require("vec2")

-- Distance

local function dist_origin()
    return dist_point(param.origin.x, param.origin.y)
end

local function dist_lastPlayerPos()
    return dist_point(param.lastPlayerPos.x, param.lastPlayerPos.y)
end

-- Range Checking

local function range_close(target)
  local me, player = getMe(), (target or getPlayer())
  --
  local x, y = dist_sprite(target)
  if x <= param.sight_close.x and x >= -param.sight_close.x
  and y >= param.sight_close.bottom and y <= param.sight_close.top
  then
      return true 
  end
  return false
end

local function range_far(target)
  local me, player = getMe(), (target or getPlayer())
  --
  local x, y = dist_sprite(target)
  if x <= param.sight_far.x and x >= -param.sight_far.x
  and y >= param.sight_close.bottom and y <= param.sight_close.top
  then
      return true 
  end
  return false
end

-- Checks n' Balances
-- Attacks

local function attack_swipe(data)
    local me = getMe()
    --
    if data ~= nil then 
        data.dir = dirs.N
        data.pressAction = false
    end
    doMotion("X+Attack")
end

local function attack_bowl(data)
    local me = getMe()
    --
    if data ~= nil then 
        data.dir = dirs.U
        data.pressAction = false
    end
    doMotion("U+Attack")
end

local function attack_eat(data)
    local me = getMe()
    --
    if data ~= nil then 
        data.dir = dirs.D
        data.pressAction = false
    end
    doMotion("D+Attack")
end

local function attack_showoff(data)
    local me = getMe()
    --
    if data ~= nil then 
        data.dir = dirs.N
        data.pressAction = true
    end
    doMotion("X+Heruka")
end

local function attack_kick(data)
    local me = getMe()
    --
    if data ~= nil then 
        data.dir = dirs.U
        data.pressAction = true
    end
    doMotion("U+Heruka")
end

local function attack_hide(data)
    local me = getMe()
    --
    if data ~= nil then 
        data.dir = dirs.D
        data.pressAction = true
    end
    doMotion("D+Heruka")
end

local function attack_hideThrow(data)
    local me = getMe()
    --
    if data ~= nil then 
      data.dir = dirs.N
      data.pressAction = true
    end
    doMotion("X+Heruka")
end

local function attack_hideEmerge(data)
    local me = getMe()
    --
    if data ~= nil then 
        data.dir = dirs.U
        data.pressAction = true
    end
    doMotion("X+Jump")
end

local function _drawDebug()
    if has_spawn_command("Taunt") then
        draw_bar(param.time, 0, 120, 2, {0, 255,0})
    else
        draw_bar(param.hideCooldown, 0, 120, 2, {0, 255,0})
    end
    draw_bar(param.attackStartup, 0, 20, 1)
    draw_bar(param.attackCooldown, 0, 120, 0, {255, 0, 0 })
    draw_point(param.lastPlayerPos.x, param.lastPlayerPos.y)
    draw_point(param.origin.x, param.origin.y)
end

local function move_ok()
  local me = getMe()

  if me.ColorNum ~= 3
  then
    return true
  end

  return false
end

-- the initialization function for this AI
function OnInit()
  -- Accessible anywhere in this file after Initialization
  local me, player = getMe(), getPlayer()

  --
  param.origin = {
    x = me.XPos,
    y = me.YPos
  }
  param.lastPlayerPos = {
    x = player.XPos,
    y = player.YPos
  }
  param.time = 0
  param.attackStartup = 0 
  param.attackCooldown = 0
  param.hideCooldown = 0 
  --
  param.triedJumpAttack = false
  param.hasShownOff = false
  param.hasTriedHidingOnce = false
  param.wantsToHide = true
  param.wantsToInvestigate = false
  param.allowedToPopOut = true
  param.allowedToHideAttack = true
  param.isATaunter = false
  --
  param.sight_close = { 
    x = 160 * 2,
    top = 160 * 2.5,
    bottom = 160 * -1
  }
  param.sight_far = {
    x = 160 * 6,
  }
  param.baseHideTime = 50
  --
  param.forceBattleOnHit = false
  --
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  if has_spawn_command("Sentry") then
    trace("Sentry")
    param.allowedToPopOut = false
  elseif has_spawn_command("ChaseOnly") then
    trace("ChaseOnly")
    param.allowedToHideAttack = false
    param.baseHideTime = 25
  elseif has_spawn_command("Taunt") then
    param.isATaunter = true
  end

  -- COOP
  param.COOP_newTargetOK = true
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()
    if param.COOP_newTargetOK then
      target = getClosestPartyMember()
    end
    --
    if aiDebug() then
      _drawDebug()
    end

    if got_hit() then
      param.lastPlayerPos.x, param.lastPlayerPos.y = target.XPos, target.YPos
    end

    if is_alerted() then
      param.wantsToInvestigate = true
    end

    if hit_player(target) and param.forceBattleOnHit then
      me:startBattle(player)
    end

    if me:hasTag("Hide_And_Throw_Wrench")
    then
      me:setEnemyLeaderTo(target)
      param.COOP_newTargetOK = false
    else
      param.COOP_newTargetOK = true
    end

    if update_ok() then
      if tick_ok() then
        param.attackCooldown = math.max(0, param.attackCooldown - 1)
      end
      param.hideCooldown = math.max(0, param.hideCooldown - 1)
        
      debugPalette()
      if param.isATaunter then
        _drawDebug()
        param.time = param.time - 1
        if not me:hasTag("ShowOffItem_Loop")
        and tick_ok()
        and param.time <=0 
        then
          attack_showoff()
          param.time = 100
          param.time = param.time + random(300)
        elseif sprite_vulnerable(target)
        and dist_sprite(target) < 0
        and select(2, dist_sprite(target)) <= 160 * 4
        and select(2, dist_sprite(target)) >= 0 
        then
          turn_around(11)
        end
      elseif target.Lifebar <= 0
      and me.OnGround 
      and not param.hasShownOff 
      and not me:hasTag("Attack") 
      then
        debugPalette("Debug_Magenta")
        attack_showoff()
        wait(30)
        param.hasShownOff = true
      elseif obstacle_in_front(160) then
        turn_around(22)
      elseif me:hasTag("Hide_and_Throw") then
        if range_close(target) then -- math.abs(dist_sprite()) <= 160 * 1.5
          debugPalette("Debug_Cyan")
          param.time = 0
          if math.abs(dist_sprite(target)) <= 240 then
            param.attackStartup = param.attackStartup + 1
          end
          if param.attackStartup > 10 and param.allowedToPopOut  
          then     
            attack_hideEmerge()
            faceSprite(target)
            param.attackCooldown = 27
          end
        elseif range_far(target)
        and param.attackCooldown <= 0
        and param.allowedToHideAttack
        then
          debugPalette("Debug_Yellow")
          attack_hideThrow()
          faceSprite(target)
          me:setEnemyLeaderTo(target)
          param.COOP_newTargetOK = false
          param.attackCooldown = 80
          param.time = 0
        else
          param.time = param.time + 1
          if param.time >= 200
          then
            param.time = 0 - (60 + random(120))
          elseif param.time >= 110
          then
            addInput(dirs.U)
          end
        end
      elseif math.abs(dist_sprite(target)) <= param.sight_far.x 
      then
        debugPalette("Debug_Blue")
        param.wantsToInvestigate = false
        if param.attackCooldown <= 0 then
          if dist_sprite(target) <= 160 * 1 and dist_sprite(target) > 0 
          and range_close(target) then
            attack_swipe()
            param.attackCooldown = 32
          elseif range_close() then
            attack_kick()
            me:setEnemyLeaderTo(target)
            param.COOP_newTargetOK = false
            
            param.attackCooldown = 40
          else
            approach(target.XPos)
            param.hideCooldown = param.baseHideTime
          end
        end
      elseif param.wantsToInvestigate
      and me.OnGround
      and not me:hasTag("Hide_and_Throw") then
        debugPalette("Debug_Red")
        if math.abs(dist_lastPlayerPos()) <= me.WorldCollisionWidth then
            param.wantsToInvestigate = false 
        end
        approach(param.lastPlayerPos.x)
        param.hideCooldown = param.baseHideTime + 20 
      elseif not param.hasTriedHidingOnce 
      and param.wantsToHide
      and me.OnGround 
      and param.hideCooldown <= 0
      --and math.abs( dist_origin() ) <= me.WorldCollisionWidth / 2
      and not me:hasTag("Hide_and_Throw")
      then
        repeat 
          attack_hide()
          param.hasTriedHidingOnce = true
          param.attackStartup = 0 
          wait()
        until me:hasTag("Hide_and_Throw")
      else
        debugPalette("Debug_Green")
        -- let's try this!
        if true --math.abs( dist_origin() ) >= me.WorldCollisionWidth / 2
        and param.hideCooldown <= 0
        then
          approach(param.origin.x)
        end
        param.hasTriedHidingOnce = false
        param.wantsToHide = true
      end
    end

    wait() -- try to keep this at the end
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
    local toReturn = DefaultOnBattleAttack(battleData)
    local me = getMe()

    if me:hasTag("Hide_And_Throw")
    then
        attack_hideEmerge(toReturn)
    end

  -- Pop out if we're not already present in battle
  --[[
  if me:hasTag("Hide_And_Throw") then
    if battleData.turnNum >= 4 then
      attack_hideThrow(toReturn)
    else
      attack_hideEmerge(toReturn)
    end
  end
  ]]

    return toReturn
end --OnBattleAttack()