require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("math_ext")
require("movement")
require("util")

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

-- These are all forward facing, instead of both sides, watch OUT
local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())

  local dx, dy = dist_sprite(target)

  if dx >= 0
  and dx <= param.sightRangeX
  and dy <= param.sightRangeY.top
  and dy >= param.sightRangeY.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange(target)
  local me, player = getMe(), (target or getPlayer())

  local dx, dy = dist_sprite(target)

  if dx >= 0
  and dx <= param.attackRangeX
  and dy <= param.attackRangeY.top
  and dy >= param.attackRangeY.bottom
  then
    return true
  end
  return false
end

local function hasFlag15()
  return me.Flag15 > 0 
end

local function attack_nm_projectile(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
  end
  doMotion("U+Heruka")
end

local function attack_flamethrower(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
  end
  doMotion("X+Heruka")
end

local function attack_laser(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
  end
  doMotion("X+Attack")
end

local function attack_spin(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
  end
  doMotion("D+Attack")
end


-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  param.time = 0
  param.spawnDir = me.FacingDir
  param.origin = { x = me.XPos, y = me.YPos }
  param.attackCooldown = 0
  --
  param.setUpSentry = false
  param.hitPlayer = false
  param.continueCharging = false
  param.startCharging = false
  param.enteredViewpoint = false
  param.hugeDistCheck = false
  param.turretPrimed = false
  param.isASummon = false
  param.isFastOnFirstTurn = false
  param.noFastCheck = false
  --
  param.sightRangeX = 160 * 11
  param.sightRangeY = { top = 160 * 4.7, bottom = 160 * -4 }
  param.attackRangeX = 160 * 7
  param.attackRangeY = { top = 160 * 4.7, bottom = 160 * -4 }

  if has_spawn_command("HugeBeam") then
    param.sightRangeX = 160 * 30
    param.hugeDistCheck = true
  end

  if has_spawn_command("NoFastAcquisition") then
    param.noFastCheck = true
  end

  if has_spawn_command("Summoned") 
  or has_spawn_command("Summoned_FastShot")
  then
    param.isASummon = true
  end

  if has_spawn_command("Summoned_FastShot") then
    param.isFastOnFirstTurn = true
    --me.BattleSpeedMult = 1000
  end

  param.isNMGemstone = false
  if me.WeaponType == hash("Notorious_Minion")
  then
    param.isNMGemstone = true
  end

  if param.noFastCheck
  then
    param.attackRangeX = 160 * 5
  end
end

function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember() 

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end

  while true do  
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --

    debugPalette()

    if me:hasTag("Laser_Loop") then
      param.time = param.time + 1
    else
      param.time = 0 
    end
    
    -- arrow ranges
    local arrowDetectL = (me.FacingDir == facingLeft) and -param.sightRangeX * 2 or 0
    local arrowDetectR = (me.FacingDir == facingRight) and param.sightRangeX * 2 or 0

    me:setEnemyLeaderTo(target)

    -- MAKE SURE WE DO THIS FIRST REALLY -- 
    if not param.setUpSentry then
      while not me:hasTag("Sentry") do
        doMotion("D+Heruka")
        wait()
      end
      trace("Setup Sentry!")
      param.setUpSentry = true
    elseif me:hasTag("Stunned") then
      --beep boop--
      debugPalette("Debug_Blink")
    elseif (dist_arrow(160 * 10) and dist_arrow(160 * 10) <= arrowDetectR)
    and (dist_arrow(160 * 10) and dist_arrow(160 * 10) >= arrowDetectL)
    and (dist_arrow(160 * 10) and math.abs(dist_arrow(160 * 10)) >= me.WorldCollisionWidth * 2)
    and (dist_arrow(160 * 10) and math.abs(select(2, dist_arrow(160 * 10))) <= param.sightRangeY.top)
    and (dist_arrow(160 * 10) and math.abs(select(2, dist_arrow(160 * 10))) >= param.sightRangeY.bottom)
    then
      debugPalette("Debug_Magenta")
      --

      if not me:hasTag("Laser_Charge") then
        trace("0")
        debugPalette("Debug_Green")
        doMotion("U+Attack")
      elseif not me:hasTag("Laser_Loop") then
        trace("1")
        debugPalette("Debug_White")
        doMotion("U+Heruka")
      elseif me:hasTag("Laser_Loop") then
        trace("2")
        debugPalette("Debug_Blue")
        doMotion("U+Heruka")
      end
    elseif param.hugeDistCheck then
      -- in this version, we're going to just shoot asap
      if isInSightRange() 
      and not param.turretPrimed then
        debugPalette("Debug_Red")
        doMotion("U+Attack")
        wait(4)
        param.turretPrimed = true
      elseif param.turretPrimed then
        debugPalette("Debug_Blue")
        wait(11)
        doMotion("U+Heruka")
        wait(33)
        param.turretPrimed = false
        param.hugeDistCheck = false
      else
        debugPalette("Debug_Green")
      end
    elseif isInAttackRange(target) 
    then
      debugPalette("Debug_Red")
      if me:hasTag("Laser_Charge")
      and not param.noFastCheck
      then
        doMotion("U+Heruka")
      elseif me:hasTag("Sentry_Idle")
      then
        doMotion("U+Attack")
      end
    elseif isInSightRange(target)
    then
      debugPalette("Debug_Orange")
      if me:hasTag("Laser_Charge") then
        doMotion("X+Heruka")
      elseif me:hasTag("Sentry_Idle")
      then
        doMotion("U+Attack")
      end
    elseif me.Flag15 then -- 14 AU btw!
      debugPalette("Debug_Yellow")
      if me:hasTag("Sentry_Idle") then
        trace("Laser Warning Start")
        doMotion("U+Attack")
      elseif me:hasTag("Laser_Charge")
      and isInSightRange()
      then
        trace("Laser Hold")
        doMotion("U+Attack")
      elseif me:hasTag("Laser_Loop")
      and param.time > 50
      then
        trace("Stop the laser")
        doMotion("X+Dash")
      else
        debugPalette("Debug_Cyan")
      end
    else
      if not me.Flag15 then
        if me:hasTag("Laser_Charge") 
        and me.Slot1 <= 0
        then
          trace("Laser Warning Stop")
          doMotion("D+Heruka")
        elseif me:hasTag("Laser_Loop") 
        then
          trace("Laser Loop Stop")
          doMotion("D+Heruka")
        end
      end
    end
    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
    local toReturn = DefaultOnBattleAttack(battleData)
    local me = getMe()

    local dirChance = random(1000)
    local actionChance = random(1000)

    if param.isNMGemstone then
      --trace("Gemstone attack!!")
      if actionChance > 850
      then
        attack_flamethrower(toReturn)
      elseif dirChance > 850
      then
        attack_spin(toReturn)
      else
        attack_nm_projectile(toReturn)
      end
    elseif param.isASummon then
      attack_flamethrower(toReturn)

      if dirChance > 500 then
        attack_laser(toReturn)
      end
    else
      if dirChance > 500 then
        attack_flamethrower(toReturn)
      else
        attack_spin(toReturn)
      end
    end
    
    if param.isFastOnFirstTurn 
    and battleData.turnNum > 1 
    then
      attack_laser(toReturn)
      me.BattleSpeedMult = 0.8
      param.isFastOnFirstTurn = false
    elseif not param.isFastOnFirstTurn 
    then
      me.BattleSpeedMult = 0.8
    end

    return toReturn
end -- OnBattleAttack()