require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function attack_headbutt(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  addInput(dirs.N, btns.Attack)
end

local function attack_strikeBall(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  addInput(dirs.N, btns.Attack)
end

local function attack_superHiFive(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  addInput(dirs.N, btns.Heruka)
end

local function ballHitOK()
  local me = getMe()

  if me.LightR > 0
  and true
  then
    return true
  else
    return false
  end

  return false
end

local function ballIsClose()
  local me = getMe()

  if me:getFlagValue(8)
  then
    return true
  else
    return false
  end

  return false
end

local hash_ballName = hash("HX_Ghost")
local function getBall()
  local me = getMe()

  local c = me:getChildren()

  for idx, spr in pairs(c)
  do
    if spr ~= nil
    then
      if spr.MyName == hash_ballName
      then
        return spr
      end
    end
  end

  return nil
end

-- the initialization function for this AI
function OnInit()
  param.time = 0
  param.attackCooldown = 0
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  while true do
    me, player = getMe(), getPlayer()

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  local dirChance = random(1000)
  local actionChance = random(1000)

  --
  if battleData.numBallHits == nil
  then
    battleData.numBallHits = 0 
  end
  if battleData.maxBallHits == nil
  then
    battleData.maxBallHits = 3
  end
  if battleData.turnsSinceSuper == nil
  then
    battleData.turnsSinceSuper = 0
  end
  if battleData.ballOK == nil
  then
    battleData.ballOK = false
  end
  if battleData.prevStateWasSplat == nil
  then
    battleData.prevStateWasSplat = false
  end
  if battleData.battleMaxTimeout == nil
  then
    battleData.battleMaxTimeout = 60 * 5
  end
  --battleData.battleMaxTimeout = 60 * 5 -- ensure that we don't hold the battle forever :)
  param.time = 0 

  if me.LifePct <= 0.1
  then
    battleData.maxBallHits = 7
  elseif me.LifePct <= 0.25
  then
    battleData.maxBallHits = 5
  elseif me.LifePct <= 0.55
  then
    battleData.maxBallHits = 4
  else
    battleData.maxBallHits = 3
  end

  if actionChance > 375
  and battleData.turnsSinceSuper >= 3
  then
    attack_superHiFive(toReturn)
    battleData.ballOK = false
    battleData.numBallHits = 0
    battleData.turnsSinceSuper = 0
  else
    attack_headbutt(toReturn)
    battleData.ballOK = true
    battleData.numBallHits = 0 
    battleData.turnsSinceSuper = battleData.turnsSinceSuper + 1
  end

  return toReturn
end --OnBattleAttack()

function OnBattleTurnUpdate(battleData)
  local me = getMe()
  local ball = getBall()
  
  if battleData.ballOK
  and param.time <= battleData.battleMaxTimeout
  then
    if ball
    and ball:hasTag("Ball")
    then
      local dist = vec_dist( vec_pos(me), vec_pos(ball) )
      local vx, vy = ball.XVel, ball.YVel
      local dx, dy = dist_sprite(ball)

      local hitThisFrame = false
      local bonusY = math.min(0, vy)
      local checkInFront = (160 * 4) + (vx/3) + bonusY/3 -- This is the range to check if we should JUMP!
      local checkInFront_close = 160 * 3
      local ball_in_front = dx > 0 -- If it hasn't passed us!
      local ballTooFar = dx > 160 * 11   

      if ball:hasTag("BallSplat_Small")
      or ball:hasTag("BallSplat_Big")
      then
        if not battleData.prevStateWasSplat
        then
          hitThisFrame = true
          battleData.numBallHits = battleData.numBallHits + 1
          trace("We got a HIT! Hits now are %i"  % { battleData.numBallHits})
        end
        -- Set it now!
        battleData.prevStateWasSplat = true
      else
        battleData.prevStateWasSplat = false
      end

      -- Always...
      param.attackCooldown = math.max(0, param.attackCooldown - 1)

      -- skips --
      if me:hasTag("Super")
      or ballTooFar then
        param.time = battleData.battleMaxTimeout
      elseif battleData.numBallHits < battleData.maxBallHits
      then
        -- If we're close enough to the ball, On the ground, and not attempting, and rebounding, let's give it a go!
        if ball_in_front
        and dx <= checkInFront
        and ball:hasTag("Rebound") -- No jumping early!
        and me.OnGround > 0 
        and me.Hitstun <= 0
        and me.Hitpause <= 0
        and not me:hasTag("StrikeJump")
        and param.attackCooldown <= 0
        then
          debugPalette("Debug_Red")
          attack_headbutt()
          param.attackCooldown = 3
        -- If we're close enough to the ball, in the air, and strike jumping, and rebounding, let's give it a go!
        elseif ball_in_front
        and dx <= checkInFront_close
        and ball:hasTag("Rebound") -- No jumping early!
        and me.OnGround <= 0 
        and me.Hitstun <= 0
        and me.Hitpause <= 0
        and me:hasTag("StrikeJump")
        and param.attackCooldown <= 0
        then
          attack_strikeBall()
          debugPalette("Debug_Blue")
          param.attackCooldown = 3
        -- Keep us alive if the ball is in play in front of us
        elseif ball_in_front
        then
          debugPalette("Debug_Orange")
          addInput(dirs.N)
        end
      end
    else
      -- We don't have a ball, don't DO ANYTHING
    end
  end

  -- update only if not hitstop
  if update_ok()
  then
    param.time = param.time + 1
  end
end