require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("math_ext")
require("movement")
require("util")

local function attack_eyeballToss(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.meterBurn = false
  end
  doMotion("U+Heruka")
end

local function attack_swipe(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.meterBurn = false
  end
  param.attackCooldown = 25
  doMotion("X+Heruka")
end

local function attack_grab(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.meterBurn = false
  end
  param.attackCooldown = 40
  doMotion("U+Attack")
end

local function attack_tangle(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_buryHands(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function dist_origin()
  local me = getMe()
  return param.origin.x - me.XPos, param.origin.y - me.Ypos 
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)

  local w = param.attackRangeX
  local h = param.attackRangeY
  if param.bossBurstedMode then
    w = param.attackRangeRageX
    h = param.attackRangeRageY
  end

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

function OnInit()
  local me, player = getMe(), getPlayer()
  local players = getPlayers()
  
  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end

  param.time = 0
  param.turnWaitTime = 0
  param.roamWaitTime = 0 
  param.blockTime = 0
  param.attackCooldown = 0
  param.attackStartup = 0
  param.attackStartupThreshold = 7

  param.roamRangeX = 160 * 8
  param.roamRangeFudgeDist = 75
  param.sightRangeX = 160 * 12
  param.sightRangeY = { top = 160 * 1.5, bottom = 160 * -1 }
  param.attackRangeX = 160 * 7
  param.attackRangeY = { top = 160 * 2.5, bottom = 160 * -1.5 }
  param.moveDeadzone = 160 * 2
  param.originDeadzone = 160 * 0.5
  --
  param.sightRangeRageX = 160 * 20
  param.sightRangeRageY = { top = 160 * 6, bottom = 160 * -1 }
  param.attackRangeRageX = 160 * 6 -- this doesn't need to change that much
  param.attackRangeRageY = { top = 160 * 4, bottom = 160 * -1.5 }
  
  param.bossBurstedMode = false
end

function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    if tick_ok()
    then
      param.attackCooldown = math.max(param.attackCooldown - 1, 0)
      param.roamWaitTime = math.max(param.roamWaitTime - 1, 0)
    end
    
    -- co-op 
    if bossBursted() 
    then
      param.turnWaitTime = 0
      param.perchWaitTime = 0 
      param.moveDecay = 0 
      param.attackCooldown = 32 -- no instant grab
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 8 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    
    if update_ok()
    then
      if isInSightRange(target)
      then
        debugPalette("Debug_Orange")
        param.roamWaitTime = 100
        if ( param.bossBurstedMode or sprite_vulnerable(target) )
        then
          if param.attackStartup > param.attackStartupThreshold
          and param.attackCooldown <= 0
          and math.abs(me.XVel) <= 3 -- Cancelling out of walk looks TERRIBLE
          then
            debugPalette("Debug_Red")
            if dist_sprite(target) > 0
            then
              local r = random(1000)
              if r > 500
              then
                attack_grab()
              else
                attack_swipe()
              end
              param.attackStartup = 0 
            else
              turn_around(15)
            end
          elseif isInAttackRange(target)
          then
            param.attackStartup = param.attackStartup + 1
          elseif math.abs(dist_sprite(target)) > param.moveDeadzone
          then
            approach_withTurn(target.XPos, 6)
          end
        end
      else
        debugPalette("Debug_White")
        if math.abs(dist_origin()) > param.originDeadzone
        and param.attackCooldown <= 0
        then
          approach_withTurn(param.origin.x)
        end
      end
    end

    wait()
  end -- The AI Function ending
end -- OnUpdate()

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  -- Moves:
  -- Bury        X+Attack
  -- Grab        U+Attack
  -- Wall        D+Attack - is okay now!
  -- Swipe       X+Heruka
  -- Toss        U+Heruka -- NO

  local dirChance = random(1000)

  if dirChance < 250 then
    attack_buryHands(toReturn)
  elseif dirChance > 750 then
    attack_eyeballToss(toReturn)
  else
    attack_swipe(toReturn)
  end

  return toReturn
end -- OnBattleAttack()