require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local me, player = nil
local param = {
  origin = { x = 0, y = 0 },
  time = 0
}

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function attack_in_battle(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_world_random(data)
  local dirChance = random(1000)
  local actionChance = random(1000)
  
  if actionChance < 700 then
    if dirChance < 333 then
      doMotion("D+Atk")
    elseif dirChance > 667 then
      doMotion("U+Atk")
    else
      doMotion("N+Atk")
    end
  else
    if dirChance < 333 then
      doMotion("D+Atk+Hrk")
    elseif dirChance > 667 then
      doMotion("U+Atk+Hrk")
    else
      doMotion("N+Atk+Hrk")
    end
  end
  
  param.attackCooldown = 300
end

local function attack_headVortex(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("N+Dash")
  param.attackCooldown = 30
end

local function attack_worldDestroy(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
  param.attackCooldown = 200
end

local function action_organDeath(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local AIFlyingHash = hash("AIFlying")

-- the initialization function for this AI
function OnInit()
  -- Accessible anywhere in this file after Initialization
  me, player = getMe(), getPlayer()
  local players = getPlayers()

  -- Set origin
  param.time = 0
  param.movementTime = 0 
  param.attackCooldown = 240  -- to start with
  param.turnCooldown = 0
  param.playerTurnDist = 280
  param.useClassicMovement = false
  --

  param.head_worldDestroyTier = 0 
  param.head_worldDestroyTierHPVals = {
    [0] = 0.75, -- 75%
    [1] = 0.50, -- 50%
    [2] = 0.25, -- 25%
    [3] = 0.01, -- 1%
  }

  param.organ_wantToExplode = false
  param.organ_explodedAlready = false 

  param.useWilyMovement = true

  param.entityKind = ""
  if has_spawn_command("Organ")
  then
    param.entityKind = "Organ"
  elseif has_spawn_command("Boss")
  then
    param.entityKind = "Boss"
  end
  --
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end


  param.wilyOrbitBase = { x = 0, y = 0 }
  -- use the defaults for now --
  param.wilyOrbitBase.x = param.origin.x
  param.wilyOrbitBase.y = param.origin.y

  -- if we find where we want her to orbit to, let's override --

  local _x, _y, _z = getObjectPos("Spawn_Kala_OrbitBase")
  if _x ~= nil and _y ~= nil and _z ~= nil
  then
    param.wilyOrbitBase.x = _x
    param.wilyOrbitBase.y = _y
  end

  --[[ For Wily Pints
  param.wilyPoints = {}
  param.wilyPointIndex = 0

  local function MakeWilyPoints( angle )
    local _x = (math.cos(angle/80) ) * (160 * 6)
    local _y = (math.sin(angle/40) ) * (160 * 4)
    --
    local _v = { x = _x, y = _y } 
    return _v
  end
  param.wilyPoints[#param.wilyPoints+1] = MakeWilyPoints(0)
  for i = 1, 80 do
    local angle = (i / 80) * 503 -- 360 (why is it 503 oh my god)
    param.wilyPoints[#param.wilyPoints+1] = MakeWilyPoints(angle)
  end
  ]]
end

local function attack_ok()
  local me = getMe()
  
  if false --param.attackCooldown <=0 and canAttack(me)
  then
    return true
  else
    return false
  end
  
  ---------------

  return false
end

-- the main update function for this AI
function OnUpdate()
  while true do
    me, player = getMe(), getPlayer()
    --
    --[[ MOVEMENT OFFSETS ]]--
    local move_x_offset = (math.cos(param.movementTime/80) ) * (160 * 6)
    local move_y_offset = (math.sin(param.movementTime/40) ) * (160 * 4)

    if aiDebug() then
      trace("Movement Time %i" % { param.movementTime })

      draw_bar( param.attackCooldown, 0, 60 )
      draw_bar( param.turnCooldown, 0, 60, 1 )
      
      draw_point( param.wilyOrbitBase.x + move_x_offset, param.wilyOrbitBase.y + move_y_offset )
      --[[ Wily Points
      for i, point in ipairs(param.wilyPoints) do
        draw_point( param.origin.x + point.x, param.origin.y + point.y ) 
      end
      ]]
    end

    if tick_ok() then
      param.attackCooldown = math.max(0, param.attackCooldown - 1)
      param.movementTime = param.movementTime + 1
      param.turnCooldown = math.max(0, param.turnCooldown - 1)
    end

    -- Actually doing stuff
    if update_ok() then
      -- [[ KALA HERSELF ]] --
      if param.entityKind == "Boss"
      then
        debugPalette("Debug_DarkMagenta")

        local pDistX = dist_player()
        local pDistY = select(2, dist_player())
        local pDistVec = vec_dist(
          { x = player.xpos, y = player.ypos },
          { x = me.xpos, y = me.ypos }
        )
        local yOffset = 800
        
        if param.useWilyMovement
        and tick_ok()
        then
          --[[ For Points
          local pointWeWant = param.wilyPoints[param.wilyPointIndex + 1]

          local dist_offset_x =  me.XPos - ( param.origin.x + pointWeWant.x )
          local dist_offset_y =  me.YPos - ( param.origin.y + pointWeWant.y )
          local distFromPoint = vec_dist(
            { x = me.XPos, y = me.YPos}, { x = param.origin.x + pointWeWant.x, y = param.origin.y + pointWeWant.y }
          )
          ]]
          local dist_offset_x =  me.XPos - ( param.wilyOrbitBase.x + move_x_offset )
          local dist_offset_y =  me.YPos - ( param.wilyOrbitBase.y + move_y_offset )
          local xvel_max = 20
          local yvel_max = 20

          --trace( dist_offset_x )
          --trace( dist_offset_y )

          me.XVel = (-dist_offset_x / 2) * me.FacingDir
          me.XVel = clamp(me.XVel, -xvel_max, xvel_max)
          me.YVel = -dist_offset_y / 8
          me.YVel = clamp(me.YVel, -yvel_max, yvel_max)

          me.LightZ = AIFlyingHash
          me.Friction = 0
          me.Weight = 0
          me.PhysicsType = 0
          
          if pDistX < -param.playerTurnDist 
          and param.turnCooldown <= 0 
          then
            trace("Turning around")
            addInput(dirs.B)
            param.turnCooldown = 55
          end

          -- THE REAL WAY TO DO VORTEX, when you've taken enough damage! --
          
          if player_vulnerable() 
          and param.attackCooldown <= 0
          and (
               (me.LightG < 1 and me.ExtraDamage >= 11500)
               or (me.LightG == 1 and me.ExtraDamage >= 18500)
               or (me.LightG == 2 and me.ExtraDamage >= 25000)
               or (me.LightG > 2 and me.ExtraDamage >= 32000)
               )
          then
            attack_headVortex()
          end
          
          --  DEBUG VORTEX - Uncomment This, Mike! --
          if player.Debug1 > 0
          and player:hasTag("ChakraGateOK")
          and player_vulnerable() 
          -- CAN ALWAYS DO IT -- and param.attackCooldown <=0
          then
            attack_headVortex()
          end
          -- To Here 
          
          -- Attack? --
          if player_vulnerable() and param.attackCooldown <= 0 and me:hasTag("Idle")
          then
            attack_world_random()
          end

          --[[ For Points
          if distFromPoint <= 30
          then
            debugPalette("Debug_Blue")
            param.wilyPointIndex = param.wilyPointIndex + 1
            if param.wilyPointIndex >= #param.wilyPoints
            then
              param.wilyPointIndex = 0
            end
          end
          ]]

        elseif math.abs( pDistX ) < 160 * 7
        and param.useClassicMovement
        then
          debugPalette("Debug_Magenta")
          if pDistX < -3 
          then
            local prevFacingDir = me.FacingDir
            for i = 1, 4 do
              if me.FacingDir ~= prevFacingDir
              then
                break
              end
              trace("Turning from pDist block")
              addInput(dirs.B)
              wait(1)
            end
          else
            if player_vulnerable()
            and attack_ok()
            and player.DEBUG1<=0
            then
              if me.HealthPct <= param.head_worldDestroyTierHPVals[param.head_worldDestroyTier]
              then
                trace("Attack World Destroy")
                trace(param.head_worldDestroyTier)
                attack_worldDestroy()
                if aiDebug()
                then
                  if param.head_worldDestroyTier >= 3
                  then
                    debugPalette("Debug_Yellow")
                  elseif param.head_worldDestroyTier >= 2
                  then
                    debugPalette("Debug_Green")
                  elseif param.head_worldDestroyTier >= 1
                  then
                    debugPalette("Debug_Blue")
                  else
                    debugPalette("Debug_Red")
                  end
                end
                param.head_worldDestroyTier = param.head_worldDestroyTier + 1
              else
                attack_headVortex()
              end
            end
          end
        elseif math.abs( pDistX ) >= 160 * 4
        and param.useClassicMovement
        then
          debugPalette("Debug_LightMagenta")
          if math.abs( pDistY ) >= 160 * 3
          then
            fly_to( player.XPos, player.YPos + yOffset )
          else
            fly_to( player.XPos, me.YPos )
          end
        end
      -- [[ FOLLOWTHROUGH ]] --
      else
        debugPalette("Debug_Blink")
      end
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  
  attack_in_battle(toReturn)
  
  --if me.HealthPct <= param.head_worldDestroyTierHPVals[param.head_worldDestroyTier]
  --then
  --  attack_worldDestroy()
  --  param.head_worldDestroyTier = param.head_worldDestroyTier + 1
  --else
  --  attack_headVortex(toReturn)
  --end

  return toReturn
end --OnBattleAttack()