require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  
  local dx, dy = dist_sprite(target)

  if dx >= -param.sightRangeX/2
  and dx <= param.sightRangeX/2
  and dy <= param.sightRangeY.top
  and dy >= param.sightRangeY.bottom
  then
    return true
  end
  return false
end

local function attack_arrowStun( data )
  if data ~= nil then
      data.dir = dirs.U
      data.pressAction = true
      data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("U+Heruka")
end

local function attack_callForHelp( data )
  if data ~= nil then
      data.dir = dirs.D
      data.pressAction = true
      data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("D+Heruka")
end

local function attack_screw( data )
  if data ~= nil then
      data.dir = dirs.D
      data.pressAction = false
      data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("D+Attack")
end

local function attack_spin( data )
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = false
      data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("X+Attack")
end

local function action_scan( data )
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = true
      data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function action_arrowStun( data )
  if data ~= nil then
      data.dir = dirs.U
      data.pressAction = true
      data.meterBurn = false
  end
  doMotion("U+Heruka")
end

function OnInit()
  local me, player = getMe(), getPlayer()

  param.attackCooldown = 0 
  param.customHitstun = 0 
  param.caughtPlayer = false
  param.hasSnitched = false
  param.hasSpawnCommand = false
  if has_spawn_command("UShallNotPass")
  then
    param.hasSpawnCommand = true
  end

  param.isNotoriousMonster = false
  if has_spawn_command("Notorious")
  or me.WeaponType == hash("Notorious")
  then
    param.isNotoriousMonster = true
  end


  param.isHazard = false
  if has_spawn_command("SpinHazard")
  or has_spawn_command("HazardSpin")
  or has_spawn_command("DrillHazard")
  or has_spawn_command("Collision")
  then
    param.isHazard = true
  end

  param.sightRangeX = 160 * 3.5
  param.sightRangeY = { top = 160 * 2, bottom = 160 * -5 }
end

function OnUpdate()
    local me, player = getMe(), getPlayer()

    local players = getPlayers()

    -- SCARY HACK WE SHOULD PROBABLY NOT DO --
    while shouldNeverUpdate() do
    
    end

    -- AI loop
    while true do
      me, player = getMe(), getPlayer()
      --
      if param.isHazard
      then
        wait(600)
      else
        if tick_ok() then
          param.attackCooldown = math.max(0, param.attackCooldown - 1) 
          param.customHitstun = math.max(0, param.customHitstun - 1) 
        end

        if param.customHitstun > 0
        then
          trace("yep")
          debugPalette("Debug_Blink")
          if me:hasTag("Scan")
          then
            action_arrowStun()
            wait()
          elseif not me:hasTag("Arrow_Stun")
          then
            action_arrowStun()
          else
            --action_arrowStun()
          end
        elseif param.isNotoriousMonster
        then
          wait() -- do nothing
        else
          if not me:hasTag("Scan")
          then
            action_scan()
          elseif me:hasTag("Scan")
          then
            -- whoa --
            action_scan()
          end
        end

        if got_hit(me)
        or me.Alerted > 0 
        then
          trace("got hit")
          param.customHitstun = 60 * 2
        end

        if isInSightRange() 
        and param.attackCooldown <= 0 
        and player_vulnerable()
        and not is_hurt(me)
        and canAttack()
        and me.WasInBattle <= 0 
        and not me:hasTag("Arrow_Stun")
        then
          param.caughtPlayer = true
          me:startBattle(player)
        end
      end
      wait()
    end -- The AI Function ending
end -- OnUpdate()

function OnBattleAttack(battleData)
  local me = getMe()
  local toReturn = DefaultOnBattleAttack(battleData)

  local dirChance = random(1000)
  local specialChance = random(1000)
  
  -- Default

  if specialChance > 500 then
    trace("Screw")
    attack_screw(toReturn)
  else
    trace("Spin")
    attack_spin(toReturn)
  end

  -- Snitch on 'em
  if param.caughtPlayer 
  and not param.hasSnitched 
  then
    trace("Snitch")
    param.hasSnitched = true
    attack_callForHelp(toReturn)
  end

  return toReturn
end -- OnBattleAttack()