-- Here are some global variables we'll use for our AI!
-- for the AI we just spawned:
newRoutine = nil
newOnInit = nil
newOnUpdate = nil
-- for whichever AI we're currently updating:
curOnInit = nil
curOnUpdate = nil
curRoutine = nil
binding = nil
param = nil

-- We'll log any new global variables created, and optionally fail if we accidentally declare a global.
do
  local strict = true
  local whitelistedNames = { OnInit=true, OnUpdate=true, OnBattleAttack=true, OnBattleTurnUpdate=true, newRoutine=true, newOnInit=true, newOnUpdate=true, curOnInit=true, curOnUpdate=true, curRoutine=true, binding=true, param=true }

  setmetatable(_G, {
    __newindex = function(t, k, v)
      if whitelistedNames[k] then
        rawset(t, k, v);
      elseif strict == false then
        print(string.format("WARNING - Set a global variable '%s'", tostring(k)))
        rawset(t, k, v);
      else
        error(string.format("Tried to set a global variable '%s'", tostring(k)))
      end
    end
  })
  rawset(_G, "declare", function(k, v)
    rawset(_G, k, v)
  end)
  declare("setStrict", function(setting)
    if setting == false then
      strict = false
    elseif setting == true then
      strict = true
    end
  end)
end

-- This block adds the % operator to strings, allowing easy access to string.format
-- Example "%d: %s" % { 1, "foo" } returns "1: foo"
do
  local stringmeta = getmetatable("")
  stringmeta.__mod = function(a, b)
    if not b then
      return a
    elseif type(b) == "table" then
      return string.format(a, unpack(b))
    else
      return string.format(a, b)
    end
  end
  stringmeta.__index.count = function(str, substr)
    local found = 0
    local start = 1
    while str:find(substr, start) do
      found = found + 1
      start = str:find(substr, start) + 1
    end
    return found
  end
end

-- Helpers to put callbacks on F11 debug toggles
---[[--
do
  local debugOn = 0
  local debugHooks = {}
  
  declare("onDebug", function(on, off, debugtext)
    local handle = {}
    handle.cancel = function()
      debugHooks[handle] = nil
    end
    debugHooks[handle] = { on = on, off = off, debugtext = debugtext }
  end)
  setUpdate(function()
    local p1 = getMe()
    local debugNow = p1.Debug1
    if debugOn == 0 and debugNow == 1 then
      --print("Debug on")
      for k, v in pairs(debugHooks) do
        if v.on and type(v.on) == "function" then
          local status, msg = pcall(v.on)
          if status == false then
            if v.debugtext then
              print("ERROR during debug on in " .. v.debugtext)
            else
              print("ERROR during debug on")
            end
            print(msg)
            print(debug.traceback())
          end
        end
      end
    end
    if debugOn == 1 and debugNow == 0 then
      --print("Debug off")
      for k, v in pairs(debugHooks) do
        if v.off and type(v.off) == "function" then
          local status, msg = pcall(v.off)
          if status == false then
            if v.debugtext then
              print("ERROR during debug off in " .. v.debugtext)
            else
              print("ERROR during debug off")
            end
            print(msg)
          end
        end
      end
    end
    debugOn = debugNow
  end)
end
--]]--

----DEBUGGING HELPER FUNCTIONS----
declare("TableCount", function(T)
  local count = 0
  for _ in pairs(T) do
    count = count + 1
  end
  return count
end)

declare("PrintTrueFalse", function(key, bool)
  print("%s: %s" % { key, tostring(bool) })
end)

declare("PrintSpriteTag", function(sprite, tag)
  sprite = sprite or getMe()
  PrintTrueFalse(tag, sprite:hasTag(tag))
end)

declare("printSlots", function(sprite, ...)
  sprite = sprite or getMe()
  for _, slot in ipairs({...}) do
    print("%s: %s" % { slot, sprite[slot]})
  end
end)

declare("printTags", function(sprite, ...)
  sprite = sprite or getMe()
  for _, tag in ipairs({...}) do
    print("%s: %s" % { tag, sprite:hasTag(tag)})
  end
end)

------HELPER FUNCTIONS------
-- Here's a bunch of useful coroutine ai functons
declare("holdDirFor", function(dir, len)
  for i = 1, len do
    addInput(dir)
    coroutine.yield()
  end
end)

declare("wait", function(len)
  len = len or 1
  if len == 0 then return false end
  for i = 1, len do 
    coroutine.yield()
  end
end)


declare("waitForGround", function(sprite)
  sprite = sprite or getMe()
  while sprite.grd == 0 do
    coroutine.yield()
  end
end)

declare("waitForHitstop", function()
  --while getMe().hitpause > 0 or getPlayer().hitpause > 0 do
  while getMe().hitpause > 0 or getPlayer().hitpause > 0 do
    coroutine.yield()
  end
end)

declare("waitForHitstun", function(sprite)
  sprite = sprite or getMe()
  while sprite.hitstun > 0 do
    print "Waiting for Hitstun"
    coroutine.yield()
  end
end)

declare("waitUntil", function(fn, timeout)
  local curframe = 0
  while not fn() do
    coroutine.yield()
    curframe = curframe + 1
    if(timeout ~= nil and curframe > timeout) then return false end
  end
  return true
end)

declare("waitForSequenceComplete", function()
  while not sequenceComplete() do
    coroutine.yield()
  end
end)

declare("waitForIdle", function(p)
  p = p or getMe()
  while not p:hasTag("Idle") do
    coroutine.yield()
  end
end)

declare("facePlayer", function(me, player)
  me = me or getMe()
  player = player or getPlayer()
  
  if me.XPos > player.XPos then
    me.facingDir = facingLeft
  else
    me.facingDir = facingRight
  end
end)

declare("adjustedDiff", function(difficulty, basediff, base, delta)
  if difficulty == nil then difficulty = getDifficulty() end
  local difference = math.abs(difficulty - basediff)
  return base - (difference * delta)
end)

declare("isDebug", function()
  local p = getPlayer()
  if p ~= nil then
    return ( p.DEBUG1 > 0)
  end
  return false
end)

declare("aiDebug", function()
  local p = getPlayer()
  local wantDebug = false
  if p ~= nil then
    return ( p.DEBUG1 > 0) and wantDebug
  end
  return false
end)

-- constants - Should probably be moved in to code, but, uh, yeah.
declare("animhint", {
  none = 0,
  trip = 1,
  sprawl = 2,
  knockdown = 3,
  high = 4, mid = 5, low = 6,
  floorbounce = 7, wallbounce = 8, floorslide = 9,
  crumple = 10, stagger = 11,
  snapout1 = 12, snapout2 = 13
})

declare("facingLeft", -1)
declare("facingRight", 1)

-- some default states

declare("childHasTag", function(sprite, ...)
  for _, child in ipairs( sprite:getChildren() ) do
    if child:hasTag(...) then return true end
  end
  return false
end)

declare("distance", function(e, pt)
  e = e or getPlayer()
  pt = pt or getMe()
  return (e.xpos - pt.xpos) * -pt.facingdir, e.ypos - pt.ypos
end)

-- Constants - This is a shitty way to do this, but, hey.
declare("dirs", {
  F = 0,
  DF = 1,
  D = 2,
  DB = 3,
  B = 4,
  UB = 5,
  U = 6,
  UF = 7,
  N = 8
})

declare("btns", {
  atk = getButtonIndex("Attack"),
  Atk = getButtonIndex("Attack"),
  attack = getButtonIndex("Attack"),
  Attack = getButtonIndex("Attack"),
  heruka = getButtonIndex("Heruka"),
  Heruka = getButtonIndex("Heruka"),
  jump = getButtonIndex("Jump"),
  Jump = getButtonIndex("Jump"),
  dash = getButtonIndex("Dash"),
  Dash = getButtonIndex("Dash"),
  shootbtn = getButtonIndex("ShootBtn"),
  ShootBtn = getButtonIndex("ShootBtn"),
  ShootButton = getButtonIndex("ShootBtn"),
  worldguard = getButtonIndex("WorldGuard"),
  WorldGuard = getButtonIndex("WorldGuard"),
  start = getButtonIndex("Start"),
  Start = getButtonIndex("Start"),
  select = getButtonIndex("Select"),
  Select = getButtonIndex("Select"),
  allguard = getButtonIndex("AllGuard"),
  AllGuard = getButtonIndex("AllGuard"),
  meterburn = getButtonIndex("MeterBurn"),
  MeterBurn = getButtonIndex("MeterBurn"),
  battleinterrupt = getButtonIndex("BattleInterrupt"),
  BattleInterrupt = getButtonIndex("BattleInterrupt"),
  forcewalk = getButtonIndex("ForceWalk"),
  ForceWalk = getButtonIndex("ForceWalk"),
})

declare("animhint", {
  none = 0,
  trip = 1,
  sprawl = 2,
  knockdown = 3,
  high = 4,
  mid = 5,
  low = 6,
  floorbounce = 7,
  wallbounce = 8,
  sliding = 9,
  crumple = 10,
  stagger = 11,
  snapouta1 = 12,
  snapouta2 = 13,
  techsprawl = 14,
  techknockdown = 15,
  
  [0] = "none",
  [1] = "trip",
  [2] = "sprawl",
  [3] = "knockdown",
  [4] = "high",
  [5] = "mid",
  [6] = "low",
  [7] = "floorbounce",
  [8] = "wallbounce",
  [9] = "sliding",
  [10] = "crumple",
  [11] = "stagger",
  [12] = "snapouta1",
  [13] = "snapouta2",
  [14] = "techsprawl",
  [15] = "techknockdown",
})