-- We need Lua and Monster lib locally
require("LuaLib")

require("vec2")
require("math_ext")

-- Rule 1: Find the Mass centers --
declare("boids_mass", function(sprites)
  local center = {
    x = 0, y = 0, z = 0
  }

  -- return nothing -- 
  if sprites == nil then 
    return center
  end

  local count = 0 
  for i, spr in ipairs(sprites) do
    center.x = center.x + spr.XPos
    center.y = center.y + spr.YPos
    center.z = center.z + spr.ZPos
    count = count + 1
  end

  if count > 0 then
    center.x = center.x / count
    center.y = center.y / count
    center.z = center.z / count
  end

  return center
end)


-- Rule 2: Keep everyone away from eachother --
declare("boids_bias", function(base, sprites, checkDist)
  local bias = { 
    x = 0, y = 0, z = 0 
  }

  -- return nothing --
  if base == nil
  or sprites == nil 
  then 
    return bias
  end

  local count = 0
  local vecMe = {
    x = base.XPos,
    y = base.YPos,
    z = base.ZPos
  }

  for i, spr in ipairs(sprites) do
    local vecSpr = {
      x = spr.XPos,
      y = spr.YPos,
      z = spr.ZPos
    }
    
    local dist = vec_dist(vecSpr, vecMe)

    local _checkDist = checkDist or 300
    if (dist < checkDist) then
      bias.x = bias.x - (vecSpr.x - vecMe.x)
      bias.y = bias.y - (vecSpr.y - vecMe.y)
      bias.z = bias.z - (vecSpr.z - vecMe.z)
      count = count + 1
    end
  end

  return bias
end)


-- Rule 3: Try to match velocity with near boids
declare("boids_vel", function(sprites)
  local vel = { x = 0, y = 0, z = 0 }
  if sprites == nil then 
    return vel 
  end

  local count = 0
  for i, spr in ipairs(sprites) do
    vel.x = vel.x + (spr.XVel * spr.FacingDir)
    vel.y = vel.y + spr.YVel
    vel.z = vel.z + spr.ZVel
    count = count + 1
  end

  if count > 0 then
    vel.x = vel.x / count
    vel.y = vel.y / count
    vel.z = vel.z / count
  end

  return vel
end)

-- Rule 4: Add push if we're past the boundaries
declare("boids_bounds", function(sprite, l, r, t, b)
  local bonus = { x = 0, y = 0, z = 0 }
  if sprite == nil then return bonus end
  local spr = sprite or getMe()

  if spr.YPos > t then
      bonus.y = -1
  elseif spr.YPos < b then
      bonus.y = 1
  elseif spr.XPos < l then
      bonus.x = 1
  elseif spr.XPos > r then
      bonus.x = -1
  end

  return bonus
end)

declare("boids_tendency", function(vec, _spr)
  local tendency = { x = 0, y = 0, z = 0 }

  if vec == nil then 
    return tendency 
  end

  local spr = _spr or getMe()
  local tendency = vec

  tendency.x = (tendency.x - spr.XPos)
  tendency.y = (tendency.y - spr.YPos)
  tendency.z = (tendency.z - spr.ZPos)

  return tendency
end)