-- We need Lua and Monster lib locally
require("LuaLib")

require("math_ext")

declare("trace", function(s)
  if aiDebug() then
    print(s)
  end
end)

declare("draw_physbox", function()
  local me = getMe()
  if aiDebug() then
    drawBox(
      me.XPos - me.WorldCollisionWidth / 2 - 1,
      me.XPos + me.WorldCollisionWidth / 2 + 1,
      me.YPos + me.WorldCollisionHeight + 1,
      me.YPos,
      1, -1,
      255, 255, 255
    )
  end
end)

declare("draw_point", function(px, py, color)
  if aiDebug() then
    local size = 80
    local r, g, b = 255, 0, 255

    if color
    then
      r = color[1]
      g = color[2]
      b = color[3]
      if color.r and color.g and color.b
      then
        r = color.r
        g = color.g
        b = color.b
      end
    end
    drawBox(
      px - size / 2,
      px + size / 2,
      py + size / 2,
      py - size / 2,
      1, -1,
      r, g, b 
    )
  end
end)

declare("draw_bar", function(v, low, hi, offset, color)
  local me = getMe()
  local value = clamp(scale(v, low, hi, 0, 1), 0, 1)
  local bar_size = 320
  local _offset = offset or 0
  local _color = color or {255, 255, 0}

  if aiDebug() then
    -- Black Fill
    drawBox(
      me.XPos - bar_size/2,
      me.XPos + bar_size/2,
      me.YPos + me.WorldCollisionHeight + 64 + 16 + (_offset * 48),
      me.YPos + me.WorldCollisionHeight + 64 - 16 + (_offset * 48),
      1, -1,
      0,0,0
    )
    -- Value
    drawBox(
      me.XPos - bar_size/2,
      me.XPos - bar_size/2 + (value * bar_size) + 1,
      me.YPos + me.WorldCollisionHeight + 64 + 16 + (_offset * 48),
      me.YPos + me.WorldCollisionHeight + 64 - 16 + (_offset * 48),
      1, -1,
      unpack(_color)
    )
  end
end)

declare("debugPalette", function( name )
  local me = getMe()
    local show_palFX = false
    if aiDebug() 
    and show_palFX then
    if name == nil then
      me.PalFXSelf = 0
      me.PalFXSelfTime = 0 
    end
    me.PalFXSelf = hash(tostring(name))
    me.PalFXSelfTime = 2
  end
end)

-- Alias
declare("debugPal", function(name)
  return debugPalette(name)
end)

declare("_color", {
  alerted = {224, 60, 40},
  stun = {255, 231, 55}
})