-- We need Lua and Monster lib locally
require("LuaLib")

declare("dist_player", function(spr)
    local me = getMe()
    local player = getPlayer()

    if spr ~= nil 
    and spr.XPos ~= nil
    and spr.YPos ~= nil
    then 
      player = spr
    end
    local x_dist = player.XPos - me.XPos
    local y_dist = player.YPos - me.YPos
    x_dist = x_dist * me.FacingDir
    --
    return x_dist, y_dist
end)

-- Alias
declare("dist_sprite", function(spr)
  return dist_player(spr)
end)

declare("dist_point", function(px, py)
    local me = getMe()
    local x_dist = px - me.XPos
    local y_dist = py - me.YPos
    x_dist = x_dist * me.FacingDir
    --
    return x_dist, y_dist
end)

declare("faceSprite", function(sprite)
  local me = me or getMe()
  local player = sprite or getPlayer()
  
  if me.XPos > player.XPos then
    me.facingDir = facingLeft
  else
    me.facingDir = facingRight
  end
end)

declare("_ARROW_NOTICE_RANGE", 160 * 8)
declare("dist_arrow", function(_r)
    local me = getMe()
    local range = _r or _ARROW_NOTICE_RANGE
    local l,r,u,d = me.XPos - range, me.XPos + range, me.YPos + range, me.YPos - range
    local sprites = getSpritesInBox(
        l, r, u, d,
        0, me
    )
    local found_arrow = false
    local _x, _y = nil, nil
    
    for _, spr in ipairs(sprites) do 
        if spr.MyName == hash("AjnaBasicArrow") then
            if spr:hasTag("AjnaArrow")
            or (spr:hasTag("ArrowExplode") and spr.TimeInState <= 10) then
                _x, _y = spr.XPos, spr.Ypos
                found_arrow = true
            end
        end
    end

    if not found_arrow then return nil end
    return _x - me.XPos, _y - me.YPos
end)

declare("dist_spear", function(_r)
    local me = getMe()
    local range = _r or _ARROW_NOTICE_RANGE
    local l,r,u,d = me.XPos - range, me.XPos + range, me.YPos + range, me.YPos - range
    local sprites = getSpritesInBox(
        l, r, u, d,
        0, me
    )
    local found_spear = false
    local _x, _y = nil, nil
    
    for _, spr in ipairs(sprites) do 
        if spr.MyName == hash("AjnaSpear") 
        or spr.MyName == hash("AjnaSpearHARD") then
            _x, _y = spr.XPos, spr.Ypos
            found_spear = true
        end
    end

    if not found_spear then return nil end
    return _x - me.XPos, _y - me.YPos
end)

declare("dist_blang", function(_r)
    local me = getMe()
    local range = _r or _ARROW_NOTICE_RANGE
    local l,r,u,d = me.XPos - range, me.XPos + range, me.YPos + range, me.YPos - range
    local sprites = getSpritesInBox(
        l, r, u, d,
        0, me
    )
    local found_fx = false
    local _x, _y = nil, nil
    
    for _, spr in ipairs(sprites) do 
        if spr.MyName == hash("BlangFX") then
            _x, _y = spr.XPos, spr.Ypos
            found_fx = true
        end
    end
    

    if not found_fx then return nil end
    return _x - me.XPos, _y - me.YPos
end)

declare("player_vulnerable", function(spr)
    local player = spr or getPlayer()
    if player.Hitstun <= 0 
    and player.Alerted <= 0 
    and player.invinFlashTime <= 0
    and player.inBattle == 0 
    and player.Lifebar > 0 -- can't hit you if you're dead.
    then
        return true
    end
    return false
end)
declare("sprite_vulnerable", function(spr)
  return player_vulnerable(spr)
end)

declare("is_dead", function(spr)
  local _spr = _spr or getMe()
  --
  return _spr.Lifebar <= 0 
end)

declare("is_alerted", function(spr)
  local _spr = _spr or getMe()

    -- 30 is the value provided from code
    return _spr.Alerted >= 30 
end)

declare("shouldNeverUpdate", function()
  local me, player = getMe(), getPlayer()

  if has_spawn_command("StoryMidboss")
  then
    local _rangeX = 160 * 12
    local _rangeY = 160 * 6

    if isInRange(player, me.XPos, me.YPos,
      _rangeX, _rangeY)
    and tick_ok()
    then
      if select(1,dist_player()) < 0 then
        addInput(dirs.B)
        wait()
        addInput(dirs.B)
        wait(3)
        --turn_around(9)
      end
    end
    wait()
    return true
  elseif has_spawn_command("NEVER_UPDATE_WORLD") then
    wait(6000)
    return true 
  else
    return false
  end
end)

declare("is_hurt", function(sprite)
  local _spr = sprite
  if _spr == nil then _spr = getMe() end
  --
  if _spr.Hitstun > 0
  --or _spr.Alerted >= 30
  or _spr:hasTag("Hitstun")
  or _spr:hasTag("Dead")
  or _spr:hasTag("Thrown")
  then 
    return true
  end

  return false
end)

declare("got_hit", function(spr)
  local _spr = spr or getMe()
  --
  return is_alerted(spr) or _spr.Hitstun > 0 
end)

declare("blocked_hit", function(sprDef, sprAtk)
  local defender = sprDef or getMe()
  local attacker = sprAtk or getPlayer()
  --
  if defender:hasTag("Blockstun") and attacker.NumTouches > 0 then
    return true
  end
  return false
end)

declare("hit_player", function(spr)
  local me = getMe()
  local player = spr or getPlayer()
  --
  return me.NumTouches > 0 and player:hasTag("Hitstun")
end)

declare("obstacle_in_front", function(_w)
  local me = getMe()
  local width = math.min(70, me.WorldCollisionWidth)
  if _w ~= nil then 
    width = _w 
  end
  --
  if me.FloorInfo > 0
  or me.WallDistAhead <= width
  then
    trace("width %i dist %i" % { width, me.WorldCollisionWidth } )
    return true
  end
  return false
end)

declare("obstacle_behind", function()
  local me = getMe()
  local width = math.min(120, me.WorldCollisionWidth)
  --
  if me.FloorInfo ~= 0 or me.WallDistBehind <= width then
      return true
  end
  return false
end)

declare("floor_in_front", function()
  local me = getMe()
  --
  if me.FloorInfo > 0 then
    return false
  end
  return true
end)

declare("floor_behind", function()
  local me = getMe()
  --
  if me.FloorInfo < 0 then
    return false
  end
  return true
end)

declare("has_spawn_command", function(cmd)
  local me = getMe()
  --
  if me.SpawnCommand == hash(cmd)
  or me.SpawnCommand2 == hash(cmd) then
    return true
  end
  return false
end)

-- We 
declare("isTangled", function(_spr)
  local spr = _spr or getMe()

  local hitHash = hash("Tangled_Hit")
  local wearingOffHash = hash("Tangled_WearingOff")

  if spr.PalFXHit == hitHash
  or spr.PalFXHit == wearingOffHash
  then
    return true
  end
  return false
end)


declare("update_ok", function()
  local me = getMe()
  if me.Hitpause <= 0
  and not isTangled(me)
  then
      return true
  end
  return false
end)

declare("tick_ok", function()
  local me = getMe()
  if not me:hasTag("Attack") and not me:hasTag("Blockstun")
  and not me:hasTag("Hitstun")
  and me.Hitpause <= 0
  and not isTangled(me)
  then
      return true
  end
  return false
end)

declare("_BOSS_BURST_STUN_MAX", 60 * 4.5)
declare("_ALERTED_MAX", 100)
declare("bossBursted", function()
  local me = getMe()
  
  if me.Alerted >= 99
  then
    return true
  else
    return false
  end

  return false
end)

declare("palEffect", function( name, time )
  local me = getMe()
  local _time = time or 60
  if name == nil then
    me.PalFXSelf = 0
    me.PalFXSelfTime = 0 
  end
  me.PalFXSelf = hash(tostring(name))
  me.PalFXSelfTime = _time
end)

declare("getPlayers", function()
  local players = {}
  players[#players+1] = getPlayer()
  
  return players
end)

declare("_INCARNATION_APPEAR_PALFX", hash("Incarnation_AddedToBattle"))
declare("_INCARNATION_DISAPPEAR_PALFX", hash("Incarnation_InWorld_Disappear"))

declare("partyMemberOK", function(spr)
  if spr == nil
  or spr.XPos == nil
  or spr.YPos == nil
  then
    return false
  elseif spr.PalFXSelf2 == _INCARNATION_DISAPPEAR_PALFX
  or spr.PalFXSelf == _INCARNATION_DISAPPEAR_PALFX
  or spr.PalFXHit == _INCARNATION_DISAPPEAR_PALFX
  then
    return false
  elseif spr.PalFXSelf2 == _INCARNATION_APPEAR_PALFX
  or spr.PalFXSelf == _INCARNATION_APPEAR_PALFX
  or spr.PalFXHit == _INCARNATION_APPEAR_PALFX
  then
    return false
  else
    return true
  end

return false
end)

declare("getClosestPartyMember", function()
  local me = getMe()

  local _spr = getPlayer()

  local new = getClosestPlayer(me.XPos, me.YPos)
  if new ~= nil
  and partyMemberOK(new)
  then
    return new
  else
    return _spr
  end

  return _spr
end)

declare("getClosestPartyMemberOnY", function()
  local me = getMe()

  local _spr = getPlayer()
  
  local new = getClosestPlayer(me.XPos, me.YPos, false, true)
  if new ~= nil
  and partyMemberOK(new)
  then
    return new
  else
    return _spr
  end

  return _spr
end)

declare("getClosestPartyMemberOnX", function()
  local me = getMe()

  local _spr = getPlayer()

  local new = getClosestPlayer(me.XPos, me.YPos, true, false)
  if new ~= nil
  and partyMemberOK(new)
  then
    return new
  else
    return _spr
  end

  return _spr
end)

declare("applyGuts", function( _scale )
  local me = getMe()

  local scale = _scale or 1
  local prevHP = me.Lifebar
  local prevMax = me.MaxLife

  local newHP = math.floor( prevHP * scale )
  local newMax = math.floor( prevMax * scale )

  me.Lifebar = newHP
  me.MaxLife = newMax
end)

declare("isInRange", function( _spr, x, y, w, h )
  local spr = _spr or getPlayer()

  -- Our delta
  local dx, dy = spr.XPos - x, spr.YPos - y 


  if dx >= -w/2
  and dx <= w/2
  and dy >= -h/2
  and dy <= h/2
  then
    return true
  end

  return false
end)

declare("isOnScreen", function( sprite )
  local me = getMe()
  local target = me
  if sprite ~= nil
  then
    target = sprite
  end

  if target ~= nil
  then
    if target.TouchScreenLR <= 1
    and target.TouchScreenLR >= -1
    and target.TouchScreenUD <= 1
    and target.TouchScreenUD >= -1
    then
      return true
    else
      return false
    end
  else
    return false
  end

  return false
end)

---
declare("canAttack", function(spr)
  local me = getMe()
  if spr and spr ~= nil then 
    me = spr
  end

  local ALLOWED_TO_ATTACK = true
  
  if ALLOWED_TO_ATTACK
  and not isTangled(me)
  then
    return true
  end

  return false
end)


---