require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function attack_meditation( data )
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = true
      data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function attack_trap( data )
  if data ~= nil then
      data.dir = dirs.U
      data.pressAction = true
      data.meterBurn = false
  end
  doMotion("U+Heruka")
end

local function attack_arcLaser( data )
  if data ~= nil then
      data.dir = dirs.D
      data.pressAction = true
      data.meterBurn = false
  end
  doMotion("D+Heruka")
end

local function attack_punch( data )
  if data ~= nil then
      data.dir = dirs.X
      data.pressAction = false
      data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_roar( data )
  if data ~= nil then
      data.dir = dirs.U
      data.pressAction = false
      data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_knockout( data )
  if data ~= nil then
      data.dir = dirs.D
      data.pressAction = true
      data.meterBurn = false
  end
  doMotion("D+Heruka")
end

local function attack_popoutStart( data )
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = false
      data.meterBurn = false
  end
  doMotion("X+ShootBtn")
end

local function attack_popoutPunchStart( data )
  local me = getMe()
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = false
      data.meterBurn = false
  end
  -- This has to be done, because we were ORIGINALLY facing SOME direction,
  -- and we want LEFT to match the cutscene
  me.FacingDir = facingLeft
  doMotion("D+ShootBtn")

end

local function fireball_right()
  local me = getMe()
  me.FacingDir = facingRight
  doMotion("X+Attack")
  wait()
end

local function fireball_left()
  local me = getMe()
  me.FacingDir = facingLeft
  doMotion("X+Attack")
  wait()
  me.FacingDir = facingRight
end

local function fireball_down()
  doMotion("X+Heruka")
end

local function fireball_up()
  doMotion("X+Jump")
end

local function GetSpawnpointInfo(s)
  local _x, _y, _z = getObjectPos(s)
  -- We want a table
  return { x = _x, y = _y }
end

local function isInSightRange()
  local me, player = getMe(), getPlayer()

  if dist_player() > -param.sightRangeX/2
  and dist_player() < param.sightRangeX/2
  and select(2,dist_player()) < param.sightRangeY.top
  and select(2,dist_player()) > param.sightRangeY.bottom
  then
    return true
  end
  return false
end

local function ApplyRank(force)
  local me = getMe()
  local r = param.rank 

  if force ~= nil then 
    r = force
  else
    r = math.min(r, 2) -- Just in the case we mess up.
  end
  local dsm = 1 + (0.15 * r)
  dsm = dsm * param.baseDamageMult 
  me.DamageMult = dsm
  me.BattleSpeedMult = 1 + (0.15 * r)

  -- debug here too --
  if aiDebug() then
    if me.InBattle > 0 then
      local s = string.format("Rank %i, Dmg %1.2f Speed %1.2f",
        r, me.DamageMult, me.BattleSpeedMult
      )
      --trace(s)
    end
  end
end

-- FlyTo2 is a clone of fly_to, but adds inputs manually
local function FlyTo2(px, py, fast)
  local me = getMe()

  local vT = { x = px, y = py }
  local vP = { x = me.XPos, y = me.YPos }

  local diff = { x = (vT.x - vP.x), y = (vT.y - vP.y) }
  diff.x = diff.x * me.FacingDir

  local angle = vec_angle(diff)

  -- We have bonus parameters to use.
  local add1 = -1
  local add2 = -1
  if type(fast) == "boolean"
  and fast == true 
  then
    add1 = btns.Dash
    add2 = btns.MeterBurn
  elseif type(fast) == "number"
  then
    -- INSANE SPEED --
    if fast == 2 then
      add1 = btns.Dash
      add2 = btns.MeterBurn
    -- OK SPEED --
    elseif fast == 1 then
      add1 = btns.Dash
    elseif fast == 0 then
      add1 = btns.MeterBurn
    end
  end

  if inRange(angle, 22.5, 67.5) then addInput(dirs.UF, add1, add2) end -- UF
  if inRange(angle, -22.5, 22.5) then addInput(dirs.F, add1, add2) end -- F
  if inRange(angle, -67.5, -22.5) then addInput(dirs.DF, add1, add2) end --DF
  if inRange(angle, -112.5, -67.5) then addInput(dirs.D, add1, add2) end --D
  if inRange(angle, -157.5, -112.5) then addInput(dirs.DB, add1, add2) end --DB
  if inRange(angle, -202.5, -157.5)
  or inRange(angle, 157.5, 202.5) then addInput(dirs.B, add1, add2) end --B
  if inRange(angle, 112.5, 157.5) then addInput(dirs.UB, add1, add2) end --UB
  if inRange(angle, 67.5, 112.5) then addInput(dirs.U, add1, add2) end --U
end

-- Name, phase, pattern, type
local spawnpoints = {
  { name = "Marker_RavFireballs1_Start",
  phase = 1, pattern = 1, type = 1},
  { name = "Marker_RavFireballs1_End",
  phase = 0, pattern = 1, type = 1},
  { name = "Marker_RavFireballs2_Start",
  phase = 2, pattern = 2, type = 2},
  { name = "Marker_RavFireballs2_End",
  phase = 0, pattern = 2, type = 2},
  { name = "Marker_RavFireballs3_Start",
  phase = 3, pattern = 3, type = 1},
  { name = "Marker_RavFireballs3_End",
  phase = 0, pattern = 3, type = 3},
  { name = "Marker_RavFireballs4_Start",
  phase = 4, pattern = 4, type = 2},
  { name = "Marker_RavFireballs4_End",
  phase = 0, pattern = 4, type = 2},
  { name = "Marker_RavFireballs5_Start",
  phase = 5, pattern = 5, type = 1},
  { name = "Marker_RavFireballs5_End",  -- This is the last real point
  phase = 0, pattern = 5, type = 4},
  { name = "Marker_RavFireballs5_End",  -- In case we go past the limit.
  phase = 0, pattern = 5, type = 4}
}

-- PatternIndex
--- Index, {x,y}
local patternOffsets = {
  [1] = {
    [0] = { 160 * 0, 160 * 0.25 },
    [1] = { 160 * 0, 160 * 3.25 },
    [2] = { 160 * 0, 160 * 4 } -- LAST
    --[3] = { 160 * 0, 160 * 0.55 },
  },
  --
  [2] = {
    [0] = { 160 * 0, 160 * 0 },
    [1] = { 160 * 0, 160 * 0 }
  },
  --
  [3] = {
    [0] = { 160 * 0, 160 * 3 }, -- High
    [1] = { 160 * 0, 160 * 0 }, -- Low
    [2] = { 160 * 0, 160 * 1.5 }, -- Medium
    [3] = { 160 * 0, 160 * -1.5 }, -- Lowest
    [4] = { 160 * 0, 160 * 0 }
    --[3] = { 160 * 0, 160 * 0.55 }, -- up 3?
    --[4] = { 160 * 0, 160 * 1.5 }, -- Slide Under!
    --[5] = { 160 * 0, 160 * 0 },
  },
  -- Each Repeats 3 times!
  [4] = {
    [0] = { 160 * -0.5, 160 * 0 }, -- Left
    [1] = { 160 * 0.75, 160 * 0 }, -- Mid
    [2] = { 160 * 2, 160 * 0 }, -- Right
    [3] = { 160 * 0, 160 * 0 }, -- Mid
    [4] = { 160 * 0, 160 * 0 },
  },
  [5] = {
    [0] = { 160 * 0, 160 * 0 }, -- From Bottom To Top
    [1] = { 160 * 0, 160 * 7 }, -- From Top To Bottom
    [2] = { 160 * 0, 160 * 0 }, -- NO!
    [3] = { 160 * 0, 160 * 2 }, -- Highest
    [4] = { 160 * 0, 160 * 0 }, -- End
  },
}

local function PlayOutPattern()
  local me = getMe()

  -- 1 --
  if param.fireballPhase == 1 then
    if (param.time % 70) == 0 
    and param.time >= 0 
    then
      fireball_left()
      param.fireballShots = param.fireballShots + 1
      if param.fireballShots >= 2 then
        param.fireballShots = 0 
      end
      param.time = param.time % 70
    else
      local base = GetSpawnpointInfo("Marker_RavFireballs1_End")
      local _offsets = patternOffsets[param.fireballPattern]
      local offset = _offsets[param.fireballShots]

      local distToNextPoint = vec_dist(
        { x = base.x + offset[1], y = base.y + offset[2] },
        { x = me.XPos, y = me.YPos } 
      )
      if distToNextPoint > 200 then
        FlyTo2(base.x + offset[1], base.y + offset[2], 1)
      elseif distToNextPoint > 50 then
        FlyTo2(base.x + offset[1], base.y + offset[2])
      elseif distToNextPoint > 15 then
        FlyTo2(base.x + offset[1], base.y + offset[2], 0)
      end
      draw_point(base.x + offset[1], base.y + offset[2])
    end
  -- 2 --
  elseif param.fireballPhase == 2 
  then
    if (param.time % 60) == 0 
    and param.time > 0 
    then
      for i = 1, 3 do
        fireball_down()
        for j = 1, 13 do
          wait()
        end
      end
      param.fireballShots = param.fireballShots + 1
      if param.fireballShots >= 1 then
        param.fireballShots = 0 
      end
      param.time = -40 -- Offset wrongly so we can get an early shot
    else
      local base = GetSpawnpointInfo("Marker_RavFireballs2_End")
      local _offsets = patternOffsets[param.fireballPattern]
      local offset = _offsets[param.fireballShots]

      local distToNextPoint = vec_dist(
        { x = base.x + offset[1], y = base.y + offset[2] },
        { x = me.XPos, y = me.YPos } 
      )
      if distToNextPoint > 50 then
        FlyTo2(base.x + offset[1], base.y + offset[2])
      elseif distToNextPoint > 25 then
        FlyTo2(base.x + offset[1], base.y + offset[2], 0)
      end
      draw_point(base.x + offset[1], base.y + offset[2])
    end
  -- 3 --
  elseif param.fireballPhase == 3 then
    if (param.time % 70) == 0
    and param.time > 0 
    then
      for i = 1, 3 do
        fireball_right()
        wait()
        for j = 1, 6 do
          wait()
        end
      end
      param.fireballShots = param.fireballShots + 1
      if param.fireballShots >= 4 then
        param.fireballShots = 0 
      end
      param.time = 0
    else
      local base = GetSpawnpointInfo("Marker_RavFireballs3_End")
      local _offsets = patternOffsets[param.fireballPattern]
      local offset = _offsets[param.fireballShots]

      local distToNextPoint = vec_dist(
        { x = base.x + offset[1], y = base.y + offset[2] },
        { x = me.XPos, y = me.YPos } 
      )
      if distToNextPoint > 50 then
        FlyTo2(base.x + offset[1], base.y + offset[2])
      elseif distToNextPoint > 25 then
        FlyTo2(base.x + offset[1], base.y + offset[2], 0)
      end
      draw_point(base.x + offset[1], base.y + offset[2])
    end
  -- 4 --
  elseif param.fireballPhase == 4 then
    if (param.time % 40) == 0
    and param.time > 0 
    then
      for i = 1, 3 do 
        fireball_down()
        for j = 1, 7 do
          wait()
        end
      end
      param.fireballShots = param.fireballShots + 1
      if param.fireballShots > 3 then
        param.fireballShots = 0 
      end
      param.time = -20
    else
      local base = GetSpawnpointInfo("Marker_RavFireballs4_End")
      local _offsets = patternOffsets[param.fireballPattern]
      local offset = _offsets[param.fireballShots]

      local distToNextPoint = vec_dist(
        { x = base.x + offset[1], y = base.y + offset[2] },
        { x = me.XPos, y = me.YPos } 
      )
      if distToNextPoint > 50 then
        FlyTo2(base.x + offset[1], base.y + offset[2])
      elseif distToNextPoint > 25 then
        FlyTo2(base.x + offset[1], base.y + offset[2], 0)
      end
      draw_point(base.x + offset[1], base.y + offset[2])
    end
  -- 5 --
  elseif param.fireballPhase == 5 then
    if (param.time % 60) == 0
    and param.time > 0 
    then
      -- Bottom To Top
      if param.fireballShots == 0 then
        fireball_left()
        for j = 1, 2 do
          wait()
        end
        for i = 1, 14 do
          for j = 1, 2 do
            addInput(dirs.U)
            wait()
          end
          wait(3)
          fireball_left()
        end
      -- Top To Bottom
      elseif param.fireballShots == 1 then
        fireball_left()
        for j = 1, 2 do
          wait()
        end
        for i = 1, 9 do
          for j = 1, 2 do
            addInput(dirs.D)
            wait()
          end
          wait(3)
          fireball_left()
        end
      else
        for i = 1, 4 do 
          fireball_left()
          for j = 1, 3 do
            wait()
          end
        end
      end
      param.fireballShots = param.fireballShots + 1
      if param.fireballShots >= 2 then
        param.fireballShots = 0 
      end
      param.time = -20
    else
      local base = GetSpawnpointInfo("Marker_RavFireballs5_End")
      local _offsets = patternOffsets[param.fireballPattern]
      local offset = _offsets[param.fireballShots]

      local distToNextPoint = vec_dist(
        { x = base.x + offset[1], y = base.y + offset[2] },
        { x = me.XPos, y = me.YPos } 
      )
      if distToNextPoint > 50 then
        FlyTo2(base.x + offset[1], base.y + offset[2])
      elseif distToNextPoint > 25 then
        FlyTo2(base.x + offset[1], base.y + offset[2], 0)
      end
      draw_point(base.x + offset[1], base.y + offset[2])
    end
  else
    param.time = 0 
  end
end

local function attack_surprisePunch(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  --[[
  local point = GetSpawnpointInfo(spawnpoints[param.fireballDestPoint][1])
  local xDiff = point.x - me.XPos

  xDiff = xDiff * me.FacingDir
  ]]
  if param.fireballType == 3
  or param.fireballType == 4
  then 
    me.FacingDir = facingLeft
  else
    me.FacingDir = facingRight
  end
  doMotion("X+Attack")
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  param.time = 0
  param.waitForFirstBattle = true
  param.kickedPlayer = false
  param.kickedPlayerAgain = false
  param.isFireballMode = false
  param.safeToDoPattern = false
  param.goToPopout = false
  param.timesPoppedOut = 0
  param.fireballPhase = 0
  param.fireballPattern = 0
  param.fireballDestPoint = 1
  param.fireballShots = 0
  param.fireballType = 0
  param.fireballStartDecay = 25
  param.goToPopout = false
  param.isFinalFight = false

  param.kickedPlayerCooldown = 0 

  param.debug_startBoss = false

  param.isShadowBoss = false
  if has_spawn_command("ShadowBoss")
  then
    param.isShadowBoss = true
  end

  param.rank = 0 -- THIS IS GONNA BE NEAT
  param.baseDamageMult = me.DamageMult

  param.despawnDist = 160 * 5 * 40 -- 1AU * 5 * 40 = MANY megablocks
  param.sightRangeX = 160 * 16
  param.sightRangeY = { top = 160 * 8, bottom = 160 * -2 }

  me.DespawnDistX = param.despawnDist
  me.DespawnDistY = param.despawnDist
  ApplyRank()

  --[[
  if aiDebug() then
    attack_knockout(toReturn)
    param.waitForFirstBattle = false
    param.kickedPlayer = true
    param.isFireballMode = true
    --
    param.fireballPhase = 0
    param.fireballPattern = 0
  end
  ]]
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  local players = getPlayers()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end

  while true do
    me, player = getMe(), getPlayer()

    -- Debug
    if aiDebug() then
      drawBox(
        me.XPos - 200, me.XPos + 200,
        me.YPos + 200, me.YPos - 200,
        1, -1, 255, 255, 255
      )
      if isInSightRange() 
      and player_vulnerable()
      and not param.kickedPlayer
      then
        if not param.debug_startBoss
        then
          param.debug_startBoss = me:startBattle(player)
        end
      end
    end

    ApplyRank()


    --------------------------------------------------------------------
    -- CHECKING IF WE EJECTED PLAYER OUT OF BATTLE
    --------------------------------------------------------------------
    if me:hasTag("Knockout") then
      -- we should check :) --
      if param.kickedPlayerCooldown <= 0
      then
        trace("Knockout Check")
        if player.Hitstun > 0 -- me.NumHits > 0
        then
          if not param.kickedPlayer then 
            trace("Kicked Player 1")
            param.kickedPlayer = true 
            param.kickedPlayerCooldown = 200
          elseif not param.kickedPlayerAgain then
            trace("Kicked Player 2")
            param.kickedPlayerAgain = true
            param.kickedPlayerCooldown = 200
          end
        end
      end
    else
      param.kickedPlayerCooldown = math.max(0, param.kickedPlayerCooldown - 1)
    end

    --------------------------------------------------------------------
    -- WAITING FOR FIRST BATTLE
    --------------------------------------------------------------------
    if param.waitForFirstBattle then
      debugPalette("Debug_White")
      --
      if isInSightRange() 
      and player_vulnerable()
      then
        param.waitForFirstBattle = false
      elseif me.WasInBattle > 0 then
        param.waitForFirstBattle = false
      end
    --------------------------------------------------------------------
    -- ACTUAL BEHAVIOR FOR DOWN HERE
    --------------------------------------------------------------------
    elseif param.isFireballMode then
      debugPalette("Debug_Blue")
      -- Set up all the data

      local nextPoint = GetSpawnpointInfo(spawnpoints[param.fireballDestPoint].name)
      local playerDistToNextPoint = vec_dist(
        nextPoint,
        { x = player.XPos, y = player.YPos } 
      )
      local myDistToNextPoint = vec_dist(
        nextPoint,
        { x = me.XPos, y = me.YPos } 
      )
      local playerDistToMe = vec_dist(
        { x = player.XPos, y = player.YPos },
        { x = me.XPos, y = me.YPos }
      )

      -- Transition to the next point
      if isInRange( player, nextPoint.x, nextPoint.y, 160 * 10, 160 * 10 )
      --if playerDistToNextPoint <= 600 
      and player.WasInBattle <= 0 
      then
      -- Get the information from the table
        if spawnpoints[param.fireballDestPoint].phase > 0 then
          param.fireballPhase = spawnpoints[param.fireballDestPoint].phase
          param.fireballPattern = spawnpoints[param.fireballDestPoint].pattern
          param.fireballShots = 0
          param.fireballType = spawnpoints[param.fireballDestPoint].type
          param.safeToDoPattern = false
          param.time = param.fireballStartDecay -- Can't shoot fireballs until > 0
        else
          param.fireballPhase = 0 
          param.fireballShots = 0
          param.fireballType = 0 
          param.safeToDoPattern = false
          param.time = param.fireballStartDecay -- Can't shoot fireballs until > 0
        end

        -- Set next point
        param.fireballDestPoint = math.min(param.fireballDestPoint + 1, #spawnpoints)

        if param.fireballDestPoint >= #spawnpoints then
          param.isFinalFight = true
        else
          nextPoint = GetSpawnpointInfo(spawnpoints[param.fireballDestPoint].name)
        end
      elseif me:hasTag("PopOut_Meditation")
      or me:hasTag("PopOut_StartBattle") then
        if param.goToPopout then
          param.timesPoppedOut = param.timesPoppedOut + 1
          param.goToPopout = false
        end
        --trace("holding down Attack from chomp")
        doMotion("X+Attack")
      elseif me:hasTag("PopOut_Wait") then
        if param.fireballType == 4 
        or param.timesPoppedOut > 0 then
          --trace("holding down Attack from waiting")
          doMotion("X+Attack")
        end
      elseif param.goToPopout then
        if not me:hasTag("PopOut_Wait") then
          -- fudge our ypos a smidge
          debugPalette("Debug_Yellow")
          for i = 1 , 6 do
            FlyTo2( nextPoint.x, nextPoint.y + 40) 
            wait()
          end
          if spawnpoints[param.fireballDestPoint].type == 4 then
            attack_popoutPunchStart()
          elseif spawnpoints[param.fireballDestPoint].type == 3 then
            attack_popoutStart()
          end
          param.rank = param.rank + 1
        end
      -- Home towards the goal!
      elseif me:hasTag("Offscreen") then
        -- Do not apply rank in this position
        ApplyRank(0)

        if (param.fireballType == 1 and (me.TouchScreenLR >= 2 or me.TouchScreenLR <= -2) )
        or (param.fireballType == 2 and (me.TouchScreenUD >= 2 or me.TouchScreenUD <= -2) )
        then
          param.safeToDoPattern = true
          param.time = param.time + 1
          PlayOutPattern()
        elseif myDistToNextPoint <= 60 then
          if param.isFinalFight 
          or spawnpoints[param.fireballDestPoint].type >= 3 then
            --trace("HEY UH THIS IS PROBABLY THE TIME YOU GO TO POP!")
            param.goToPopout = true
            FlyTo2(nextPoint.x, nextPoint.y, 0)
          end
        else
          local _fast = -1
          if myDistToNextPoint >= 1500 then
            --trace("Fast = 2")
            _fast = 2
          elseif myDistToNextPoint >= 600 then
            --trace("Fast = 1")
            _fast = 1
          elseif myDistToNextPoint <= 60 then
            --trace("Fast = 0")
            _fast = 0 
          else
            --trace("Fast = -1")
            _fast = -1
          end
          FlyTo2(nextPoint.x, nextPoint.y, _fast)
        end
      end
      local s = string.format("pD %08.2f mD %08.2f dest %i phase %i pattern %i shot %i safe %5s time %03i pops %02i final %s",
        playerDistToNextPoint,
        myDistToNextPoint,
        param.fireballDestPoint,
        param.fireballPhase,
        param.fireballPattern,
        param.fireballShots,
        tostring(param.safeToDoPattern),
        param.time,
        param.timesPoppedOut,
        param.isFinalFight and "true" or "false"
      )
      --trace(s)
    else
      debugPalette("Debug_Blink")
      -- hey hey nana
    end
    --------------------------------------------------------------------
    -- SMALL ASS DESPAWN RANGE IF THE PLAYER DIED
    --------------------------------------------------------------------
    if player.Lifebar <= 0
    then
      me.DespawnDistX = 1
      me.DespawnDistY = 1
    end

    wait() -- try to keep this at the end
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  ApplyRank()

  param.waitForFirstBattle = false
  attack_punch(toReturn)

  local dirChance = random(1000)
  local actionChance = random(1000)

  if dirChance < 333 then
    attack_roar(toReturn)
  end

  -- Kickems
  if not param.isFinalFight then
    -- First time kick!
    if not param.kickedPlayer and me.HealthPct <= 0.5
    then
      attack_knockout(toReturn)
      param.waitForFirstBattle = false
      --param.kickedPlayer = true
      param.isFireballMode = true
      --
      param.fireballPhase = 0
      param.fireballPattern = 0
      --
      applyGuts(2)
    elseif me.HealthPct <= 0.25
    and not param.kickedPlayerAgain then
      attack_knockout(toReturn)
      param.isFireballMode = true
      --param.kickedPlayerAgain = true
      --
      if param.timesPoppedOut > 0 then
        applyGuts(1.2)
      end
      --param.rank = param.rank + 1
    end
  end

  if param.isShadowBoss
  then
    attack_punch(toReturn)
  end

  return toReturn
end --OnBattleAttack()