require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function dist_grazePoint()
  local me = getMe()
  local dx = param.grazePoint.x - me.XPos
  local dy = param.grazePoint.y - me.YPos
  return dx, dy 
end

local function player_inRange(target)
  local spr = target or getPlayer()
  local dx, dy = dist_sprite(spr)
  
  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  return math.abs(dx) <= w
  and dy <= h.top
  and dy >= h.bottom
end 

local function attack_nm_spit(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    --
    trace("Attack - nm_spit")
  end
  doMotion("U+Heruka")
  param.attackCooldown = 25
end

local function attack_nm_swallow(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    --
    trace("Attack - nm_swallow")
  end
  doMotion("D+Heruka")
  param.attackCooldown = 25
end

local function attack_mouthNaga(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    --
    trace("Attack - mouthNaga")
  end
  doMotion("X+Attack")
  param.attackCooldown = 25
end

local function attack_twister(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    --
    trace("Attack - twister")
  end
  doMotion("U+Attack")
  param.attackCooldown = 22
end

local function attack_naga(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    --
    trace("Attack - naga")
  end
  doMotion("D +Attack")
  param.attackCooldown = 45
end

local function attack_snoutGrab(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    --
    trace("Attack - snoutGrab")
  end
  doMotion("X+Heruka")
  param.attackCooldown = 33
end

local prog_newGamePlus = "NewGamePlus"
-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.time = 0
  param.attackCooldown = 0 
  param.blockTime = 0
  param.aggroTime = 0
  param.origin = { x = me.XPos, y = me.YPos }

  param.grazePoint = param.origin
  param.grazeOffsetRange = 160 * 6
  param.grazePoint.x = param.origin.x - (param.grazeOffsetRange) + random(param.grazeOffsetRange * 2)
  --
  if me:getSpawnpointPos() ~= nil then
      param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  param.sightRangeX = 160 * 8
  param.sightRangeY = { top = 160 * 2, bottom = 160 * -1 }
  param.attackRangeX = 160 * 5
  param.blockRangeX = 160 * 6
  param.blockRangeY = { top = 160 * 3.5, bottom = 160 * -1 }
  --
  param.sightRangeRageX = 160 * 8
  param.sightRangeRageY = { top = 160 * 2, bottom = 160 * -1 }
  param.attackRangeRageX = 160 * 5
  
  param.isNotorious = false
  if me.ColorNum == 7
  and getProgression(prog_newGamePlus)
  then
    param.isNotorious = true
  end
  --
  param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end

  while true do
    me, player = getMe(), getPlayer()

    target = getClosestPartyMember() or getPlayer()
    --
    if aiDebug() then
      draw_bar(param.aggroTime, 0, 40, 2, {255, 255, 0})
      draw_bar(param.attackCooldown, 0, 60, 1, {255, 0, 0})
      draw_bar(param.blockTime, 0, 60, 0, {0, 255, 255})
    end
    --
    if not me:hasTag("Attack") and not me:hasTag("Blockstun")
    and me.Hitpause <= 0 then
      param.time = param.time + 1
      param.blockTime = math.max(0, param.blockTime - 1)
      param.attackCooldown = math.max(0, param.attackCooldown - 1)
    end
    if param.time > 240 then
      param.grazePoint.x = param.origin.x - (param.grazeOffsetRange) + random(param.grazeOffsetRange * 2)
      param.time = 0 - random(60 * 10)
    end

    -- co-op 
    if bossBursted() 
    then
      param.aggroTime = 60 * 5
      param.attackCooldown = 32 
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 3 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    
    -- Blocking Near Front of Me
    if target:hasTag("Attack")
    and select(2, dist_sprite(target)) <= param.blockRangeY.top
    and select(2, dist_sprite(target)) >= param.blockRangeY.bottom
    and dist_sprite(target) <= param.blockRangeX 
    and dist_sprite(target) > 0
    and not me:hasTag("Blocking")
    and not target:hasTag("Heruka_1")
    and not target:hasTag("Heruka_2")
    and not target:hasTag("Heruka_3") then 
      param.blockTime = 22
    -- Blocking Arrow in Front
    elseif dist_arrow()
    and select(2, dist_arrow()) <= param.blockRangeY.top
    and select(2, dist_arrow()) >= param.blockRangeY.bottom
    and math.abs(dist_arrow()) <= param.blockRangeX then
      param.blockTime = 15
    elseif blocked_hit() then
      param.blockTime = math.max(20, param.blockTime)
    end
    --
    if param.blockTime > 0 then
      addInput(dirs.N, btns.WorldGuard)
    elseif player_inRange(target)
    and math.abs(dist_sprite(target)) <= ( param.bossBurstedMode and param.attackRangeRageX or param.attackRangeX ) then
      if dist_sprite(target) < 0 then
        turn_around(7)
      elseif param.attackCooldown <= 0 then
        if dist_sprite(target) <= 160 * 2 
        and dist_sprite(target) > 0 then
          attack_mouthNaga()
          if param.bossBurstedMode
          then
            attack_twister()
          end
        elseif dist_sprite(target) > 0 then
          attack_snoutGrab()
          if param.bossBurstedMode
          then
            attack_twister()
          end
        end
      end
    elseif math.abs(dist_sprite(target)) <= param.sightRangeX 
    and select(2, dist_sprite(target)) <= param.sightRangeY.top
    and select(2, dist_sprite(target)) >= param.sightRangeY.bottom then
      if param.aggroTime > 40 then
        approach(target.XPos)
      else
        if dist_sprite(target) < 0 then
          turn_around()
        end
        param.aggroTime = param.aggroTime + 1
      end
    elseif math.abs(dist_grazePoint()) >= me.WorldCollisionWidth * 0.4 then
      param.aggroTime = math.max(0, param.aggroTime - 1)
      if param.aggroTime <= 0 then
        approach(param.grazePoint.x)
      end
    end
    
    wait() -- try to keep this at the end
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  
  attack_naga(toReturn)
  
  local dirChance = random(1000)
  local actionChance = random(1000)

  trace("dir %i" % { dirChance })
  trace("act %i" % { actionChance })

  if param.isNotorious then
    trace("Notorious")
    if actionChance > 650
    then
      attack_nm_spit(toReturn)
    elseif dirChance > 500
    then
      attack_mouthNaga(toReturn)
    else
      attack_twister(toReturn)
    end
  else
    if (dirChance < 333) then
      attack_snoutGrab(toReturn)
    elseif (dirChance > 667) then
      attack_mouthNaga(toReturn)
    end

    local actionChance = random(1000) 
    if me.HealthPct <= 0.33
    and actionChance > 500 then
      attack_twister(toReturn)
    end
  end

  return toReturn
end --OnBattleAttack()