require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

-- Attacks --
local function attack_attack(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_strike(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_flashKick(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_riderKick(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_superStrike(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = true
  end
  doMotion("X+Heruka")
end

local function attack_superFlashKick(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = true
  end
  doMotion("D+Heruka")
end

local function attack_superRiderKick(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = true
  end
  doMotion("U+Heruka")
end

-- sequence keys
local sequence = {
  default = 0,

  triple = 0,
  flashKick = 1,
  riderKick = 2,
  
  super = 10
}

-- the initialization function for this AI
function OnInit()
  param.time = 0
  --
  param.battleSeq = 0  
  param.battleSubSeq = 0
  param.battleMaxTimeout = 60 * 5 -- just in case
  param.battleTimeout = 60 * 1 -- default for our own sequence
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  while true do
    me, player = getMe(), getPlayer()

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  local dirChance = random(1000)
  local actionChance = random(1000)
  local superOK = (me.LifePct <= 50/100)
  --
  param.battleSeq = 0
  param.battleSubSeq = 0 
  param.time = 0

  -- clear motions
  toReturn.meterBurn = false
  toReturn.pressAction = false
  toReturn.pressDash = false


  if superOK
  and actionChance >= 500
  then
    trace("Super")
    param.battleSeq = sequence.super
    if dirChance > 667
    then
      attack_superFlashKick(toReturn)
    elseif dirChance < 334
    then
      attack_superRiderKick(toReturn)
    else
      attack_superStrike(toReturn)
    end
  else
    trace("Normal")
    if dirChance > 667
    then
      attack_flashKick(toReturn)
      param.battleSeq = sequence.flashKick
    elseif dirChance < 334
    then
      attack_riderKick(toReturn)
      param.battleSeq = sequence.riderKick
    else
      attack_strike(toReturn)
      param.battleSeq = sequence.triple
    end
  end

  return toReturn
end --OnBattleAttack()

function OnBattleTurnUpdate(battleData)
  local me = getMe()

  if aiDebug()
  then
    --trace("%003i:%i:%i" % { param.time, param.battleSeq, param.battleSubSeq })
    draw_bar(param.time, 0, param.battleMaxTimeout, 0)
    draw_bar(param.time, 0, param.battleTimeout, 1, {0, 255, 0})
  end

  -- the loop --
  if param.time <= param.battleMaxTimeout
  then
    -- Triple Punch
    if param.battleSeq == sequence.triple
    then
      if param.time >= param.battleTimeout
      then
        debugPalette("Debug_Black")
      elseif param.battleSubSeq == 0 
      then
        debugPalette("Debug_LightBlue")
        if param.time >= 5
        and param.time <= 80
        and (((param.time + 5) % 11) == 0)
        then
          attack_attack(toReturn)
        end
      else
        debugPalette("Debug_DarkBlue")
        addInput(dirs.N)
      end
    -- Flash Kick
    elseif param.battleSeq == sequence.flashKick
    then
      if param.time >= param.battleTimeout
      then
        debugPalette("Debug_Black")
      elseif param.battleSubSeq == 0 
      then
        debugPalette("Debug_LightRed")
        --
        if me.NumHits > 0
        and me.NumTouches == me.NumTouches -- if you didn't hit everyone 
        then
          param.time = 0
          param.battleSubSeq = 1
        end
      elseif param.battleSubSeq == 1
      then
        debugPalette("Debug_BlinkRed")
        --
        if param.time >= 23
        then
          attack_riderKick(toReturn)
          param.time = param.battleMaxTimeout
        end
      else
        debugPalette("Debug_DarkRed")
        addInput(dirs.N)
      end
    -- Rider Kick
    elseif param.battleSeq == sequence.riderKick
    then
      if param.time >= param.battleTimeout
      then
        debugPalette("Debug_Black")
      elseif param.battleSubSeq == 0 
      then
        debugPalette("Debug_Orange")
        --
        if me:hasTag("RiderKick_Dive_Hit")
        then
          param.time = 0
          param.battleSubSeq = 1
        end
      elseif param.battleSubSeq == 1
      then
        debugPalette("Debug_BlinkOrange")
        --
        if me:hasTag("RiderKick_FallLoop")
        then
          attack_riderKick(toReturn)
          --param.time = param.time + 60
        end
      else
        debugPalette("Debug_DarkOrange")
        addInput(dirs.N)
      end
    elseif param.battleSeq == sequence.super
    then
      if param.time >= param.battleTimeout
      then
        debugPalette("Debug_Black")
      elseif param.battleSubSeq == 0 
      then
        debugPalette("Debug_Blink")
      else
        debugPalette("Debug_BlinkFast")
        addInput(dirs.N)
      end
    end
  end
  
  -- update only if not hitstop
  if update_ok()
  then
    param.time = param.time + 1
  end
end