require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local string_isBoss = "Boss"

-- attacks and actions
local function attack_attack(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_wallHit(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_wallHitFAST(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = true
  end
  doMotion("X+Attack+MeterBurn")
end

local function attack_wallSplat(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_tailFlap(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_wingFlapFront(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Heruka")
end

local function attack_wingFlapBack(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Heruka")
end

local function attack_wingFlapDouble(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function attack_beam(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function action_enterOffscreenIdle(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("U+Dash")
end

local function action_offscreenIdleTeleport(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("X+WorldGuard")
end

local function setTeleportInfo(_x, _y)
  local me = getMe()
  local x = _x or me.XPos
  local y = _y or me.YPos

  me.WeaponType = x
  me.DHCNum = y
end

local function action_exitOffscreenIdle(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("D+Dash")
end

-- utilities
local function isInStartBattleRange()
  local me, player = getMe(), getPlayer()

  local w = 160 * 10
  local h = 160 * 10

  return isInRange(player, me.XPos, me.YPos, w, h)
end

local function isInFirstStartBattleRange()
  local me, player = getMe(), getPlayer()

  local w = 160 * 10
  local h = 160 * 10

  return isInRange(player, me.XPos, me.YPos, w, h)
end

local function isInWorldAttackRange()
  local me, player = getMe(), getPlayer()

  local w = 160 * 26
  local h = 160 * 18

  return isInRange(player, me.XPos, me.YPos, w, h)
end

local function isInOffscreenIdle()
  local me = getMe()

  if me:hasTag("Offscreen_Idle")
  or me:hasTag("Offscreen_Idle_Teleport")
  or me:hasTag("Offscreen_Idle_Exit")
  then
    return true
  else
    return false
  end

  return false
end

local function WorldAttackOK()
  local prog = getProgression("JingfeiRecruit")
  if prog >= 30
  and prog <= 999
  then
    return true
  else
    return false
  end
  
  return false
end

local Mode = {
  Init = 0,
  IntroWait = 100,
  LeaveBattle = 200,
  WaitForBattle = 300,
}

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()
  param.time = 0
  --
  param.origin = { x = me.XPos, y = me.YPos, z = 0 }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  param.isBoss = false
  if has_spawn_command(string_isBoss)
  then
    param.isBoss = true
  end
  --
  param.curMode = 0
  param.prevMode = 0
  param.curCmd = 0
  param.prevCmd = 0
  -- 
  param.phase = 0 
  param.startedBattle = false
  param.approchPoint = false
  param.goToTopStairs = true
  param.hasDoneIntroPattern = false
  param.hasDoneFirstWallHit = false
  --
  param.teleportCooldown = 0
  --
  param.points = {
    BottomStairs = vec_object("Spawn_Boss_Okosoko_BottomStairs"),
    TopStairs = vec_object("Spawn_Boss_Okosoko_TopStairs")
  }
end

local function debugDraw()
  local me = getMe()

  draw_bar(param.time, 0, 120, -20, {255, 255, 255})
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end

  while true do
    me, player = getMe(), getPlayer()

    if aiDebug()
    then
      debugDraw()
    end

    if update_ok() then
      if param.isBoss
      then
        if isInWorldAttackRange()
        then
          if not aiDebug() 
          then
            param.time = param.time +1
          end
        else
          param.time = 0
        end

        if not isInOffscreenIdle()
        then
          param.teleportCooldown = math.max(0, param.teleportCooldown - 1)
        end

        if isInOffscreenIdle()
        and param.teleportCooldown <= 0
        then
          debugPalette("Debug_Yellow")
          trace("isInOffscreenIdle()")
          --
          local target = param.points.BottomStairs
          if param.goToTopStairs then
            target = param.points.TopStairs
          end
          
          local distToPoint = vec_dist( target, vec_pos(me) )
          trace("distToPoint %i" % { distToPoint })
          trace("goToTopStairs %s" % { tostring(param.goToTopStairs) })

          if me:hasTag("Offscreen_Idle")
          then
            trace("In OffScreen_Idle")
            
            if param.goToTopStairs then
              param.goToTopStairs = false
            else
              param.goToTopStairs = true
            end

            action_offscreenIdleTeleport()
            setTeleportInfo(target.x, target.y)

            param.teleportCooldown = 30
            param.time = 0 

          elseif me:hasTag("Offscreen_Idle_Teleport")
          then
            trace("In Offscreen_Idle_Teleport")
          elseif me:hasTag("Offscreen_Idle_Exit")
          then
            trace("oh ok we're going BACK now")
          else
            trace("Waiting for WallHit_Loop tag!")
          end
        elseif not param.hasDoneIntroPattern
        then
          if isInWorldAttackRange()
          and WorldAttackOK()
          then
            attack_wallHitFAST()
            param.hasDoneIntroPattern = true
            param.approchPoint = true
          end
        elseif not param.startedBattle
        then
          debugPalette("Debug_Red")

          if isInStartBattleRange()
          and player_vulnerable()
          and not me:hasTag("Attack")
          then
            param.startedBattle = me:startBattle(player)
          end
        else
          debugPalette("Debug_Blue")

          if param.time >= 110
          then
            attack_wallHit()
            param.approchPoint = true
          end
        end
      end
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  local dirChance = random(1000)
  local actionChance = random(1000)

  -- MAKE SURE WE DON'T CHOOSE SOMETHING SPECIAL --
  attack_beam(toReturn)

  if me.Flag11
  then
    trace("hey whats up")
    print("hey why not what the hell")
    attack_wallSplat(toReturn)
    param.hasDoneFirstWallHit = true
  elseif me.HealthPct <= 33/100
  and param.phase >= 2
  then
    trace("33%")
    attack_wallHit(toReturn)
    param.approchPoint = true
    param.hasDoneFirstWallHit = true
  elseif me.HealthPct <= 66/100
  and param.phase <= 0
  then
    trace("66%")
    attack_wallHit(toReturn)
    param.approchPoint = true
    param.hasDoneFirstWallHit = true
  -- Attacks --
  else
    trace("Normal")
    if battleData.turnNum >= 5
    then
      attack_wallHit(toReturn)
      battleData.turnNum = 0 --
      param.approchPoint = true
      param.hasDoneFirstWallHit = true
    else
      attack_wingFlapBack(toReturn)

      -- Flap --
      if actionChance >= 725
      and param.hasDoneFirstWallHit 
      then
        attack_wingFlapDouble(toReturn)
      elseif actionChance <= 275
      and param.hasDoneFirstWallHit 
      then
        attack_wingFlapFront(toReturn)
      -- Beamu --
      elseif dirChance >= 667
      then
        attack_beam(toReturn)
      elseif dirChance <= 334
      then
        attack_tailFlap(toReturn)
      end

      if aiDebug()
      then
        attack_beam(toReturn)
        --attack_wingFlap(toReturn)
        attack_tailFlap(toReturn)

        if battleData.turnNum == 1
        then
          attack_wingFlapBack(toReturn)
        elseif battleData.turnNum == 2
        then
          attack_wingFlapFront(toReturn)
        elseif battleData.turnNum == 3
        then
          attack_wingFlapDouble(toReturn)
        elseif battleData.turnNum == 4
        then
          attack_tailFlap(toReturn)
        else
          attack_wallSplat(toReturn)
        end
        
        attack_wingFlapFront(toReturn)
      end
    end
  end

  return toReturn
end --OnBattleAttack()

--[[
function OnBattleTurnUpdate(battleData)
  local me = getMe()
end
]]