require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function dist_origin()
  local me = getMe()
  return param.origin.x - me.XPos, param.origin.y - me.Ypos 
end

local function bulletDist()
  local _x, _y = nil, nil
  local _ret = nil
  if select(1,dist_blang()) ~= nil then
    return dist_blang()
  elseif select(1,dist_spear()) ~= nil then
    return dist_spear()
  elseif  select(1,dist_arrow()) ~= nil then
    return dist_arrow()
  end
  return -256^2, -256^2
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())

  local dx, dy = dist_sprite(target)

  if dx >= -param.sightRangeX/2
  and dx <= param.sightRangeX/2
  and dy <= param.sightRangeY.top
  and dy >= param.sightRangeY.bottom
  then
    return true
  end
  return false
end

local function isInCloseRange(target)
  local me, player = getMe(), (target or getPlayer())

  local dx, dy = dist_sprite(target)

  if dx >= -param.closeRangeX/2
  and dx <= param.closeRangeX/2
  and dy <= param.closeRangeY.top
  and dy >= param.closeRangeY.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange(target)
  local me, player = getMe(), (target or getPlayer())

  local dx, dy = dist_sprite(target)

  if dx >= -param.attackRangeX/2
  and dx <= param.attackRangeX/2
  and dy <= param.attackRangeY.top
  and dy >= param.attackRangeY.bottom
  then
    return true
  end
  return false
end

local function isBulletInFront()
  local me = getMe()
  local bx, by = bulletDist()

  bx = bx * me.FacingDir

  if bx >= 0
  and bx < param.sightRangeX
  and by <= param.sightRangeY.top
  and by > param.sightRangeY.bottom
  then 
    return true
  else 
    return false
  end
end

local function attack_disappear( data )
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Jump")
end

local function attack_riseFromGrave( data )
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.attemptedShot = true
end

local function attack_staffLift( data )
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function attack_allySummon( data )
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_chargeCast( data )
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Heruka")
end

local function attack_spinSwipe( data )
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_poseBlock( data )
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("U+Dash")
end

local function attack_poseSummon( data )
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("D+Dash")
end

local function attack_poseUpward( data )
  if data ~= nil then
    data.dir = dirs.DF
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("DF+Dash")
end

local function attack_poseStaffSpin( data )
  if data ~= nil then
    data.dir = dirs.DB
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("DB+Dash")
end

local function attack_poseLeaveBattle( data )
  if data ~= nil then
    data.dir = dirs.F
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("F+Dash")
end

local function action_introAppear( data )
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function action_turnInvisible(data)
  doMotion("X+WorldGuard")
end

local function action_turnVisible(data)
  doMotion("X+WorldGuard")
end

local function action_faceLeft(data)
  addInput(dirs.B)
end

local function action_faceRight(data)
  addInput(dirs.F)
end

local function action_teleport(x, y)
  local me = getMe()
  
  doMotion("X+Dash")
  me.ThrowX = x
  me.ThrowY = y
end

local function isInvisible()
  local me = getMe()
  if me:hasTag("Invisible")
  then
    return true
  else
    return false
  end

  return false
end

local function getPoint(s)
  local name = s 
  local _x, _y, _z = getObjectPos(name, false)
  --trace("We requested point %s" % { name})
  if type(_x) == "number" and type(_y) == "number"
  then
    --trace("(%i %i)" % { _x, _y })
    return { x = _x, y = _y }
  else
    --trace("Failed to find point %s" % { s })
    return nil
  end

  return nil
end

local prog_attackOKValue = "OontBoss_AttackOK"
local prog_attackValue = "OontBoss_Command"
local prog_rage = "OontBoss_RageMode"
local prog_inBattle = "OontBoss_InBattle"

local function poseAttackOK()
  local pv = getProgression(prog_attackValue)

  --trace(pv)
  if pv == 0
  then
    return true
  else
    return false
  end

  return false
end

local function worldAttackOK()
  local pv = getProgression(prog_attackOKValue)

  --trace(pv)
  if pv == 1
  then
    return true
  else
    return false
  end

  return false
end

local function isOnScreen()
  local me = getMe()

  if me.TouchScreenLR < 1
  and me.TouchScreenLR > -1
  and me.TouchScreenUD < 1
  and me.TouchScreenUD > -1
  then
    return true
  else
    return false
  end

  return false
end

local function applyRank(r)
  local me = getMe()
  local _rank = 0
  if _r then _rank = _r end

  -- World Version
  if r <= -1
  then
    me.DamageMult = param.baseDamageMult * 0.5
    me.BaseBattleSpeed = param.baseBattleSpeed
  else
    local factor = (1) + (r * param.rankDamageMultFactor)
    me.DamageMult = param.baseDamageMult + (factor * param.baseDamageMult)

    factor = (1) + (r  * param.rankBattleSpeedMultFactor )
    me.BaseBattleSpeed = param.baseBattleSpeed + (factor * param.rankBattleSpeedMultFactor)
  end
end

local function sideOfArena()
  local me = getMe()
  if param.points.BottomCenter ~= nil
  then
    if me.XPos <= param.points.BottomCenter.x
    then
      return -1
    else
      return 1
    end
  else
    return 0
  end

  return 0
end

local function pickRandomOffset( battle )
  local me = getMe()
  local isBattle = battle or false

  -- If we're more than halfway from our origin, we should look the other way
  local zoneX = sideOfArena()
  local facingDir = me.FacingDir

  -- If we were in battle, we're going to pick somewhere we want to be facing.

  if isBattle
  then
    zoneX = zoneX * facingDir
  end

  -- Pick an area not near the player...
  local _offset = 500 -- the default...
  if zoneX == 1 
  then
    _offset = random(50, 100)
  elseif zoneX == -1
  then
    _offset = random(900, 950)
  else
    -- Roll again, so we don't end up near the middle again
    _offset = random(0, 1000)
    if _offset > 500
    then
      _offset = random(50, 100)
    else
      _offset = random(900, 950)
    end
  end

  return _offset
end

function OnInit()
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  param.lastPlayerPos = {x = 0, y = 0 }
  --
  param.time = 0
  param.timeOnScreen = 0 
  param.disappearTime = 110
  param.introWaitTime = 0 
  param.isFirstSpawn = false
  param.attemptedShot = false
  --
  param.isOontCommander = false
  param.commandCooldown = 0  -- No Poses when spawning!
  param.attackCooldown = 100 -- No Attacking when spawning!
  param.evadeTime = 0 
  param.hurtCheckTime = 0
  param.canAttackCheckTime = 0 -- Warmup value when we haven't had a chance to attack
  param.startBattleOK = true
  param.evadePoint = { x = 0, y = 0 }
  param.evadePointMaxRange = 160 * 52 -- 26 x 2 
  param.startedBattle = false
  param.shouldPickFarEvade = false
  param.shouldPickEvadePoint = false
  param.forceEvade = false
  param.hideUntilWormDone = false
  param.maxHideTime = 60 * 6
  param.hasDoneIntroPose = false
  param.timesHitByArrow = 0 
  param.turnsEscaped = 0 
  --
  param.sightRangeX = 160 * 14
  param.sightRangeAlertedX = 160 * 16
  param.sightRangeY = { top = 160 * 1.5, bottom = 160 * -1 }
  param.attackRangeX = 160 * 7
  param.attackRangeCloseX = 160 * 3
  param.attackRangeY = { top = 160 * 2.5, bottom = 160 * -1.5 }
  param.closeRangeX = 160 * 7
  param.closeRangeY = { top = 160 * 2.5, bottom = 160 * -2.5 }
  --
  param.playerCloseTime = 0 
  param.playerCloseTimeMax = 15
  --
  if has_spawn_command("FirstSpawn") then
    param.isFirstSpawn = true
  end
  --
  if has_spawn_command("WormCommander")
  then
    param.isOontCommander = true
    -- For Evading --
    param.attackRangeX = 160 * 12
    param.attackRangeY = { top = 160 * 9, bottom = 160 * -9 }
    -- For firing magic --
    param.sightRangeX = 160 * 20
    param.sightRangeY = { top = 160 * 9, bottom = 160 * -9 }
    --
    param.points = {
      BottomCenter = getPoint("Spawn_Worm_BottomAnchor")
    }
  end

  -- Reset this, because we issue the commands
  setProgression(prog_attackValue, 0)
  setProgression(prog_inBattle, 0)

  -- Battle
  param.phase = 0  -- This one is for what you CAN do in battle WHEN
  param.maxPhases = 0
  param.rank = 0  -- This one is for adjusting how fast or mean you are
  param.maxRank = 5
  param.baseDamageMult = me.DamageMult
  param.baseBattleSpeed = me.BaseBattleSpeed
  param.rankDamageMultFactor = 0.05  -- Add this much per rank
  param.rankBattleSpeedMultFactor = 0.02  -- Add this much per rank
  param.battleTurnsSinceSpecial = 0
  param.battleForceLeaveTurns = 4
  param.battleForceLeaveTurnsDefault = 4
  param.leaveBattleTiers = {
    [0] = 0.67,
    [1] = 0.37,
    [2] = 0.15,
    [3] = 0.09,
    [4] = 0
  }
  param.maxPhases = #param.leaveBattleTiers - 1

  param.bossBurstInstall = false

  -- Notorious
  param.isNMBeastmaster = false
  if has_spawn_command("NM_Beastmaster")
  then
    param.isNMBeastmaster = true
  end
  -- Debug
  param.debugDemoMode = false
end

function OnUpdate()
  local me, player = getMe(), getPlayer()

  local players = getPlayers()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end

  while true do
    me, player = getMe(), getPlayer()
    --
    -- Debug
    if aiDebug() then
      draw_bar(0, -30, 120, 0, {255, 0, 0})
      --draw_bar(param.timeOnScreen, -30, 300, 0, {255, 255, 255})
      --
      draw_bar(param.time, 0, 120, 0, {255, 255, 255})
      draw_bar(param.hurtCheckTime, 0, 10, 1, {255, 255, 0})
      draw_bar(param.evadeTime, 0, 120, 2, {255, 128, 128})
      draw_bar(param.commandCooldown, 0, 150, 3, {128, 128, 255})
      draw_bar(param.attackCooldown, 0, 150, 4, {255, 0, 0})
      draw_bar(param.canAttackCheckTime, 0, 30, 5, {0, 255, 255})
      draw_bar(param.playerCloseTime, 0, 30, 6, {255, 128, 32})
    end

    -- Checks
    if tick_ok() then
      if me.TouchScreenLR < 1 and me.TouchScreenLR > -1 then
        param.timeOnScreen = param.timeOnScreen + 1
      else
        param.timeOnScreen = 0 
      end
    end

    -- 
    if param.isOontCommander
    then
      -- Personal Updates --
      if tick_ok() then
        param.time = param.time + 1
        param.evadeTime = math.max(0, param.evadeTime - 1)
      end

      if worldAttackOK()
      then
        param.canAttackCheckTime = param.canAttackCheckTime + 1
      else
        param.canAttackCheckTime = 0 
      end

      if not me:hasTag("Pose")
      and not me:hasTag("Intro")
      then
        param.commandCooldown = math.max(0, param.commandCooldown - 1)
      end
      
      if not me:hasTag("Attack")
      and not me:hasTag("Intro")
      then
        param.attackCooldown = math.max(0, param.attackCooldown - 1)
      end

      -- Force Ourselves to leave if Boss Burst
      if bossBursted() 
      then
        trace("BOSS BURST")
        param.forceEvade = true
        param.hurtCheckTime = 0 
        me.InvinFlashTime = 300  -- longer for us
        param.bossBurstInstall = true 
        param.evadeTime = 60
      end

      -- Special checks that make sure you can't CHEEZE oont with arrows! --
      if update_ok()
      then
        param.hurtCheckTime = math.max(0, param.hurtCheckTime - 1)
        --
        if param.hurtCheckTime <= 0
        then
          param.startBattleOK = true
        end

        --
        if isInCloseRange()
        then
          param.playerCloseTime = param.playerCloseTime + 1
        else
          param.playerCloseTime = 0
        end

        --
        if dist_arrow(160 * 4)
        and param.timesHitByArrow > 0 
        and not me:hasTag("Attack")
        then
          param.shouldPickEvadePoint = true
          if getProgression(prog_attackOKValue) == 0
          -- and me:hasTag("Pose")
          then
            param.hideUntilWormDone = true
          end
        end
        --
        setProgression(prog_inBattle, 0)
      end

      -- Logic --
      if update_ok()
      then
        -- Apply Rank --
        applyRank(-1)

        -- Got shot by an arrow! -- 
        if is_alerted()
        and param.hurtCheckTime <= 0 
        and param.startBattleOK
        then
          trace("Got Shot!")

          -- Only advance this in Escape 3
          param.hurtCheckTime = 4
          if param.turnsEscaped >= 2
          then
            param.timesHitByArrow = param.timesHitByArrow + 1 
          end
          local dist = vec_dist( vec_pos(me), vec_pos(player) )

          local range = 160 * 14 
          if dist <= range
          and isInRange(player, me.XPos, me.YPos, range, range)
          then
            -- Force a battle 
            param.startedBattle = player:startBattle(me)
            param.startBattleOK = false
          end
        end

        -- INTRO MODE --
        if not param.hasDoneIntroPose
        then
          trace("[INFO] Oont waiting for Intro")
          if param.introWaitTime >= 30
          then
            action_introAppear()
            wait()
            facePlayer()
            wait(5)
            param.hasDoneIntroPose = true
          elseif isOnScreen()
          then
            param.introWaitTime = param.introWaitTime + 1
          end
          -- Demo Mode --
        elseif param.debugDemoMode
        then
          param.time = param.time + 1
          --
          if param.time >= 30
          then
            param.time = 0
            local r = random(1000)

            --
            if r >= 667
            then
              attack_poseBlock()
            elseif r <= 333
            then
              attack_poseSummon()
            else
              attack_poseUpward()
            end
            --attack_poseBlock() -- 1
            --attack_poseUpward() -- 3
            --attack_poseSummon() -- 2
          end
        -- Don't ever be invisible --
        elseif me:hasTag("Invisible_Idle_Loop")
        and param.hasDoneIntroPose
        then
          param.forceEvade = true
          param.shouldPickEvadePoint = true
        -- FORCEFULLY EVADE --
        elseif param.forceEvade
        then
          param.forceEvade = false
          param.shouldPickEvadePoint = true
          trace("yo what's up we came from battle")
        -- Boss Burst Install --
        elseif param.bossBurstInstall
        and param.evadeTime <= 0
        and math.abs(me.XVel) <= 7
        and not isInAttackRange()
        and not me:hasTag("Intro")
        and not me:hasTag("MoveFwd_End")
        and not me:hasTag("Pose")
        and not me:hasTag("Attack")
        then
          trace("Boss Burst Command!")
          param.hasDoneIntroPose = true -- JUST IN CASE
          param.bossBurstInstall = false 

          attack_poseLeaveBattle()

          param.commandCooldown = 60 * 4
          param.attackCooldown = 60 * 2.75
          param.time = 0
        -- Issue a command! --
        elseif param.time >= 30
        and poseAttackOK()
        --and isOnScreen()
        and param.commandCooldown <= 0 
        and param.evadeTime <= 0
        and param.playerCloseTime <= 0 
        and math.abs(me.XVel) <= 7
        and not isInAttackRange()
        and not me:hasTag("Intro")
        and not me:hasTag("MoveFwd_End")
        and not me:hasTag("Pose")
        and not me:hasTag("Attack")
        then
          trace("Command!")
          param.hasDoneIntroPose = true -- JUST IN CASE
          local r = random(1000)

          if param.phase >= 2
          then
            if r > 675
            then
              attack_poseBlock()
            elseif r < 325
            then
              attack_poseSummon()
            else
              attack_poseUpward()
            end
          else
            if r > 500
            then
              attack_poseBlock()
            else
              attack_poseSummon()
            end
          end

          param.commandCooldown = 60 * 4
          param.attackCooldown = 60 * 2.75
          param.time = 0
        -- Leave if the player is close to us, IMMEDIATELY
        elseif isInCloseRange()
        and param.turnsEscaped >= 2
        and param.playerCloseTime >= param.playerCloseTimeMax
        --and not me:hasTag("Pose")
        and not param.shouldPickEvadePoint
        then
          trace("CloseRange TURN 2 Leave!")
          param.shouldPickEvadePoint = true
        -- Leave if the player is close to us!
        elseif isInAttackRange()
        and param.turnsEscaped >= 1
        and not me:hasTag("Attack")
        and param.playerCloseTime >= param.playerCloseTimeMax
        --and not me:hasTag("Pose")
        and not param.shouldPickEvadePoint
        then
          trace("CloseRange Leave!")
          param.shouldPickEvadePoint = true
        -- Leave logic for real
        elseif param.shouldPickEvadePoint
        and param.hasDoneIntroPose
        then
          trace("Start Evade")
          param.shouldPickEvadePoint = false
          local _offset = pickRandomOffset()
          _offset = _offset / 1000
          param.shouldPickFarEvade = false
          --
          trace("%003.2f%%" % { _offset * 100 })
          param.evadeTime = 300
          local _x = param.points.BottomCenter.x - param.evadePointMaxRange/2
          param.evadePoint.x = _x + (_offset * param.evadePointMaxRange)
          param.evadePoint.y = param.origin.y + 1280  -- Just in case
          --
          param.evadeTime = 0 
          param.time = 0
          local _process = true
          while _process do
            --
            local wantVisible = false
            trace("[INFO] Proc1")
            trace("Invisible? %s" % { tostring(isInvisible()) })
            while not isInvisible()
            do
              trace("we are not invisible, we're going to be now!")
              action_turnInvisible()
              wait(13)
            end
            --
            wait(30)
            param.time = 0 
            while param.hideUntilWormDone do
              param.time = param.time + 1
              wait()
              trace("STILL WAITING UNTIL WORM DONE! (%i/%i)" % {
                param.time, param.maxHideTime
              })
              if getProgression(prog_attackOKValue) > 0
              or param.time >= param.maxHideTime
              then
                param.hideUntilWormDone = false
              end
            end
            trace("[INFO] Proc2")
            trace("Invisible? %s" % { tostring(isInvisible()) })
            action_teleport(param.evadePoint.x, param.evadePoint.y)
            wait(5)
            --
            trace("[INFO] Proc3")
            trace("Invisible? %s" % { tostring(isInvisible()) })
            action_faceRight()
            trace("dist_player() %i" % { dist_player() })

            if dist_player() < 0 -- other way around
            then
              action_faceLeft()
            end

            trace("[INFO] Proc4")
            trace("Invisible? %s" % { tostring(isInvisible()) })
            wantVisible = true
            action_turnVisible()
            wait(15)

            if isInvisible()
            and wantVisible
            then
              trace("Invisible? %s" % { tostring(isInvisible()) })
              trace("we are invisible, why?! turn it OFF")
              action_turnVisible()
              trace("Invisible? %s" % { tostring(isInvisible()) })
              wait(15)
              trace("Invisible? %s" % { tostring(isInvisible()) })
            end
            param.shouldPickEvadePoint = false
            param.hideUntilWormDone = false
            param.forceEvade = false
            param.evadeTime = 0 
            _process = false
          end
        -- Attempt to attack the player if we're chillin, otherwise --
        elseif param.attackCooldown <= 0
        --and isOnScreen()
        --and worldAttackOK()
        and param.attackCooldown <= 0
        and param.playerCloseTime <= 0 
        --and param.canAttackCheckTime >= 30
        and not me:hasTag("MoveFwd_End")
        and not me:hasTag("Pose")
        and not me:hasTag("Attack")
        then
          trace("Attack!")
          if dist_player() <= 0
          then
            turn_around(3)
          end
          --
          local r = random(1000)
          if r > 500
          then
            attack_staffLift()
          else
            attack_riseFromGrave()
          end
          --
          param.attackCooldown = 72
        end
      end
    -- Normal Oont Behavior --
    elseif me.OnGround > 0 then
      -- FirstSpawn first!
      if param.isFirstSpawn then
        if param.timeOnScreen > param.disappearTime 
        or isInSightRange() 
        or isBulletInFront()
        then
          debugPalette("Debug_Red")
          attack_disappear()
        end
      -- Vanish from bullet
      elseif isBulletInFront() then
        attack_disappear()
      -- Vanish only if you're close
      elseif isInAttackRange() then
        debugPalette("Debug_Green")
        attack_disappear()
      -- Vanish AFTER Attacks
      elseif param.attemptedShot
      and me.NumTouches > 0
      then
        attack_disappear()
      elseif param.attemptedShot
      and param.timeOnScreen > 2
      then
        attack_disappear()
      -- Attacks
      elseif player_vulnerable()
      and isInSightRange()
      then
        debugPalette("Debug_Yellow")
        attack_riseFromGrave()
        param.timeOnScreen = 0
        -- Do.. nothing
      elseif player_vulnerable()
      and param.timeOnScreen > param.disappearTime
      then
        debugPalette("Debug_Orange")
        attack_riseFromGrave()
        param.timeOnScreen = 0
      end
    else
      -- whoa nice
    end

    wait() -- 
  end -- loop
end -- OnUpdate()

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  local dirChance = random(1000)
  local actionChance = random(1000)

  if param.isOontCommander
  then
    setProgression(prog_inBattle, 1)
    applyRank(param.rank)
    param.battleForceLeaveTurns = param.battleForceLeaveTurns - 1
    param.battleTurnsSinceSpecial = param.battleTurnsSinceSpecial + 1
    trace("Turns until leaving: %i" % { param.battleForceLeaveTurns })

    -- Command Leave -- 
    local shouldLeave = false 
    if (param.battleForceLeaveTurns <= 0) 
    or me.LifePct <= param.leaveBattleTiers[param.phase]
    then
      shouldLeave = true
    end

    if shouldLeave
    then
      param.battleForceLeaveTurns = param.battleForceLeaveTurnsDefault
      param.phase = math.min(param.phase + 1, param.maxPhases)
      param.rank = math.min(param.rank + 1, param.maxRank) -- Why not..
      trace("[INFO] Oont Commander LEAVING battle! Now on Phase %i" % { param.phase })
      
      attack_poseLeaveBattle(toReturn)
      
      -- Ensure we clear our values if we're going to go back to the world
      param.time = 0 
      param.attackCooldown = 200 -- No attacking for a WHILE
      param.commandCooldown = 10
      
      param.turnsEscaped = param.turnsEscaped + 1
      param.evadeTime = 1
      param.shouldPickEvadePoint = true
      param.forceEvade = true
      param.hideUntilWormDone = false
    else
      if actionChance >= 500
      and param.battleTurnsSinceSpecial > 2
      then
        attack_spinSwipe(toReturn)
        param.battleTurnsSinceSpecial = 0
      elseif dirChance <= 500
      then
        attack_staffLift(toReturn)
      else
        attack_riseFromGrave(toReturn)
      end
    end
  end

  if param.isNMBeastmaster
  then
    attack_poseStaffSpin(toReturn)
  end

  return toReturn
end -- OnBattleAttack()