require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function dist_origin()
  local me = getMe()
  return param.origin.x - me.XPos, param.origin.y - me.Ypos 
end

local function dist_goal()
  local me = getMe()
  return param.goalX - me.XPos, param.origin.y - me.Ypos 
end

local function dist_lastPlayerPos()
  local me, player = getMe(), getPlayer()
  local x_dist = param.lastPlayerPos.x - me.XPos
  local y_dist = param.lastPlayerPos.y - me.YPos
  --
  return x_dist, y_dist
end

local function isInSightAlerted(target)
  local me, player = getMe(), (target or getPlayer())
  
  local w = param.sightRangeAlertedX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  local dx, dy = dist_sprite(target)
  if dx >= -w/2 and dx <= w/2
  and dy >= h.bottom and dy <= h.top
  then
      return true 
  end
  return false
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)
  
  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)
  
  local w = param.attackRangeX
  local h = param.attackRangeY
  if param.bossBurstedMode then
    w = param.attackRangeRageX
    h = param.attackRangeRageY
  end

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function attack_3hitCombo( data )
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("X+Attack")
end

local function attack_jumpSlash( data )
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("U+Attack")
end


local function attack_flurry( data )
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("D+Attack")
end

local function attack_singleSlash( data )
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("X+Heruka")
end

local function attack_shieldBang( data )
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("U+Heruka")
end

local function attack_shieldRush( data )
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("D+Heruka")
end

local function attack_statueRecover( data )
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  param.attackCooldown = 30
  doMotion("X+Attack")
end

local function attack_paletteAttack( data )
  local me = getMe()

  if me.ColorNum >= 5 then
    local r = random(1000)
    if r > 500 
    then
      return attack_jumpSlash(data)
    else
      return attack_flurry(data)
    end
  else
    return attack_singleSlash(data)
  end
end

local function attack_interrupt(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+BattleInterrupt")
end

local function specialAttackOK()
  local me = getMe()

  if me.ColorNum >= 5
  then
    return true
  else
    return false
  end

  return false
end

local function isStationaryGuard()
  local me = getMe()

  if me.ColorNum < 5
  then
    return true
  else
    return false
  end

  return false
end

-- sequence keys
local sequence = {
  none = 0,
  test = 1,
  combo = 10,
  overhead = 11,
  singleSlash = 12,
  specialAtk = 13
}

function OnInit()
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos, z = me.ZPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end

  --
  param.time = 0
  param.roamWaitTime = 0
  param.roamTurnTime = 0
  param.attackCooldown = 0
  param.alertTime = 0
  param.alertBaseTime = 60 * 3
  param.enteredStatueRange = false
  param.searchingForPlayer = false
  param.summonedFriends = false

  param.roamRangeX = 160 * 4
  param.roamRangeFudgeDist = 75
  --
  param.sightRangeX = 160 * 12
  param.sightRangeAlertedX = 160 * 16
  param.sightRangeY = { top = 160 * 1.5, bottom = 160 * -1 }
  param.attackRangeX = 160 * 6
  param.attackRangeCloseX = 160 * 2.5
  param.attackRangeY = { top = 160 * 2.5, bottom = 160 * -1.5 }
  --
  param.sightRangeRageX = 160 * 22
  param.sightRangeRageY = { top = 160 * 6, bottom = 160 * -1 }
  param.attackRangeRageX = 160 * 6
  param.attackRangeRageY = { top = 160 * 4, bottom = 160 * -1.5 }

  param.isPaidalRanger = false
  if me.WeaponType == hash("Challenge_Ranger")
  then
    param.isPaidalRanger = true
  end
  
  param.bossBurstedMode = false

  param.whatRanger = 0
  -- Red
  if me.ColorNum == 11
  or me.LightX == hash("Ranger_Red")
  then
    param.whatRanger = "Red"
  end
  -- Blue
  if me.ColorNum == 12
  or me.LightX == hash("Ranger_Blue")
  then
    param.whatRanger = "Blue"
  end
  -- Yellow
  if me.ColorNum == 13
  or me.LightX == hash("Ranger_Yellow")
  then
    param.whatRanger = "Yellow"
  end
  -- Black
  if me.ColorNum == 14
  or me.LightX == hash("Ranger_Black")
  then
    param.whatRanger = "Black"
  end
  -- Pink
  if me.ColorNum == 15
  or me.LightX == hash("Ranger_Pink")
  then
    param.whatRanger = "Pink"
  end

  -- Gold guards only look around.
  if isStationaryGuard() then
    param.roamRangeX = 160 * 0.375
  end

  param.goalX = param.origin.x
  param.goalX = param.goalX + ((param.roamRangeX/2) * (me.FacingDir * -1))

  param.lastPlayerPos = {x = 0, y = 0 }

  -- Battle 
  param.battleSeq = 0  
  param.battleSubSeq = 0
  param.battleTurnTimeout = 60 * 5 -- just in case
  param.battleSeqTimeout = 60 * 1 -- default for our own sequence
end

function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()
  
  while shouldNeverUpdate() do
    
  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    -- Debug
    if aiDebug() then
      draw_point(param.goalX, me.YPos)
      draw_bar(param.roamWaitTime, 0, 240)
      draw_bar(param.alertTime, 0, 180, 1, {255, 0, 255})
    end

    -- Checks
    if tick_ok() then
      if not isInSightRange() then
        param.alertTime = math.max(0, param.alertTime - 1)
        param.roamWaitTime = math.max(0, param.roamWaitTime - 1)
      end
      if not me:hasTag("Attack") then
        param.attackCooldown = math.max(0, param.attackCooldown - 1)
      end
    end

    if dist_sprite(target) <= 160 * 3
    and dist_sprite(target) >= 160 * -2 
    and select(2, dist_sprite(target)) <= 160 * 3
    and select(2, dist_sprite(target)) > 160 * -1
    then
      param.enteredStatueRange = true
    end

    -- Change goal pos when you get there
    local dist_goalPoint = dist_point(param.goalX, me.YPos)
    if math.abs( dist_goalPoint ) <= 20
    or me.WallDistAhead < 80 then
      local side = param.origin.x - me.XPos
      if side >= 0 then side = 1 else side = -1 end 
      param.goalX = param.origin.x + (param.roamRangeX/2 * (side))
      param.roamWaitTime = 60 * 4
      param.roamTurnTime = param.roamWaitTime / 2
    end    

    local _dist = dist_point(param.lastPlayerPos.x, param.lastPlayerPos.y)
    if me.Alerted >= 30 or me.Hitstun > 0 then
      param.alertTime = param.alertBaseTime -- whoa man
      param.roamWaitTime = 0
      param.lastPlayerPos = { x = target.XPos, y = target.YPos}
      param.searchingForPlayer = true
    end

    -- co-op 
    if bossBursted() 
    then
      param.roamWaitTime = 0
      param.attackCooldown = 20
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 4 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end

    -- Statue modes
    debugPalette()
    if me:hasTag("Statue_Loop_Forever") then
      wait()
    elseif me:hasTag("Statue_Loop") 
    and me.OnGround > 0 
    then
      if math.abs(dist_sprite(target)) <= 160 * 8
      and select(2, dist_sprite(target)) <= 160 * 3
      and select(2, dist_sprite(target)) > 160 * -1
      and
      (
      (me.FacingDir == player.FacingDir and dist_sprite(target) > 160 * 3  and dist_sprite(target) < 160 * 4)
      or (me.FacingDir ~= player.FacingDir and dist_sprite(target) <= 160 * -2 )
      )
      and sprite_vulnerable(target)
      then
        if param.enteredStatueRange 
        and not me:hasTag("Statue_Loop_Forever")
        then
          doMotion("X+Attack")
          wait(3)
        else
          -- !
        end
      end
    -- Normal Loop
    elseif me.OnGround > 0 then
      if isInAttackRange(target) then
        debugPalette("Debug_Red")
        if ( param.bossBurstedMode or sprite_vulnerable(target) ) 
        and dist_sprite(target) > 0
        and param.attackCooldown <= 0
        then
          -- Above
          if select(2,dist_sprite(target)) >= 160 * 1.5 then
            trace("Single Slash")
            attack_singleSlash()
          -- Close
          --[[elseif dist_player() <= param.attackRangeCloseX then
            attack_singleSlash()
            --attack_flurry()
            -- Far!
          elseif me.ColorNum ~= 1 then
          ]]
          else
            trace("Palette Attack")
            attack_paletteAttack()
          end
        elseif dist_sprite(target) < 0
        or ( dist_sprite(target) > 160 * 3 and dist_sprite(target) <= param.attackRangeX )
        then
          debugPalette("Debug_Magenta")
          approach_withTurn(target.XPos, 11)
        end
      elseif isInSightRange(target) then
        debugPalette("Debug_Blue")
        if ( param.bossBurstedMode or sprite_vulnerable(target) ) then
          approach_withTurn(target.XPos, 11)
        end
      elseif param.searchingForPlayer then
        debugPalette("Debug_Yellow")
        approach_withTurn(param.lastPlayerPos.x, 11)
        -- Let's just check the point we saw the player last
        if math.abs( dist_sprite(target) ) <= 100 then
          param.searchingForPlayer = false
        elseif obstacle_in_front() 
        or (
          math.abs( dist_lastPlayerPos() ) <= 100
        and not isInSightAlerted(target) 
        )
        then
          -- Hand off
          param.searchingForPlayer = false
          param.roamWaitTime = 90
        end
      else
        debugPalette("Debug_Green")
        ---
        if param.roamWaitTime > 0 then
          if param.roamWaitTime < param.roamTurnTime then
            if param.goalX < me.XPos 
            and me.FacingDir >= 1 then
              turn_around(4)
            elseif param.goalX >= me.XPos 
            and me.FacingDir <= -1 then
              turn_around(4)
            end
          end
        else
          approach_withTurn(param.goalX, 11)
        end
      end
    end
    wait() -- yield
  end
end -- OnUpdate()

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  -- Battle Sequences --
  battleData.sequenceTime = 0 
  param.battleSeq = 0
  param.battleSubSeq = 0 

  -- Default Atk
  attack_3hitCombo(toReturn)
  param.battleSeq = sequence.combo

  -- Pick Specials
  local specialChance = random(1000)
  if specialChance > 667
  and specialAttackOK()
  then
    attack_paletteAttack(toReturn)
    param.battleSeq = sequence.specialAtk
  elseif specialChance < 334
  and specialAttackOK()
  then
    attack_singleSlash(toReturn)
    param.battleSeq = sequence.singleSlash
  end
  
  -- Pick Procs
  if me.HealthPct <= 50/100
  and not param.summonedFriends
  then
    debugPalette("Debug_Blink")
    attack_shieldBang(toReturn)
    param.battleSeq = 0 
    param.summonedFriends = true
  end

  -- SPECIAL CASES 
  if me:hasTag("Statue_Loop") then
    attack_statueRecover(toReturn)
  end

  if param.isPaidalRanger
  then
    trace("Paidal Ranger")
    -- Red
    attack_singleSlash(toReturn)
    -- Blue
    if param.whatRanger == "Blue" then
      attack_flurry(toReturn)
    end
    -- Yellow
    if param.whatRanger == "Yellow" then
      attack_3hitCombo(toReturn)
    end
    -- Black
    if param.whatRanger == "Black" then
      attack_shieldRush(toReturn)
    end
    -- Pink
    if param.whatRanger == "Pink" then
      attack_jumpSlash(toReturn)
    end
    --
    param.battleSeq = 0 -- none!
  end

  -- Debug
  if aiDebug()
  then
    attack_singleSlash(toReturn)
    param.battleSeq = sequence.singleSlash
  end

  return toReturn
end -- OnBattleAttack()

function OnBattleTurnUpdate(battleData)
  local me = getMe()

  local updateOK = me.ColorNum >= 5 

  -- the loop --
  if battleData.sequenceTime <= param.battleTurnTimeout
  and updateOK
  then
    -- Scythe Followup --
    if param.battleSeq == sequence.test
    then
      if battleData.sequenceTime >= param.battleSeqTimeout
      then
        debugPalette("Debug_Black")
      elseif param.battleSubSeq == 0 
      then
        debugPalette("Debug_LightBlue")
      else
        debugPalette("Debug_White")
        addInput(dirs.N)
      end
    elseif param.battleSeq == sequence.singleSlash
    then
      if battleData.sequenceTime >= param.battleSeqTimeout
      then
        debugPalette("Debug_Black")
      elseif param.battleSubSeq == 0 
      then
        debugPalette("Debug_LightRed")
        --
        if me.NumTouches > 0 
        then
          param.battleSubSeq = 1
          battleData.sequenceTime = 0
        end
      elseif param.battleSubSeq == 1 
      then
        debugPalette("Debug_Red")
        --
        if not me:hasTag("BattleNewAttackInterrupt")
        and battleData.sequenceTime <= 0
        then
          debugPalette("Debug_BlinkRed")
          attack_interrupt()
        elseif not me:hasTag("Attack")
        then
          debugPalette("Debug_BlinkOrange")
          attack_shieldRush(toReturn)
          --
          battleData.sequenceTime = math.max(battleData.sequenceTime, param.battleSeqTimeout)
        else -- Hold 
          debugPalette("Debug_Blue")
          addInput(dirs.N)
        end
      else
        debugPalette("Debug_White")
        addInput(dirs.N)
      end
    end
  end
  
  -- update only if not hitstop
  if update_ok()
  then
    trace("(%i:%i) Time: %i" % { param.battleSeq, param.battleSubSeq, battleData.sequenceTime })
    battleData.sequenceTime = battleData.sequenceTime + 1
  end
end