require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function attack_turnIntoBottle(data)
  if data ~= nil then
      data.dir = dirs.U
      data.pressAction = false
  end
  doMotion("U+Attack")
end

local function attack_emergeFromBottle(data)
  if data ~= nil then
      data.dir = dirs.U
      data.pressAction = false
  end
  doMotion("U+Attack")
end

local function attack_killerBee(data)
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = false
  end
  doMotion("X+Attack")
end

local function attack_stab(data)
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = true
  end
  doMotion("X+Heruka")
end

local function attack_snap(data)
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = false
  end
  doMotion("X+Attack")
end

local function sightRangeX()
  local w = param.sightRangeX
  if param.bossBurstedMode then
    w = param.sightRangeRageX
  end
  return w
end

local function sightRangeY()
  local h = param.sightRangeY
  if param.bossBurstedMode then
    h = param.sightRangeRageY
  end
  return h
end

local function attackRangeX()
  local w = param.attackRangeX
  if param.bossBurstedMode then
    w = param.attackRangeRageX
  end
  return w
end

local function is_bottle()
  local me = getMe()
  --
  --if me:hasTag("Bottle")                  -- All
  if me:hasTag("Bottle_Idle")             -- As
  --or me:hasTag("Bottle_Snap")             -- As
  --or me:hasTag("Bottle_Transform")        -- From, To 
  or me:hasTag("Bottle_Transform_Start")  -- To
  --or me:hasTag("Bottle_Transform_End")    -- From
  then
      return true
  end
  return false
end

local function is_transforming()
  local me = getMe()
  --
  if me:hasTag("Bottle_Idle")             -- As
  or me:hasTag("Bottle_Transform_Start")  -- To
  then
      return true
  end
  return false
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()
  --
  param.turnedIntoBottleOnce = false
  param.forceGoIntoBottle = false
  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }
  --
  param.sightRangeX = 160 * 4
  param.transformRangeX = 160 * 5.5
  param.sightRangeY = { top = 160 * 3, bottom = 160 * -1 }
  param.turnAroundRangeXMin = 160 * 2 -- use sightRangeX for max
  param.attackRangeX = 160 * 2
  --
  param.sightRangeRageX = 160 * 11
  param.sightRangeRageY = { top = 160 * 4, bottom = 160 * -1 }
  param.attackRangeRageX = 160 * 4
  
  param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()
  
  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end
  
  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    debugPalette()
    --
    if got_hit() then 
      param.forceGoIntoBottle = true 
    end
    --
    -- co-op 
    if bossBursted() 
    then
      param.attackCooldown = 20
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 3 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    --
    if is_bottle() then
      debugPalette("Debug_Orange")
      --
      param.forceGoIntoBottle = false
      --
      if param.bossBurstedMode then
        attack_emergeFromBottle()
      elseif math.abs( dist_sprite(target) ) <= sightRangeX()
      and select(2,dist_sprite(target)) <= sightRangeY().top
      and select(2,dist_sprite(target)) >= sightRangeY().bottom then
        -- We're not facing the player, let's get em
        if dist_sprite(target) > 0 and dist_sprite(target) <= param.attackRangeX
        and me.FacingDir == target.FacingDir then
          attack_emergeFromBottle()
        elseif dist_sprite(target) < -param.turnAroundRangeXMin
        and dist_sprite(target) >= -param.sightRangeX
        and me.FacingDir ~= player.FacingDir then
          faceSprite(target)
          attack_emergeFromBottle()
        end
      end
    elseif not is_bottle() then
      debugPalette("Debug_Magenta")
      -- Startin' off
      if not param.turnedIntoBottleOnce 
      and not param.bossBurstedMode then
        repeat 
          debugPalette("Debug_White")
          attack_turnIntoBottle()
          param.turnedIntoBottleOnce = true
          if got_hit() then break end
          wait()
        until is_bottle()
      -- We had to have gotten hit somehow
      elseif param.forceGoIntoBottle 
      and not param.bossBurstedMode then
        debugPalette("Debug_White")
        attack_turnIntoBottle()
      -- We're sneaking up on the player
      elseif dist_sprite(target) > 0
      and dist_sprite(target) <= sightRangeX()
      and select(2,dist_sprite(target)) <= sightRangeY().top
      and select(2,dist_sprite(target)) >= sightRangeY().bottom
      and me.FacingDir == target.FacingDir then
        debugPalette("Debug_Yellow")
        if math.abs(dist_sprite(target)) <= attackRangeX() then
          faceSprite(target)
          -- we can pick one at random, really
          if random(1000) > 500 then
            attack_stab()
          else
            attack_killerBee()
          end
        else
          approach(target.XPos)
        end
      -- We're facing the player
      elseif dist_sprite(target) > sightRangeX()
      and select(2,dist_sprite(target)) <= sightRangeY().top
      and select(2,dist_sprite(target)) >= sightRangeY().bottom 
      and not param.bossBurstedMode then
        debugPalette("Debug_Cyan")
        attack_turnIntoBottle()
      elseif dist_sprite(target) < 0
      and select(2,dist_sprite(target)) <= sightRangeY().top
      and select(2,dist_sprite(target)) >= sightRangeY().bottom
      and not param.bossBurstedMode
      and me.FacingDir ~= target.FacingDir then
        debugPalette("Debug_Green")
        attack_turnIntoBottle()
      elseif math.abs( dist_sprite(target) ) <= sightRangeX()
      and select(2,dist_sprite(target)) <= sightRangeY().top
      and select(2,dist_sprite(target)) >= sightRangeY().bottom 
      and not param.bossBurstedMode then
        debugPalette("Debug_Red")
        attack_turnIntoBottle()
      end
    else
      debugPalette("Debug_BlinkFast")
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  -- battle info --
  if battleData.bottleTurns == nil
  then
    battleData.bottleTurns = 0
  end

  if battleData.turnsSinceBottle == nil
  then
    battleData.turnsSinceBottle = 0 
  end

  -- battle info update
  if me:hasTag("Bottle_Idle")
  then
    battleData.turnsSinceBottle = 0
    battleData.bottleTurns = battleData.bottleTurns + 1
  else
    battleData.turnsSinceBottle = battleData.turnsSinceBottle + 1
    battleData.bottleTurns = 0
  end

  if battleData.bottleTurns > 2
  then
    attack_emergeFromBottle(toReturn)
  elseif me:hasTag("Bottle_Idle")
  then
    attack_snap(toReturn)
    -- if we're also past this limit, pass meterBurn to exit quick --
    toReturn.meterBurn = false
    if battleData.bottleTurns >= 2 
    then
      toReturn.meterBurn = true
    end
  end

  if aiDebug()
  and not me:hasTag("Bottle_Idle")
  then
    attack_turnIntoBottle(toReturn)
  end

  return toReturn
end --OnBattleAttack()