require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

-- Attacks --
local function attack_arrow(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_launcher(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_grab(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_super(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = true
  end
  doMotion("X+Heruka")
end

-- sequence keys
local sequence = {
  none = 0,
  arrow = 1,
  launcher = 2,
  -- supers --
  super = 3
}

-- the initialization function for this AI
function OnInit()
  param.time = 0
  --
  param.battleSeq = 0  
  param.battleSubSeq = 0
  param.battleMaxTimeout = 60 * 5 -- just in case
  param.battleTimeout = 60 * 1 -- default for our own sequence
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  while true do
    me, player = getMe(), getPlayer()

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  -- sequence picker
  local maxSequences = 2
  param.battleSeq = random(maxSequences)
  param.battleSubSeq = 0 
  trace(param.battleSeq)

  if aiDebug()
  then
    param.battleSeq = sequence.launcher
  end
  
  param.time = 0

  -- attack difficulty
  if battleData.attackLevel == nil
  then
    battleData.attackLevel = 0 
  end

  if me.LifePct <= 25/100
  then
    battleData.attackLevel = 2 
  elseif me.LifePct <= 66/100
  then
    battleData.attackLevel = 1
  else
    battleData.attackLevel = 0 
  end

  -- reset commands
  toReturn.meterBurn = false
  toReturn.pressAction = false
  toReturn.pressDash = false

  -- sequence assignment
  if param.battleSeq == sequence.arrow
  then
    attack_arrow(toReturn)
  elseif param.battleSeq == sequence.launcher
  then
    attack_launcher(toReturn)
  else
    attack_arrow(toReturn)
  end

  return toReturn
end --OnBattleAttack()

function OnBattleTurnUpdate(battleData)
  local me = getMe()
  local level = battleData.attackLevel

  -- the loop --
  if param.time <= param.battleMaxTimeout
  then
    -- Scythe Followup --
    if param.battleSeq == sequence.arrow
    then
      if param.time >= param.battleTimeout
      then
        debugPalette("Debug_Black")
        -- Do nothing, which exits --
      -- [0] Arrow 1 --
      elseif param.battleSubSeq == 0 
      then
        debugPalette("Debug_LightBlue")
        --
        if level >= 1
        and me:hasTag("BowShot_Release")
        then
          param.battleSubSeq = 1
          param.time = 0
        end 
      -- [1] Arrow 2
      elseif param.battleSubSeq == 1
      then
        debugPalette("Debug_Blue")
        --
        if param.time <= 1
        then
          attack_arrow()
        end

        if level >= 2
        and me:hasTag("BowShot_Release")
        and param.time >= 30 -- ensure we had time to go from Release to Start
        then
          param.battleSubSeq = 2
          param.time = 0
        end 
      -- [2] Arrow 3
      elseif param.battleSubSeq == 2 
      then
        debugPalette("Debug_DarkBlue")
        --
        if param.time <= 1
        then
          attack_arrow()
        end
      else
        debugPalette("Debug_White")
        addInput(dirs.N)
      end
    elseif param.battleSeq == sequence.launcher
    then
      if param.time >= param.battleTimeout
      then
        debugPalette("Debug_Black")
        -- Do nothing, which exits --
      elseif param.battleSubSeq == 0 
      then
        debugPalette("Debug_LightRed")
        --
        if me.NumHits > 0
        and me.NumHits == me.NumTouches -- if you didn't hit everyone
        and level >= 1
        then
          debugPalette("Debug_Green")
          --
          param.time = 0
          param.battleSubSeq = 1
        elseif me.NumTouches > 0
        and level >= 1
        then
          debugPalette("Debug_Red")
          --
          param.time = 0
          param.battleSubSeq = -1
        end
      -- [1](1) On Hit - Grab
      elseif param.battleSubSeq == 1
      then
        debugPalette("Debug_Orange")
        --
        if not me:hasTag("BattleNewAttackInterrupt")
        and param.time <= 2
        then
          doMotion("X+BattleInterrupt")
        elseif not me:hasTag("AirGrab")
        and param.time > 2
        and param.time <= 15
        then
          attack_grab()
        elseif level >= 2
        and param.time > 15
        then
          param.time = -80 -- bonus time
          param.battleSubSeq = 2
        end
      -- [1](2) Follow-up Arrow
      elseif param.battleSubSeq == 2
      then
        debugPalette("Debug_DarkOrange")
        --
        if not me:hasTag("BattleNewAttackInterrupt")
        and me:hasTag("AirGrab_Toss")
        and me.Throwing == 0 
        then
          doMotion("X+BattleInterrupt")
        elseif not me:hasTag("BowShot")
        and param.time > 0 
        and param.time <= param.battleTimeout
        then
          debugPalette("Debug_BlinkOrange")
          attack_arrow()
        end
      -- [1](-1) On Block - Arrow
      elseif param.battleSubSeq == -1
      then
        debugPalette("Debug_Magenta")
        --
        if not me:hasTag("BattleNewAttackInterrupt")
        and param.time <= 2
        then
          doMotion("X+BattleInterrupt")
        elseif param.time > 2
        and param.time <= 7
        then
          addInput(dirs.N)
        elseif not me:hasTag("BowShot")
        and param.time > 7
        and param.time <= param.battleTimeout
        then
          attack_arrow()
        end
      else
        debugPalette("Debug_White")
        addInput(dirs.N)
      end
    end
  end
  
  -- update only if not hitstop
  if update_ok()
  then
    param.time = param.time + 1
  end
end