require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function trace( str )
  if aiDebug() then print( str ) end 
end

local function hurtbox(alt)
  local me = getMe()
  local width = math.max(me.WorldCollisionWidth,320)
  local height = math.max(me.WorldCollisionHeight,320)
  local color = alt and {0,255,0} or {255,255,255}
  drawBox(
    me.XPos - width/2,
    me.XPos + width/2,
    me.YPos + 5120,
    me.YPos - 5120,
    0,
    -256,
    80,80,80
  )
  drawBox(    
    me.XPos - width/2,
    me.XPos + width/2,
    me.YPos + height,
    me.YPos,
    1,
    -1,unpack(color)
  )
  drawBox(    
    me.XPos - 16,
    me.XPos + 16,
    me.YPos + 32,
    me.YPos,
    1,
    -1,
    255,
    0,
    255
  )
end

local function attack_swipe(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_arrowShot(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_acidSpit(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_slam(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function attack_trample(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Heruka")
end

local function attack_scream(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Heruka")
end

-- DASH --

local function attack_escape(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("X+Dash")
end

local function attack_escapeQuick(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("U+Dash")
end

-- OFF SCREEN IDLE --

local function attack_screamOffscreen(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_shot(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_tripleShot(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_createMindayar(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+ShootBtn")
end

local function attack_fallDown(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+ShootBtn")
end

local function ResetBattle()
  local me, player = getMe(), getPlayer()
  
  param.time = 0
  param.cooldown = 0
  param.boss_phase = 0
  param.attack = 1
  --
  param.origin = { x = me.XPos, y = me.YPos }
  --
  param.initiatedFight = false
  param.hasLeftBattleOnce = false
  param.canMoveInCanopy = false
  param.canDropFromCanopy = false
  param.canAttackFromCanopy = false

  param.canSpawnMindayar = false
  param.spawnedMindayar = false
  param.numMindayarCreated = 0
  param.maxMindayarToCreate = 8

  param.trampleFlag = false
  param.hasDoneOneAttackLoop = false
  param.canScream = true

  -- 
  param.wantsQuickRecovery = false
  param.weDied = false 

  --
  param.landingWaitTime = 33
  param.leapToOffscreenTime = 41 
  param.quickLeapToOffscreenTime = 25 

  param.total_dmg = 0
  param.prev_HealthPct = me.HealthPct
  param.prev_life = me.Lifebar
  param.prev_bossPhase = 0
  --
  param.leftBattleAtTiers = {
    [0] = false,
    false, 
    false,
    false
  }
  -- This value changes the damage of Anchor.
  me.LightY = 0
end

local function SetDamageMult()
  local me = getMe()

  -- we don't really need to track anything except base damage
  local tiers = {
    1,
    1,
    1,
    1
  }
  if param.isHardMode then
    tiers = { 
      1,
      1.2,
      1.5,
      1.75
    }
  end
  me.DamageMult = tiers[param.boss_phase+1] * param.base_dmg
end

local function SetSpeedMult()
  local me = getMe()
  -- we don't really need to track anything except base damage
  local tiers = {
    1,
    1.1,
    1.2,
    1.5
  }
  if param.isHardMode then
    -- VERY OBVIOUSLY mean
    tiers = {
      1,
      1.35,
      1.7,
      2
    }
  end
  me.BattleSpeedMult = tiers[param.boss_phase+1] * param.base_speed
end

local function SetAnchorLife()
  -- We use LightG to set the Anchor life
  local me = getMe()
  -- Ajna Neutral Axe is 850
  local hp_tiers = {
    51,    -- 1 hit
    50 * 3,   -- 3 hits
    50 * 3,   -- 3 hits
    50 * 5,   -- 5 hits
  }
  me.LightG = hp_tiers[1]
  if param.canDropFromCanopy then
    me.LightG = hp_tiers[param.boss_phase+1]
  end
end

local function ApplyPalette()
  local me = getMe()
  if me.HealthPct < param.phase3_healthCap and me.HealthPct > 0 then
    me.PalFXSelf = hash("Boss_Weakened")
    me.PalFXSelfTime = 60 * 60
  elseif me:hasTag("DeathAnimation")
  or me.Lifebar == 1
  or me.HealthPct <= 0
  then
    me.PalFXSelf = 0
    me.PalFXSelfTime = 0
  end
end

local function HasWaitSpawnCommand()
  local me = getMe() 
  local hash = hash("WaitForBattle")
  if me.SpawnCommand == hash 
  or me.SpawnCommand2 == hash
  then return true end
  return false
end

local function MindayarCount()
  local me, player = getMe(), getPlayer()

  local _size = 8000
  local sprs = getSpritesInBox(
    me.XPos - _size, me.XPos + _size,
    me.YPos + _size, me.YPos - _size,
    0, me
  )

  local _count = 0
  for i, spr in ipairs(sprs) do
    if spr.MyName == hash("Mindayar") then
      _count = _count + 1
    end
  end

  return _count
end

function OnInit()
  param.isHardMode = false
  param.phase1_healthCap = 0.67        -- 66.67%
  param.phase2_healthCap = 0.33        -- 33.33%
  param.phase3_healthCap = 0.12        -- 12.00%
  param.phase2_turns = 4
  param.phase3_turns = 3
  param.phase4_turns = 3

  ResetBattle()
  --
  local me = getMe()

  param.base_hp = me.MaxLife
  param.base_dmg = me.DamageMult
  param.base_speed = me.BattleSpeedMult

  SetDamageMult()
  SetSpeedMult()

  if aiDebug()
  then
    param.initiatedFight = true
    param.hasLeftBattleOnce = true
    param.canSpawnMindayar = false
    param.canMoveInCanopy = true
    param.canDropFromCanopy = true
    param.canScream = true
    --
    param.leftBattleAtTiers[0] = true
    --param.leftBattleAtTiers[1] = true
    --
    param.boss_phase = 1
    --
    me.Lifebar = me.MaxLife
    me.Lifebar = me.Lifebar * 0.55
  end
end

function OnUpdate()
  local me, player = getMe(), getPlayer()
  local players = getPlayers()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end

  local dist_playerX = function() 
    local dist = player.XPos - me.XPos
    dist = dist * me.FacingDir
    return dist     -- world space
  end

  local dist_playerY = function() 
    return player.YPos - me.YPos
  end
  
  local dist_wall = function()
    return ( 
      ((me.WorldCollisionWidth / 2) + param.wallDistCheck) * me.FacingDir ) 
      + me.WallDistAhead
  end
  local function anchor_pos()
    for _, child in ipairs( me:getChildren() ) do
      if child:hasTag("AnchorIdle") then 
        return child.XPos, child.YPos
      end
    end
    return 0
  end

  local function anchor_exists()
    for _, child in ipairs( me:getChildren() ) do
      if child:hasTag("AnchorIdle") then 
        return true
      end
    end
    return false
  end

  local function anchor_dist()
    if anchor_exists() then
      local ax, ay = anchor_pos()
      return ax - me.XPos
    end
    return 0
  end

  local dist_playerX_anchorX = function()
    local player = player or getPlayer()
    if not anchor_exists() then return nil end
    local ax = anchor_pos()
    return ax - player.XPos
  end

  local function approach(xpos, cond)
    local me = getMe()
    local far = cond or false
    local dist = xpos - me.XPos
    dist = dist * me.FacingDir
    if dist > 0 then
      if far then 
        addInput(dirs.F, btns.Dash)
      else
        addInput(dirs.F)
      end
    else
      if far then 
        addInput(dirs.B, btns.Dash)
      else
        addInput(dirs.B)
      end
    end
  end

  local function approach_quick(xpos, cond)
    local far = cond or false
    local dist = xpos - me.XPos
    dist = dist * me.FacingDir
    if dist > 0 then
      if far then 
        addInput(dirs.F, btns.MeterBurn)
      else
        addInput(dirs.F, btns.Dash)
      end
    else
      if far then 
        addInput(dirs.B, btns.MeterBurn)
      else
        addInput(dirs.B, btns.Dash)
      end
    end
  end

  local function attack_cooldown()
    param.cooldown = 90
    if param.isHardMode then 
      param.cooldown = 67
    end
  end

  local function attack_pattern_to_get()
    local pattern = {
      2,1,2,3,1,5,1,10
    }
    return pattern[param.attack % (#pattern+1)]
  end

  -- The actual AI process
  -- Debug stuff
  local show_hitbox = false
  local show_anchor = false
  local show_cooldown = false
  local start_debug = false

while true do
    me, player = getMe(), getPlayer()
    -- Debug 
    if aiDebug() then
      --player.DamageMult = 3
      if show_hitbox then hurtbox() end
      if show_anchor then
        if anchor_exists() then
          local x, y = anchor_pos()
          drawBox(x - 64, x + 64,
                  y + 128, y,
                  1,-1,
                  255,255,255)
        end
      end
      if show_cooldown then
        local width = 480
        local max_shown = 60 
        local fill = math.max(0,math.min(max_shown, param.cooldown)) / max_shown
        drawBox(
          player.XPos - width/2,
          player.XPos + width/2,
          player.YPos + player.WorldCollisionHeight + 64, 
          player.YPos + player.WorldCollisionHeight + 32,
          1,-1,
          0,0,0
        )
        if fill > 0 then
          drawBox(
            player.XPos - width/2,
            player.XPos - width/2 + (width * fill),
            player.YPos + player.WorldCollisionHeight + 64, 
            player.YPos + player.WorldCollisionHeight + 32,
            1,-1,
            255, 255 - fill*255, 255 - fill*255
          )
        end
      end
      if start_debug then
        trace("[WARNING] Debug information SET")
        param.boss_phase = 2
        param.canAttackFromCanopy = true
        param.hasLeftBattleOnce = true
        param.canDropFromCanopy = true
        param.canMoveInCanopy = true 
        start_debug = false
      end
      draw_bar(me.TotalStun, 0, 240)
    end
    -- Update per-frame ticks
    SetDamageMult()
    SetSpeedMult()

    param.time = param.time + 1
    param.prev_life = me.Lifebar
    param.prev_bossPhase = param.boss_phase 
    me.LightY = 0
    if me:hasTag("Off_Screen_Idle") and me.Hitstun <= 0 then
      param.cooldown = math.max(0,param.cooldown - 1)
    end
    -- If the player is dead, we set out despawn dist very low so we can
    -- respawn
    if ( player.HealthPct <= 0 or player:hasTag("Dead") ) and me.WasInBattle <= 0 then
      me.DespawnDistX = 10 
    end

    if has_spawn_command("TestSpawn") then
      param.time = param.time + 1
      if me:hasTag("Off_Screen_Idle") then
        if param.numMindayarCreated < param.maxMindayarToCreate
        and param.cooldown <= 0 
        then
          trace("Spawning")
          attack_createMindayar()
          param.numMindayarCreated = param.numMindayarCreated + 1
          param.cooldown = 6
        elseif param.numMindayarCreated >= param.maxMindayarToCreate
        and not param.spawnedMindayar 
        then
          trace("Done spawn")
          param.spawnedMindayar = true
        elseif param.spawnedMindayar 
        and MindayarCount() <= 0
        then
          trace("SCREAM")
          attack_fallDown()
        end
      elseif me.OnGround > 0 then
        attack_escapeQuick()
        param.spawnedMindayar = false
        param.numMindayarCreated = 0 
        param.cooldown = 80
      else
        --attack_escapeQuick()
      end
    -- Spawn setup 
    elseif not param.initiatedFight then
      if  math.abs(dist_playerX()) < 160 * 3
      and math.abs(dist_playerY()) < 160 * 56
      and not me:hasTag("Attack")
      then
        attack_slam()
        param.cooldown = 300
        param.initiatedFight = true
      end
    -- When we're in off screen mode
    elseif me:hasTag("Off_Screen_Idle") then
        local anchor_x, anchor_y = nil, nil
        if anchor_exists() then anchor_x, anchor_y = anchor_pos() end
        if  param.canAttackFromCanopy
        and param.cooldown <= 0 then
          -- spam info for now
          local attack = attack_pattern_to_get()
          trace("[INFO] attack is " .. tostring(attack))
          -- single shot x 5
          if attack == 1 then
            local time = 0
            -- startup
            for i = 1, (param.isHardMode and 7 or 5) do 
              time = 0 
              while time < (param.isHardMode and 22 or 35) do
                time = time + 1
                approach_quick(player.XPos,math.abs( dist_playerX() ) >= 160 * 1)
                wait()
              end
              attack_shot()
              wait()
            end
            param.attack = param.attack + 1
            attack_cooldown()
          -- triple shot x 4
          elseif attack == 2 then
            local time = 0
            for i = 1, 4 do 
              time = 0 
              while time < (param.isHardMode and 45 or 60) do
                time = time + 1
                approach_quick(player.XPos, math.abs( dist_playerX() ) >= 160 * 1)
                wait()
              end
              attack_tripleShot()
              wait()
            end
            param.attack = param.attack + 1
            attack_cooldown()
          -- slamdown
          elseif attack == 3 then
            local time = 0
            while time < 60 do
              time = time + 1
              approach_quick(player.XPos, math.abs( dist_playerX() ) >= 160 * 1 )
              wait()
            end
            wait(35)
            attack_slam()
            param.wantsQuickRecovery = true
            param.attack = param.attack + 1
            attack_cooldown()
          -- sweeping shot
          elseif attack == 4 then
            local time = 0
            local flipOK = 0
            local dir = dirs.F
            while time < 15 do
              time = time + 1
              if math.abs( dist_playerX() ) > 160 * 3 then
                approach(player.XPos)
              end
              wait()
            end
            if dist_playerX() < 0 then dir = dirs.B end  
            time = 0 
            for i = 1, (param.isHardMode and 15 or 9) do
              attack_shot()
              wait()
              while time < 13 do
                if (me.WallDistAhead < 50 or me.WallDistBehind < 50)
                and flipOK <= 0 then
                  if dir == dirs.F then 
                    dir = dirs.B
                  elseif dir == dirs.B then 
                    dir = dirs.F 
                  end 
                  flipOK = 15
                end
                addInput(dir, btns.Dash)
                wait()
                time = time + 1
              end
              time = 0
              flipOK = math.max(0, flipOK - 1)
            end
            param.attack = param.attack + 1
            attack_cooldown()
          -- fullscreen splat
          elseif attack == 5 then
            local time = 0
            local shots = 0
            local safe_shot = 0
            local lockout = 10 -- This keeps us from caring about a wall for a short period of time
            while time < 10 do wait() time = time + 1 end
            time = 0 
            local dir = dirs.N
            -- Find a corner!
            if dist_playerX() > 0 then dir = dirs.F else dir = dirs.B end
            -- If we picked right or left, is the player character there?
            repeat
              addInput(dir, btns.MeterBurn)
              wait()
              lockout = math.max(0, lockout - 1)
            until ( me.WallDistAhead < 50 or me.WallDistBehind < 50 ) 
            --
            while time < 10 do wait() time = time + 1 end
            -- Flip the direction
            if dir == dirs.F then 
              dir = dirs.B
            elseif dir == dirs.B then 
              dir = dirs.F 
            end 
            -- Go the OTHER way!
            -- Fudge our input.. for 5f just to make sure we're not in the corner!
            time = 0
            for i = 1, 3 do
              addInput(dir)
              wait()
            end
            -- Start rushing. Choose a random safe spot HERE
            safe_shot = 2 + random(4)
            local safe_shot2 = safe_shot + 1 + random(4)
            repeat
              time = time + 1
              if time < (param.isHardMode and 4 or 7) then
                addInput(dir, btns.MeterBurn)
              else
                time = 0
                if shots ~= safe_shot and shots ~= safe_shot2 then 
                  attack_shot()
                end
                shots = shots + 1
              end
              wait()
            until me.WallDistAhead <= 40 or me.WallDistBehind <= 40 
            -- just ONE more shot too
            wait(3)
            attack_tripleShot()
            --
            param.attack = param.attack + 1
            attack_cooldown()
          elseif attack == 10 then
            -- This will be SLAMMO
            wait(15)
            local dir = dirs.N
            --
            if dist_playerX() > 0 then dir = dirs.F else dir = dirs.B end
            repeat
              addInput(dir, btns.MeterBurn)
              wait()
            until me.WallDistAhead < 50 or me.WallDistBehind < 50 
            --
            wait(30)
            attack_slam()
            wait()
            param.trampleFlag = true
            param.wantsQuickRecovery = anchor_exists()
          else
            wait(60)
            param.attack = param.attack + 1 
            attack_cooldown()
          end
        elseif param.canSpawnMindayar then
          if param.numMindayarCreated < param.maxMindayarToCreate 
          and param.cooldown <= 0 
          and not param.spawnedMindayar 
          then
            trace("Spawning")
            attack_createMindayar()
            param.numMindayarCreated = param.numMindayarCreated + 1
            param.cooldown = 11
          elseif param.spawnedMindayar 
          and MindayarCount() <= 0
          then
            trace("SCREAM")
            attack_fallDown()
          elseif MindayarCount() < param.maxMindayarToCreate 
          and param.cooldown <= 0 
          and param.spawnedMindayar 
          then
            trace("Spawning EXTRA")
            attack_createMindayar()
            param.numMindayarCreated = param.numMindayarCreated + 1
            param.cooldown = 180
          elseif param.numMindayarCreated >= param.maxMindayarToCreate
          and not param.spawnedMindayar 
          then
            trace("Done spawn")
            param.spawnedMindayar = true
          end
        elseif param.canMoveInCanopy then
          if param.canDropFromCanopy
          and param.cooldown <= 0 
          and math.abs( dist_playerX() ) <= 160 * 2 then
            -- Fudge the startup
            local time = 0
            -- Scream! And make sure we can scream on ground
            attack_screamOffscreen()
            param.canScream = true
            wait()
            
            -- Delay
            wait(10) 
            attack_slam()
            wait()
            if anchor_exists() then
              param.wantsQuickRecovery = true
            end
          elseif math.abs( dist_playerX() ) >= 160 * 0.5 then
            approach(player.XPos, math.abs( dist_playerX() ) >= 160 * 4 )
          end
        end
    -- Escape is a good time to unflag anything.
    elseif me:hasTag("Escape") or me:hasTag("DropSlam_HopUp") then
      SetAnchorLife()
      param.numMindayarCreated = 0
      param.spawnedMindayar = false

      if param.wantsQuickRecovery then 
        param.wantsQuickRecovery = false 
      end
    -- I must be on the ground
    elseif me.OnGround > 0 and not me:hasTag("Hitstun") and not me:hasTag("Escape") then
        local function EarlyBail() 
          if me:hasTag("Hitstun") then return true end
          if me.OnGround <= 0 then return true end
          if me.WasInBattle > 0 then return true end
          if me:hasTag("Escape") then return true end
          if bossBursted() then
            trace("BOSS BURST ESCAPE")
            return true 
          end
          return false
        end
        -- Land and wait a tiny bit
        local time = 0 
        while time < param.landingWaitTime do
          time = time + 1
          if EarlyBail() then break end
          wait()
        end
        if param.trampleFlag then
          while not me:hasTag("Trample_Loop") do
            attack_trample()
            wait()
            if me:hasTag("Hitstun") then break end
            if me.OnGround <= 0 then break end
            if me.WasInBattle > 0 then break end
          end
          -- If we haven't unlocked Drop, let's do that now!
          param.hasDoneOneAttackLoop = true
          param.trampleFlag = false
          param.attack = 1
        elseif param.wantsQuickRecovery then
          debugPalette("Debug_Blink")
          time = 0 
          attack_scream()
          while time <= param.quickLeapToOffscreenTime do
            time = time + 1
            if EarlyBail() then break end
            wait()
          end
          SetAnchorLife()
          attack_escapeQuick()
        else
          time = 0
          -- Let's scream here
          if param.canScream
          then
            attack_scream()
            param.canScream = false
          end
          wait(3)
          while time <= param.leapToOffscreenTime
          do
            time = time + 1
            if EarlyBail() then break end
            wait()
          end
          SetAnchorLife()
          -- This used to be slow 
          attack_escapeQuick()
        end
        if param.boss_phase < 2 then
          param.cooldown = 104
        else
          param.cooldown = 60
          if param.isHardMode then 
            param.cooldown = 39
          end
        end
    end
    wait()
  end -- while
end -- OnUpdate()

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  SetDamageMult()
  SetSpeedMult()
  SetAnchorLife()

  -- Make sure we don't chain any attacks, at first
  me.LightY = 0 

  -- Single check to make sure we can Drop from balcony in Phase 1
  if param.hasLeftBattleOnce and me.HealthPct <= param.phase1_healthCap then
    trace("[INFO] We can NOW drop from canopy")
    param.canDropFromCanopy = true
  end


  -- Phase 4 Trigger
  if me.HealthPct <= param.phase3_healthCap
  and not param.leftBattleAtTiers[2]
  then
    debugPalette("Debug_Red")
    trace("Phase 4 Trigger!")
    -- Quick Leave
    attack_escapeQuick(toReturn)
    --toReturn.dir = dirs.U
    --toReturn.pressAction = true
    -- Ensure we're in Phase 4
    if param.boss_phase < 3
    then
      param.boss_phase = 3 
    end

    param.cooldown = 100
    -- Just In Case
    param.canAttackFromCanopy = true
    param.canSpawnMindayar = true
    param.canMoveInCanopy = true
    param.canDropFromCanopy = true
    param.hasLeftBattleOnce = true
    param.canScream = true
    param.leftBattleAtTiers[2] = true
    param.leftBattleAtTiers[1] = true
    param.leftBattleAtTiers[0] = true
  -- Phase 3 Trigger
  elseif me.HealthPct <= param.phase2_healthCap
  and not param.leftBattleAtTiers[1]
  then
      debugPalette("Debug_Orange")
      trace("Phase 3 Trigger!")
      -- Quick Leave
      attack_escapeQuick(toReturn)
      --toReturn.dir = dirs.U
      --toReturn.pressAction = true
      -- Ensure we're in Phase 3
      if param.boss_phase < 2
      then
        param.boss_phase = 2 
      end
      -- Just In Case
      param.canSpawnMindayar = true
      param.canMoveInCanopy = true
      param.canDropFromCanopy = true
      param.hasLeftBattleOnce = true
      param.canScream = true
      param.leftBattleAtTiers[1] = true
      param.leftBattleAtTiers[0] = true
  -- Phase 2 Trigger
  elseif me.HealthPct <= param.phase1_healthCap
  and not param.leftBattleAtTiers[0]
  --and param.boss_phase < 1
  then
    debugPalette("Debug_Green")
    trace("Phase 2 Trigger!")
    -- Escape
    attack_escapeQuick(toReturn)
    --toReturn.dir = dirs.U
    --toReturn.pressAction = false
    -- We SHOULD be able to do this now!!!!!!!
    if param.boss_phase < 1
    then
      param.boss_phase = 1
    end 
    param.canMoveInCanopy = true
    param.canDropFromCanopy = true
    param.hasLeftBattleOnce = true
    param.canScream = true
    param.leftBattleAtTiers[0] = true
  else
    debugPalette("Debug_Cyan")
    trace("Regular Attacks!")

    local rand_direction = random(1000)
    local rand_special = random(1000)

    -- Base Choices: Arrow Shots
    if rand_direction <= 500
    then
      attack_swipe(toReturn)
    else
      attack_arrowShot(toReturn)
    end

    -- Special Boss junk
    if (param.boss_phase >= 1 and rand_special > 500) then
      debugPalette("Debug_Yellow") -- WE ATTEMPTED to GO HERE
      -- Down
      if rand_direction <= 333 
      and param.boss_phase > 1
      then
        attack_swipe(toReturn)
      -- Phase 2 Spit ( REPLACE WITH DROPSLAM )
      elseif rand_direction >= 667
      and param.boss_phase >= 1
      then
        attack_acidSpit(toReturn)
      end
    end
    
    -- If we're at a turn limit, let's ensure we have the ability to leave
    -- after this attack
    --[[
    if param.boss_phase >= 2 and battleData.turnNum >= param.phase3_turns
    or param.boss_phase >= 3 and battleData.turnNum >= param.phase4_turns
    then
        me.LightY = 1
        local rand_doHop = random(1000)
        if rand_doHop > 500 then
            me.LightY = 2
        end
    end
    ]]

    param.cooldown = 100
  end

  trace(string.format("HP %03i/%03i (%2.2f%%) Turn %i Phase %i Speed %2.2fx Dmg %2.2fx LightY %i",
    me.Lifebar,
    me.MaxLife,
    me.HealthPct * 100,
    battleData.turnNum,
    param.boss_phase,
    me.BattleSpeedMult,
    me.DamageMult,
    me.LightY
    )
  )

  return toReturn

end -- OnBattleAttack()