require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)
  
  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function setScaredPos()
  param.lastScaredPos = {x = me.XPos, y = me.YPos }
end

local function getScaredPos()
  return param.lastScaredPos.x, param.lastScaredPos.y
end

local function dist_scared()
  return (param.lastScaredPos.x - me.XPos) * me.FacingDir, param.lastScaredPos.y - me.YPos
end

-- Attacks --
local function attack_slap(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
  param.canAttackTimer = 15
end

local function attack_buttKick(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.canAttackTimer = 15
end

local function attack_footballKick(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
  param.canAttackTimer = 15
end

local function attack_kiss(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
  param.canAttackTimer = 15
end

local function attack_buttStomp(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Heruka")
  param.canAttackTimer = 15
end

local function slapAttackOK()
  local me = getMe()

  if me.ColorNum >= 2
  then
    return true
  else
    return false
  end

  return false
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()
  -- World AI
  param.origin = { x = me.XPos, y = me.YPos }
  --
  param.origin = { x = 0, y = 0 }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  param.scared = false
  param.scaredTimer = 0
  param.scaredBaseTime = 45
  param.startled = false 
  param.canAttackTimer = 0
  param.canTurnAroundTimer = 0 
  param.lastScaredPos = { x = 0, y = 0 }
  --
  param.roamRangeX = 160 * 4
  param.sightRangeX = 160 * 10
  param.sightRangeY = { top = 160 * 3.3, bottom = 160 * -1.5 }
  param.attackRangeX = 160 * 4
  param.attackRangeY = { top = 160 * 3.3, bottom = 160 * -1.5 }
  --
  param.sightRangeRageX = 160 * 19
  param.sightRangeRageY = { top = 160 * 6, bottom = 160 * -1.5 }
  param.attackRangeRageX = 160 * 4
  param.attackRangeRageY = { top = 160 * 3.3, bottom = 160 * -1.5 }
  -- Battle Attack Tracking
  param.numHandslaps = 0
  param.numAdditionalAttacks = 0
  param.turnsTracked = 0
  -- SpawnCommands
  param.cmd_runAway = hash("RunAway")
  
  param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end

  -- Core loop
  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    if aiDebug() then
      draw_bar(param.scaredTimer, 0, 120)
      draw_bar(param.canAttackTimer, 0, 120, 1, { 255, 255, 255})
    end

    if got_hit(me) then
      param.scared = true
      param.scaredTimer = param.scaredBaseTime
    end

    if tick_ok() then
      param.canTurnAroundTimer = math.max(0, param.canTurnAroundTimer - 1)
      param.canAttackTimer = math.max(0, param.canAttackTimer - 1)
    end

    if tick_ok()
    and not isInSightRange()
    and me.OnGround > 0
    then
      param.scaredTimer = math.max(0, param.scaredTimer - 1)
    end

    -- co-op 
    if bossBursted() 
    then
      param.scaredTimer = 0
      param.canAttackTimer = 0 
      param.canTurnAroundTimer = 0 
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 5 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    
    -- Start Here
    if update_ok() then
      debugPalette("Debug_Blue")
      if me.SpawnCommand == param.cmd_runAway
      then
        -- don't DO anything
        wait()
      -- Not Scared, Back
      elseif param.scared then
        debugPalette("Debug_Cyan")
        -- Jump first
        if param.startled then
          --
          param.startled = false
          turn_around(10)
          wait(7)
          addInput(dirs.N, btns.Jump)
          local _t = 0
          while _t < 5 
          and tick_ok()
          do
            _t = _t + 1
            addInput(dirs.N, btns.Jump)
            wait()
          end
          -- wait even longer --
          while me.OnGround <= 0  
          and tick_ok()
          do
            wait()
          end
        elseif obstacle_in_front()
        then
          debugPalette("Debug_White")
          turn_around(10)
          param.canTurnAroundTimer = 15
        -- Let's move
        elseif isInSightRange(target)
        then
          -- if in front, turn around
          if dist_sprite(target) > 0
          and dist_sprite(target) <= 160 * 2.75
          and param.canTurnAroundTimer > 0
          and param.canAttackTimer <= 0
          then
            attack_slap()
            -- we shouldn't cheat but ok
            param.canAttackTimer = 2
            param.canTurnAroundTimer = 4
          elseif dist_sprite(target) > 0
          and param.canTurnAroundTimer <= 0
          then
            debugPalette("Debug_Orange")
            turn_around(10)
            param.scaredTimer = param.scaredTimer + 2
          else
            debugPalette("Debug_Green")
            addInput(dirs.F, btns.Dash)
          end
        elseif param.scaredTimer > 0
        then
          debugPalette("Debug_Dark")
          addInput(dirs.F, btns.Dash)
        else
          param.scared = false
        end
      elseif isInSightRange(target)
      then
        debugPalette("Debug_Yellow")
        --
        param.scared = true
        param.scaredTimer = param.scaredBaseTime
        --
        if dist_sprite(target) < 0
        and not param.bossBurstedMode 
        then
          debugPalette("Debug_Magenta")
          param.startled = true
        end
      else
      end
    else
      debugPalette("Debug_Blink")
    end
    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  local dirChance = random(1000)
  local actionChance = random(1000)

  -- Default the attacks to Kicks
  attack_footballKick(toReturn)

  if actionChance > 625
  and slapAttackOK()
  then
    attack_slap(toReturn)
  elseif dirChance > 667
  then
    attack_buttKick(toReturn)
  elseif dirChance < 334
  then
    attack_buttStomp(toReturn)
  end

  return toReturn
end --OnBattleAttack()