require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function dist_grazePoint()
  local me = getMe()
  local dx = param.grazePoint.x - me.XPos
  local dy = param.grazePoint.y - me.YPos
  return dx, dy 
end

local function attack_hornStab(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
  end
  doMotion("X+Attack")
end

local function attack_shofar(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
  end
  doMotion("U+Attack")
end

local function attack_graze(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
  end
  doMotion("D+Attack")
end

local function attack_stomp(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
  end
  doMotion("X+Heruka")
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()
  local players = getPlayers()
  -- Set origin
  param.time = 0
  param.blockTime = 0 
  param.attackCooldown = 60 * 3 + random(60 * 5)
  param.origin = { x = me.XPos, y = me.YPos }

  param.arrowBlockDist = 160 * 10
  param.counterDist = 160 * 5
  param.grazeOffsetRange = 160 * 6

  param.grazePoint = { x = param.origin.x, y = param.origin.y } 
  param.grazePoint.x = param.origin.x - (param.grazeOffsetRange) + random(param.grazeOffsetRange * 2)
  
  param.timesBlocked = 0 
  param.counterAttack = false
  
  param.bossBurstedMode = false

  --
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()
  
  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end
  
  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    if aiDebug() then
      local time = 60 * 10
      draw_bar(param.timesBlocked, 0, 5, 4, {255, 255, 0})
      draw_bar(me.Alerted, 0, 30, 3, {255, 255, 255})
      draw_bar(param.blockTime, 0, 60, 2, {0, 0, 255})
      draw_bar(param.time, 0, 300, 1, {0, 255, 255})
      draw_bar(param.attackCooldown, 0, 60 * 8, 0, {255, 0, 0})
      --
      draw_point(param.origin.x, param.origin.y)
      draw_point(param.grazePoint.x, param.grazePoint.y)
    end
    --
    if tick_ok() then
      param.timesBlocked = 0
      param.time = param.time + 1
      param.attackCooldown = math.max(0, param.attackCooldown - 1)
      param.blockTime = math.max(0, param.blockTime - 1)
    end

    if player.NumBlockedHits > 0 or me.Blocked ~= 0 then
      param.counterAttack = true
      param.timesBlocked = param.timesBlocked + 1
    end

    if me.OnGround
    and ( ( dist_arrow() and math.abs( dist_arrow() ) <= param.arrowBlockDist ) 
    or ( dist_spear() and math.abs( dist_spear() ) <= param.arrowBlockDist )
    )
    then
      trace("Projectile Attack!")
      param.blockTime = 15
    elseif me.OnGround
    and ( player:hasTag("Attack") and math.abs(dist_sprite(target)) <= param.counterDist and dist_sprite(target) > 0 )
    and ( select(2, dist_sprite(target)) >= 0 and select(2, dist_sprite(target)) < 160 * 3 ) 
    then
      trace("Normal Ground Attack!")
      param.blockTime = 60
    end

    -- co-op 
    if bossBursted() 
    then
      param.attackCooldown = 20
      param.counterAttack = true
      param.blockTime = 100
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 3 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    
    if param.counterAttack and dist_sprite(target) > 0 and dist_sprite(target) <= param.counterDist 
    and not me:hasTag("Blocking") 
    and not me:hasTag("Postblock")
    then
      trace("nah")
      --facePlayer()
      attack_hornStab()
      param.blockTime = 0
      param.counterAttack = false
    elseif param.blockTime > 0 
    then
      addInput(dirs.N, btns.WorldGuard)
    elseif param.time > 300 
    then
      param.grazePoint.x = param.origin.x - (param.grazeOffsetRange) + random(param.grazeOffsetRange * 2)
      param.grazePoint.x = clamp(param.grazePoint.x, param.origin.x - param.grazeOffsetRange, param.origin.x + param.grazeOffsetRange)
      param.time = random(60 * 1.5)
    else
      if param.attackCooldown <= 0 then
        attack_graze()
        param.counterAttack = false
        param.attackCooldown = 60 * 3 + random(60 * 5)
      elseif obstacle_in_front(50) then
        turn_around(11)
        param.grazePoint.x = param.origin.x - (param.grazeOffsetRange) + random(param.grazeOffsetRange * 2)
        param.time = random(60 * 1.5)
      elseif math.abs(dist_grazePoint()) >= me.WorldCollisionWidth * 0.4 then
        approach(param.grazePoint.x)
      end
    end
    --
    param.blockedLastFrame = param.blockedThisFrame
    --
    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  
  attack_hornStab(toReturn)

  local r = random(1000)
  if r > 667
  and me.HealthPct < 0.4 then
    attack_graze(toReturn)
  elseif r >= 666 then
    attack_stomp(toReturn)
  elseif r <= 333 then
    attack_shofar(toReturn)
  end

  return toReturn
end --OnBattleAttack()