require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

-- Attacks and stuff to Do

local function isInYRange(target)
    local me, player = getMe(), (target or getPlayer())
    local diff = player.YPos - me.YPos
    if diff >= param.noticeYMin and diff <= param.noticeYMax then
        return true
    end
    return false
end

local function isInAttackYRange(target)
    local me, player = getMe(), (target or getPlayer())
    local diff = player.YPos - me.YPos
    if diff >= param.attackYMin and diff <= param.attackYMax then
        return true
    end
    return false
end

local function dist_origin()
    local me = getMe()
    local dx = me.XPos - param.origin.x
    local dy = me.YPos - param.origin.y 
    return dx, dy 
end

local function dist_player(target)
    local me, player = getMe(), (target or getPlayer())
    local dx = player.XPos - me.XPos
    dx = dx * me.FacingDir
    local dy = player.YPos - me.YPos
    return dx, dy
end

local function attackOK()
    if param.cooldown <= 0 then 
        return true 
    end
    return false 
end

local function attack_dash( data )
    if data ~= nil then
        data.dir = dirs.N
        data.pressAction = false
        data.meterBurn = false
    end
    doMotion("X+Attack")
end

local function attack_whip( data )
    if data ~= nil then
        data.dir = dirs.N
        data.pressAction = true
        data.meterBurn = false
    end
    doMotion("X+Heruka")
end

local function attack_dig( data )
    if data ~= nil then
        data.dir = dirs.D
        data.pressAction = false
        data.meterBurn = false
    end
    doMotion("D+Attack")
end

local function attack_evade( data )
    if data ~= nil then
        data.dir = dirs.D
        data.pressAction = true
        data.meterBurn = false
    end
    doMotion("D+Heruka")
end

local function attack_bite( data )
    if data ~= nil then
        data.dir = dirs.U
        data.pressAction = true
        data.meterBurn = false
    end
    doMotion("U+Heruka")
end

local function attack_howl( data )
    if data ~= nil then
        data.dir = dirs.U
        data.pressAction = false
        data.meterBurn = false
    end
    doMotion("U+Attack")
end

local function attackX()
  local w = param.attackXRange
  if param.bossBurstedMode then
    w = param.attackXRageRange
  end
  return w
end

local function noticeX()
  local w = param.noticeXRange
  if param.bossBurstedMode then
    w = param.noticeXRageRange
  end
  return w
end

-- the initialization function for this AI
function OnInit()
    -- Accessible anywhere in this file after Initialization
    local me, player = getMe(), getPlayer()
    --
    param.time = 0 
    param.cooldown = 0 
    param.idleWaitTime = 80 -- Not random the first time
    param.randomX = 0
    --
    param.origin = { x = me.XPos, y = me.YPos, z = me.ZPos }
    --[[
    if me:getSpawnpointPos() then
        param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
    end
    ]]
    --
    param.attackXRange = 160 * 3
    param.attackXRageRange = 160 * 3
    param.attackYMax = 160 * 2.5
    param.attackYMin = 160 * -1
    --
    param.noticeXRange = 160 * 6
    param.noticeXRageRange = 160 * 6
    param.noticeYMax = 160 * 3
    param.noticeYMin = 160 * -1
    
    param.isPackLeader = false
    if me.WeaponType == hash("PackLeader")
    or has_spawn_command("PackLeader")
    then
      param.isPackLeader = true
    end
    
    param.isPackMember = false
    if me.ColorNum == 15
    --and has_spawn_command("Summoned")
    then
      param.isPackMember = true
    end
    
    param.isWerewolf = false
    if me.ColorNum == 16
    or me.WeaponType == hash("Notorious")
    then
      param.isWerewolf = true
    end

    param.isChallengeDig = false
    if me.WeaponType == hash("Challenge_DigSummon") --me.ColorNum == 17
    then
      param.isChallengeDig = true
    end
  
    param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    if aiDebug() then 
      draw_bar(param.cooldown, 0, 60)
    end

    -- Ticks
    if not me:hasTag("Attack") and not me:hasTag("Hitstun") then
      param.time = param.time + 1
    end

    if not me:hasTag("Attack") then
      param.cooldown = math.max(0, param.cooldown - 1 )
    end

    local idle_dx, idle_dy = dist_point(param.origin.x + param.randomX, param.origin.y)
    if aiDebug()
    then
      draw_point(param.origin.x + param.randomX, param.origin.y)
    end

    -- co-op 
    if bossBursted() 
    then
      param.time = 0 
      param.cooldown = 20
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 5 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end

    if dist_sprite(target) < attackX()
    and dist_sprite(target) > 0
    and attackOK()
    and isInAttackYRange(target)
    then
      param.time = 0 
      debugPalette("Debug_Yellow")
      attack_whip()
      param.cooldown = 30
    elseif math.abs( dist_sprite(target) ) <= noticeX()
    and isInYRange(target)
    and ( param.bossBurstedMode and sprite_vulnerable(target) )
    and not got_hit()
    then
      debugPalette("Debug_Orange")
      approach(target.XPos)
    elseif dist_origin() > 160 * 2 then
      debugPalette("Debug_Magenta")
      approach_withTurn(param.origin.x, 11)
    elseif param.time > param.idleWaitTime then 
      debugPalette("Debug_Cyan")
      param.time = 0 
      param.idleWaitTime = 120 + random(120)
      param.randomX = -160*8 + random(160*8)
    elseif math.abs(idle_dx) >= 160 then
      debugPalette("Debug_Green")
      approach(param.origin.x + param.randomX)
      --approach_withTurn(param.origin.x + param.randomX, 11)
    else
      debugPalette("Debug_Dark")
    end
    wait() -- try to keep this at the end
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  local dirChance = random(1000)
  local actionChance = random(1000)

  if battleData.turnsSinceHowl == nil
  then
    battleData.turnsSinceHowl = 0 
  end

  -- Dig Forever --
  if param.isChallengeDig
  then
    trace("Challenge - Dig")
    attack_dig(toReturn)
  -- Werewolf --
  elseif param.isWerewolf
  then
    if actionChance > 600
    then
      attack_bite(toReturn)
    elseif dirChance > 500
    then
      attack_whip(toReturn)
    else
      attack_dash(toReturn)
    end
    -- Pack Leader --
  elseif param.isPackLeader
  then
    if aiDebug()
    and false
    then
      attack_howl(toReturn)
    elseif battleData.turnsSinceHowl >= 3
    then
      trace("Howl")
      attack_howl(toReturn)
      --
      battleData.howledOnce = true 
      battleData.turnsSinceHowl = 0 
    else
      if dirChance > 500 then
        attack_whip( toReturn )
      else
        attack_dash(toReturn)
      end
    end
  -- Pack Member --
  elseif param.isPackMember
  then
    trace("Summoned Attack")
    if dirChance > 500
    then
      attack_dig( toReturn )
    else
      attack_bite( toReturn )
    end
  -- Normal --
  else
    if dirChance > 500 then
      attack_whip( toReturn )
    else
      attack_dash(toReturn)
    end
  end

  --
  battleData.turnsSinceHowl = battleData.turnsSinceHowl + 1

  return toReturn
end --OnBattleAttack()