require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("math_ext")
require("movement")
require("util")

local function action_changeStance(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function action_wallJump()
  --[[
  local _dir = dirs.N
  if airChase_dir ~= nil then
    _dir = airChase_dir
  end
  -- Jumplock here!
  param.jumpLock = 5
  -- No joystick movement for 1f
  trace("pt1")
  addInput( dirs.N )
  wait(1)
  trace("pt2")
  -- Walljump
  addInput( _dir, btns.Jump )
  wait(1)
  ]]
  doMotion("F+~Jump")
  debugPalette("Debug_White")
end

local function action_dig()
  doMotion("D+Dash")
end

local function action_digReappear()
  doMotion("X+Attack")
end

local function attack_punch(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_kick(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_cheer(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_stopCheering(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function attack_goAwayForever(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Heruka")
end
local function FloorAnglesOK( angle )
  if not angle then return false end
  if angle <= -45
  and angle >= 45
  then
    return true
  else
    return false
  end
end

local function isInStopChaseRange()
  local me, player = getMe(), getPlayer()
  
  if isInRange( player,
    me.XPos, me.YPos,
    param.stopChaseX,
    param.chaseRangeY)
  then
    return true
  end
  
  return false
end

local function isInTurnRange()
  local me, player = getMe(), getPlayer()
  
  if isInRange( player,
    me.XPos, me.YPos,
    param.turnRangeX,
    param.turnRangeY)
  then
    return true
  end

  return false
end

local function isInChaseRange()
  local me, player = getMe(), getPlayer()
  
  if isInRange( player,
    me.XPos, me.YPos,
    param.chaseRangeX,
    param.chaseRangeY)
  then
    return true
  end
  
  return false
end

local function isInFwdChaseRange()
  local me, player = getMe(), getPlayer()
  
  if isInRange( player,
    me.XPos, me.YPos,
    param.chaseRangeX,
    param.chaseRangeY)
  then
    return true
  end
  
  return false
end

local function isInActionStance()
  local me = getMe()

  if me.Flag13 then
    return true
  else
    return false
  end
end

local function isPlayerAiming()
  local me, player = getMe(), getPlayer()

  if player:hasTag("BowMode")
  or player:hasTag("BowShot")
  then
    return true
  else
    return false
  end

  return false
end

local function ChaseOK()
  local me, player = getMe(), getPlayer()

  if not isPlayerAiming()
  then
    return true
  else
    return false
  end

  return false
end

local function turn_around_failsafe()
  local me = getMe()

  trace("FacingDir %i" % { me.FacingDir })
  local _facingDir = me.FacingDir
  addInput(dirs.B)
  while me.FacingDir == _facingDir
  do
    addInput(dirs.B)
    wait(3)
    trace("FacingDir %i" % { me.FacingDir })
    if me.FacingDir == _facingDir
    then
      -- Don't do this, seriously, but we're going to
      me.FacingDir = me.FacingDir * -1
    end
  end
end

local function log(s)
  if param.verboseMode
  then
    print(s)
  end
end

local function playerAvgPos()
  local _x, _y = 0, 0 

  if param.lastPlayerPositions
  then
    for i = 1, #param.lastPlayerPositions do
      _x = _x + param.lastPlayerPositions[i].x
      _y = _y + param.lastPlayerPositions[i].y
    end
  end

  _x = _x / #param.lastPlayerPositions
  _y = _y / #param.lastPlayerPositions

  return _x, _y
end

-- the initialization function for this AI
local hash_dogName = hash("Pet_Dog")
function OnInit()
  local me, player = getMe(), getPlayer()

  param.time = 0 

  param.turnRangeX = 160 * 7
  param.turnRangeY = 160 * 6
  param.stopChaseX = 160 * 7
  param.chaseRangeX = 160 * 22
  param.chaseRangeY = 160 * 11
  param.playerAirborneTime = 0
  param.chaseJumpTime = 15
  param.mustOvershootInChase = false
  if me.MyName == hash_dogName
  then
    -- dog uses this :)
    --param.stopChaseX = 160 * 3
    --param.mustOvershootInChase = false
  end

  param.jumpLock = 0
  param.startChasing = true

  param.airChaseDir = dirs.F
  param.leapOverWall = false

  param.vecBattleStartBattlePos = {  x = 0, y = 0 }
  param.offscreenTime = 0 
  param.cheerTime = 0 
  param.stillInBattleTime = 0 
  param.isInvisibleTime = 0 
  param.awayFromPlayerTime = 0
  param.minRunSpeed = 16.5 -- 22 by default, 33 h1, 40 h2
  param.maxRunSpeedH1 = 40
  param.maxRunSpeedH2 = 55
  param.maxRunSpeedH3 = 70
  param.maxRunSpeed = 72
  param.maxRunSpeedBonus = 15 -- ADDITIONAL on top of EVERYTHING ELSE
  param.runSpeedLeashDist = 160 * 4 -- looks right
  param.lastPlayerPositions = {}
  param.waitWhileDugTime = 80
  param.waitWhileDugTime_Quick = 15

  param.isABuddy = false
  if has_spawn_command("Friendly")
  then
    param.isABuddy = true
    me.DespawnDistX = -1
    me.DespawnDistY = -1
  end
  
  param.isChallengeGiant = false
  if me.WeaponType == hash("Challenge_GiantRoti")
  then
    param.isChallengeGiant = true
  end

  param.forceDig = false
  if has_spawn_command("ForceDig")
  then
    param.forceDig = true
  end

  param.justSpawnedForReal = true

  param.verboseMode = false
end

-- the main update function for this AI
function OnUpdate()
  local imScared = false
  local startedBattle = false

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end

  while true do
    local me, player = getMe(), getPlayer()
    local players = getPlayers()

    if aiDebug() then
      -- Turn --
      drawBox( me.XPos - param.turnRangeX/2, me.XPos + param.turnRangeX/2,
        me.YPos + param.turnRangeY/2, me.YPos - param.turnRangeY/2,
        1, -1,
        255, 255, 0
      )
      -- Chase --
      drawBox( me.XPos - param.chaseRangeX/2, me.XPos + param.chaseRangeX/2,
        me.YPos + param.chaseRangeY/2, me.YPos - param.chaseRangeY/2,
        1, -1,
        255, 0, 255
      )
      -- Start / Stop Chasing --
      drawBox( me.XPos - param.stopChaseX/2, me.XPos + param.stopChaseX/2,
        me.YPos + param.turnRangeY/2, me.YPos - param.turnRangeY/2,
        1, -1,
        0, 255, 255
      )

      --trace("Time %i" % { param.time })
      --trace("InBattle %i" % { player.InBattle })
      -- Info
      --trace( "Ahead %3.2f Behind %3.2f Offscreen %i" % { me.WallDistAhead, me.wallDistBehind, param.offscreenTime } )

      -- Airborne
      draw_bar( param.playerAirborneTime, 1, 120, 0, { 255, 255, 255 } )

      -- JumpLock
      draw_bar( param.jumpLock, 1, 10, 1, { 128, 255, 255 } )
      
      -- Cheer
      draw_bar( param.cheerTime, 1, 300, 2, { 128, 0, 255 } )

      -- Away From Player
      draw_bar( param.awayFromPlayerTime, 1, 300, 3, { 128, 0, 255 } )

      -- Offscreen
      draw_bar( param.offscreenTime, 1, 300, 4, { 255, 0, 255 } )

      -- Battle
      draw_bar( param.stillInBattleTime, 1, 300, 5, { 128, 128, 255 } )

      -- Invisible Time
      draw_bar( param.isInvisibleTime, 1, 300, 6, { 128, 128, 128 } )

      -- ChaseOK
      draw_bar( ChaseOK() and 1 or 0 , 0, 1, 7, { 255, 0, 128 } )
    end

    -------------------
    -- Roti Chasing! --
    -------------------

    -- Check if we should be chasing before we do anything
    if isInStopChaseRange() 
    then
      -- In front of us, facing the same way!
      --if dist_player() > 0 
      --and me.FacingDir == player.FacingDir
      if me.FacingDir == player.FacingDir
      and player.XVel <= 12
      then
        if param.mustOvershootInChase
        and dist_player() < -3
        then
          param.startChasing = false
        elseif not param.mustOvershootInChase
        then
          param.startChasing = false
        end
      elseif me.FacingDir ~= player.FacingDir
      then
        param.startChasing = false
      end
    elseif not isInTurnRange() then
      param.startChasing = true
    end
    
    -- How long is player airborne
    if player_vulnerable()
    and player.OnGround <= 0
    then
      param.playerAirborneTime = param.playerAirborneTime + 1
    else
      param.playerAirborneTime = 0
    end

    if update_ok() then
      param.jumpLock = math.max(0, param.jumpLock - 1)
      param.time = param.time + 1

      if (me.TouchScreenLR >= 1 or me.TouchScreenLR <= -1)
      or (me.TouchScreenUD >= 1 or me.TouchScreenUD <= -1)
      or me:hasTag("Dig_Offscreen")
      then
        param.offscreenTime = param.offscreenTime + 1
      else
        param.offscreenTime = 0
      end

      if me:hasTag("Dig")
      then
        param.isInvisibleTime = param.isInvisibleTime + 1
      else
        param.isInvisibleTime = 0 
      end

      -- add to our list of positions
      param.lastPlayerPositions[#param.lastPlayerPositions+1] = {
        x = player.XPos, y = player.YPos
      }
      if #param.lastPlayerPositions >= 5
      then
        while #param.lastPlayerPositions >= 5 do
          table.remove(param.lastPlayerPositions, 1)
        end
      end
    end

    -- BETTER CHECK --
    if has_spawn_command("ForceDig")
    and not param.forceDig
    then
      param.forceDig = true
    end
    
    if me:hasTag("Cheer")
    then
      param.cheerTime = param.cheerTime + 1
    else
      param.cheerTime = 0 
    end

    if isInTurnRange()
    then
      param.awayFromPlayerTime = 0 
    else
      param.awayFromPlayerTime = param.awayFromPlayerTime + 1
    end

    if math.abs( me.XVel ) <= 0
    and player.InBattle > 0 
    then
      param.stillInBattleTime = param.stillInBattleTime + 1
    else
      param.stillInBattleTime = 0
    end

    -------------------------
    -- Follow the player!! --
    -------------------------
    if param.isABuddy then
      -- We don't get Air control, unless we're close
      if me.OnGround > 0
      and math.abs( dist_player() ) >= 160 * 0.5 
      then
        param.airChaseDir = dist_player() >= 0 and dirs.F or dirs.B
        --
      end

      if param.forceDig
      then
        trace("ForceDig!")
        --
        if me:hasTag("Dig_Offscreen")
        then
          attack_goAwayForever()
        else
          action_dig()
          wait(3)
        end
      -- Ajna is in Battle
      elseif player.InBattle ~= 0 
      then
        trace("We're in BATTLE")
        debugPalette("Debug_Orange")

        local _basePoint = player.XPos
        local playerXOffset = -160 
        playerXOffset = playerXOffset * player.FacingDir

        if me.LightB ~= 0
        then
          trace("Accepting battle pos info from LightB!")
          _basePoint = me.LightB
        else
          _basePoint = _basePoint + playerXOffset
        end
        
        local targetPoint = { x = _basePoint, y = player.YPos }


        local distToPoint = dist_point(targetPoint.x, targetPoint.y)
        trace("Distance %i" % { distToPoint })
        --[[
        local distToPoint = vec_dist(
          { x = me.XPos, y = me.YPos }, targetPoint )
        trace("Distance %i" % { distToPoint })
        ]]
        draw_point( targetPoint.x, targetPoint.y ) 
      
        if math.abs(distToPoint) >= 160
        then
          --if distToPoint > 0
          --then
            approach( _basePoint )
          --else
          --  trace("WE WANT TO TURN AROUND!")
          --  turn_around_failsafe()
          --end
          --
          --if me.XVel >= 10
          --then
          --  me.XVel = 10
          --elseif me.XVel <= -10
          --then
          --  me.XVel = -10
          --end
        else
          debugPalette("Debug_Magenta")
          if me.XVel > 10
          and not me:hasTag("Attack")
          then
            me.XVel = 10
          elseif me.XVel < -10
          and not me:hasTag("Attack")
          then
            me.XVel = -10
          end
          
          --local dirToPlayer = player.XPos - me.XPos
          --if dirToPlayer >= 0 then
          --  dirToPlayer = 1
          --else
          --  dirToPlayer = -1
          --end
          
          --if me.FacingDir ~= dirToPlayer and me.XVel < 1 and me.XVel > -1 and me.OnGround
          --then
          --  turn_around_failsafe()
          --end
        end
        --[[
        if math.abs(distToPoint) <= 160 * 2
        --and player_vulnerable()
        then
          trace("Free To Act")
          debugPalette("Debug_Cyan")
          -- I wanna see Ajna fight!!!
          if param.stillInBattleTime >= 3
          and me.FacingDir ~= player.FacingDir
          then
            turn_around(7)
          
          elseif param.stillInBattleTime >= 10
          and not isInActionStance()
          then
            action_changeStance()
          elseif param.stillInBattleTime >= 11
          and isInActionStance()
          and player.NumHits > 0 
          then
            debugPalette("Debug_Yellow")
            attack_cheer()
            wait(20)
          elseif param.stillInBattleTime >= 11
          and isInActionStance()
          and player.NumBlockedHits > 0 
          then
            debugPalette("Debug_Yellow")
            attack_punch()
            wait(20)
          elseif param.stillInBattleTime >= 11
          and isInActionStance()
          and player.NumTouches > 0
          then
            attack_punch()
            wait(20)
          else
            debugPalette("Debug_BlinkYellow")
          end
        else
          trace("Move!")
          debugPalette("Debug_BlinkMagenta")
          approach( targetPoint.x )
          --approach_withTurn( targetPoint.x, 7 )
        end
        ]]
      -- Jump Towards Ajna if we're Airborne
      elseif player.OnGround <= 0
      and player_vulnerable()
      and param.playerAirborneTime >= param.chaseJumpTime
      and player.InBattle <= 0 
      and not me:hasTag("Dig")
      --select(2, dist_player()) >= 160 * 0.5
      then
        trace("Airborne For Walljumps!")
        debugPalette("Debug_Yellow")
        if me.OnGround > 0 
        then
          addInput( param.airChaseDir, btns.Jump )
        else
          -- Walljump Attempt
          if me.WallDistAhead <= 35 
          and not param.leapOverWall
          then
            trace("Walljump")
            action_wallJump()
          -- Hold the Airjump as long as the player is above us!
          elseif select(2, dist_player()) >= 160 * 0.5
          then
            addInput( dirs.F, btns.Jump )
          else
            addInput( param.airChaseDir )
          end
        end
      elseif player.OnGround > 0 
      and player_vulnerable()
      and select(2, dist_player()) >= 160 * 1.75
      and player.InBattle <= 0 
      --and FloorAnglesOK( player.FloorAngle )
      and me.WallDistAhead <= 100
      and me.YVel >= 0
      and not me:hasTag("Dig")
      then
        trace("Airborne For Following!")
        debugPalette("Debug_Dark")
        addInput( airChase_dir, btns.Jump )
        param.leapOverWall = true
      elseif isInTurnRange()
      and not me:hasTag("Dig")
      and player.InBattle <= 0 
      then
        debugPalette("Debug_Red")
        if me.OnGround <= 0
        and me.YVel >= 0
        then
          debugPalette("Debug_Magenta")
          if me.WallDistAhead <= 55
          and param.leapOverWall 
          then
            addInput( param.airChaseDir, btns.Jump )
          else
            addInput( dirs.N, btns.Jump )
          end
        elseif player_vulnerable() 
        and param.startChasing 
        then
          debugPalette("Debug_Magenta")
          approach(player.XPos)
        elseif player.OnGround > 0
        then
          if isInActionStance()
          then
            action_changeStance()
          elseif dist_player() < 0
          then
            trace("turn around")
            turn_around(7)
          end
        end
      elseif isInChaseRange() 
      and not me:hasTag("Dig")
      and player.InBattle <= 0 
      then
        trace("Chase Range")
        debugPalette("Debug_Blue")
        -- Jump Towards them on Flat Ground
        if player.OnGround > 0
        and select(2, dist_player()) >= 160 * 0.5
        and FloorAnglesOK( player.FloorAngle )
        then
          debugPalette("Debug_Green")
          addInput( airChase_dir, btns.Jump )
          -- Hold Input in Air
        elseif me.OnGround <= 0 
        then
          debugPalette("Debug_Green")
        elseif player_vulnerable() 
        and param.startChasing 
        and math.abs(dist_player()) >= 160 * 1.5 
        then
          approach(player.XPos)
          param.leapOverWall = false
        end
      elseif me:hasTag("Dig")
      and player.InBattle <= 0 
      then
        trace("Dig State")
        --
        local targetRespawnTime = param.waitWhileDugTime_Quick
        if param.justSpawnedForReal
        then
          targetRespawnTime = param.waitWhileDugTime_Quick
        end
        --
        if param.awayFromPlayerTime >= targetRespawnTime
        and math.abs( player.XVel ) <= 3 
        and player.OnGround > 0
        and me:hasTag("Dig_Offscreen")
        then
          trace("Reappearing from dig!")
          action_digReappear()
          wait(40)
          param.justSpawnedForReal = false
        elseif param.offscreenTime >= targetRespawnTime
        and player.OnGround > 0
        and me:hasTag("Dig_Offscreen")
        then
          trace("Reappearing from dig, OVERTIME!")
          action_digReappear()
          wait(40)
          param.justSpawnedForReal = false
        elseif param.isInvisibleTime >= targetRespawnTime
        and player.OnGround > 0
        and me:hasTag("Dig_Offscreen")
        then
          trace("Reappearing from dig, INVISIBLE")
          action_digReappear()
          wait(40)
          param.justSpawnedForReal = false
        end
      else
        trace("Start Dig")
        action_dig()
        wait(40)
      end
      trace("Hey! %i / %i / %i" % { param.awayFromPlayerTime, param.offscreenTime, param.isInvisibleTime })
    ------------------------
    -- get scared and run --
    ------------------------
    elseif not imScared then
      trace("we're not here right?")
      local playerDistX = player.XPos - me.XPos  -- in world space
      playerDistX = playerDistX * me.FacingDir  -- now in local space
      local playerDistY = player.YPos - me.YPos
      
      -- print("Roti-player xdist %d ydist %d" % {playerDistX, playerDistY})
      
      -- if they're within attack distance, attack!
      if playerDistX < 750 and playerDistX > -750
      and playerDistY < 750 and playerDistY > -450 then
        imScared = true
        addInput(dirs.B)
         -- print("Pwayer tuu cwose, Ms. Roti is scuwwwed!")
      end
    else
      trace("not here either, yeah?")
      addInput(dirs.F)
      if me.TouchScreenLR > 1 or me.TouchScreenLR < -1 
        or me.TouchScreenUD > 1   -- can rise off the top
        -- or me.TouchScreenUD < -1   -- cannot fall off the bottom...for now at least?
        then
        --print("Ms. Roti is completely offscreen, disappearing now!")
        --doGameEvent("YoWhatsUpDUDE")
        doMotion("X+Heruka")  -- goodbye....
      end
    end -- the if-on-ground
    
    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
    local toReturn = DefaultOnBattleAttack(battleData)
    local me = getMe()

    local dirChance = random(1000)
    local actionChance = random(1000)

    if param.isChallengeGiant
    then
      if dirChance > 500
      then
        attack_kick(toReturn)
      else
        attack_punch(toReturn)
      end
    elseif not me.Flag13 then
      trace("Whoa")
      action_changeStance(toReturn)
    elseif me.LightR >= me.LightG 
    and not me:hasTag("Cheer")
    then
      -- let's celebrate?!
      toReturn.dir = dirs.U
      toReturn.pressAction = false
    elseif me:hasTag("Cheer") then
      trace("Yay!!")
      toReturn.dir = dirs.N
      toReturn.pressAction = true
    else
      --
      if toReturn.dir == dirs.U then
        trace("not this time, dude")
        local r = random(1000)
        if r > 500 then
          toReturn.dir = dirs.N
        else
          toReturn.dir = dirs.D
        end
      end
      --
      toReturn.pressAction = false
      trace("Yeah")
    end

    return toReturn
end -- OnBattleAttack()