require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function block()
  doMotion("X+WorldGuard")
end

local function attack_ball(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 35
end

local function attack_spikes(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 40
end

local function attack_spikesRow(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Heruka")
  param.attackCooldown = 52
end

local function attack_clap(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
  param.attackCooldown = 30
end

local function attack_cube(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
  param.attackCooldown = 30
end

local function isInAttackRange()
  local me, player = getMe(), getPlayer()

  local _w, _h = param.attackRangeX, param.attackRangeY
  return isInRange( player,
    me.XPos, me.YPos,
    _w, _h
  )
end

local function isInSeekRange()
  local me, player = getMe(), getPlayer()

  local _w, _h = param.seekRangeX, param.seekRangeY
  return isInRange( player,
    me.XPos, me.YPos,
    _w, _h
  )
end
local function isACube()
  local me = getMe()

  if me:hasTag("Cube")
  or me:hasTag("Cube_Start")
  or me:hasTag("Cube_Loop")
  or me:hasTag("Cube_Recovery")
  then
    return true
  else
    return false
  end

  return false
end

-- the initialization function for this AI
function OnInit()
  -- Accessible anywhere in this file after Initialization
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end

  param.time = 0
  param.attackCooldown = 0

  param.isBoss = true
  if has_spawn_command("Boss")
  then
    param.isBoss = true
  end
  param.boss_WorldMoveOK = false

  param.cubeHPPercent = 0.8
  param.hasTurnedIntoCube = false
  param.wantsCube = false
  param.forceAttack = false

  param.attackRangeX = 160 * 10
  param.attackRangeY = 160 * 7
  param.seekRangeX = 160 * 10
  param.seekRangeY = 160 * 9
  param.sightRangeX = 160 * 10
  param.sightRangeY = 160 * 9

  if aiDebug()
  then
    param.boss_WorldMoveOK = true
    param.hasTurnedIntoCube = true
    --
    param.wantsCube = false
    param.forceAttack = true
    me.InvinFlashTime = 400
  end
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end

  while true do
    me, player = getMe(), getPlayer()
    --
    if aiDebug() then
      draw_bar(param.attackCooldown, 0, 100)
      --draw_bar(me.InvinFlashTime, 0, 400, 1)
    end

    if tick_ok() then
      if not me:hasTag("Attack") then
        param.attackCooldown = math.max(0, param.attackCooldown - 1)
      end
      --param.time = param.time + 1
    end

    -- The Loop
    if true
    then
      if param.isBoss then
        debugPalette() 

        local _xdist = dist_player()
        local _ydist = select(2, dist_player())

        if got_hit(me)
        then
          if isACube()
          then
            param.boss_WorldMoveOK = true
          end

          if param.wantsCube
          then
            param.wantsCube = false
          end
        end
        
        if me.InvinFlashTime <= 0
        then
          param.forceAttack = false
        end
        
        --[[
        if param.forceAttack then
          debugPalette("Debug_BlinkOrange")
        end
        ]]
        
        if param.boss_WorldMoveOK
        then
          if param.wantsCube
          and not isACube()
          then
            debugPalette("Debug_Red")
            attack_cube()
          else
            if bossBursted() then
              trace("BOSS BURST")
              param.forceAttack = true
              me.InvinFlashTime = 400  -- longer for us
            -- Boss Burst attack clone, that is more frequent 
            elseif param.forceAttack
            and math.abs(me.XVel) <= 8 -- About our acceleration..
            and not me:hasTag("Attack")
            and isInSeekRange()
            then
              -- Wait
              wait(3)

              -- We can pick anything!
              local r = random(1000)

              if _ydist > 160 * 2
              then
                trace("YDist High")
                if r >= 500
                then
                  attack_ball()
                else
                  attack_clap()
                end
              elseif _xdist > 160 * 3.25
              then
                trace("XDist Far")
                attack_ball()
              else
                trace("Normal")
                if r >= 667
                then
                  attack_spikesRow()
                elseif r <= 333
                then
                  attack_clap()
                else
                  attack_spikes()
                end
              end
            elseif player_vulnerable()
            and param.attackCooldown <= 0
            and math.abs(me.XVel) <= 8 -- about our Acceleration.. 
            and isInAttackRange()
            and not me:hasTag("Attack")
            then
              debugPalette("Debug_Blue")

              -- Wait
              wait(30)

              -- We can pick anything!
              local r = random(1000)

              if _ydist > 160 * 2
              then
                trace("YDist High")
                if r >= 500
                then
                  attack_ball()
                else
                  attack_clap()
                end
              elseif _xdist > 160 * 4.5
              then
                trace("XDist Far")
                attack_spikesRow()
              else
                trace("Normal")
                if r >= 667
                then
                  attack_ball()
                elseif r <= 333
                then
                  attack_clap()
                else
                  attack_spikes()
                end
              end
              --attack_spikes()
            elseif me:hasTag("Attack")
            then
              debugPalette("Orange")
            elseif isInSeekRange()
            then
              debugPalette("Debug_Dark")
              if _xdist < 0 
              then
                turn_around(20)
              end
            elseif not isInSeekRange()
            then
              approach_withTurn(player.XPos, 20)
            else
              trace("Fallthrough..")
              debugPalette("Debug_Blink")
            end
          end
        end
      else
        -- Nothing
      end
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local me = getMe()
  local toReturn = DefaultOnBattleAttack(battleData)

  local dirChance = random(1000)
  local actionChance = random(1000)

  param.boss_WorldMoveOK = true

  -- Default
  attack_spikes(toReturn)

  -- Directions
  if dirChance > 667
  then
    attack_ball(toReturn)
  elseif dirChance < 334
  then
    attack_clap(toReturn)
  end

  if me.LifePct <= param.cubeHPPercent 
  and not param.hasTurnedIntoCube 
  then
    param.hasTurnedIntoCube = true
    param.wantsCube = true
    param.boss_WorldMoveOK = true
    attack_cube(toReturn)
  elseif param.hasTurnedIntoCube
  and battleData.turnNum >= 5
  then
    param.hasTurnedIntoCube = true
    param.wantsCube = true
    param.boss_WorldMoveOK = true
    attack_cube(toReturn)
  end

  if aiDebug()
  then
    --attack_clap(toReturn)
    attack_cube(toReturn)
  end

  return toReturn
end --OnBattleAttack()