require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function attack_punch(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_point(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_summon(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_bite(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function action_jumpToFlying_withBarrier(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Heruka")
end

local function action_jumpToFlying(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Heruka")
end

local function action_toggleBarrier(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Heruka")
end

local function action_jumpFromFlying(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Heruka")
end

local function attack_pillar(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_pillarBattle(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_fireball(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_fireballBattle(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function action_summonPose(data)
  local me = getMe()

  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function action_summonPose_hold(data)
  local me = getMe()

  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function action_FakeHit(data)
  local me = getMe()
  --
  local final = param.bossPhase >=3 and true or false

  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("X+Dash")

  if final
  then
    --me.Lifebar = 1000 -- I actually think we need this, but.. yeah
  end

  --[[
  me.ArmorHits = 0 
  me.GotHit = 1
  me.Hitpause = 11

  if final
  then
    me.Hitpause = 50
  end
  me.Hitstun = 50
  me.AnimHint = hash("KnockdownAnim")
  ]]
end

local hash_snakesName = hash("Snakes")
local function getChild_platform()
  local me = getMe()
  local c = me:getChildren()

  for k, v in pairs(c) do
    if v.MyName == hash_snakesName
    then
      return v
    end
  end

  return nil
end

local hash_barrierName = hash("BarrierFX")
local function getChild_barrier()
  local me = getMe()
  local c = me:getChildren()

  local _target = nil
  for k, v in pairs(c) do
    if v.MyName == hash_snakesName
    then
      _target = v
    end
  end

  if v == nil 
  then 
    return nil
  end

  c = v:getChildren()
  for k, v in pairs(c) do
    if v.MyName == hash_barrierName
    then
      return v
    end
  end

  return nil
end

local function isPlayerInChakraGateRange()
  local me, player = getMe(), getPlayer()
  local x, y = param.origin.x, param.origin.y
  local w, h = 160 * 6, 160 * 3

  -- Override X/Y with Spawnpoints Later!
  x = param.chakraGatePoint.x
  y = param.chakraGatePoint.y

  if isInRange(player, x, y, w, h)
  then
    return true
  else
    return false
  end
end

local function isPlatformAlive()
  local me = getMe()
  local platform = getChild_platform()

  if platform
  then
    if platform:hasTag("Alive") then
      return true
    end
  end
end


local function isPlayerInFocus()
  local me, player = getMe(), getPlayer()

  if player:hasTag("ChakraGateOK")
  and isPlayerInChakraGateRange()
  then
    return true
  else
    return false
  end
end

local function getBarrierOnly()
  local me = getMe()
  local c = me:getChildren()

  for i, child in ipairs(c) do
    if child.MyName == hash("BarrierFX")
    then
      return child
    end
  end
  
  return false
end

local function isBarrierAlive()
  local me, player = getMe(), getPlayer()
  return me:getFlagValue(9) 
end

local function isBarrierVisible()
  local me, player = getMe(), getPlayer()
  return me:getFlagValue(15) 
end


local function OneTimeSetup()
  if not param._oneTimeSetup then
    -- Override X/Y with Spawnpoints Later!
    local _x, _y, _z = getObjectPos("Spawn_BossInfo_SnakeQueen_ChakraGateCenter", false)
    if _x ~= nil then
      param.chakraGatePoint.x = _x
      trace(_x)
      param.chakraGatePoint.y = _y
    end
  end
  param._oneTimeSetup = true
end

local function platform_ok()
  local me = getMe()

  if not me:hasTag("Hitstun")
  --and me.Hitstun <= 0 
  --and me.Hitpause <= 0
  -- and not isTangled(me)
  then
    return true
  end

  return false
end

local function snakes_ok()
  local me = getMe()

  if me.Flag13
  then 
    return true
  end

  return false
end

local function barrierHP()
  local me = getMe()

  if me.LightR ~= nil
  then
    return me.LightR
  else
    return nil
  end

  return nil
end

local function doFocusCheck()
  local me = getMe()
  if not isBarrierVisible()
  then
    param.focusTime = math.min(param.focusTime + 0.875, param.focusMax)
  elseif isBarrierVisible()
  then
    param.focusTime = math.max(0, param.focusTime - 2)
  end
end

local function debugDraw()
  if aiDebug()
  then
    local me, player = getMe(), getPlayer()
    --
    -- Time without a barrier
    draw_bar(param.timeWithoutBarrier, 0, 120, -3, {255, 255, 255})
    -- Focus
    draw_bar(param.focusTime, 0, param.focusMax, -2, {99, 155, 255})
    -- Time (BACKER)
    draw_bar(0, 60 * -8, 60 * 10, -1, {255, 255, 255})
    -- Time (FWD)
    draw_bar(param.time, 60 * -8, 60 * 10, -1, {153, 229, 80})
    -- Barrier
    draw_bar(barrierHP(), 0, 100, 0, {88, 211, 50})
    -- Phase
    drawText("FlyingTime: %i\nBoss Phase: %i\nTime Without Barrier: %i (%2.2f)\nHasBarrier: %s" % { param.flyingTime, param.bossPhase, param.timeWithoutBarrier, (param.timeWithoutBarrier / 60), tostring(isBarrierAlive()) },
             player.XPos, player.YPos, 0)
  end
end

local AIFlyingHash = hash("AIFlying")

-- the initialization function for this AI
function OnInit()
  -- Accessible anywhere in this file after Initialization
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end

  param.chakraGatePoint = { x = 0, y = 0 }
  param.flightYOffset = 160 * 10 -- 8, but we want some give.
  param.time = 0
  param.flyingTime = 0 
  param.turnCooldown = 0 
  param.attackCooldown = 0
  param.timeWithoutBarrier = 0
  param.barrierRegainThresholds = {
    [0] = 40,
    [1] = 40,
    [2] = 15,
    [3] = 10,
    [4] = 0
  }

  -- free variables for pattern
  param.patternAtkType = 0
  param.patternAtkSubType = 0 
  param.patternTime = 0 

  param.focusTime = 0
  param.focusMax = 60 * 5 -- WE INCREASE THIS LATER!
  param.summonPoseTime = 0 

  param.attackRangeX = 160 * 18
  param.attackRangeY = 160 * 7
  param.sightRangeX = 160 * 30
  param.sightRangeY = 160 * 9

  param.goToPatternSetup = false
  param.forcedKDByPlayer = false
  param.wantsToRechargeBarrier = false

  param.event_TriggerVFXEvent = false

  param.isBoss = false
  if has_spawn_command("Boss")
  then
    param.isBoss = true
  end
  param.bossPhase = 0 -- ah, yes, the golden variable
  param.battleExitThresholds = {
    [0] = 0.8,
    [1] = 0.6,
    [2] = 0.3,
    [3] = 0.01 -- Shouldn't ever get here
  }

  param.maxFramesToAscend = 52 -- 49 -- 50

  param.speedBonus_tracking =
  { min = 11, max = 28, range = 160 * 7 }
  param.speedBonus_oscillate =
  { min = 5, max = 55, range = 160 * 2 }
  param.speedBonus_returning =
  { min = 3, max = 55, range = 160 * 7 }

  param.manuallyLeftFromBattle = false
  param._oneTimeSetup = false
  
  param.palFX_playerFocus = hash("Focus_InnerRealmOK")
  param.palFX_recharging = hash("AxePickupGlow")
  param.palFX_time = 3

  if aiDebug()
  then
    trace("[INFO] Started from DEBUG!")
    param.bossPhase = 1
  end
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end

  while true do
    me, player = getMe(), getPlayer()
    --
    OneTimeSetup()

    if aiDebug() then
      debugDraw()
    end

    if param.isBoss
    then
      if tick_ok() then
        -- Turns
        param.turnCooldown = math.max(0, param.turnCooldown - 1)

        -- AttackCooldown
        if not me:hasTag("Attack") then
          param.attackCooldown = math.max(0, param.attackCooldown - 1)
        end

        -- Barrier Recharge
        if param.wantsToRechargeBarrier
        and barrierHP() >= 100
        then
          param.wantsToRechargeBarrier = false
        elseif not param.wantsToRechargeBarrier
        and not isBarrierVisible()
        -- and barrierHP() <= param.barrierRegainThresholds[param.bossPhase]
        and param.time > 3
        then
          param.wantsToRechargeBarrier = true
        end

        -- Summon Time
        if me:hasTag("Platform_SummonPose")
        then
          param.summonPoseTime = param.summonPoseTime + 1
        else
          param.summonPoseTime = 0 
        end

        -- I need a BARRIER
        if barrierHP() < 100
        and param.wantsToRechargeBarrier
        then
          param.timeWithoutBarrier = param.timeWithoutBarrier + 1
        else
          param.timeWithoutBarrier = 0 
        end
      end

      -- This actually may need a better check
      doFocusCheck()

      if param.forcedKDByPlayer
      then
        -- wait until we're on the ground again
        -- delay doing anything until the player is near us
        -- and if the player IS near us during this state for some amount of time
        -- we'll force start battle
        debugPalette("Debug_DarkMagenta")

        local _t = 0
        repeat
          -- Manual break
          if me.OnGround > 0
          and got_hit(me)
          and _t > 16 -- just in case!
          then
            trace("ouch!")
            --break
          end

          if me.OnGround > 0
          and me:hasTag("Standing")
          then
            trace(_t)
            _t = _t + 1
          end

          debugPalette("Debug_Magenta")
          wait()
        until _t > 60 * 1
        or me.InBattle > 0
        or me.WasInBattle > 0 

        -- No more of this loop!
        param.forcedKDByPlayer = false
      elseif not param.goToPatternSetup
      then
        debugPalette("Debug_Red")
        trace("[1]")

        local function bossCheck()
          if bossBursted() then
            me.InvinFlashTime = 500  -- longer for us
          end
        end

        -- first lets wait until grounded --
        repeat
          debugPalette("Debug_DarkRed")
          wait()
          -- Quick Break --
          bossCheck()
          if param.manuallyLeftFromBattle then
            debugPalette("Debug_LightRed")
            break
          end
        until me.OnGround > 0
        and snakes_ok()
        --and isPlatformAlive()

        -- Delay it, but with a visual instead 
        trace("[2]")
        for t = 0, 60 do
          debugPalette("Debug_White") 
          wait()
          -- Quick Break --
          bossCheck()
          if param.manuallyLeftFromBattle then
            debugPalette("Debug_Dark")
            break
          end
        end

        -- then let's JUMP --
        trace("[3]")
        debugPalette("Debug_Green")
        repeat
          action_jumpToFlying_withBarrier()
          wait()
          -- Quick Break --
          bossCheck()
          if param.manuallyLeftFromBattle then
            debugPalette("Debug_LightGreen")
            break
          end
        until me.OnGround <= 0 

        -- then let's go up for 56f --
        trace("[4]")
        debugPalette("Debug_Blue")
        local _t = 0
        while _t <= param.maxFramesToAscend
        and platform_ok()
        do
          addInput(dirs.U)
          wait()
          bossCheck()

          if not platform_ok()
          then 
            break
          end

          if me:hasTag("Platform_Idle")
          then
            _t = _t + 1
          end

          trace( _t )
          debugPalette("Debug_Blue")
          
          if _t == param.maxFramesToAscend
          then
            param.goToPatternSetup = true
          end

          local oy = param.origin.y + param.flightYOffset
          local dx, dy = dist_point(me.XPos, oy)

          trace("disty: %i" % { dy })
          if math.abs(dy) <= 40
          then
            param.goToPatternSetup = true
            param.wantsToRechargeBarrier = false
            param.manuallyLeftFromBattle = false
            -- dangerous... --
            me.LightR = 100
            --
            param.focusTime = 0 
            param.patternTime = 0 
            param.patternAtkType = 0
            param.patternAtkSubType = 0 
            break
          end
        end
        trace("[5]")

        -- This is probably where we hand the logic off
        debugPalette("Debug_Yellow")
      elseif param.goToPatternSetup
      then
        debugDraw()
        param.manuallyLeftFromBattle = false

        -- HERE COMES THE COMPLEX JUNK --
        local xBasePoint = param.chakraGatePoint.x
        local xLimit = 160 * 22  -- 18.5
        local xDist = ( player.XPos - me.XPos ) 
        local xDistLS = xDist * me.FacingDir
        local xWaveAdd = math.cos(math.rad(param.flyingTime * (180/360))) * 160 * 12
        -- draw_point( xBasePoint + xWaveAdd, me.YPos )

        local xGoal = clamp(xBasePoint + xWaveAdd, xBasePoint - xLimit, xBasePoint + xLimit )
        local xHomeDist = ( xBasePoint - me.XPos )
        local xMargin = 32
        local yBase = param.origin.y + param.flightYOffset
        local ySink = clamp(param.focusTime, 0, param.focusMax) 
        ySink = scale(ySink, 0, param.focusMax, 0, 1)
        ySink = ySink * 160 * -6.5
        local yTarget = yBase + ySink
        local pDistX, pDistY = dist_player()

        -- Flying Overrides, From Kala --
        local dist_x =  ( xGoal ) - me.XPos
        local dist_y =  ( yTarget ) - me.YPos
        local xvel_max = 20
        local yvel_max = 20

        local function manualFly( allowTurns )
          local turnOK = allowTurns or false
          me.XVel = (dist_x / 2) * me.FacingDir
          me.XVel = clamp(me.XVel, -xvel_max, xvel_max)
          me.YVel = dist_y / 8
          me.YVel = clamp(me.YVel, -yvel_max, yvel_max)

          me.LightZ = AIFlyingHash
          me.Friction = 0
          me.Weight = 0
          me.PhysicsType = 0
          
          if pDistX < -3
          and param.turnCooldown <= 0
          and turnOK
          then
            trace("Turning around")
            addInput(dirs.B)
            param.turnCooldown = 15
          end
        end

        draw_point( me.XPos + dist_x, me.YPos + dist_y )

        -- Timers --
        if not isTangled(me)
        and me.Hitpause <= 0 
        then 
          param.time = param.time + 1

          if isPlayerInFocus()
          then
            if not param.event_TriggerVFXEvent 
            then
              doGameEvent("Event_SnakeQueen_ChakraGate_On")
              param.event_TriggerVFXEvent = true
            end
            me.PalFXSelf = param.palFX_playerFocus -- hash("Incarnation_BattleEnded_Disappear")
            me.PalFXSelfTime = param.palFX_time
          else
            if param.event_TriggerVFXEvent 
            then
              doGameEvent("Event_SnakeQueen_ChakraGate_Off")
              param.event_TriggerVFXEvent = false
            end
            me.PalFXSelf = 0 
          end
        end

        -- Flying Timer --
        if not isTangled(me)
        and me.Hitstun <= 0
        and me.Hitpause <= 0 
        and not me:hasTag("Platform_SummonPose")
        then
          param.flyingTime = param.flyingTime + 1
        end

        -- Event Emits --
        if param.focusTime >= param.focusMax
        and false
        then
          while not me:hasTag("Hitstun")
          do
            --action_FakeHit()
            break
            --wait()
          end

          -- Manual hit!
          local dX = dist_player()
          --me.XVel = dX / 80
          --me.YVel = 32

          param.forcedKDByPlayer = true
          param.goToPatternSetup = false
          param.focusTime = 0

          doGameEvent("Event_SnakeQueen_ChakraGate_Hurt")
          param.event_TriggerVFXEvent = false
          me.PalFXSelf = param.palFX_playerFocus -- hash("Incarnation_BattleEnded_Disappear")
          me.PalFXSelfTime = 11
        end

        -- The Pattern Itself --
        if me.Hitstun > 0
        and me:hasTag("Hitstun")
        then
          trace("!oh no!")
          param.goToPatternSetup = false
        elseif param.bossPhase <= 1
        then
          debugPalette("Debug_Orange")
          local moveOK = true
          --
          if not isBarrierAlive()
          then
            action_toggleBarrier()
          elseif isBarrierAlive()
          then
            if param.wantsToRechargeBarrier
            and param.timeWithoutBarrier > 60 * 2.5
            then
              if not me:hasTag("Platform_SummonPose")
              then
                action_summonPose()
              elseif me:hasTag("Platform_SummonPose")
              and param.summonPoseTime <= 60 * 7
              then
                action_summonPose_hold()
              elseif me:hasTag("Platform_SummonPose")
              and (param.summonPoseTime > 60 * 7
              or barrierHP() >= 100
              )
              then
                param.timeWithoutBarrier = 0 
                param.patternAtkType = 0 
                param.wantsToRechargeBarrier = false
              end
            elseif not me:hasTag("Attack")
            and param.attackCooldown <= 0
            then
              if param.timeWithoutBarrier <= 0
              and param.patternAtkType >= 3
              -- and barrierHP() <= param.barrierRegainThresholds[param.bossPhase]
              -- and false
              then
                param.wantsToRechargeBarrier = true
                param.attackCooldown = 120
                param.patternAtkType = 0 
              elseif (param.flyingTime % 140) == 0
              and param.timeWithoutBarrier <= 0
              and param.patternAtkType < 3
              then
                trace("Regular Pillar")
                me.LightX = 0 
                attack_pillar()
                param.attackCooldown = 33
                param.patternAtkType = param.patternAtkType + 1
              end
            end
          end

          if me.Hitstun > 0
          --or me.Hitpause > 0 
          then
            moveOK = false
          end

          if moveOK
          then
            local turnOK = not me:hasTag("Platform_SummonPose")
            manualFly(turnOK) 
          end
        elseif param.bossPhase == 2
        then
          debugPalette("Debug_Blue")
          local moveOK = true
          --
          if not isBarrierAlive()
          then
            action_toggleBarrier()
          elseif isBarrierAlive()
          then
            if param.wantsToRechargeBarrier
            then
              if param.timeWithoutBarrier > 60 * 2.2
              then
                if not me:hasTag("Platform_SummonPose")
                then
                  action_summonPose()
                elseif me:hasTag("Platform_SummonPose")
                and param.summonPoseTime <= 60 * 10
                then
                  action_summonPose_hold()
                elseif me:hasTag("Platform_SummonPose")
                and (param.summonPoseTime > 60 * 10
                or barrierHP() >= 100
                )
                then
                  param.timeWithoutBarrier = 0 
                  param.wantsToRechargeBarrier = false
                end
              elseif player:hasTag("ChakraGateOK")
              and param.timeWithoutBarrier <= 60 * 1
              then
                if not me:hasTag("Attack")
                and not me:hasTag("Fireball")
                and param.attackCooldown <= 0
                then
                  trace("PANIC! WaveOut Pillar")
                  attack_pillar()
                  me.LightX = hash("Wave_MiddleOut")
                  me.LightR = math.min(me.LightR, 0)
                  param.attackCooldown = 120
                  param.patternAtkType = 0 
                end
              end
            elseif not me:hasTag("Attack")
            and param.attackCooldown <= 0
            then
              if param.timeWithoutBarrier <= 0
              and param.patternAtkType >= 3
              then
                trace("WaveOut Pillar")
                attack_pillar()
                me.LightX = hash("Wave_MiddleOut")
                me.LightR = math.min(me.LightR, 0)
                param.attackCooldown = 120
                param.patternAtkType = 0 
              elseif (param.flyingTime % 140) == 0
              and param.timeWithoutBarrier <= 0
              and param.patternAtkType < 3
              then
                trace("Regular Pillar")
                me.LightX = 0 
                attack_pillar()
                param.attackCooldown = 33
                param.patternAtkType = param.patternAtkType + 1
              end
            end
          end

          if me.Hitstun > 0
          --or me.Hitpause > 0 
          then
            moveOK = false
          end

          if moveOK
          then
            local turnOK = not me:hasTag("Platform_SummonPose")
            manualFly(turnOK) 
          end
        elseif param.bossPhase == 3
        then
          debugPalette("Debug_Yellow")
          local moveOK = true
          --
          if not isBarrierAlive()
          then
            action_toggleBarrier()
          elseif isBarrierAlive()
          then
            if param.wantsToRechargeBarrier
            and param.timeWithoutBarrier > 60 * 1
            then
              if param.focusTime > 0
              and player:hasTag("ChakraGateOK")
              then
                if not me:hasTag("Attack")
                and not me:hasTag("Fireball")
                and param.attackCooldown <= 0
                then
                  me.LightX = 0 
                  attack_pillar()
                end
              elseif not me:hasTag("Platform_SummonPose")
              then
                action_summonPose()
              elseif me:hasTag("Platform_SummonPose")
              and param.summonPoseTime <= 60 * 10
              then
                action_summonPose_hold()
              elseif me:hasTag("Platform_SummonPose")
              and (param.summonPoseTime > 60 * 10
              or barrierHP() >= 100
              )
              then
                param.timeWithoutBarrier = 0 
                param.wantsToRechargeBarrier = false
              end
            elseif not me:hasTag("Attack")
            and param.attackCooldown <= 0
            then
              if param.timeWithoutBarrier <= 0
              and param.patternAtkType >= 3
              then
                trace("WaveOut Pillar")
                attack_pillar()
                me.LightX = hash("Wave_MiddleOut")
                me.LightR = math.min(me.LightR, 0)
                param.attackCooldown = 120
                param.patternAtkType = 0 
              elseif (param.flyingTime % 140) == 0
              and param.timeWithoutBarrier <= 0
              and param.patternAtkType < 3
              then
                trace("Regular Pillar")
                me.LightX = 0 
                attack_pillar()
                param.attackCooldown = 33
                param.patternAtkType = param.patternAtkType + 1
              end
            end
          end

          if me.Hitstun > 0
          --or me.Hitpause > 0 
          then
            moveOK = false
          end

          if moveOK
          then
            local turnOK = not me:hasTag("Platform_SummonPose")
            manualFly(turnOK) 
          end
        end
      else
        -- UH OH --
        trace("WE SHOULD NEVER GET HERE?!")
        debugPalette("Debug_Blink")
        wait()
      end
    else
      trace("NOT A BOSS")
      debugPalette("Debug_Blink")
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local me = getMe()
  local toReturn = DefaultOnBattleAttack(battleData)

  if has_spawn_command("Boss")
  then
    param.isBoss = true
  end

  local dirChance = random(1000)
  local actionChance = random(1000)

  trace("HP %02.0f" % { me.LifePct * 100} .. "%%")

  -- Don't ever go back to being hit
  param.forcedKDByPlayer = false

  -- Default
  attack_punch(toReturn)

  -- Directions
  if actionChance > 667
  then
    attack_bite(toReturn)
  end


  -- Going to be dumb instead of smart
  -- Use the lowest tier and actually check inside
  local prevPhase = param.bossPhase

  if me.LifePct <= param.battleExitThresholds[0]
  then
    -- 3rd 
    if me.LifePct <= param.battleExitThresholds[2]
    --and param.bossPhase == 2
    then
      trace("Phase 3 we OUT")
      param.bossPhase = 3
    elseif me.LifePct <= param.battleExitThresholds[1]
    --and param.bossPhase == 1
    then
      trace("Phase 2 we OUT")
      param.bossPhase = 2
    elseif  me.LifePct <= param.battleExitThresholds[0]
    --and param.bossPhase == 0
    then
      trace("Phase 1 we OUT")
      param.bossPhase = 1
    end

    if prevPhase ~= param.bossPhase
    then
      action_jumpToFlying_withBarrier(toReturn)

      param.manuallyLeftFromBattle = true
      param.forcedKDByPlayer = false
      param.goToPatternSetup = false
      param.time = 0 
      --
      param.flyingTime = 0
      param.attackCooldown = 0 
      param.summonPoseTime = 0 
      param.timeWithoutBarrier = 0
      param.focusTime = 0 
      --
      param.patternTime = 0 
      param.patternAtkType = 0
      param.patternAtkSubType = 0 
    end
  end

  return toReturn
end --OnBattleAttack()