require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function block()
  doMotion("X+WorldGuard")
end

local function attack_kick( data )
    if data ~= nil then 
        data.dir = dirs.D
        --data.dir = dirs.N
        data.pressAction = false
    end
    doMotion("D+Attack")
    param.attackCooldown = 20
end

local function attack_salute( data )
    if data ~= nil then
        data.dir = dirs.U
        data.pressAction = false
    end
    doMotion("U+Attack")
    param.attackCooldown = 40
end

local function attack_gunshot( data )
    if data ~= nil then 
        data.dir = dirs.N
        data.pressAction = false
    end
    doMotion("X+Attack")
    param.attackCooldown = 30
end

local function attack_grenadeToss( data )
    if data ~= nil then 
        data.dir = dirs.N
        data.pressAction = true
    end
    doMotion("X+Heruka")
    param.attackCooldown = 30
end

local function dist_origin()
    return dist_point(param.origin.x, param.origin.y)
end

local function dist_goal()
  local me = getMe()
  return param.goalX - me.XPos, param.origin.y - me.Ypos 
end

local function dist_lastPlayerPos()
    local me, player = getMe(), getPlayer()
    local x_dist = param.lastPlayerPos.x - me.XPos
    local y_dist = param.lastPlayerPos.y - me.YPos
    --
    return x_dist, y_dist
end

local function dist_friendly()
    local me = getMe()
    local range = _ARROW_NOTICE_RANGE
    local l,r,u,d = me.XPos - range, me.XPos + range, me.YPos + range, me.YPos - range
    local sprites = getSpritesInBox(
        l, r, u, d,
        0, me
    )
    local found_buddy = false
    local _x, _y, _spr = nil, nil, nil
    
    for _, spr in ipairs(sprites) do 
        if spr.MyName == hash("Soldier") then
            _x, _y = spr.XPos, spr.Ypos
            _spr = spr
            found_buddy = true
        end
    end

    if not found_buddy then return nil end
    return _x - me.XPos, _y - me.YPos, _spr
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)
  
  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(target)
  
  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end

  param.time = 0
  param.turnWaitTime = 0
  param.blockTime = 0
  param.attackCooldown = 0
  param.attackStartup = 0

  param.roamRangeX = 160 * 8
  param.roamRangeFudgeDist = 75
  param.sightRangeX = 160 * 10
  param.sightRangeY = { top = 160 * 1.5, bottom = 160 * -1 }
  param.closeAttackRangeX = 160 * 5
  param.attackRangeX = 160 * 8.5
  param.attackRangeY = { top = 160 * 2.5, bottom = 160 * -1.5 }
  param.guardRangeX = 160 * 4
  --
  param.sightRangeRageX = 160 * 21
  param.sightRangeRageY = { top = 160 * 3, bottom = 160 * -1 }
  param.attackRangeRageX = 160 * 16
  param.attackRangeRageY = { top = 160 * 2.5, bottom = 160 * -1.5 }
  

  param.goalX = param.origin.x
  param.goalX = param.goalX + ((param.roamRangeX/2) * (me.FacingDir * -1))
  
  param.bossBurstedMode = false
end

-- The world update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    if aiDebug() then
      draw_point(param.origin.x, param.origin.y)
      draw_point(param.goalX, param.origin.y)
      draw_bar(param.attackStartup, 0, 15, 3, {255, 255, 0})
      draw_bar(param.attackCooldown, 0, 60, 2, {255, 0, 0})
      draw_bar(param.turnWaitTime, 0, 180, 1, {255, 0, 255})
      draw_bar(param.blockTime, 0, 60, 0, {0, 255, 255})
    end

    if tick_ok() then
      if not isInSightRange(target) then
        param.turnWaitTime = math.max(0, param.turnWaitTime - 1)
      end
      param.attackCooldown = math.max(0, param.attackCooldown - 1)
      param.blockTime = math.max(0, param.blockTime - 1)
    end

    -- co-op 
    if bossBursted() 
    then
      param.turnWaitTime = 0
      param.blockTime = 0
      param.attackCooldown = 20
      param.attackStartup = 0
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 5 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end

    local dist_goalPoint = dist_point(param.goalX, me.YPos)
    if math.abs( dist_goalPoint ) <= me.WorldCollisionWidth/2 then
      local side = param.origin.x - me.XPos
      if side >= 0 then side = 1 else side = -1 end 
      param.goalX = param.origin.x + (param.roamRangeX/2 * (side))
      param.turnWaitTime = 45 + random(67)
    end
    if update_ok()
    then
      if (dist_arrow() and math.abs( dist_arrow() ) <= param.guardRangeX)
      then
        param.blockTime = math.max(45, param.blockTime)
      elseif (dist_sprite(target) > 0 and dist_sprite(target) <= param.guardRangeX)
      and player:hasTag("Attack")
      then
        param.blockTime = math.max(20, param.blockTime)
      end

      if param.blockTime > 0 
      then
        block()
        param.attackStartup = 0 
      elseif blocked_hit() then 
        param.blockTime = math.max(15, param.blockTime) 
        param.attackStartup = 0
      elseif isInSightRange(target)
      then
        if player:hasTag("Attack")
        and not me:hasTag("Attack")
        and dist_sprite(target) > 0 then
          param.blockTime = 25
          param.attackStartup = 0 
        elseif ( param.bossBurstedMode or sprite_vulnerable(target) ) 
        then
          if isInAttackRange(target)
          and dist_sprite(target) >= 0
          and param.attackCooldown <= 0
          and param.attackStartup >= 5
          then
            local attackChance = random(1000)
            if dist_sprite(target) > 300
            then
              attack_gunshot()
            else
              attack_kick()
            end
            param.attackStartup = 0 
          elseif isInAttackRange(target)
          and dist_sprite(target) >= 0
          and param.attackCooldown <= 0
          then
            param.attackStartup = param.attackStartup + 1
          else -- if math.abs(dist_player()) > 160 * 3 then
            param.attackStartup = 0
            if dist_sprite(target) > param.closeAttackRangeX
            or dist_sprite(target) < -3
            then
              approach_withTurn(target.XPos, 6)
            end
          end
        end
      elseif param.turnWaitTime > 0 then
        -- we're just waitin'
      else
        approach_withTurn(param.goalX, 8)
      end
    end
    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
    local toReturn = DefaultOnBattleAttack(battleData)
    local me = getMe()

    local dirChance = random(1000)
    local actionChance = random(1000)

    local gunUnloaded = me:getFlagValue(12)
    local gotInterrupted = me:getFlagValue(14)

    if gunUnloaded
    then
      debugPalette("Debug_Red")
      trace("Gun Unloaded")
      if dirChance >= 500
      then
        attack_kick(toReturn)
      else
        attack_gunshot(toReturn)
      end
    else
      trace("Gun Loaded")
      debugPalette("Debug_Green")
      if dirChance >= 750
      then
        attack_kick(toReturn)
      else
        attack_gunshot(toReturn)
      end
    end

    if aiDebug() then
      attack_grenadeToss(toReturn)
    end

    return toReturn
end --OnBattleAttack()