require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function isInSightRange(_spr)
  local me = getMe()
  local spr = _spr or getPlayer()
  if isInRange(spr, me.XPos, me.YPos, param.sightRangeX, param.sightRangeY)
  then
    return true
  end
  return false
end

local function isInAttackRange(_spr)
  local me = getMe()
  local spr = _spr or getPlayer()
  if isInRange(spr, me.XPos, me.YPos, param.attackRangeX, param.attackRangeY)
  then
    return true
  end
  return false
end

local function attack_shotgun(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
  trace("Shotgun")
  param.attackCooldown = 120
end

local function attack_baseRam(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 120
end

local function attack_teslaZap(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  trace("Tesla Zap")
  doMotion("X+Attack")
end

local function attack_mortar(data)
  local me = getMe()
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  trace("Mortar")
  doMotion("U+Attack")
end

local function attack_mortarMulti(data)
  local me = getMe()
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  trace("Multi Mortar")
  doMotion("D+Attack")
end

local function isArtilleryTopper()
  local me = getMe()

  if me.LightG == 1
  then
    return true
  else
    return false
  end

  return false
end

local function isTeslaTopper()
  local me = getMe()

  if me.LightG == 3
  then
    return true
  else
    return false
  end

  return false
end

local function isBossTank()
  local me = getMe()

  local boss_hash = hash("BossVersion")

  if me.SpawnCommand == boss_hash
  or me.SpawnCommand2 == boss_hash
  then
    return true
  else
    return false
  end

  return false
end

-- the initialization function for this AI
function OnInit()
  -- Accessible anywhere in this file after Initialization
  local me, player = getMe(), getPlayer()
  local players = getPlayers()

  -- Set origin
  param.time = 0
  param.attackCooldown = 0 

  param.origin = { x = me.XPos, y = me.YPos, z = me.ZPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end
  --
  param.sightRangeX = 160 * 18
  param.sightRangeY = 160 * 15
  --param.sightRangeY = { top = 160 * 1.5, bottom = 160 * -1 }
  param.attackRangeX = 160 * 18
  param.attackRangeY = 160 * 9
  --param.attackRangeY = { top = 160 * 2.5, bottom = 160 * -1.5 }

  param.wantedToStartBattle = false
  param.startBattleCooldown = 0 
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  
  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end

  while true do
    me, player = getMe(), getPlayer()

    if update_ok()
    then
      -- Tick --
      param.startBattleCooldown = param.startBattleCooldown - 1

      if not player_vulnerable(me)
      and param.startBattleCooldown <=0
      and not param.wantedToStartBattle
      and isInAttackRange()
      and not isBossTank()
      and not player:hasTag("HerukaHustle")
      then
        trace("OK, Going to attempt battle!")
        param.wantedToStartBattle = me:startBattle(player)
        param.startBattleCooldown = 30
      end
    end
  
    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  
  local dirChance = random(1000)
  local actionChance = random(1000)

  --[[ 
  ---- Boss Tank
  --]]
  if isBossTank()
  then
    trace("Boss Tank")
    -- Starts with this by default
    if isArtilleryTopper()
    then
      -- Default to shotgun
      attack_shotgun(toReturn)

      if dirChance > 667 -- 1/3
      then
        attack_mortar(toReturn)
      elseif dirChance < 334 -- 1/3
      then
        attack_mortarMulti(toReturn)
      end
    -- ~30%? health, or some such
    elseif isTeslaTopper()
    then
      attack_teslaZap(toReturn)
    -- In case we have no Top
    else
      attack_baseRam(toReturn)
    end
  --[[ 
  ---- Tesla
  --]]
  elseif isTeslaTopper()
  then
    trace("Tesla Choice")
    attack_teslaZap(toReturn)
  --[[ 
  ---- Artillery
  --]]
  elseif isArtilleryTopper()
  then
    trace("Artillery Choice")
    if actionChance > 700
    then
      attack_mortar(toReturn)
    elseif dirChance > 600
    then
      attack_mortarMulti(toReturn)
    else
      attack_shotgun(toReturn)
    end
  else
    trace("NoTopper Choice")
    attack_baseRam(toReturn)
  end

  return toReturn
end --OnBattleAttack()