require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function block()
  doMotion("X+WorldGuard")
end

local function action_armoredFloat(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = true
    data.meterBurn = false
  end
  doMotion("U+Dash")
end

local function attack_burst(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Heruka")
end

local function attack_rekka(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Heruka")
  param.attackCooldown = 33
end

local function attack_throw(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Heruka")
  param.attackCooldown = 50
end

local function attack_wingFlap(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_vanishDash(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 47
end

local function attack_summon(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_newButterfly(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+ShootBtn")
end

local function GetPoint( s )
  local _x, _y, _z = getObjectPos(s)
  -- We want a table!
  return { x = _x, y = _y }
end

local function isInSurveyRange()
  local me, player = getMe(), getPlayer()

  if dist_player() > -param.surveyRangeX/2
  and dist_player() < param.surveyRangeX/2
  and select(2,dist_player()) < param.surveyRangeY.top
  and select(2,dist_player()) > param.surveyRangeY.bottom
  then
    return true
  end
  return false
end

local function isInSightRange()
  local me, player = getMe(), getPlayer()

  if dist_player() > -param.sightRangeX/2
  and dist_player() < param.sightRangeX/2
  and select(2,dist_player()) < param.sightRangeY.top
  and select(2,dist_player()) > param.sightRangeY.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange()
  local me, player = getMe(), getPlayer()

  if dist_player() > -param.sightRangeX/2
  and dist_player() < param.sightRangeX/2
  and select(2,dist_player()) < param.sightRangeY.top
  and select(2,dist_player()) > param.sightRangeY.bottom
  then
    return true
  end
  return false
end

local function bulletDist()
  local _x, _y = nil, nil
  local _ret = nil
  if select(1,dist_blang()) ~= nil then
    return dist_blang()
  elseif select(1,dist_spear()) ~= nil then
    return dist_spear()
  elseif  select(1,dist_arrow()) ~= nil then
    return dist_arrow()
  end
  return -256^2, -256^2
end

local function vec_dist_origin()
  local me = getMe()
  local vm = { x = me.XPos, y = me.YPos }
  local vo = { x = param.origin.x, y = param.origin.y }

  local result = vec_dist( vm, vo )

  return result
end

local function vec_dist_finalPoint()
  local me = getMe()
  local vm = { x = me.XPos, y = me.YPos }
  local vo = { x = param.finalPoint.x, y = param.finalPoint.y }

  local result = vec_dist( vm, vo )

  return result
end

local function dist_origin()
  local me = getMe()
  return param.origin.x - me.XPos, param.origin.y - me.Ypos 
end

local function isTotemSummonPose()
  local me = getMe()

  if me:hasTag("Totems_Summon_Pose")
  then
    return true
  else
    return false
  end

  return false
end

local function isTotemSlashesPose()
  local me = getMe()

  if me:hasTag("Totems_Slashes_Pose")
  then
    return true
  else
    return false
  end

  return false
end

local function isObsidianPose()
  local me = getMe()

  if me:hasTag("Obsidian")
  --and me:hasTag("Standing")
  then
    return true
  else
    return false
  end

  return false
end

local function hasFlag13()
  local me = getMe()

  if me.LightZ > 0
  then
    return true
  else
    return false
  end

  return false
end

local butterfly_hash = hash("Butterflies")
local function howManyButterflies()
  local me = getMe()

  local c = me:getChildren()
  local count = 0
  for _, c in ipairs(c) do
    if c.MyName == butterfly_hash then 
      count = count + 1
    end
  end

  return count 
end

-- the initialization function for this AI
function OnInit()
  -- Accessible anywhere in this file after Initialization
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end

  param.finalPoint = { x = 0, y = 0, z = 0 }
  param.finalPoint = GetPoint("Spawn_Tzitzi_FinalPoint")

  param.time = 0
  param.attackCooldown = 0

  --
  param.surveyTime = 0 
  param.surveyRangeX = 160 * 13
  param.surveyRangeY = { top = 160 * 4, bottom = 160 * -9 }
  param.sightRangeX = 160 * 10
  param.sightRangeY = { top = 160 * 1.5, bottom = 160 * -1 }
  param.attackRangeX = 160 * 7
  param.attackRangeY = { top = 160 * 2.5, bottom = 160 * -1.5 }

  param.distSeekOrigin = 30
  param.speedBonus_goToOrigin =
  { min = 2, max = 20, range = 160 * 3 }
  param.speedBonus_goToOrigin_FAST =
  { min = 2, max = 320, range = 160 * 16 }
  --
  param.isTotemHarassment = false
  if has_spawn_command("TotemHarassment")
  then
    param.isTotemHarassment = true
  end

  param.isObsidianWind = false
  if has_spawn_command("ObsidianWind")
  or has_spawn_command("Obsidian")
  then
    param.isObsidianWind = true
  end

  param.burstStage = 0
  param.burstTiers = {
    [0] = 0.80,
    [1] = 0.60,
    [2] = 0.40,
    [3] = 0.20,
    [4] = 0.01,
    [5] = 0.001,
    [6] = 0.0001,
  }

  param.vanishAcqTime = 90 -- This is what takes to vanish to you
  param.vanishedOnce = false
end

-- The world update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end

  while true do
    me, player = getMe(), getPlayer()
    if aiDebug() then
      draw_bar(param.time, 0, 180, 2, {128, 192, 128})
      draw_bar(param.attackCooldown, 0, 60, 1, {255, 0, 0})
      draw_bar(param.surveyTime, 0, 120, 0, {255, 128, 255})
    end

    if tick_ok() and not is_hurt(me) then
      param.attackCooldown = math.max(0, param.attackCooldown - 1)
    end

    if isInSurveyRange() then
      param.surveyTime = param.surveyTime + 1
    else
      param.surveyTime = math.max(0, param.surveyTime - 1)
    end

    if tick_ok() then

    if param.isObsidianWind then
      debugPalette("Debug_Red")
      if isObsidianPose()
      then
        if hasFlag13()
        then
          trace("HEYO!")
          if vec_dist_finalPoint() > 24
          then
            fly_to(param.finalPoint.x, param.finalPoint.y, param.speedBonus_goToOrigin_FAST)
          end
        elseif vec_dist_origin() > 24
        then
          debugPalette("Debug_DarkRed")
          fly_to(param.origin.x, param.origin.y, param.speedBonus_goToOrigin)
        end
      elseif not isObsidianPose()
      then
        -- WE SHOULD BE IN THAT POSE
        action_armoredFloat()
      else
        -- NOTHING -- 
      end
    elseif param.isTotemHarassment
    then
      trace( howManyButterflies() ) 
      if isTotemSummonPose()
      then
        param.time = param.time + 1
        if player:hasTag("ChakraGateOK")
        and param.time >= 60
        and howManyButterflies() >= 1 
        then
          attack_newButterfly()
          param.time = 0
        elseif howManyButterflies() <= 0
        then
          attack_newButterfly()
          param.time = 0
        end
      else
        trace("No...")
      end
    else
    -- We're looking for bad friends
        if not param.vanishedOnce then
          if isInSurveyRange() then
            if param.attackCooldown <= 0
            and param.surveyTime > param.vanishAcqTime
            and me.WasInBattle <= 0 
            then
              attack_vanishDash()
              param.vanishedOnce = true
            else
              if dist_player() < 0 then
                turn_around(7)
              end
            end
          end
        -- We can chase you on the ground now!
        else
          if isInSightRange()
          and me.WasInBattle <= 0 
          then
            if isInAttackRange()
            and param.attackCooldown <= 0 
            and me.OnGround > 0 
            then
              local actionChance = random(1000)
              if actionChance > 500 then
                attack_throw()
              else
                attack_rekka()
              end
            end
          else
            -- nah son
            param.time = param.time + 1
            if param.time % 72 == 0 
            and param.time > 0 
            then
              param.time = 0
              param.time = param.time - random(80)
              turn_around(7)
            end
          end
        end
      end
    end
    wait()
  end -- the while loop
end -- the function

-- The battle action function for this AI 
function OnBattleAttack(battleData)
  local me = getMe()
  local toReturn = DefaultOnBattleAttack(battleData)

  local dirChance = random(1000)
  local actionChance = random(1000)

  -- Special
  if actionChance > 850 
  and param.isObsidianWind
  then
    attack_wingFlap(toReturn)
  elseif actionChance > 725
  then
    attack_throw(toReturn)
  elseif dirChance > 500
  then
    attack_vanishDash(toReturn)
  else
    attack_rekka(toReturn)  
  end

  if aiDebug()
  then
    --attack_throw(toReturn)
    attack_rekka(toReturn)
  end
  
  -- Obsdian Wind stuff
  if param.isObsidianWind
  then
    if me.LifePct <= param.burstTiers[4]
    and param.burstStage <= 4
    then
      attack_burst(toReturn)
      me.LightZ = 1
      param.burstStage = 5
    elseif me.LifePct <= param.burstTiers[3]
    and param.burstStage <= 3
    then
      attack_burst(toReturn)
      param.burstStage = 4
    elseif me.LifePct <= param.burstTiers[2]
    and param.burstStage <= 2
    then
      attack_burst(toReturn)
      param.burstStage = 3
    elseif me.LifePct <= param.burstTiers[1]
    and param.burstStage <= 1
    then
      attack_burst(toReturn)
      param.burstStage = 2
    elseif me.LifePct <= param.burstTiers[0]
    and param.burstStage <= 0
    then
      attack_burst(toReturn)
      param.burstStage = 1
    end

    trace(param.burstStage)
  end
  return toReturn
end --OnBattleAttack()