require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")
require("boids")

--
local hash_AIFlying = hash("AIFlying")

local function vec_dist_player()
    local me, player = getMe(), getPlayer()
    -- 
    local vec_me = { x = me.XPos, y = me.YPos }
    local vec_pl = { x = player.XPos, y = player.YPos }

    return vec_dist( vec_me, vec_pl)
end

local function vec_dist_origin()
    local me = getMe()
    -- 
    local vec_me = { x = me.XPos, y = me.YPos }
    local vec_pt = { x = param.origin.x, y = param.origin.y }

    return vec_dist( vec_me, vec_pt)
end

local function sight_attackRange()
    local me, player = getMe(), getPlayer()

    if math.abs( dist_player() ) <= param.attackRangeX
    and select(2,dist_player()) <= param.attackRangeY.top
    and select(2,dist_player()) >= param.attackRangeY.bottom then
        return true
    end
    return false
end

local function attack_splorch(data)
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = false
      data.pressDash = false
      data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function attack_stabShort(data)
  if data ~= nil then
      data.dir = dirs.D
      data.pressAction = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 15
end

local function attack_stab(data)
  if data ~= nil then
      data.dir = dirs.N
      data.pressAction = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 21
end

local function attack_stabLong(data)
  if data ~= nil then
      data.dir = dirs.U
      data.pressAction = false
  end
  doMotion("U+Attack")
  param.attackCooldown = 28
end

local _SPRITE_DETECTION_RANGE = 160 * 4
local function getSpritesOfMe()
  local me = getMe()
  local range = _SPRITE_DETECTION_RANGE / 2
  local l, r, u, d = me.XPos - range, me.XPos + range, me.YPos + range, me.YPos - range

  local sprites = getSpritesInBox(l, r, u, d, nil, me)
  local sprites_toReturn = {}
  for i, spr in ipairs(sprites) do
    if spr.MyName == me.MyName
    and not is_hurt(spr) 
    and not isTangled(spr) then
      sprites_toReturn[#sprites_toReturn+1] = spr
    end
  end

  return sprites_toReturn
end

local function getHurtSpritesOfMe()
  local me = getMe()
  local range = _SPRITE_DETECTION_RANGE / 2
  local l, r, u, d = me.XPos - range, me.XPos + range, me.YPos + range, me.YPos - range

  local sprites = getSpritesInBox(l, r, u, d, nil, me)
  local sprites_toReturn = {}
  for i, spr in ipairs(sprites) do
    if spr.MyName == me.MyName
    and (is_hurt(spr) 
    or isTangled(spr)) then
      sprites_toReturn[#sprites_toReturn+1] = spr
    end
  end

  return sprites_toReturn
end

local function BoundRanges()
  local bL, bR, bT, bB = 0
  bL = param.origin.x - param.flockBoundsX 
  bR = param.origin.x + param.flockBoundsX
  bT = param.origin.y + param.flockBoundsY.top
  bB = param.origin.y + param.flockBoundsY.bottom
  return bL, bR, bT, bB
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }

  if me:getSpawnpointPos() ~= nil then
      param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end

  --
  param.time = 0
  param.randomOffset = random(1000) % 360
  param.attackCooldown = 0
  --
  param.searchingForPlayer = false
  param.useBoids = true

  param.isHazard = false
  if has_spawn_command("AscentHazard") then
    param.isHazard = true
  end
  param.randomPosOffset = { x = 0, y = 0 }
  --
  if has_spawn_command("Stationary") then
      param.stationaryMode = true
  end
  --
  param.flockBoundsX = 160 * 8
  param.flockBoundsY = { top = 160 * 8, bottom = 160 * -3}

  if param.isHazard
  then
    param.flockBoundsX = 160 * 2
    param.flockBoundsY = { top = 160 * 2, bottom = 160 * -2 }
  end

  param.sightRange = 160 * 6.5
  param.attackRangeX = 160 * 3.5
  param.attackRangeY = { top = 160 * 4, bottom = 160 * -4 }
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
 
  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end

  while true do
    me, player = getMe(), getPlayer()

    -- Ticks --
    if tick_ok()
    and not is_hurt()
    then
      param.time = param.time + 1
      param.attackCooldown = math.max(0, param.attackCooldown - 1)
    end

    -- Stationary --
    if param.stationaryMode then
      if vec_dist_origin() > 80
      then
        fly_to(param.origin.x, param.origin.y)
      else
        -- We're chillin'
      end
    elseif param.isHazard
    and tick_ok()
    then
      -- every 20f? 
      if ( param.time % 40 ) == 0
      then
        local _x = -160 * 2
        _x = _x + random(160 * 2)
        local _y = -160 * 2
        _y = _y + random(160 * 2)
        param.randomPosOffset.x = _x
        param.randomPosOffset.y = _y
      end

      if update_ok()
      then
        fly_to(param.origin.x + param.randomPosOffset.x, param.origin.y + param.randomPosOffset.y)
      end
    -- Boids -- 
    elseif tick_ok() 
    and param.useBoids
    then
      -- Set up the vars --
      local vecCenter = boids_mass(sprites)
      local vecAvgVel = boids_vel(sprites)

      local sprites = getSpritesOfMe()
      local hurtSprites = getHurtSpritesOfMe()

      local vecBiasFactor = 300
      if param.isHazard
      then
        vecBiasFactor = 120
      end

      local vecBias = boids_bias(me, sprites, vecBiasFactor)
      local vecBounds = boids_bounds(me, BoundRanges() ) 
      local vecTendency = boids_tendency(
        { x = param.origin.x, y = param.origin.y, z = param.origin.z},
        me
      )

      local vecRepulse = boids_tendency(
        { x = player.XPos, y = player.YPos, z = player.ZPos},
        me
      )
      local vecHurtAvoid = boids_mass(hurtSprites)

      local avoidPointFactor = 22
      local avoidTendencyFactor = 80
      local biasFactor = 100
      local boundsFactor = 30

      if param.isHazard
      then
        avoidTendencyFactor = 300
        biasFactor = 50
      end

      -- Let's do some prep math --
      local vecAvoidPoint = {x = 0, y = 0, z = 0}
      vecAvoidPoint.x = vecAvoidPoint.x / 1
      vecAvoidPoint.y = vecAvoidPoint.y / 1
      vecAvoidPoint.z = vecAvoidPoint.z / 1
  
      vecCenter.x = (vecCenter.x - me.XPos) / 100
      vecCenter.y = (vecCenter.y - me.YPos) / 100
      vecCenter.z = (vecCenter.z - me.ZPos) / 100
  
      local xVelWS = me.XVel * me.FacingDir
      vecAvgVel.x = (vecAvgVel.x - xVelWS)  / 8
      vecAvgVel.y = (vecAvgVel.y - me.yVel) / 8
      vecAvgVel.z = (vecAvgVel.z - me.zVel) / 8

      -- Apply --
      if not is_hurt(me)
      then
        local xDiff = vecBounds.x * boundsFactor
        + vecTendency.x / avoidPointFactor
        - vecAvoidPoint.x / avoidTendencyFactor

        local yDiff = vecBounds.y * boundsFactor
        + vecTendency.y / avoidPointFactor --300
        - vecAvoidPoint.y / avoidTendencyFactor

        if #sprites > 0 then
          xDiff = xDiff
          + vecCenter.x
          + vecBias.x / biasFactor
          + vecAvgVel.x

          yDiff = yDiff
          + vecCenter.y
          + vecBias.y / biasFactor
          + vecAvgVel.y
        end

        -- 
        local playerVecDist = vec_dist(
          { x = player.XPos, y = player.YPos},
          { x = me.XPos, y = me.YPos, }
        )
        local facingDestPoint = { x = param.origin.x, y = param.origin.y } 

        -- Seek player --
        if playerVecDist <= 160 * 5 
        and not param.isHazard
        then
          facingDestPoint.x = player.XPos
          facingDestPoint.y = player.YPos

          xDiff = xDiff + vecRepulse.x * 0.15
          yDiff = yDiff + vecRepulse.y * 0.15
        end

        -- Apply the stuff
        local velDelta = { x = xDiff, y = yDiff }
        local deltaLen = vec_len(velDelta)

        local MAX_DELTA = 1.75
        if deltaLen > MAX_DELTA then
          xDiff = xDiff * (MAX_DELTA / deltaLen)
          yDiff = yDiff * (MAX_DELTA / deltaLen)
        end

        xVelWS = xVelWS + xDiff
        me.XVel = xVelWS * me.FacingDir
        me.YVel = me.YVel + yDiff

        local velVec = { x = me.XVel, y = me.YVel }
        local velLen = vec_len(velVec)

        local MAX_VELOCITY = 26
        if velLen > MAX_VELOCITY then
          me.XVel = me.XVel * (MAX_VELOCITY / velLen)
          me.YVel = me.YVel * (MAX_VELOCITY / velLen)
        end

        -- Face the tendency
        local shouldTurnAround = (facingDestPoint.x - me.XPos) * me.FacingDir
        if shouldTurnAround < 0 then
          --turn_around(7)
          addInput(dirs.B)
        end

        -- Setup useful variables on the sprite --
        me.LightZ = hash_AIFlying
        me.Friction = 0
        me.Weight = 0
        me.PhysicsType = 0

        if aiDebug()
        then
          trace( tostring(param.isHazard) ) 
          draw_point( me.XPos + vecAvoidPoint.x, me.YPos + vecAvoidPoint.y )
          draw_point( me.XPos + vecTendency.x, me.YPos + vecTendency.y )
          draw_point( me.XPos + vecBias.x, me.YPos + vecBias.y ) 
        end

        --[[ Attack --
        if playerVecDist < 160 * 2 then
          wait(7)
          attack_stabShort()
        end
        ]]
      end
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()

  local dirChance = random(1000)

  attack_stab(toReturn)

  if dirChance > 667 then
    attack_stabShort(toReturn)
  elseif dirChance < 334 then
    attack_stabLong(toReturn)
  end

  return toReturn
end --OnBattleAttack()