require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function dist_origin()
  local me = getMe()
  local dx = param.origin.x - me.XPos
  local dy = param.origin.y - me.YPos
  return dx, dy 
end

local function isInSightRange(_spr)
  local me = getMe()
  local spr = _spr or getPlayer()
  
  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end
  
  if isInRange(spr, me.XPos, me.YPos, w, h)
  then
    return true
  end
  return false
end

local function isInAttackRange(_spr)
  local me = getMe()
  local spr = _spr or getPlayer()
  
  local w = param.attackRangeX
  local h = param.attackRangeY
  if param.bossBurstedMode then
    w = param.attackRangeRageX
    h = param.attackRangeRageY
  end
  
  if isInRange(spr, me.XPos, me.YPos, w, h)
  then
    return true
  end
  return false
end

local function attack_nm_summon(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function attack_nm_sacrifice(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
end

local function attack_death(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.meterBurn = false
  end
  doMotion("X+Heruka")
end

local function attack_doom(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("U+Attack")
end

local function attack_sacrifice(data)
  if data ~= nil then
    data.dir = dirs.D
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("D+Attack")
  param.attackCooldown = 47
end

local function attack_cast(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
    data.meterBurn = false
  end
  doMotion("X+Attack")
end

local function debugStuff()
  draw_bar(param.attackCooldown, 0, 100)
end

-- the initialization function for this AI
function OnInit()
  -- Accessible anywhere in this file after Initialization
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.time = 0
  param.attackCooldown = 0 

  param.origin = { x = me.XPos, y = me.YPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end

  param.sightRangeX = 160 * 18
  param.sightRangeY = 160 * 15
  --param.sightRangeY = { top = 160 * 1.5, bottom = 160 * -1 },
  param.attackRangeX = 160 * 13
  param.attackRangeY = 160 * 9
  --param.attackRangeY = { top = 160 * 2.5, bottom = 160 * -1.5 },
  
  param.sightRangeRageX = 160 * 27
  param.sightRangeRageY = 160 * 15
  param.attackRangeRageX = 160 * 22
  param.attackRangeRageY = 160 * 13

  param.isNotorious = false
  if me.WeaponType == hash("Necromancer")
  then
    param.isNotorious = true
  end

  param.isNecromancer = false
  if me.WeaponType == hash("Necromancer")
  then
    param.isNecromancer = true
  end
  
  param.bossBurstedMode = false
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()
    
  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end
  
  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    if aiDebug() then
      debugStuff()
    end

    if tick_ok() then
      if not me:hasTag("Attack") then
        param.attackCooldown = math.max(0, param.attackCooldown - 1)
      end
      param.time = param.time + 1
    end

    -- co-op 
    if bossBursted() 
    then
      param.attackCooldown = 20
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 7 
    end

    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    
    if ( param.bossBurstedMode or sprite_vulnerable(target) ) then
      if param.attackCooldown <= 0
      and isInAttackRange(target)
      and dist_point(target.XPos, target.YPos) >= 0 
      then
        attack_sacrifice()
        debugPalette("Debug_Red")
      elseif isInSightRange(target)
      and dist_point(target.XPos, target.YPos) < 0 
      then
        turn_around(11)
        debugPalette("Debug_Blue")
      end
    end

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  
  local dirChance = random(1000)
  local actionChance = random(1000)

  --
  if battleData.turnsSinceSummon == nil
  then
    battleData.turnsSinceSummon = 0
  end
  --

  if param.isNotorious
  then
    trace("turns %i" % { battleData.turnsSinceSummon })
    battleData.turnsSinceSummon = battleData.turnsSinceSummon + 1
    
    if battleData.turnsSinceSummon >= 3
    then
      trace("Sacrifice")
      attack_nm_sacrifice(toReturn)
      battleData.turnsSinceSummon = 0 
    else
      trace("Summon")
      attack_nm_summon(toReturn)
    end
  -- Normal 
  else
    attack_sacrifice(toReturn)
  end

  --if actionChance > 500 then
  --  attack_doom(toReturn)
  --end

  return toReturn
end --OnBattleAttack()