require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function draw_point(px, py, _size)
  local me = getMe()

  local size2 = 640
  local size = 400

  if _size ~= nil 
  then
    size = _size
  end

  if aiDebug() then
    drawBox(
      px - 400/2,
      px + 400/2,
      py + 400/2,
      py - 400/2,
      160, -160,
      255, 255, 255
    )
    drawBox(
      px - 400/2,
      px + 400/2,
      py + 400/2,
      py - 400/2,
      160, -160,
      255, 255, 255
    )
    drawBox(
      px - size / 2,
      px + size / 2,
      py + size / 2,
      py - size / 2,
      160, -160,
      255, 0, 32
    )
  end 
end

-- This function don't want visual entities!!
local function GetPoint( s )
  local _x, _y, _z = getObjectPos(s)
  -- We want a table
  return { x = _x, y = _y }
end

local function ArenaSide()
  local me, player = getMe(), getPlayer()
  -- Requires this to be defined anyways
  local _x = param.origin.x
  local diff = me.XPos - _x
  if diff >= 0 then 
    return 1
  elseif diff < 0 then
    return -1
  else
    return 0 
  end
end

local function drawAngle(_angle, _dist, _color)
  local me = getMe()

  local color = { 255, 255, 0 }
  if _color ~= nil then 
    if _color[1] and _color[2] and _color[3] then
      color = _color
    end
  end

  if aiDebug() then
    local distance = _dist or 512
    local _a = math.rad(_angle)
    local _x = distance * math.cos(_a)
    local _y = distance * math.sin(_a)
    local _size = 128

    drawBox(
      (me.XPos + _x) - _size/2, (me.XPos + _x) + _size/2,
      (me.YPos + _y) - _size/2, (me.YPos + _y) + _size/2,
      -1, 1, unpack(color)
    )
  end

end

local function steerToPoint(px, py, _addButtons, _params )   
  local me = getMe()

  local vT = { x = px, y = py }
  local vP = { x = me.XPos, y = me.YPos }

  local vMeToTargetWS = { x = (vT.x - vP.x), y = (vT.y - vP.y) }
  local vMeToTargetLS = { x = vMeToTargetWS.x * me.FacingDir, y = vMeToTargetWS.y }
  local distToTarget = vec_len(vMeToTargetWS)

  local angleToTargetLS = vec_angle(vMeToTargetLS)
  local facingAngleLS = me.JumpsLeft
  local _XSpeed  = 0 

  -- calculate difference between the two angles
  local dAngleLS = angleToTargetLS - facingAngleLS
  
  --correct this to +180 / -180 so we know which hemisphere to aim at
  
  if(dAngleLS > 180) then
    dAngleLS = dAngleLS - 360
  elseif(dAngleLS < -180) then
    dAngleLS = dAngleLS + 360
  end
  
  local correctedFacingAngleLS = facingAngleLS
  if ( correctedFacingAngleLS > 360 ) then
    correctedFacingAngleLS = correctedFacingAngleLS - 360
  elseif ( correctedFacingAngleLS < 0 ) then
    correctedFacingAngleLS = correctedFacingAngleLS + 360
  end
  
  local dAngle2LS = angleToTargetLS - correctedFacingAngleLS 

  local choice = nil 
  local addButtons = {}

  if _addButtons ~= nil then 
    addButtons = _addButtons
  end

  if inRange(dAngleLS, -270, -5)
  then 
    if false
    then
      addInput(dirs.F, unpack(addButtons))
    else
      addInput(dirs.UF, unpack(addButtons))
    end
    choice = "Right" -- UF
    _XSpeed = -1
  elseif inRange(dAngleLS, 5, 270)
  then
    if false
    then
      addInput(dirs.B, unpack(addButtons))
    else
      addInput(dirs.UB, unpack(addButtons))
    end
    choice = "Left" -- UB
    _XSpeed = 1
  else   
    addInput(dirs.U, unpack(addButtons))
    choice = "Forward"  -- U
  end

  -- If we're in-between the speed, just snapit
  if ( math.abs( dAngleLS ) < me.DashesLeft/10 ) 
  and tick_ok()
  then
    local _r = me.ArtRotation

    _r = _r - dAngleLS
    me.ArtRotation = _r
    --
    me.XVel = math.cos(_r) * me.DashesLeft
    me.YVel = math.sin(_r) * me.DashesLeft
  end

  if aiDebug() then
    --trace("dAngle  %03.2f" % { dAngleLS})
    --trace("dAngle2 %03.2f" % { dAngle2LS})
    --trace(dAngle2LS)

    drawAngle(facingAngleLS, 900, {255, 0, 0}) -- red
    drawAngle(angleToTargetLS, 900, {0, 255, 0}) -- green
    --drawAngle(dAngleLS, 900, {255, 255, 0}) -- yellow
    drawAngle(facingAngleLS + dAngle2LS, 1280, {0, 0, 255}) -- blue
    drawAngle(facingAngleLS + (me.DashesLeft * _XSpeed)/10 , 900, {255, 255, 0}) -- yellow
    --drawBox( px - 200, px + 200, py - 200, py + 200,
    --    512, 512 - 128, 255, 255, 255)
  end

  return distToTarget
end

local function HasSpawnCommand(cmd)
  local me = getMe()
  --
  if me.SpawnCommand == hash(cmd)
  or me.SpawnCommand2 == hash(cmd) 
  then
    return true
  end
  return false
end

local function HasSpawnCommand2(cmd)
  local me = getMe()
  --
  if me.SpawnCommand2 == hash(cmd)
  then
    return true
  end
  return false
end

local function gotHurt()
  local me, player = getMe(), getPlayer()
  if me.Alerted >= 30 or me.Hitstun > 0 then
      return true
  end
  return false
end

local function vec_dist_origin()
  local me = getMe()
  local vm = { x = me.XPos, y = me.YPos }
  local vo = { x = param.origin.x, y = param.origin.y }

  local result = vec_dist( vm, vo )

  return result
end

local function dist_origin()
  local me = getMe()
  return param.origin.x - me.XPos, param.origin.y - me.Ypos 
end

local function vec_dist_player()
  local me, player = getMe(), getPlayer()
  local vM = { x = me.XPos, y = me.YPos }
  local vP = { x = player.XPos, y = player.YPos }

  return vec_dist( vP, vM )
end

local function action_headPop(toReturn)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
    data.pressDash = false
    data.meterBurn = false
  end
  doMotion("X+SpawnCommand=\"HeadPopOut\"")
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  --
  param.Modes = {
    Init = -1,
    Start = 0,
    AtOrigin = 100,
    ReturnToOrigin = 101,
    RetreatFromOrigin = 110,
    Harass = 500,
    SetupForPattern = 600,
    DoPattern = 610,
    Retreat = 700,
    Die = 800,
    GoToOrigin = 1000
  }

  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }
  param.lastHitPoint = { x = param.origin.x, y = param.origin.y}
  param.time = 0
  param.deadTime = 0
  param.patternTime = 0
  -- 
  param.stageChanged = false
  param.prevStage = 0
  param.stage = 1
  param.stageTiers = {
    1.00,  -- 100%
    0.70,  -- 70%
    0.50,  -- 50%
    0.30,  -- 30%
    0.001,  -- 1%
  }

  param.damageThresholdMax = 0
  param.rageTiers = {
    [0] = 0,
    850,
    250 * 3,  -- 750dmg
    250 * 2,  -- 500dmg
    250 * 2,  -- 500dmg
    250 * 9999,  -- never..
  }

  param.prevHealthbar = me.Lifebar
  param.damageTaken = 0
  param.rageTurns = 0
  param.enraged = false
  param.forceEnraged = false
  param.forceGoToOrigin = false

  param.startedBattle = false
  param.moveMode = param.Modes.Init
  param.prevMoveMode = param.Modes.Init

  param.pointToVisit = 1
  param.patternToPick = 0
  param.patternWanted = 0
  param.patternsAttempted = 0
  param.maxPatternsBeforeHead = 2
  param.timesHitWhileOut = 0

  param.finalizedDeath = false

  param.sideAtStartOfPattern = 0
  param.patternModifier = 0

  param.prevSpeed = 0 
  param.prevAngle = me.JumpsLeft

  me.JumpsLeft = 90 -- Change our angle.

  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end

  param.debug_shortVulnPause = false

  param.hash_TailSwitch = hash("Tail_Switch")
  if me.SpawnCommand == param.hash_TailSwitch then
    me.IsBoss = 0 
  end

  if aiDebug()
  then
    --me.Lifebar = (me.MaxLife) * param.stageTiers[0]
    --param.stage = 1
    --param.prevStage = 1
    --param.enraged = true
    --param.stageChanged = true
  end
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  
  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do
  
  end
  
  while true do
    -- Debug
    if aiDebug() then
      local function GetMode()
        for k, v in pairs(param.Modes) do 
          if param.moveMode == v then return tostring(k) end
        end
        return -1
      end
      draw_physbox(me)
      drawBox( me.XPos - 350/2, me.XPos + 350/2, me.YPos - 350/2, me.YPos + 350/2,
        512, 386, 255, 255, 128)
      if param.moveMode ~= param.prevMoveMode then
        trace("======== "  .. GetMode() .. " ========" )
      end

      -- WE DEBUGGIN'
      if player:hasTag("ChakraGateOK")
      then
        me.Lifebar = 1
        player.Lifebar = 1
      end
    end

    -- Points
    -- Inside middle of wall = 40au R
    local Point = {
      OriginBottom = { x = param.origin.x, y = param.origin.y + (800 * -5)},

      --
      --Arena_BottomLeft = GetPoint("Mark_Arena_BottomLeft"),
      --Arena_WallMidLeft = GetPoint("Mark_Arena_WallLeftMiddle"),
      Arena_EntryLeft = GetPoint("Mark_Arena_EntryLeft"),
      --Arena_BottomRight = GetPoint("Mark_Arena_BottomRight"),
      --Arena_WallMidRight = GetPoint("Mark_Arena_WallRightMiddle"),
      Arena_EntryRight = GetPoint("Mark_Arena_EntryRight"),
      
      --
      SetupPatternPoint = {
        { x = param.origin.x, y = param.origin.y + (800 * -5) }
      },
      OriginBottomToSideRight = {
        { x = param.origin.x + (160 * 40), y = param.origin.y },
        { x = param.origin.x + (160 * 40), y = param.origin.y + (160 * 7) },
      },
      OriginBottomToSideLeft = {
        { x = param.origin.x + (160 * -40), y = param.origin.y },
        { x = param.origin.x + (160 * -40), y = param.origin.y + (160 * 7) },
      },

      -- NORMAL PATTERN! --
      AttackPattern_Normal_LTR = {
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 8)},
        { x = param.origin.x + (160 * -24), y = param.origin.y + (160 * 3), MaxDelta = 4 },
        { x = param.origin.x + (160 * -16), y = param.origin.y + (160 * 8), MaxDelta = 4 }, 
        { x = param.origin.x + (160 * -8), y = param.origin.y + (160 * 0), MaxDelta = 8 },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 4), MaxDelta = 4 },
        { x = param.origin.x + (160 * 8), y = param.origin.y + (160 * 1), MaxDelta = 8 },
        { x = param.origin.x + (160 * 16), y = param.origin.y + (160 * 8), MaxDelta = 4 },
        { x = param.origin.x + (160 * 24), y = param.origin.y + (160 * 3), MaxDelta = 4 },
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 8), MaxDelta = 4 },
      },
      AttackPattern_Normal_RTL = {
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 8)},
        { x = param.origin.x + (160 * 24), y = param.origin.y + (160 * 3), MaxDelta = 4 },
        { x = param.origin.x + (160 * 16), y = param.origin.y + (160 * 8), MaxDelta = 4 }, 
        { x = param.origin.x + (160 * 8), y = param.origin.y + (160 * 0), MaxDelta = 8 },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 4), MaxDelta = 4 }, -- BOTTOM
        { x = param.origin.x + (160 * -8), y = param.origin.y + (160 * 1), MaxDelta = 8 },
        { x = param.origin.x + (160 * -16), y = param.origin.y + (160 * 8), MaxDelta = 4 },
        { x = param.origin.x + (160 * -24), y = param.origin.y + (160 * 3), MaxDelta = 4 },
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 8), MaxDelta = 4 },
      },

      -- DASH OVERLOADS --
      AttackPattern_Normal_Dashing_LTR = {
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 5) },
        { x = param.origin.x + (160 * -24), y = param.origin.y + (160 * 7.5) },
        { x = param.origin.x + (160 * -16), y = param.origin.y + (160 * 7.5) }, 
        { x = param.origin.x + (160 * -8), y = param.origin.y + (160 * 5) },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 0) }, -- BOTTOM
        { x = param.origin.x + (160 * 8), y = param.origin.y + (160 * 5) },
        { x = param.origin.x + (160 * 16), y = param.origin.y + (160 * 7.5) },
        { x = param.origin.x + (160 * 24), y = param.origin.y + (160 * 7.5) },
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 5) },
      },
      AttackPattern_Normal_Dashing_RTL = {
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 5) },
        { x = param.origin.x + (160 * 24), y = param.origin.y + (160 * 7.5) },
        { x = param.origin.x + (160 * 16), y = param.origin.y + (160 * 7.5) }, 
        { x = param.origin.x + (160 * 8), y = param.origin.y + (160 * 5) },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 0) }, -- BOTTOM
        { x = param.origin.x + (160 * -8), y = param.origin.y + (160 * 5) },
        { x = param.origin.x + (160 * -16), y = param.origin.y + (160 * 7.5) },
        { x = param.origin.x + (160 * -24), y = param.origin.y + (160 * 7.5) },
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 5) },
      },

      -- DASHING ONLY -- 
      AttackPattern_Dashing_LTR = {
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 5) },
        { x = param.origin.x + (160 * -16), y = param.origin.y + (160 * 10) }, 
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * -3) }, -- BOTTOM
        { x = param.origin.x + (160 * 16), y = param.origin.y + (160 * 9) }, 
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 5) },
      },

      AttackPattern_Dashing_RTL = {
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 5) },
        { x = param.origin.x + (160 * 16), y = param.origin.y + (160 * 10) }, 
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * -3) }, -- BOTTOM
        { x = param.origin.x + (160 * -16), y = param.origin.y + (160 * 9) }, 
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 5) },
      },

      -- SCALES (HIGH) -- 
      AttackPattern_Scales_High_LTR = {
        { x = param.origin.x + (160 * -27), y = param.origin.y + (160 * 15) },
        { x = param.origin.x + (160 * -12), y = param.origin.y + (160 * 16) }, 
        { x = param.origin.x + (160 * -4), y = param.origin.y + (160 * -5) }, -- MIDDLE
        { x = param.origin.x + (160 * 4), y = param.origin.y + (160 * -5) }, -- MIDDLE
        { x = param.origin.x + (160 * 12), y = param.origin.y + (160 * 16) }, 
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 13) },
      },

      AttackPattern_Scales_High_RTL = {
        { x = param.origin.x + (160 * 27), y = param.origin.y + (160 * 15) },
        { x = param.origin.x + (160 * 12), y = param.origin.y + (160 * 16) }, 
        { x = param.origin.x + (160 * 4), y = param.origin.y + (160 * -5) }, -- MIDDLE
        { x = param.origin.x + (160 * -4), y = param.origin.y + (160 * -5) }, -- MIDDLE
        { x = param.origin.x + (160 * -12), y = param.origin.y + (160 * 16) }, 
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 13) },
      },

      -- SCALES (LOW) -- 
      AttackPattern_Scales_Low_LTR = {
        { x = param.origin.x + (160 * -24), y = param.origin.y + (160 * 0) },
        { x = param.origin.x + (160 * -12), y = param.origin.y + (160 * 10) }, 
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 10) }, -- MIDDLE
        { x = param.origin.x + (160 * 12), y = param.origin.y + (160 * 9) }, 
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 2) },
      },

      AttackPattern_Scales_Low_RTL = {
        { x = param.origin.x + (160 * 24), y = param.origin.y + (160 * 0) },
        { x = param.origin.x + (160 * 12), y = param.origin.y + (160 * 10) }, 
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 10) }, -- MIDDLE
        { x = param.origin.x + (160 * -12), y = param.origin.y + (160 * 9) }, 
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 2) },
      },

      -- HOMING --
      AttackPattern_Homing_LTR = {
        { x = param.origin.x + (160 * -34), y = param.origin.y + (160 * 9) },
        { x = param.origin.x + (160 * -16), y = param.origin.y + (160 * 14) },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * -4), 
          MaxDelta = 1, MaxDeltaIgnore = 160 * 10 },
        { x = param.origin.x + (160 * -16), y = param.origin.y + (160 * -16), 
          MaxDelta = 1, MaxDeltaIgnore = 160 * 10 },
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 10) },
      },

      AttackPattern_Homing_RTL = {
        { x = param.origin.x + (160 * 34), y = param.origin.y + (160 * 9) },
        { x = param.origin.x + (160 * 16), y = param.origin.y + (160 * 14) },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * -4), 
          MaxDelta = 1, MaxDeltaIgnore = 160 * 10 },
        { x = param.origin.x + (160 * -16), y = param.origin.y + (160 * -16),
          MaxDelta = 1, MaxDeltaIgnore = 160 * 10 },
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 10) },
      },

      -- DANGEROUS DRILL --
      AttackPattern_Drill_LTR = {
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 2) },
        { x = param.origin.x + (160 * -26), y = param.origin.y + (160 * 6) },
        { x = param.origin.x + (160 * -11), y = param.origin.y + (160 * 16), 
        MaxDelta = 2, MaxDeltaIgnore = 160 * 10  },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 8), 
        MaxDelta = 2, MaxDeltaIgnore = 160 * 10  },
        { x = param.origin.x + (160 * 11), y = param.origin.y + (160 * -3) },
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 0) },
      },

      AttackPattern_Drill_RTL = {
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 2) },
        { x = param.origin.x + (160 * 26), y = param.origin.y + (160 * 6) },
        { x = param.origin.x + (160 * 11), y = param.origin.y + (160 * 16), 
        MaxDelta = 2, MaxDeltaIgnore = 160 * 10  },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 8), 
        MaxDelta = 2, MaxDeltaIgnore = 160 * 10  },
        { x = param.origin.x + (160 * -11), y = param.origin.y + (160 * -3) },
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 0) },
      },

      -- DANGEROUS DRILL --
      AttackPattern_Drill_Below_LTR = {
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 2) },
        { x = param.origin.x + (160 * -20), y = param.origin.y + (160 * 2) },
        { x = param.origin.x + (160 * -11), y = param.origin.y + (160 * 7) },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 9), 
        MaxDelta = 2, MaxDeltaIgnore = 160 * 10  },
        { x = param.origin.x + (160 * 11), y = param.origin.y + (160 * 5) },
        { x = param.origin.x + (160 * 20), y = param.origin.y + (160 * 2) },
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 0) },
      },

      AttackPattern_Drill_Below_RTL = {
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 2) },
        { x = param.origin.x + (160 * 20), y = param.origin.y + (160 * 2) },
        { x = param.origin.x + (160 * 11), y = param.origin.y + (160 * 7) },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 9), 
        MaxDelta = 2, MaxDeltaIgnore = 160 * 10  },
        { x = param.origin.x + (160 * -11), y = param.origin.y + (160 * 5) },
        { x = param.origin.x + (160 * -20), y = param.origin.y + (160 * 2) },
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 0) },
      },

      -- BITE --
      AttackPattern_Bite_LTR = {
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * 19) },
        { x = param.origin.x + (160 * -24), y = param.origin.y + (160 * 29) },
        { x = param.origin.x + (160 * -20), y = param.origin.y + (160 * 27),
          Homing = true,},
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * -6), 
          MaxDelta = 4, MaxDeltaIgnore = 160 * 10, Homing = true, Grab = true},
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * -8), Grab = true},
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * -16), Grab = true },
        -- The Release Loop
        { x = param.origin.x + (160 * -9), y = param.origin.y + (160 * -9), Grab = true },
        { x = param.origin.x + (160 * -9), y = param.origin.y + (160 * 0), Grab = true },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 8), Release = true },
        -- Go Away now
        { x = param.origin.x + (160 * 9), y = param.origin.y + (160 * -9)},
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * -9)},
      },

      AttackPattern_Bite_RTL = {
        { x = param.origin.x + (160 * 32), y = param.origin.y + (160 * 19) },
        { x = param.origin.x + (160 * 24), y = param.origin.y + (160 * 29) },
        { x = param.origin.x + (160 * 20), y = param.origin.y + (160 * 27), 
          Homing = true },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * -6), 
          MaxDelta = 2, MaxDeltaIgnore = 160 * 10, Homing = true, Homing = true, Grab = true},
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * -8), Grab = true},
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * -16), Grab = true },
        -- The Release Loop
        { x = param.origin.x + (160 * 9), y = param.origin.y + (160 * -9), Grab = true },
        { x = param.origin.x + (160 * 9), y = param.origin.y + (160 * 0), Grab = true },
        { x = param.origin.x + (160 * 0), y = param.origin.y + (160 * 8), Release = true },
        -- Go Away now
        { x = param.origin.x + (160 * -9), y = param.origin.y + (160 * -9)},
        { x = param.origin.x + (160 * -32), y = param.origin.y + (160 * -9)},
      }
    }

    local Pattern = {
      Normal      = 0,
      Homing      = 10,
      Dash        = 60,
      Drill       = 20,
      DrillBelow  = 25,
      Scales      = 30,
      ScalesLow   = 40,
      ScalesHigh  = 50,
      Bite        = 100,
    }

    -- Add up our damages
    local hp_diff = me.Lifebar - param.prevHealthbar
    if hp_diff < 0 then
      param.damageTaken = param.damageTaken + math.abs(hp_diff)
      trace("HP Changed from %06i to %06i. damageTaken %i" % {
        param.prevHealthbar, me.Lifebar,
        param.damageTaken
      })
    end
    param.prevHealthbar = me.Lifebar
    --

    if param.damageTaken >= param.rageTiers[param.stage]
    and tick_ok()
    then
      param.enraged = true
      param.damageTaken = 0 

      trace("enraged!")
    end
    
    if ( player.HealthPct <= 0 or player:hasTag("Dead") ) 
    and me.WasInBattle <= 0 
    then
      me.DespawnDistX = 10 
      me.DespawnDistY = 10 
    end
    -- Add more points manually from the level
    --[[
    for i=1,7 do 
      local name = "Mark_Arena_Floor_%i" % { i }
      local key = "Arena_Floor%i" % { i }
      if getObjectPos(name) then
        Point[key] = GetPoint(name)
      end
    end
    ]]

    local extraButtons = { }

    if(HasSpawnCommand2("Dash")) then
      table.insert(extraButtons, btns.Dash)
    elseif(HasSpawnCommand2("Drill")) then
      table.insert(extraButtons, btns.Atk)
    elseif(HasSpawnCommand2("Rage")) then
      table.insert(extraButtons, btns.Dash)
      table.insert(extraButtons, btns.ShootBtn)
    elseif(HasSpawnCommand2("Scales")) then
      table.insert(extraButtons, btns.Atk)
      table.insert(extraButtons, btns.WorldGuard)
    end
  
    -- SpawnCommands, good luck!
    if HasSpawnCommand("Circle") then
      param.time = 0
      me.JumpsLeft = 90
      -- wait until full speed
      while me.DashesLeft < 40 do
        addInput(dirs.U, unpack(extraButtons) )
        param.time = param.time + 1
        wait()
      end
      --
      param.time = 0
      while true do
        param.time = (param.time + 1)
        if param.time % 2 == 0 then
          addInput(dirs.UF, unpack(extraButtons))
        else
          addInput(dirs.U, unpack(extraButtons))
        end

        if param.time % 30 == 0 
        then
          -- AAA --
        end

        wait()
      end
    elseif HasSpawnCommand("CircleLeft") then
        param.time = 0
        me.JumpsLeft = 90
        -- wait until full speed
        while me.DashesLeft < 50 do
            addInput(dirs.U, unpack(extraButtons))
            param.time = param.time + 1
            wait()
        end
        --
        param.time = 0
        while true do
            param.time = (param.time + 1)
            if param.time % 2 == 0 then
                addInput(dirs.UB, unpack(extraButtons))
            else
                addInput(dirs.U, unpack(extraButtons))
            end
            wait()
        end 
    elseif HasSpawnCommand("Fig8") then
        param.time = 1
        me.JumpsLeft = 90
        while true do
            trace("right")
            repeat  
                trace(me.JumpsLeft .. " / " .. param.time)
                addInput(dirs.UF, unpack(extraButtons))
                wait()
            until me.JumpsLeft <= 135 + 1
            and me.JumpsLeft >= 135 - 1
            --
            param.time = 0 
            while param.time < 25 do
                param.time = param.time + 1
                addInput(dirs.U, unpack(extraButtons))
                wait()
            end
            trace("left")
            repeat  
                trace(me.JumpsLeft .. " / " .. param.time)
                addInput(dirs.UB, unpack(extraButtons))
                wait()
            until me.JumpsLeft <= 45 + 1
            and me.JumpsLeft >= 45 - 1
            --
            param.time = 0 
            while param.time < 25 do
                param.time = param.time + 1
                addInput(dirs.U, unpack(extraButtons))
                wait()
            end
        end
    elseif HasSpawnCommand("Flat") then
        param.time = 1
        me.JumpsLeft = 0
        while param.time < 10 do
            trace(param.time)
            param.time = param.time + 1
            addInput(dirs.U, unpack(extraButtons))
            wait()
        end
        --
        while true do
            wait()
        end
    elseif HasSpawnCommand("Homing") then
        local dist = steerToPoint( 
            player.XPos,
            player.YPos
        )
        if false then --dist <= 500 then
            param.time = 0 
            while param.time < 30 do
                param.time = param.time + 1
                addInput(dirs.U, unpack(extraButtons))
                wait()
            end
        end
    -- Waiting To Start!
    elseif HasSpawnCommand("Boss") 
    or HasSpawnCommand2("Boss")
    then
      if not param.startedBattle then
        debugPalette("Debug_Dark")
        param.time = param.time + 1
        if param.time > 10 and player_vulnerable() then
          wait(30)
          param.time = 0
          --
          param.startedBattle = true
          param.moveMode = param.Modes.Init
        end
      else
        debugPalette()
        param.time = param.time + 1
        param.prevMoveMode = param.moveMode
        param.prevSpeed = me.DashesLeft

        for i = 1, #param.stageTiers do 
          if me.LifePct <= param.stageTiers[i] then 
            param.stage = i
          end
        end
        if param.prevStage ~= param.stage then
          param.stageChanged = true
          trace("Stage Changed! %i to %i" % { param.prevStage, param.stage})
        end

        -- 
        param.prevStage = param.stage

        -- This should really be called something else.
        if param.moveMode == param.Modes.Init then
          trace("Init")
          param.patternsAttempted = 0 
          param.moveMode = param.Modes.SetupForPattern
        -- just.. go away!
        elseif param.moveMode == param.Modes.Die then
          param.deadTime = param.deadTime + 1
          steerToPoint(me.XPos, me.YPos - 1000)
          --
          if param.deadTime >= 60 * 3
          and vec_dist_player() >= 4000
          then
            trace("Safe")
            if not param.finalizedDeath
            then
              me.LightZ = 1
              doGameEvent("Worm_Died")
              param.finalizedDeath = true
            end
          end
        elseif param.moveMode == param.Modes.SetupForPattern then
          debugPalette("Debug_Green")

          local function PickPattern()
            local function GetPatternFromTable(index)
              local i = index or 0
              trace(index)
              
              local pats = {
                [0] = Pattern.Dash,
                [1] = Pattern.Normal,
                [2] = Pattern.Normal,
                [3] = Pattern.Dash,
                [4] = Pattern.Dash,
                [5] = Pattern.Homing,
                [6] = Pattern.DrillBelow,
                [7] = Pattern.Drill,
                [8] = Pattern.ScalesLow,
                [9] = Pattern.Dash,
                [10] = Pattern.Dash,
              }
              i = i % #pats
              trace(i)
              if pats[i] ~= nil 
              then
                return pats[i]
              else
                return Pattern.Normal
              end
            end

            if me.Flag14
            then
              return Pattern.Normal
            else
              return GetPatternFromTable(param.patternsAttempted)
            end
            --[[
            if param.stage == 1 then
              if param.enraged then
                trace("stage 1 enraged")
                --param.forceEnraged = false
                if param.rageTurns <= 3 then
                  return Pattern.Homing
                else
                  param.rageTurns = 0
                  param.enraged = false
                end
              end
              return Pattern.Normal
            elseif param.stage == 2 then
              if param.stageChanged then
                param.forceEnraged = true
                param.enraged = false
                return Pattern.DrillBelow
              elseif param.enraged then
                trace("stage 2 enraged")
                if param.rageTurns <= 3 then
                  return Pattern.Homing
                else
                  param.rageTurns = 0
                  param.enraged = false
                end
              end
              return Pattern.Normal
            elseif param.stage == 3 then
              if param.stageChanged then
                param.forceEnraged = true
                param.enraged = false
                return Pattern.Bite
              elseif param.enraged then
                trace("stage 3 enraged")
                if param.rageTurns <= 3 then
                  return Pattern.Homing
                else
                  param.rageTurns = 0
                  param.enraged = false
                end
              end
              return Pattern.DrillBelow
            elseif param.stage == 4 then
              if param.stageChanged then
                param.enraged = false
                --param.forceEnraged = false
                local _r = random(1000)
                return _r > 500 and Pattern.ScalesHigh or Pattern.ScalesLow
              elseif param.enraged then
                trace("stage 4 enraged")
                if param.rageTurns <= 3 then
                  return Pattern.Homing
                else
                  param.rageTurns = 0
                  param.enraged = false
                end
              end
              return Pattern.Dash
            elseif param.stage == 5
            or me.Flag14
            then
              return Pattern.Normal
            else
              return Pattern.Normal
            end
            ]]
            --[[
            if param.stageChanged then
              param.patternsAttempted = 0 
            end
            ]]
          end

          -- Make sure to pick the side FIRST
          local side = ArenaSide()
          --trace(side)

          -- Pick a midpoint side depending on our side then go to it
          local side_point = { x = 0, y = 0 }
          if side >= 0 then
            --trace("Right")
            side_point = Point.Arena_EntryRight
          elseif side <= 0 then
            --trace("Left")
            side_point = Point.Arena_EntryLeft
          end

          -- Setup distance for later
          local dist = steerToPoint(
            side_point.x, side_point.y
          )
          --trace("%i,%i" % { side_point.x, side_point.y })

          -- We should just go
          if param.patternsAttempted >= 13
          then
            me.GotHit = 1
            me.Lifebar = 3
            param.moveMode = param.Modes.Die
          elseif me.Flag14 
          then
            param.moveMode = param.Modes.Die
          elseif dist < 400
          then
            param.moveMode = param.Modes.DoPattern
            param.pointToVisit = 1

            param.patternTime = 0 
            param.patternModifier = 0
            param.sideAtStartOfPattern = ArenaSide()

            -- Let turns know we're enraged
            if param.enraged then
              param.rageTurns = param.rageTurns + 1
            end

            -- Get the one we want
            param.patternToPick = PickPattern()
            --param.patternToPick = Pattern.Drill
            param.stageChanged = false
            param.forceEnraged = false
          end
        elseif param.moveMode == param.Modes.DoPattern then
          debugPalette()

          if me.Throwing > 0  then
            debugPalette("Debug_Yellow")
          end

          local goalPoint = 0
          local sideBoundary = 0
          local side = param.sideAtStartOfPattern

          if param.patternTime == 0 then
            side = ArenaSide()
            --trace("First Time Tick! it's %i" % { side })
          end
          

          -- Default pattern in case!
          local pattern = Point.AttackPattern_Normal_RTL

          -- Fetch it for real...
          -- Normal
          if param.patternToPick == Pattern.Normal then
            if side >= 0 then
              pattern = Point.AttackPattern_Normal_RTL
            elseif side < 0 then
              pattern = Point.AttackPattern_Normal_LTR
            end
          elseif param.patternToPick == Pattern.Dash then
            if side >= 0 then
              pattern = Point.AttackPattern_Dashing_RTL
            elseif side < 0 then
              pattern = Point.AttackPattern_Dashing_LTR
            end
          elseif param.patternToPick == Pattern.Homing then
            if side >= 0 then
              pattern = Point.AttackPattern_Homing_RTL
            elseif side < 0 then
              pattern = Point.AttackPattern_Homing_LTR
            end
          elseif param.patternToPick == Pattern.DrillBelow then
            if side >= 0 then
              pattern = Point.AttackPattern_Drill_Below_RTL
            elseif side < 0 then
              pattern = Point.AttackPattern_Drill_Below_LTR
            end
          elseif param.patternToPick == Pattern.Drill then
            if side >= 0 then
              pattern = Point.AttackPattern_Drill_RTL
            elseif side < 0 then
              pattern = Point.AttackPattern_Drill_LTR
            end
          elseif param.patternToPick == Pattern.ScalesLow then
            if side >= 0 then
              pattern = Point.AttackPattern_Scales_Low_RTL
            elseif side < 0 then
              pattern = Point.AttackPattern_Scales_Low_LTR
            end
          elseif param.patternToPick == Pattern.ScalesHigh then
            if side >= 0 then
              pattern = Point.AttackPattern_Scales_High_RTL
            elseif side < 0 then
              pattern = Point.AttackPattern_Scales_High_LTR
            end
          elseif param.patternToPick == Pattern.Bite then
            if side >= 0 then
              pattern = Point.AttackPattern_Bite_RTL
            elseif side < 0 then
              pattern = Point.AttackPattern_Bite_LTR
            end
          end  

          --pattern = Point.AttackPattern_Bite_LTR
          -- Put our information somewhere global, too
          param.patternWanted = pattern

          -- Steer torwards the point
          local _prevPointNum = math.max(1, param.pointToVisit - 1)
          local _prevX = pattern[_prevPointNum].x
          local _prevY = pattern[_prevPointNum].y
          --local _prevAngle = me.JumpsLeft

          local _goalX = pattern[param.pointToVisit].x
          local _goalY = pattern[param.pointToVisit].y

          --trace("Time %i" % { param.patternTime })

          local constructedPoint = { x = 0, y = 0 }

          local _min = pattern[param.pointToVisit].MinAngle
          local _max = pattern[param.pointToVisit].MaxAngle
          local _distFromPoint = pattern[param.pointToVisit].Dist 
          local _maxDelta = pattern[param.pointToVisit].MaxDelta
          local _maxDeltaDistIgnore = pattern[param.pointToVisit].MaxDeltaIgnore

          local params = {}

          local _dist = vec_len(
            { x = me.XPos, y = me.Ypos },
            { x= _goalX, y = _goalY }
          )

          -- Add inputs depending on what we want
          local function AddInputs()
            local _btns = {}

            if param.patternToPick == Pattern.Dash then
              --trace("Dash")
              table.insert(_btns, btns.Dash)
            elseif param.patternToPick == Pattern.Drill
            or param.patternToPick == Pattern.DrillBelow
            then
              --trace("Drill")
              table.insert(_btns, btns.Atk)
            elseif param.patternToPick == Pattern.Scales
            or param.patternToPick == Pattern.ScalesHigh
            or param.patternToPick == Pattern.ScalesLow
            then
              --trace("Scales")
              table.insert(_btns, btns.Atk)
              table.insert(_btns, btns.WorldGuard)
            end

            if pattern[param.pointToVisit].Grab
            and me.Throwing <= 0 
            then
              --trace("X+Heruka")
              table.insert(_btns, btns.Heruka)
            end
            
            if pattern[param.pointToVisit].Release
            and me.Throwing > 0 
            then
              --trace("X+~Heruka")
              doMotion("X+~Heruka")
            end

            if param.enraged
            or param.forceEnraged
            then
              table.insert(_btns, btns.Dash)
            end

            return _btns
          end

          --trace("JumpsLeft(angle) Before: %i" % { me.JumpsLeft })

          if true -- pattern[param.pointToVisit].Dist
          then
            local _x, _y = 0, 0
            local _vPlayer = {
              x = player.XPos - me.XPos,
              y = player.YPos - me.YPos 
            }

            _vPlayer.x = _vPlayer.x * me.FacingDir

            local _vGoal = { x = _goalX, y = _goalY }
            local _vGoalDiff = { x = _goalX - me.XPos, y = _goalY - me.YPos } 

            local _angleToGoal = vec_angle(_vGoal)
            local _angleToGoalDiff = vec_angle(_vGoalDiff)

            local _angleToPlayer = vec_angle(_vPlayer)
            _angleToPlayer = _angleToPlayer + 180

            --trace("Worm        : %2.2f" % { me.JumpsLeft })
            --trace("Worm +180   : %2.2f" % { me.JumpsLeft + 180 })
            --trace("Angle Goal  : %2.2f" % { _angleToGoal })
            --trace("Angle GDiff : %2.2f" % { _angleToGoalDiff })

            local _diffToGoal = _angleToGoalDiff - (me.JumpsLeft)
            local _flippedDiffToGoal = _angleToGoalDiff + (me.JumpsLeft)
            --trace("Angle Diff: %2.2f" % { _diffToGoal })

            --trace("Angle To Player: %2.2f" % { _angleToPlayer })
            --trace("Angle To PlyRad: %2.2f" % { math.rad(_angleToPlayer) })
            
            --draw_point( _vPlayer.x + me.XPos, _vPlayer.y + me.YPos ) 

            --trace("JumpsLeft(angle) Prev  : %i" % { param.prevAngle })
            --trace("JumpsLeft(angle) Raw   : %i" % { me.JumpsLeft })

            --[[
            if _distFromPoint ~= nil
            then
              local _a2 = math.rad(_angleToPlayer)
              _a2 = _a2 + math.pi

              local _x2 = math.cos( _a2 ) * _distFromPoint
              local _y2 = math.sin( _a2 ) * _distFromPoint

              trace("From Me   (%04i, %04i)" % { _x2 - me.XPos, _y2 - me.YPos })
              trace("From Prev (%04i, %04i)" % { _x2 - _prevX, _y2 - _prevY })

              -- Set it 
              if _dist >= 600
              then
                _goalX = _x2
                --_goalY = _y2
              end
              draw_point( _prevX + _x2, _prevY + _y2 )
            end
            ]]

            if _maxDelta ~= nil
            and type(_maxDelta) == "number"
            then
              if math.abs( _diffToGoal ) >= _maxDelta
              then
                -- Approach the goal by Delta!
                local _goal = _diffToGoal
                local _goalFlipped = (me.JumpsLeft - 180) - _goal

                --trace("Goal        : %i" % { _goal })
                --trace("Goal Flip   : %i" % { _goalFlipped })

                if _goal > 180 then
                  _goal = _goal - 360
                elseif _goal < -180 then
                  _goal = _goal + 360
                end

                local clampedDelta = clamp(_goal, -_maxDelta, _maxDelta)

                --trace("clampDelta  : %i" % { clampedDelta })
                --trace("Goal After  : %i" % { _goal })

                if _maxDeltaDistIgnore ~= nil
                and _dist > _maxDeltaDistIgnore
                then
                  me.JumpsLeft = param.prevAngle + clampedDelta
                  me.ArtRotation = me.JumpsLeft
                end
              elseif math.abs( _diffToGoal ) <= 5
              then
                -- Forcefully set this to the delta!
                me.JumpsLeft = me.JumpsLeft + _diffToGoal
                me.ArtRotation = me.JumpsLeft
              else
                me.JumpsLeft = me.JumpsLeft
                -- let's not do anything probably
              end
            end

          -- Manually check the distance before using it so we can tweak it
          if _dist >= 600
          and pattern[param.pointToVisit].Homing
          then
            _goalX = player.XPos
          end

            --[[
            if _min ~= nil
            and _max ~= nil
            then
              trace("why are these not triggering")
              local _a2 = clamp(math.rad(_angleToPlayer), _min, _max)
              local _delta = _angleToPlayer - me.JumpsLeft

              --_x = math.cos( _a ) * _dist
              --_y = math.sin( _a ) * _dist

              local _x2 = math.cos( _a2 ) * _dist
              local _y2 = math.sin( _a2 ) * _dist

              trace( "Min %i Max %i" % { _min, _max })
              trace( "(%i, %i)" % { _x2, _y2 } )
              trace( "Delta %02.2f" % { _delta } )

              local _maxDelta = 5
              local _dontExceedDelta = true

              if _dontExceedDelta
              and _delta >= _maxDelta
              then
                local clampedDelta = clamp(_delta, -_maxDelta, _maxDelta)
                --
                me.JumpsLeft = me.JumpsLeft + clampedDelta
              end

              draw_point( _prevX + _x2, _prevY + _y2 ) 
            end
            ]]
          end

          -- Show all the points if we want
          if aiDebug()
          then
            for i, p in pairs(pattern) do
              --draw_point(p.x, p.y)
            end
          end

          local dist = steerToPoint( 
              _goalX,
              _goalY,
              AddInputs(),
              params
          )

          -- Do the thing for real
          local _bonusDeltaRange = 0
          if _maxDelta ~= nil
          then 
            local _deltaSizeBonus = scale(_maxDelta, 1, 14, 10, 1)
            _bonusDeltaRange = 250 - (_deltaSizeBonus * 10)
          end
          if param.enraged
          or param.forceEnraged
          then
            _bonusDeltaRange = _bonusDeltaRange + 100
          end
          if dist < (400 + _bonusDeltaRange)
          or (dist <= 900 and param.patternTime >= 350 )
          or (dist > 1000 and param.patternTime >= 400 )
          then
            if param.pointToVisit == #pattern then
              trace("Pat %i Point %i / %i" % { param.patternsAttempted, #pattern, #pattern })
              param.moveMode = param.Modes.SetupForPattern
              param.patternsAttempted = param.patternsAttempted + 1
              param.pointToVisit = 1
            else
              trace("Pat %i Point %i / %i" % { param.patternsAttempted, param.pointToVisit, #pattern })
              param.pointToVisit = (param.pointToVisit % #pattern ) + 1
              param.patternTime = 0 
            end
          end

          if aiDebug()
          then
            draw_point( _goalX, _goalY, (400 + _bonusDeltaRange) )
          end

          param.patternTime = param.patternTime + 1
        --
        elseif param.moveMode == param.Modes.GoToOrigin then
          debugPalette("Debug_Cyan")
          if param.pointToVisit == 2 then
            steerToPoint( param.origin.x, param.origin.y )
          else
            local dist = steerToPoint( Point.OriginBottom.x, Point.OriginBottom.y )
            if dist <= 300 
            then 
              param.pointToVisit = param.pointToVisit + 1
            end
          end
          --
          if vec_dist_origin() <= 300 then
            param.moveMode = param.Modes.AtOrigin
            param.time = 0 
            param.pointToVisit = 1
          end
        elseif param.moveMode == param.Modes.AtOrigin 
        then
          debugPalette("Debug_Blink")

          draw_bar(param.time, 0, 60*7)
          draw_bar(vec_dist_origin(), 0, 350, 1)

          local maxTime = param.debug_shortVulnPause and 60 * 0.25 or 60 * 7
          param.patternsAttempted = 0 
          param.timesHitWhileOut = 0

          local scale_Speed = scale(vec_dist_origin(), 0, 300, 40, 0)
          if me.DashesLeft > 0
          then
            me.DashesLeft = scale_Speed
          end

          if param.time >= maxTime - 100 then
              debugPalette("Debug_BlinkFast")
          end
          if param.time >= maxTime
          then
            param.moveMode = param.Modes.RetreatFromOrigin
            param.time = 0 
          elseif me.WasInBattle == 1
          then 
            param.time = math.max(maxTime - 100, param.time)
          elseif me.Alerted >= 30
          then
            param.time = math.max(0, param.time + 30)
          elseif me.GotHit > 0
          then
            param.time = math.max(maxTime - 20, param.time)
          elseif param.enraged
          then
            param.time = math.max(maxTime - 60, param.time)
          end
        elseif param.moveMode == param.Modes.RetreatFromOrigin then
          debugPalette("Debug_Yellow")
          steerToPoint( Point.OriginBottom.x, Point.OriginBottom.y )
          if vec_dist( Point.OriginBottom, { x = me.XPos, y = me.YPos } ) < 350 
          then
            param.moveMode = param.Modes.SetupForPattern
            param.pointToVisit = 1
          end
        else
          trace('oh no')
          debugPalette("Debug_Blink")
        end

        param.prevAngle = me.JumpsLeft
      end
    else
      --
    end
    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
    local toReturn = DefaultOnBattleAttack(battleData)
    local me = getMe()
    
    return toReturn
end --OnBattleAttack()