require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("vec2")
require("util")
require("math_ext")
require("movement")

local function getFloorCurveY()
  local me = getMe()
  --
  local minX = param.points.CurveLeft.x
  local maxX = param.points.CurveRight.x
  local minY = param.points.CurveLeft.y
  local maxY = param.points.CurveCenter.y
  --
  local rangeX = maxX - minX
  local rangeY = maxY - minY
  local diffX = clamp(me.XPos, minX, maxX) - minX
  local ratio = diffX / rangeX
  local angle = ratio * 360
  local offsetY = math.abs( math.sin( math.rad(angle/2) ) )
  offsetY = offsetY * rangeY
  angle = math.rad( angle ) 

  local realY = minY + offsetY
  
  --trace("Range: %i" % { rangeX })
  --trace("Diff: %i" % { diffX })
  --trace("Ratio: %2.2f per" % { ratio * 100 })
  --trace("Angle: %f" % { math.deg(angle) })

  draw_point( me.XPos, minY + offsetY )
  
  return realY
end

local function steerToPoint(px, py, moveMode)
    local me, player = getMe(), getPlayer()

    -- Our current rotation; degrees, +-180
    local myAngle = me.ArtRotation  -- (-180, +180)
    myAngle = math.rad(myAngle)

    -- Vector for me, the target
    local vMe = vec_pos(me)

    local vTarget = { x = 0, y = 0 }

    if px and py
    then
      --trace("we found new targets, they're now %i %i" % { px, py })
      vTarget.x = px
      vTarget.y = py
    else
      --trace("Nope, gonna use the player position!")
      vTarget.x = player.XPos
      vTarget.y = player.YPoos
    end

    -- Vector for where I am facing
    local vMyFacing = { x = math.cos(myAngle), y = math.sin(myAngle) }
    vMyFacing = vec_norm(vMyFacing)

    -- |Me| - |Pl| 
    local vDiffWS = { x = vTarget.x - vMe.x, y = vTarget.y - vMe.y }
    local vDiffLS = { x = vDiffWS.x * me.FacingDir, y = vDiffWS.y }

    -- The actual angle to the target
    local playerAngle = math.atan2(vDiffLS.y, vDiffLS.x)
    --trace( "%15s: %+07.2f" % { "playerAngle before anything", math.deg(playerAngle) })

    -- Wrap around the angle, but also put it in degrees
    playerAngle = math.deg(playerAngle)
    if playerAngle > 180 then
      playerAngle = playerAngle - 360
    elseif playerAngle < -180 then
      playerAngle = playerAngle + 360
    end
    -- put it back --
    playerAngle = math.rad(playerAngle)

    -- Get the difference of the two angles, but also put it in degrees
    local angleDiff = playerAngle - myAngle
    angleDiff = math.deg(angleDiff)
    if angleDiff > 180 then
      --trace(">180")
      angleDiff = angleDiff - 360
    elseif angleDiff < -180 then
      --trace("<180")
      angleDiff = angleDiff + 360
    end
    -- put it back 
    angleDiff = math.rad(angleDiff)

    --trace( "%15s: " % { })
    --trace( "%15s: %+07.2f" % { "artRot", math.deg(myAngle) })
    --trace( "%15s: %+07.2f" % { "playerAngle", math.deg(playerAngle) })
    --trace( "%15s: %+07.2f" % { "angleDiff After", math.deg(angleDiff) })
    
    draw_point( vMe.x, vMe.y , { 255, 255, 255 }) -- White
    draw_point( vTarget.x, vTarget.y, { 0, 0, 255 }) -- Magenta

    --[[
    draw_point(
      vMe.x + math.cos(myAngle) * 320, 
      vMe.y + math.sin(myAngle) * 320,
    {255, 0, 0})

    draw_point(
      vMe.x + math.cos(playerAngle) * 320, 
      vMe.y + math.sin(playerAngle) * 320,
    {0, 255, 0})

    draw_point(
      vMe.x + math.cos(angleDiff) * 480, 
      vMe.y + math.sin(angleDiff) * 480,
    {0, 0, 255})
    ]]

    -- Add buttons to list, depending on our moveMode
    local addBtns = {}
    if moveMode == "Dash"
    then
      table.insert(addBtns, btns.Dash)
    elseif moveMode == "Drill"
    then
      table.insert(addBtns, btns.Attack)
    elseif moveMode == "Slow"
    then
      table.insert(addBtns, btns.WorldGuard)
    elseif moveMode == "Scales"
    then
      table.insert(addBtns, btns.Attack)
      table.insert(addBtns, btns.WorldGuard)
    elseif moveMode == "Grab"
    then
      table.insert(addBtns, btns.Heruka)
    elseif moveMode == "Dash+Grab"
    then
      table.insert(addBtns, btns.Dash)
      table.insert(addBtns, btns.Heruka)
    elseif moveMode == "Dash+Drill"
    then
      table.insert(addBtns, btns.Dash)
      table.insert(addBtns, btns.Attack)
    end

    -- Move for real 
    if angleDiff > 0
    then
      addInput(dirs.UB, unpack(addBtns))
    elseif angleDiff < 0
    then
      addInput(dirs.UF, unpack(addBtns))
    else
      addInput(dirs.F, unpack(addBtns))
    end
end

local function action_jumpTeleport( x, y, r )
  local me = getMe()

  doMotion("X+Jump")
  local _x, _y, _r = x, y, r
  if _x == nil or _y == nil or _r == nil
  then
    trace("Teleport needs an XYRot!")
    _x = 0
    _y = 0
    _r = 0 
  end
  me.WeaponType = _x
  me.DHCNum = _y 
  me.JumpsLeft = _r
end

-- Command ID table 
local Cmd = {
  Init              = 0,
  Retreat           = -1,
  Idle              = 100,
  Attack1           = 200,
  Attack2           = 201,
  Attack3           = 202,
  Attack4           = 203,
  AttackEscape      = 250,
  ChasePlayer       = 1000,
  LeaveForever      = 9000,
  CutsceneChomp     = 9001
}

-- Command ID strings
local function cmdToString(cmd)
  local strings = {
    [Cmd.Init]              = "Init",
    [Cmd.Retreat]           = "Retreat",
    [Cmd.Idle]              = "Idle",
    [Cmd.Attack1]           = "Attack1",
    [Cmd.Attack2]           = "Attack2",
    [Cmd.Attack3]           = "Attack3",
    [Cmd.Attack4]           = "Attack4",
    [Cmd.AttackEscape]      = "AttackEscape",
    [Cmd.ChasePlayer]       = "Chase Player"
  }

  if strings[cmd] ~= nil
  then
    return strings[cmd]
  else
    return "UNKNOWN COMMAND"
  end

  return nil
end

-- Can we take commands yet?
local function cmdOK()
  if param.curCmd == Cmd.Idle
  or param.curCmd == Cmd.Init
  then
    return true
  else
    return false
  end

  return false
end

local function getPoint(s)
  local name = s 
  local _x, _y, _z = getObjectPos(name, false)
  --trace("We requested point %s" % { name})
  if type(_x) == "number" and type(_y) == "number"
  then
    --trace("(%i %i)" % { _x, _y })
    return { x = _x, y = _y }
  else
    --trace("Failed to find point %s" % { s })
    return nil
  end

  return nil
end

local prog_attackOKValue = "OontBoss_AttackOK"
local prog_attackValue = "OontBoss_Command"
local prog_rage = "OontBoss_RageMode"
local prog_shouldDie = "OontBoss_Die"
local prog_inBattle = "OontBoss_InBattle"

-- Don't loop through anything if we're supposed to LEAVE
local function patternOK()
  local me = getMe()
  local player = getPlayer()
  local pv = getProgression(prog_shouldDie)
  local inBattle = getProgression(prog_inBattle)
  if player.InBattle > 0
  and not isOnScreen(me)
  then
    trace("GO DOWN AND AWAY")
    param.curCmd = Cmd.Retreat
    param.subCmd = 0
    return false 
  elseif pv <= 0
  then
    return true
    --[[
    if inBattle == 1
    then
      param.curCmd = Cmd.Retreat
      param.subCmd = 0 
      return false
    else
      return true
    end
    ]]
  else
    trace("NOT PatternOK")
    param.curCmd = Cmd.LeaveForever
    param.subCmd = -1
    return false
  end

  return false
end

local function lockAttackValue()
  setProgression(prog_attackOKValue, 0)
end

local function resetAttackValue()
  if patternOK() then
    setProgression(prog_attackValue, 0)
  end
end

local function allowAttackValue()
  setProgression(prog_attackOKValue, 1)
end

local function hasNewCommand()
  local _ret = false
  local progNow = getProgression(prog_attackValue)
  if progNow ~= param.prevAtkProg
  then
    _ret = true
  end

  -- update this value
  param.prevAtkProg = getProgression(prog_attackValue)

  return _ret
end

local function newCommandOK()
  local me = getMe()
  if not isOnScreen(me)
  then
    return true
  else
    return false
  end

  return false
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  param.points = {
    BottomCenter = getPoint("Spawn_Worm_BottomAnchor"),
    FloorLine = getPoint("Spawn_Worm_FloorLine"),
    BottomReset2 = getPoint("Spawn_Worm_BottomResetPoint2"),
    BottomReset = getPoint("Spawn_Worm_BottomResetPoint"),
    LeftAnchor = getPoint("Spawn_Worm_LeftAnchor"),
    RightAnchor = getPoint("Spawn_Worm_RightAnchor"),
    LeftAnchorToGround = getPoint("Spawn_Worm_LeftAnchorToGround"),
    RightAnchorToGround = getPoint("Spawn_Worm_RightAnchorToGround"),
    -- For EVALUATING the curve
    CurveLeft = getPoint("Spawn_Worm_CurveMarkLeft"),
    CurveRight = getPoint("Spawn_Worm_CurveMarkRight"),
    CurveCenter = getPoint("Spawn_Worm_CurveMarkCenter"),
    --
    CutsceneChomp = getPoint("Spawn_Oont_ChompMark")
  }

  --trace("(%i %i)" % { param.points[BottomCenter].x, param.points[bottomCenter].y })

  param.time = 0 
  --
  param.curCmd = Cmd.Init
  param.prevCmd = nil
  param.subCmd = 0
  --
  param.heading = 0
  param.floorYOffset = 160 * 0.55
  --
  param.prevAtkProg = 0 
  param.drillSpeed = 33 
  param.drillSpeedFast = 55 
  -- Reset this, because we care about RageMode
  setProgression(prog_rage, 0)
  -- Only reset this if we're not in cinematic death mode (1)
  if getProgression(prog_shouldDie) <= 1
  then
    setProgression(prog_shouldDie, 0)
  end
  allowAttackValue()
end

local function drawDebug()
  local me = getMe()

  if aiDebug()
  then
    local str = "%s %+i:%+i" % { cmdToString(param.curCmd), param.curCmd, param.subCmd }
    drawText( str, me.XPos, me.YPos)
    --
    --trace("prog value: %i" % { getProgression(prog_attackValue) })
  end
end

-- the main update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()

  while true do
    me, player = getMe(), getPlayer()
    ----------------------------------------------------------------------------
    -- Tick
    ----------------------------------------------------------------------------
    if tick_ok()
    then
      param.time = param.time + 1
    end

    ----------------------------------------------------------------------------
    -- True Update
    ----------------------------------------------------------------------------
    if update_ok()
    then
      if cmdOK()
      then
        if getProgression(prog_shouldDie) >= 2
        then
          param.curCmd = Cmd.CutsceneChomp
        -- Attack 1 --
        elseif getProgression(prog_attackValue) == 1
        then
          param.curCmd = Cmd.Attack1
        -- Attack 2 --
        elseif getProgression(prog_attackValue) == 2
        then
          param.curCmd = Cmd.Attack2
        -- Attack 2 --
        elseif getProgression(prog_attackValue) == 3
        then
          param.curCmd = Cmd.Attack4
        -- Escape Pose --
        elseif getProgression(prog_attackValue) == 10
        then
          param.curCmd = Cmd.AttackEscape
        elseif getProgression(prog_attackValue) == 1000
        then
          param.curCmd = Cmd.LeaveForever
        end
      end

      -- Pattern Behavior
      if param.curCmd == Cmd.Init
      then 
        debugPalette("Debug_White")

        -- 
        if getProgression(prog_shouldDie) <= 0
        then
          param.curCmd = Cmd.Idle
        end
      elseif param.curCmd == Cmd.Idle
      then
        debugPalette("Debug_Green")
      elseif param.curCmd == Cmd.Attack1
      then
        debugPalette("Debug_Cyan")
        param.subCmd = 0 
        -- Core Loop -- 
        while patternOK() do
          drawDebug()
          
          if hasNewCommand()
          and newCommandOK()
          then
            break
          end

          -- Variables --
          local myVec = vec_pos(me)

          -- [0] - Go to the bottom center --
          if param.subCmd == 0 
          then
            lockAttackValue()
            --
            local target = param.points.BottomReset
            --trace("%i %i" % { target.x, target.y })
            local dist = vec_dist(myVec, target)
            local moveOK = ( dist > 160 * 5 )
            --
            if moveOK
            then
              trace("yeah 1")
              action_jumpTeleport(target.x, target.y, 0)
              --steerToPoint(target.x, target.y, "Dash")
            else
              param.subCmd = 1
            end
          -- [1] - PICK a side then GO TO it
          elseif param.subCmd == 1
          then
            local target = param.points.LeftAnchor
            if param.heading > 500
            then
              target = param.points.RightAnchor
            end
            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 5)
            --
            if moveOK 
            then
              local rot = (param.heading > 500) and -180 or 0
              action_jumpTeleport(target.x, target.y, rot)
              --steerToPoint(target.x, target.y, "Dash")
            else
              param.subCmd = 2
            end
          -- [2] - Attack player from ground --
          elseif param.subCmd == 2
          then
            local playerDist = vec_dist(myVec, vec_pos(player))
            local facingDir = param.heading > 500 and -1 or 1
            local heading = { x = me.XPos + (facingDir * 300), y = getFloorCurveY() + param.floorYOffset }
            local target = { x = player.XPos, y = param.points.FloorLine.y }
            -- Do NOT seek player on Y UNTIL we're close
            if playerDist <= 160 * 15
            then
              target.y = player.YPos
              heading.y = (player.YPos + player.WorldCollisionHeight/2)
            end
            local dist = vec_dist(myVec, target)
            --
            local moveOK = (dist > 160 * 4) or ( me.YPos <= param.points.FloorLine.y )
            --
            if moveOK
            then
              steerToPoint(heading.x, heading.y)
            else
              -- Forcefully wait for a few frames for that cool pop --
              param.subCmd = 3
              param.time = 0 
              while param.time <= 14 do
                param.time = param.time + 1
                debugPalette("Debug_Orange")
                addInput(dirs.U, btns.Dash)
                wait()
              end
              -- Store our heading and go on to the next subCmd
              if ( me.ArtRotation >= -90 and me.ArtRotation <= 90 )
              then 
                param.heading = 1
              else 
                param.heading = -1
              end
            end
          -- [3] - Seek into the floor for a while --
          elseif param.subCmd == 3
          then
            -- try to gently angle ourselves down --
            param.time = param.time + 1
            local target = { x = me.XPos + ((160*2) * param.heading), y = me.YPos - 400 }
            local moveOK = ( param.time <= 240 ) or me.YPos >= param.points.BottomReset.y
            --
            if param.time >= 60 * 3
            then
              param.subCmd = 4
              param.time = 0 
              allowAttackValue()
            elseif moveOK then
              steerToPoint(target.x, target.y, "Dash")
            else
              param.subCmd = 4
              param.time = 0 
              allowAttackValue()
            end
          -- [4] - Go back to the Origin point.
          elseif param.subCmd == 4
          then
            local target = { x = param.points.BottomReset2.x, y = param.points.BottomReset2.y }
            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 5.5) or (me.YPos >= param.points.BottomReset2.y)
            --
            if param.time <= 11
            then
              --trace( param.time )
              param.time = param.time + 1
              steerToPoint(target.x, target.y, "Dash")
            else
              --trace("GOING TO -1")
              action_jumpTeleport(target.x, target.y, 0)
              param.subCmd = -1 
            end
          elseif param.subCmd == -1
          then
            --trace("-1")
            param.curCmd = Cmd.Retreat
            param.subCmd = 0
            break
          else
            --trace("We did not set a subCmd!")
            param.subCmd = -1
          end
          wait()
        end
      elseif param.curCmd == Cmd.Attack2
      then
        debugPalette("Debug_Yellow")
        param.subCmd = 0 
        -- Core Loop --
        while patternOK() do
          drawDebug()
          
          if hasNewCommand()
          and newCommandOK()
          then
            break
          end

          -- Variables --
          local myVec = vec_pos(me)

          -- [0] - Go to the Center
          if param.subCmd == 0 
          then
            lockAttackValue()
            --
            local target = param.points.BottomCenter
            --trace("%i %i" % { target.x, target.y })
            local dist = vec_dist(myVec, target)
            local moveOK = ( dist > 160 * 5 )
            --
            if moveOK
            then
              action_jumpTeleport(target.x, target.y, 0)
              --steerToPoint(target.x, target.y, "Dash")
            else
              param.subCmd = 1
              param.heading = random(1000)
            end
          -- [1] - PICK a side then GO TO it
          elseif param.subCmd == 1
          then
            local target = param.points.LeftAnchor
            if param.heading > 500
            then
              target = param.points.RightAnchor
            end
            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 5)
            --
            if moveOK 
            then
              local _rot = (param.heading > 500) and -180 or 0
              action_jumpTeleport(target.x, target.y, _rot)
              --steerToPoint(target.x, target.y, "Dash")
            else
              param.subCmd = 2
            end
          -- [2] - Travel along the Floor plane, but try to follow the YPos of the player
          elseif param.subCmd == 2
          then
            -- We actually want to seek the OPPOSING point X, but player Y
            -- Unless we're close, then we try to seek their Y
            param.time = param.time + 1
            local angle = (param.time/360)
            local target = { x = player.XPos, y = getFloorCurveY() + param.floorYOffset }
            local playerDist = vec_dist(myVec, vec_pos(player))
            local playerClose = playerDist <= 160 * 8 or (
              (math.abs(select(2,dist_player())) <= 160 * 4)
              and (math.abs(dist_player()) <= 160 * 4)
            )
            local sineOffset = math.sin(angle * 30) * (160 * 7)
            if playerClose
            then
              target.y = player.YPos
            else
              --trace(sineOffset)
              target.y = (getFloorCurveY() + param.floorYOffset + sineOffset)
            end

            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 4)
            --
            if moveOK
            then
              steerToPoint(target.x, target.y, "Drill")
            else
              param.subCmd = 3
            end
          -- [3] - Seek the opposing anchor point
          elseif param.subCmd == 3
          then
            local yOffset = 160 * 6
            local target = param.points.RightAnchor
            if param.heading > 500
            then
              target = param.points.LeftAnchor
            end
            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 5)
            --
            if moveOK
            then
              if dist > 160 * 8
              then
                steerToPoint(target.x, getFloorCurveY() + param.floorYOffset, "Dash")
              else
                steerToPoint(target.x, target.y, "Dash")
              end
            else
              param.subCmd = 4
            end
            -- [4] - Travel along the Floor plane, but try to follow the YPos of the player
            -- REPEAT OF [2]
          elseif param.subCmd == 4
          then
            -- We actually want to seek the OPPOSING point X, but player Y
            -- Unless we're close, then we try to seek their Y
            param.time = param.time + 1
            local angle = (param.time/360)
            local target = { x = player.XPos, y = getFloorCurveY() + param.floorYOffset }
            local playerDist = vec_dist(myVec, vec_pos(player))
            local playerClose = (playerDist <= 160 * 8) or (
              (math.abs(select(2,dist_player())) <= 160 * 4)
              and (math.abs(dist_player()) <= 160 * 4)
            )
            local sineOffset = math.sin(angle * 30) * (160 * 7)
            if playerClose
            then
              target.y = player.YPos
            else
              --trace(sineOffset)
              target.y = (getFloorCurveY() + param.floorYOffset + 160 + sineOffset)
            end

            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 3)
            --
            if moveOK
            then
              --action_jumpTeleport(target.x, target.y, 0)
              steerToPoint(target.x, target.y, "Drill")
            else
              param.subCmd = 5
            end
          -- [5] - Seek the opposing anchor point
          -- REPEAT OF [3], but use the OTHER heading instead
          elseif param.subCmd == 5
          then
            local yOffset = 160 * 6
            local target = param.points.LeftAnchor
            if param.heading > 500
            then
              target = param.points.RightAnchor
            end
            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 6)
            --
            if moveOK
            then
              if dist > 160 * 8
              then
                steerToPoint(target.x, getFloorCurveY() + param.floorYOffset, "Dash")
              else
                steerToPoint(target.x, target.y, "Dash")
              end
            else
              param.subCmd = -1
              allowAttackValue()
            end
          -- [-1] - FINISH
          elseif param.subCmd == -1
          then
            --trace("-1")
            wait(30)
            param.curCmd = Cmd.Retreat
            param.subCmd = 0
            break
          else
            --trace("We did not set a subCmd!")
            param.subCmd = -1
          end
          wait()
        end
      elseif param.curCmd == Cmd.Attack3
      then
        debugPalette("Debug_Magenta")
        param.subCmd = 0 
        -- Core Loop --
        while patternOK() do
          drawDebug()
          
          if hasNewCommand()
          and newCommandOK()
          then
            break
          end

          -- Variables --
          local myVec = vec_pos(me)

          -- [0] - Go to the Center
          if param.subCmd == 0 
          then
            allowAttackValue()
            --
            local target = param.points.BottomCenter
            --trace("%i %i" % { target.x, target.y })
            local dist = vec_dist(myVec, target)
            local moveOK = ( dist > 160 * 4 )
            --
            if moveOK
            then
              action_jumpTeleport(target.x, target.y, 0)
              --steerToPoint(target.x, target.y, "Dash")
            else
              param.subCmd = 1
              param.heading = random(1000)
            end
          -- [1] - PICK a side then GO TO it
          elseif param.subCmd == 1
          then
            local target = param.points.LeftAnchor
            if param.heading > 500
            then
              target = param.points.RightAnchor
            end
            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 4)
            --
            if moveOK 
            then
              local _rot = (param.heading > 500) and -180 or 0 
              action_jumpTeleport(target.x, target.y, _rot)
              --steerToPoint(target.x, target.y, "Dash")
            else
              param.time = -80
              param.subCmd = 2
              lockAttackValue()
            end
          -- [2] - Travel along the Floor plane
          elseif param.subCmd == 2
          then
            -- We actually want to seek the OPPOSING point X, but player Y
            -- Unless we're close, then we try to seek their Y
            param.time = param.time + 1
            local angle = (param.time/360)
            local yBase =  getFloorCurveY() + param.floorYOffset --tonumber(param.points.FloorLine.y) - (160 * 2) -- getFloorCurveY() -- 
            --
            local targetHeading = param.points.RightAnchor
            local facingDir = 1
            if param.heading > 500
            then
              facingDir = -1
              targetHeading = param.points.LeftAnchor
            end
            --
            local sineOffset = math.sin(angle * 24)
            sineOffset = sineOffset * (160 * 0.1)
            local targetGoto = { x = me.XPos + (facingDir * 300), y = yBase + sineOffset }
            local dist = vec_dist(myVec, targetHeading)
            local distX = math.abs( dist_point(targetHeading.x, targetHeading.y))
            local moveOK = (dist > 160 * 4.5) and (distX > ( 160 * 3 ))
            --
            if moveOK
            then
              if math.abs(dist_player()) >= 160 * 6
              then
                steerToPoint(targetGoto.x, targetGoto.y, "Dash")
              else
                steerToPoint(targetGoto.x, targetGoto.y, "Drill")
              end
            else
              param.subCmd = 3
              allowAttackValue()
            end
          -- [3] - Seek the opposing anchor point
          elseif param.subCmd == 3
          then
            local target = param.points.RightAnchor
            if param.heading > 500
            then
              target = param.points.LeftAnchor
            end
            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 4)
            --
            if moveOK
            then
              steerToPoint(target.x, target.y, "Dash+Drill")
            else
              param.time = -80
              param.subCmd = 4
              lockAttackValue()
            end
            -- [4] - Travel along the Floor plane, but try to follow the YPos of the player
            -- REPEAT OF [2]
          elseif param.subCmd == 4
          then            -- We actually want to seek the OPPOSING point X, but player Y
            -- Unless we're close, then we try to seek their Y
            param.time = param.time + 1
            local angle = (param.time/360)
            local yBase =  getFloorCurveY() + param.floorYOffset --tonumber(param.points.FloorLine.y) - (160 * 2) -- getFloorCurveY() -- 
            --
            local targetHeading = param.points.LeftAnchor
            local facingDir = 1
            if param.heading > 500
            then
              facingDir = -1
              targetHeading = param.points.RightAnchor
            end
            --
            local sineOffset = math.sin(angle * 24)
            sineOffset = sineOffset * (160 * 0.1)
            local targetGoto = { x = me.XPos - (facingDir * 300), y = yBase + sineOffset }
            local dist = vec_dist(myVec, targetHeading)
            local distX = math.abs( dist_point(targetHeading.x, targetHeading.y))
            local moveOK = (dist > 160 * 4.5) and (distX > ( 160 * 3 ))
            --
            if moveOK
            then
              if math.abs(dist_player()) >= 160 * 6
              then
                steerToPoint(targetGoto.x, targetGoto.y, "Dash")
              else
                steerToPoint(targetGoto.x, targetGoto.y, "Drill")
              end
            else
              param.subCmd = 5
              allowAttackValue()
            end
          -- [5] - Seek the opposing anchor point
          -- REPEAT OF [3], but use the OTHER heading instead
          elseif param.subCmd == 5
          then
            local yOffset = 160 * 6
            local target = param.points.LeftAnchor
            if param.heading > 500
            then
              target = param.points.RightAnchor
            end
            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 4.5)
            --
            if moveOK
            then
              steerToPoint(target.x, target.y, "Dash")
            else
              param.subCmd = -1
              allowAttackValue()
            end
          -- [-1] - FINISH
          elseif param.subCmd == -1
          then
            --trace("-1")
            param.curCmd = Cmd.Retreat
            param.subCmd = 0
            break
          else
            --trace("We did not set a subCmd!")
            param.subCmd = -1
          end
          wait()
        end
      elseif param.curCmd == Cmd.Attack4
      then
        debugPalette("Debug_Orange")
        param.subCmd = 0 
        -- Core Loop --
        while patternOK() do
          drawDebug()
          
          if hasNewCommand()
          and newCommandOK()
          then
            break
          end

          -- Variables --
          local myVec = vec_pos(me)

          -- [0] - Go to the Center
          if param.subCmd == 0 
          then
            --trace("0")
            allowAttackValue()
            --
            local target = param.points.BottomCenter
            --
            action_jumpTeleport(target.x, target.y, 0)
            wait(60)

            param.subCmd = 1
            param.heading = random(1000)
          -- [1] - PICK a side then GO TO it
          elseif param.subCmd == 1
          then
            --trace("1")
            local target = vec_pos(player)
            target.y = getFloorCurveY()

            local xOffset = (param.heading > 500) and -1 or 1
            xOffset = xOffset * (160 * -13)
            local yOffset = 160 * -1
            target.x = target.x + xOffset
            target.y = target.y + yOffset

            local dist = vec_dist(myVec, target)
            local _rot = (param.heading > 500) and 112.5 or 67.5

            action_jumpTeleport(target.x, target.y, _rot)
            param.time = -80
            param.subCmd = 2
            lockAttackValue()
          -- [2] - Commit to an arc
          elseif param.subCmd == 2
          then
            --trace("2")
            -- Attempt to curve DOWNWARDS
            param.time = param.time + 1
            local facingDir = 1
            if param.heading > 500
            then
              facingDir = -1
            end
            --
            local heading = { x = 160 * facingDir, 
            y = scale(param.time, -80, 60 * 2, 2, -3)
            }
            --trace(heading.y)
            local target = { x = me.XPos + heading.x, y = me.YPos + (heading.y * 160) }
            local moveOK = param.time <= 60 * 3
            --
            if moveOK
            then
              me.NormalJump = param.drillSpeedFast
              steerToPoint(target.x, target.y, "Scales")
            else
              me.NormalJump = 0
              param.subCmd = 3
              allowAttackValue()
            end
          elseif param.subCmd == 3
          then
            --trace("3")
            local target = param.points.BottomCenter
            --
            action_jumpTeleport(target.x, target.y, 0)
            wait(30)

            param.subCmd = 4
          -- [4] - Pick another position!
          elseif param.subCmd == 4
          then
            --trace("4")
            local target = vec_pos(player)

            local xOffset = (param.heading > 500) and -1 or 1
            xOffset = xOffset * (160 * 13)
            local yOffset = 160 * -1
            target.x = target.x + xOffset
            target.y = target.y + yOffset

            local dist = vec_dist(myVec, target)
            local _rot = (param.heading > 500) and 67.5 or 112.5

            wait(10)
            action_jumpTeleport(target.x, target.y, _rot)
            param.time = -80
            param.subCmd = 5
            lockAttackValue()
          -- [4] - Commit to an arc
          elseif param.subCmd == 5
          then
            --trace("5")
            -- Attempt to curve DOWNWARDS
            param.time = param.time + 1
            local facingDir = -1
            if param.heading > 500
            then
              facingDir = 1
            end
            --
            local heading = { x = 160 * facingDir, 
            y = scale(param.time, -80, 60 * 2, 2, -3)
            }
            --trace(heading.y)
            local target = { x = me.XPos + heading.x, y = me.YPos + (heading.y * 160) }
            local moveOK = param.time <= 60 * 3
            --
            if moveOK
            then
              me.NormalJump = param.drillSpeedFast
              steerToPoint(target.x, target.y, "Scales")
            else
              me.NormalJump = 0
              param.subCmd = -1
              allowAttackValue()
            end
          -- [5] - Seek the opposing anchor point
          -- REPEAT OF [3], but use the OTHER heading instead
          elseif param.subCmd == 5
          then
            local yOffset = 160 * 6
            local target = param.points.LeftAnchor
            if param.heading > 500
            then
              target = param.points.RightAnchor
            end
            local dist = vec_dist(myVec, target)
            local moveOK = (dist > 160 * 4.5)
            --
            if moveOK
            then
              steerToPoint(target.x, target.y, "Dash")
            else
              param.subCmd = -1
              allowAttackValue()
            end
          -- [-1] - FINISH
          elseif param.subCmd == -1
          then
            --trace("-1")
            param.curCmd = Cmd.Retreat
            param.subCmd = 0
            break
          else
            --trace("We did not set a subCmd!")
            param.subCmd = -1
          end
          wait()
        end
      elseif param.curCmd == Cmd.AttackEscape
      then
        debugPalette("Debug_Blue") 
        param.subCmd = 0 
        -- OK GO BACK TO NORMAL --
        while patternOK() do
          drawDebug()
          
          -- Variables --
          local myVec = vec_pos(me)
          
          -- [0] - 
          if param.subCmd == 0
          then
            allowAttackValue()
            debugPalette("Debug_BlinkBlue")
            --
            local target = param.points.BottomReset2
            local dist = vec_dist(myVec, target)
            local moveOK = ( dist > 160 * 4 )
            --
            if moveOK
            then
              action_jumpTeleport(target.x, target.y, 90)
              --steerToPoint(target.x, target.y)
            else
              param.subCmd = 1
            end
            -- [1] - Seek Player XPos + YUp, until we hit the Floor Lines
          elseif param.subCmd == 1
          then
            debugPalette("Debug_BlinkBlue")
            --
            local target = { x = player.XPos, y = me.YPos + 400 }
            local destY = param.points.FloorLine.y - 800
            local dist = vec_dist(myVec, target)
            local moveOK = (me.YPos <= destY)
            if moveOK
            then
              --[[
              if dist > 600 
              then
                steerToPoint(target.x, target.y)
              else
                steerToPoint(target.x, target.y, "Dash")
              end
              ]]
              action_jumpTeleport(target.x, target.y, 90)
            else
              -- Input Lock --
              param.time = 0
              while param.time <= 40
              do
                param.time = param.time + 1
                addInput(dirs.U, btns.Attack)
                wait()
              end
              param.time = 0 
              param.subCmd = 2
              param.heading = (dist_player() > 0) and 1 or -1
              lockAttackValue()
            end
          -- [2] - Emit scales while going up
          elseif param.subCmd == 2
          then
            debugPalette("Debug_BlinkMagenta")
            lockAttackValue()
            local target = { x = me.XPos + (param.heading * 80),
              y = me.YPos - 120 }
            local destY =  (param.points.FloorLine.y)
            local dist = vec_dist(myVec, target)
            local moveOK = (me.YPos >= destY) and ( param.time <= 80)
            if moveOK
            then
              me.NormalJump = param.drillSpeed
              steerToPoint(target.x, target.y, "Scales")
            else
              param.subCmd = 3
            end
          -- [3] - Head into the ground for a time
          elseif param.subCmd == 3
          then
            debugPalette("Debug_DarkMagenta")
            local target = { x = me.XPos + (param.heading * 40), y = me.YPos - 110}
            local destY = param.points.BottomCenter.y
            local dist = vec_dist(myVec, { x = me.XPos, y = destY })
            local moveOK = (dist >= 160 * 3) or (me.YPos <= destY)
            if moveOK then
              me.NormalJump = 0
              steerToPoint(target.x, target.y, "Dash")
            else
              param.subCmd = 4
              param.time = 0 
            end
          -- [4] - Let's just go down for a few frames
          elseif param.subCmd == 4
          then
            debugPalette("Debug_DarkMagenta")
            param.time = param.time + 1
            local target = { x = me.XPos, y = me.YPos - 110}
            local moveOK = param.time <= 30
            if moveOK then
              me.NormalJump = 0
              steerToPoint(target.x, target.y, "Dash")
            else
              param.subCmd = -1
              allowAttackValue()
            end
          -- [-1] - FINISH
          elseif param.subCmd == -1
          then
            debugPalette("Debug_BlinkRed")
            param.curCmd = Cmd.Retreat
            param.subCmd = 0
            break
          else
            param.subCmd = -1
          end
          wait()
        end
      elseif param.curCmd == Cmd.Retreat
      then
        debugPalette("Debug_BlinkRed")
        allowAttackValue()
        resetAttackValue()
        --
        local myVec = vec_pos(me)
        local target = param.points.BottomReset2
        local dist = vec_dist(myVec, target)
        local moveOK = (dist > 160 * 3)

        --trace("[Retreat] Teleporting to Target (%i, %i)" % {target.x, target.y})
        action_jumpTeleport(target.y, target.y, 0)
        param.curCmd = Cmd.Idle
        param.subCmd = 0
        wait()
      elseif param.curCmd == Cmd.LeaveForever
      then
        --trace("Leave Forever")
        debugPalette("Debug_BlinkBlue")
        steerToPoint( me.XPos, me.YPos - 300, "Dash")
      elseif param.curCmd == Cmd.CutsceneChomp
      then
        --trace("Cutscene Chomp")
        debugPalette("Debug_BlinkOrange")
        while true do
          local target = param.points.CutsceneChomp
          local dist = vec_dist( vec_pos(me), target )
          if dist >= 160 * 2
          then
            --trace(dist)
            steerToPoint( target.x, target.y, "Dash+Grab")
          else
            break
          end
          wait()
        end
        -- Loop 2 --
        param.time = 0
        while true do
          param.time = param.time + 1
          if param.time <= 600
          then
            addInput(dirs.U, btns.Dash)
          else
            break
          end
          wait()
        end
      else
        debugPalette()
      end
    end

    --steerToPoint(player.XPos, player.YPos + player.WorldCollisionHeight)
    --steerToPoint_manual(350)

    wait()
  end -- the while loop
end -- the function

function OnBattleAttack(battleData)
  local toReturn = DefaultOnBattleAttack(battleData)
  local me = getMe()
  return toReturn
end --OnBattleAttack()