require("LuaLib")
require("MonsterLib")

require("debug_ext")
require("math_ext")
require("util")
require("vec2")
require("movement")

local function block()
  doMotion("X+WorldGuard")
end

local function attack_tailSwipe(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = true
  end
  doMotion("X+Heruka")
  param.attackCooldown = 27
end

local function attack_scratch(data)
  if data ~= nil then
    data.dir = dirs.N
    data.pressAction = false
  end
  doMotion("X+Attack")
  param.attackCooldown = 42
end

local function attack_kiss(data)
  if data ~= nil then
    data.dir = dirs.U
    data.pressAction = false
  end
  doMotion("U+Attack")
  param.attackCooldown = 20
end

local function dist_origin()
  local me = getMe()
  return param.origin.x - me.XPos, param.origin.y - me.Ypos 
end

local function dist_goal()
  local me = getMe()
  return param.goalX - me.XPos, param.origin.y - me.Ypos 
end

local function isInSightRange(target)
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(player)
  
  local w = param.sightRangeX
  local h = param.sightRangeY
  if param.bossBurstedMode then
    w = param.sightRangeRageX
    h = param.sightRangeRageY
  end

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

local function isInAttackRange()
  local me, player = getMe(), (target or getPlayer())
  local dx, dy = dist_sprite(player)
  
  local w = param.attackRangeX
  local h = param.attackRangeY
  if param.bossBurstedMode then
    w = param.attackRangeRageX
    h = param.attackRangeRageY
  end

  if dx >= -w/2
  and dx <= w/2
  and dy <= h.top
  and dy >= h.bottom
  then
    return true
  end
  return false
end

-- the initialization function for this AI
function OnInit()
  local me, player = getMe(), getPlayer()

  -- Set origin
  param.origin = { x = me.XPos, y = me.YPos }
  if me:getSpawnpointPos() ~= nil then
    param.origin.x, param.origin.y, param.origin.z = me:getSpawnpointPos()
  end

  param.time = 0
  param.turnWaitTime = 0
  param.blockTime = 0
  param.attackCooldown = 0
  param.attackStartup = 0

  param.roamRangeX = 160 * 8
  param.roamRangeFudgeDist = 75
  param.sightRangeX = 160 * 9
  param.sightRangeY = { top = 160 * 1.5, bottom = 160 * -1 }
  param.attackRangeX = 160 * 7
  param.attackRangeY = { top = 160 * 2.5, bottom = 160 * -1.5 }
  param.moveDeadzone = 160 * 2 
  param.goalX = param.origin.x
  param.goalX = param.goalX + ((param.roamRangeX/2) * (me.FacingDir * -1))
  --
  -- co-op
  param.sightRangeRageX = 160 * 17
  param.sightRangeRageY = { top = 160 * 6, bottom = 160 * -7 }
  param.attackRangeRageX = 160 * 9
  param.attackRangeRageY = { top = 160 * 6, bottom = 160 * -6 }
  --
  param.bossBurstedMode = false
end

-- The world update function for this AI
function OnUpdate()
  local me, player = getMe(), getPlayer()
  local target = getClosestPartyMember()

  -- SCARY HACK WE SHOULD PROBABLY NOT DO --
  while shouldNeverUpdate() do

  end

  while true do
    me, player = getMe(), getPlayer()
    target = getClosestPartyMember()
    --
    if aiDebug() then
      draw_point(param.origin.x, param.origin.y)
      draw_point(param.goalX, param.origin.y)
      draw_bar(param.attackStartup, 0, 15, 3, {255, 255, 0})
      draw_bar(param.attackCooldown, 0, 60, 2, {255, 0, 0})
      draw_bar(param.turnWaitTime, 0, 180, 1, {255, 0, 255})
      draw_bar(param.blockTime, 0, 60, 0, {0, 255, 255})
    end

    if tick_ok() then
      if not isInSightRange(target) then
        param.turnWaitTime = math.max(0, param.turnWaitTime - 1)
      end
      param.attackCooldown = math.max(0, param.attackCooldown - 1)
      param.blockTime = math.max(0, param.blockTime - 1)
    end

    local dist_goalPoint = dist_point(param.goalX, me.YPos)
    if math.abs( dist_goalPoint ) <= me.WorldCollisionWidth/2
    then
      local side = param.origin.x - me.XPos
      if side >= 0 then side = 1 else side = -1 end 
      param.goalX = param.origin.x + (param.roamRangeX/2 * (side))
      param.turnWaitTime = 45 + random(67)
    end
    
    if bossBursted() 
    then
      param.turnWaitTime = 0
      param.blockTime = 0 
      param.attackCooldown = 22
      param.attackStartup = 22
      --
      param.bossBurstedMode = true
      me.InvinFlashTime = 60 * 5 
    end
    
    if param.bossBurstedMode 
    and me.InvinFlashTime <= 0
    then
      param.bossBurstedMode = false
    end
    
    if isInSightRange(target)
    then
      if param.blockTime > 0 then
        block()
        param.attackStartup = 0 
      elseif blocked_hit() then 
        param.blockTime = math.max(15, param.blockTime) 
        param.attackStartup = 0 
      elseif target:hasTag("Attack")
      and not me:hasTag("Attack")
      and dist_sprite(target) > 0 then
        param.blockTime = 25
        param.attackStartup = 0 
      elseif ( param.bossBurstedMode or sprite_vulnerable(target) ) then
        --
        trace("Test 1")
        --
        if isInAttackRange(target)
        and dist_sprite(target) >= 0
        and param.attackCooldown <= 0
        --and param.attackStartup >= 15 
        then
          trace("Test 2")
          local attackChance = random(1000)
          if attackChance > 500
          and dist_sprite(target) < 300
          then
            attack_scratch()
          else
            attack_tailSwipe()
          end
          param.attackStartup = 0 
        elseif isInAttackRange(target)
        and dist_sprite(target) >= 0
        and param.attackCooldown <= 0
        then
          param.attackStartup = param.attackStartup + 1
        else -- if math.abs(dist_player()) > 160 * 3 then
          param.attackStartup = 0 
          if math.abs(dist_sprite(target)) > param.moveDeadzone
          then
            approach_withTurn(target.XPos, 6)
          end
        end
      end
    elseif param.turnWaitTime > 0 then
      -- we're just waitin'
    else
      approach_withTurn(param.goalX, 8)
    end
    wait()
  end -- the while loop
end -- the function

-- The battle action function for this AI 
function OnBattleAttack(battleData)
  local me, player = getMe(), getPlayer()
  local toReturn = DefaultOnBattleAttack(battleData)

  local dirChance = random(1000)
  local actionChance = random(1000)

  if battleData.turnsSinceFreeze == nil
  then
    battleData.turnsSinceFreeze = 0
  end
  battleData.turnsSinceFreeze = battleData.turnsSinceFreeze + 1 
  
  if false -- battleData.turnsSinceFreeze >= 4
  then
    battleData.turnsSinceFreeze = 0
    attack_kiss(toReturn)
  elseif dirChance > 550
  then
    attack_scratch(toReturn)
  else
    attack_tailSwipe(toReturn)
  end

  if player.debug1 > 0 -- aiDebug()
  then
    trace("Debug kiss")
    attack_kiss(toReturn)
  end

  return toReturn
end --OnBattleAttack()