// GLSL Fragment Shader "main"
// Generated by XShaderCompiler

#version 330 core

in vec2 xsv_TEXCOORD0;
out vec4 SV_Target0;
//  This header includes functions for working with the G-buffer
struct Gbuffer
{
    vec3  Albedo;
    vec3  Normal;
    float Emissivity;
    float EmissiveIntensity;
};
vec3 UnpackNormal(vec3 sampled_normal)
{
    return sampled_normal * 2.0f - vec3(1.0f);
}
float EmissiveMax = 4.0f;
float EmissiveMaxInverse = 0.25f;
float EncodeEmissive(float e)
{
    return e * EmissiveMaxInverse;
}
float DecodeEmissive(float e)
{
    return e * EmissiveMax;
}
Gbuffer SampleGbuffer(sampler2D gbuffer0, sampler2D gbuffer1, vec2 texcoord)
{
    Gbuffer g;
    vec4 t = vec4(texcoord, 0.0f, 0.0f);
    vec4 xst_temp0 = (t);
    vec4 sample0 = textureLod(gbuffer0, xst_temp0.xy, xst_temp0.w);
    vec4 xst_temp1 = (t);
    vec4 sample1 = textureLod(gbuffer1, xst_temp1.xy, xst_temp1.w);
    g.Albedo = sample0.rgb;
    g.Normal = UnpackNormal(sample1.rgb);
    g.Emissivity = sample0.a;
    g.EmissiveIntensity = DecodeEmissive(sample1.a);
    return g;
}
float ConvertDepthSample(float depth)
{
    return depth;
}
vec4 GetPositionFromExponentialDepth(float depth, float near, float far, vec2 screen, mat4 ivp)
{
    vec4 p = vec4(screen.x, screen.y, depth, 1.0f);
    far;
    // avoid warnings
    float z = near / min(-0.001f, depth - 1.0f);
    p *= vec4(-z);
    vec4 pw = ((p) * (ivp));
    return pw;
}
vec4 GetPositionFromDepth(float depth, float near, float far, vec2 screen, mat4 ivp)
{
    return GetPositionFromExponentialDepth(depth, near, far, screen, ivp);
}
// *
//  *  \param position  World position of the point being shaded.
//  *  \param normal    Surface normal at the point being shaded.
//  *  \param params1   Vector containing: {z scale, z offset, overdarken, saturation}
//  *  \param params2   Vector containing: {angle fade, unused, unused, unused}
//  *  \return          Vector containing: {shadow factor, darkening factor, saturation}
//
vec3 ForegroundShading(vec3 position, vec3 normal, vec4 params1, vec4 params2)
{
    float normal_scale = pow(max(1.0f - clamp(dot(vec3(0, 1, 0), normal), float(0), float(1)), 0.0001f), params2.x);
    float t = normal_scale * clamp(position.z * params1.x + params1.y, float(0), float(1));
    float shadow = (1.0f - t);
    float darken = mix(1.0f, params1.z, t);
    float saturation = mix(1.0f, params1.w, t);
    return vec3(shadow, darken, saturation);
}
vec3 GetPlayerPathAmbient(float zpos, vec3 base_ambient, vec3 path_ambient, vec4 path_params)
{
    float tfront = clamp(zpos * path_params.x + path_params.y, float(0), float(1));
    float tback = clamp(zpos * path_params.z + path_params.w, float(0), float(1));
    float t = max(tfront, tback);
    return mix(path_ambient, base_ambient, vec3(t));
}
vec3 DecodeLightMap(vec4 encoded)
{
    float k = 3.0f;
    float scale = k * encoded.a;
    vec3 light = scale * encoded.rgb;
    return light;
}
//  converts a texture coordinate to a screen coordinate
vec2 TexToScreen(vec2 tex)
{
    return tex.xy * vec2(2.0f, -2.0f) + vec2(-1.0f, 1.0f);
}
//  computes the luminance value for an RGB color
float RgbToLuminance(vec3 color)
{
    return dot(color, vec3(0.212671, 0.715160, 0.072169));
}
//  takes screen pixel coords and return a 0 or 1 checkerboard pattern
vec2 Checkerboard(vec2 screen)
{
    vec2 checkerboard = trunc(screen);
    checkerboard = 0.5f * checkerboard;
    checkerboard = fract(checkerboard);
    checkerboard = 2.0f * checkerboard;
    return checkerboard;
}
//  Shadow map functions
//  sampling kernels
vec2 Poisson9[9] = vec2[9](vec2(0.1567495f, 0.2453307f), vec2(-0.4647257f, -0.1085932f), vec2(-0.2514141f, 0.6994573f), vec2(0.5295524f, -0.3833097f), vec2(-0.4383077f, -0.7664956f), vec2(0.957235f, 0.2121297f), vec2(0.3471507f, 0.9330071f), vec2(0.1850856f, -0.8876677f), vec2(-0.8153945f, 0.4067398f));
float GetShadowFactorAndCenterDepth(sampler2D map, vec2 uv, float depth, float bias, vec2 dims, vec2 screen, out float out_depth)
{
    out_depth = 1.0f;
    // xbox one doesn't seem to support border color correctly, so have to do this manually
    // RT_TODO: benchmark it, maybe it's faster?
    vec2 tex_scale = vec2(1.0f / dims.x, 1.0f / dims.y);
    float check_depth = depth + bias;
    depth -= bias;
    float shadow = 0.0f;
    // poisson samples
    vec2 c = Checkerboard(screen);
    float cost = 1.0f - (c.x + c.y);
    float sint = c.x - c.y;
    for (int i = 0; i < 9; ++i)
    {
        vec2 p = 1.4f * Poisson9[i];
        vec2 rp = vec2(cost * p.x - sint * p.y, sint * p.x + cost * p.y);
        vec4 s = vec4(uv + rp * tex_scale, 0, 0);
        s.y = 1.0f - s.y;
        vec4 xst_temp5 = (s);
        float map_depth = textureLod(map, xst_temp5.xy, xst_temp5.w).r;
        shadow += (depth <= map_depth ? 1.0f : 0.0f);
        if (i == 0)
        {
            out_depth = map_depth;
        }
    }
    shadow /= 9.0f;
    return mix(shadow, 1.0f, max(0.0f, sign(check_depth - 1.0f)));
}
float GetShadowFactorAndMinDepth(sampler2D map, vec2 uv, float depth, float bias, vec2 dims, vec2 screen, out float out_depth)
{
    out_depth = 1.0f;
    // xbox one doesn't seem to support border color correctly, so have to do this manually
    // RT_TODO: benchmark it, maybe it's faster?
    vec2 tex_scale = vec2(1.0f / dims.x, 1.0f / dims.y);
    out_depth = 1.0f;
    float check_depth = depth + bias;
    depth -= bias;
    float shadow = 0.0f;
    // poisson samples
    vec2 c = Checkerboard(screen);
    float cost = 1.0f - (c.x + c.y);
    float sint = c.x - c.y;
    for (int i = 0; i < 9; ++i)
    {
        vec2 p = 1.4f * Poisson9[i];
        vec2 rp = vec2(cost * p.x - sint * p.y, sint * p.x + cost * p.y);
        vec4 s = vec4(uv + rp * tex_scale, 0, 0);
        s.y = 1.0f - s.y;
        vec4 xst_temp6 = (s);
        float map_depth = textureLod(map, xst_temp6.xy, xst_temp6.w).r;
        shadow += (depth <= map_depth ? 1.0f : 0.0f);
        out_depth = min(out_depth, map_depth);
    }
    shadow /= 9.0f;
    return mix(shadow, 1.0f, max(0.0f, sign(check_depth - 1.0f)));
}
//  NOTE: debug view constants; must be kept in sync with the RenderJobDirectionalLight::DebugView enum
float DebugView_None = 0.0f;
float DebugView_Albedo = 1.0f;
float DebugView_Normals = 2.0f;
float DebugView_Emissive = 3.0f;
float DebugView_Ndotl = 4.0f;
float DebugView_Directional = 5.0f;
float DebugView_Ambient = 6.0f;
float DebugView_Ssao = 7.0f;
float DebugView_Depth = 8.0f;
float DebugView_Shadow = 9.0f;
float DebugView_AmbientMap = 10.0f;
float DebugView_LightColorMap = 11.0f;
float DebugView_ShadowColorMap = 12.0f;
uniform vec4 preg_0;
// (r, g, b, use ndotl remap)
uniform vec4 preg_1;
// (x, y, z, wrap strength)
uniform vec4 preg_2;
uniform vec4 preg_3;
uniform vec4 preg_4;
uniform vec4 preg_5;
// (debug view, ssao enabled, has shadows/max shadow z-plane, foreground darken enabled)
// CONSTANT_DECL(float4, NearFar_GoboMode, 6);  // (near, far, gobo axis, gobo enabled)
uniform vec4 preg_6;
// (near, far, gobo axis, gobo enabled)
uniform vec4 preg_7;
// (world bias, depth bias, size, unused)
uniform mat4 preg_8;
uniform mat4 preg_12;
uniform vec4 preg_17;
// (r, g, b, unused)
uniform vec4 preg_18;
// (scale, offset, overdarken, saturation)
uniform vec4 preg_19;
uniform mat4 preg_20;
uniform vec4 preg_24;
// (world, sprites, both, both/world)
uniform vec4 preg_25;
// (size, world bias, fade scale, fade offset)
uniform mat4 preg_26;
uniform vec4 preg_30;
//  This needs to preceed all texture defines for post-processing to work properly
// TextureNames;Gbuffer0:0;Gbuffer1:1;Ssao:2;DepthTexture:3;ShadowMap:4;ExtraShadowMap:5;NdotlRemap:6;AmbientMap:7;LightColorMap:8;ShadowColorMap:9;
//  global textures
//  light-specific textures
// TEX_DEFINE(Gobo, 6);
uniform sampler2D treg_0;
uniform sampler2D treg_1;
uniform sampler2D treg_2;
uniform sampler2D treg_3;
uniform sampler2D treg_4;
uniform sampler2D treg_5;
uniform sampler2D treg_6;
uniform sampler2D treg_7;
uniform sampler2D treg_8;
uniform sampler2D treg_9;


// Uniform name defines ///////////////////////////////////////
#define Color_NdotlRemap preg_0
#define Direction_Wrap preg_1
#define AmbientColor_DepthScale preg_2
#define PlayerPathAmbient preg_3
#define PlayerPathParams preg_4
#define Settings preg_5
#define NearFar preg_6
#define ShadowParams preg_7
#define ShadowViewProjection preg_8
#define CameraInvProjection preg_12
#define ShadowColor preg_17
#define Foreground preg_18
#define ForegroundAngle preg_19
#define CameraInvView preg_20
#define ShadowIntensities preg_24
#define ExtraShadowParams preg_25
#define ExtraShadowViewProjection preg_26
#define AmbientMapTransform preg_30
#define Gbuffer0 treg_0
#define Gbuffer1 treg_1
#define Ssao treg_2
#define DepthTexture treg_3
#define ShadowMap treg_4
#define ExtraShadowMap treg_5
#define NdotlRemap treg_6
#define AmbientMap treg_7
#define LightColorMap treg_8
#define ShadowColorMap treg_9
///////////////////////////////////////////////////////////////

void main()
{
    bool has_shadows = Settings.z != 0.0f;
    float shadow_z_cutoff = Settings.z;
    // const bool has_gobo = NearFar_GoboMode.w != 0.0f;
    bool has_foreground_shade = Settings.w != 0.0f;
    vec4 texcoord = vec4(xsv_TEXCOORD0, 0.0f, 0.0f);
    Gbuffer gbuffer = SampleGbuffer(Gbuffer0, Gbuffer1, texcoord.xy);
    vec2 screen_coord = TexToScreen(vec2(texcoord.x, 1.0f - texcoord.y));
    vec4 xst_temp7 = (texcoord);
    float depth = ConvertDepthSample(textureLod(DepthTexture, xst_temp7.xy, xst_temp7.w).r);
    // the base n dot L is used to prevent sprite shadows from showing on faces pointing away from the light
    float base_ndotl = -dot(Direction_Wrap.xyz, gbuffer.Normal);
    float near = NearFar.x;
    float far = NearFar.y;
    vec4 world = GetPositionFromDepth(depth, near, far, screen_coord, CameraInvProjection);
    world = ((world) * (CameraInvView));
    float ao = 1.0f;
    if (Settings.y != 0.0f)
    {
        vec4 xst_temp8 = (texcoord);
        ao = textureLod(Ssao, xst_temp8.xy, xst_temp8.w).a;
    }
    vec3 ambient = AmbientColor_DepthScale.rgb;
    vec3 light_color = Color_NdotlRemap.rgb;
    vec3 shadow_color = ShadowColor.rgb;
    if (AmbientMapTransform.x != -1.0f)
    {
        vec2 uv = world.xy * AmbientMapTransform.xy + AmbientMapTransform.zw;
        vec4 t = vec4(uv, 0, 0);
        vec4 xst_temp9 = (t);
        // ambient = TEX2DLOD(AmbientMap, t).rgb;
        // light_color = TEX2DLOD(LightColorMap, t).rgb;
        // shadow_color = TEX2DLOD(ShadowColorMap, t).rgb;
        ambient = DecodeLightMap(textureLod(AmbientMap, xst_temp9.xy, xst_temp9.w));
        vec4 xst_temp10 = (t);
        light_color = DecodeLightMap(textureLod(LightColorMap, xst_temp10.xy, xst_temp10.w));
        vec4 xst_temp11 = (t);
        shadow_color = DecodeLightMap(textureLod(ShadowColorMap, xst_temp11.xy, xst_temp11.w));
    }
    ambient = GetPlayerPathAmbient(world.z, ambient, PlayerPathAmbient.rgb, PlayerPathParams);
    ambient *= vec3(ao);
    float shadow = 1.0f;
    float foreground_darkening = 1.0f;
    float albedo_saturation = 1.0f;
    vec4 color = vec4(0, 0, 0, 1);
    float debug_view = Settings.x;
    if (has_shadows)
    {
        vec3 world_for_shadows = world.xyz + ShadowParams.x * gbuffer.Normal;
        float shadow_bias = ShadowParams.y;
        vec4 lpos = ((vec4(world_for_shadows, 1)) * (ShadowViewProjection));
        vec2 luv = lpos.xy * vec2(0.5f, -0.5f) + vec2(0.5f);
        float base_map_depth;
        float extra_map_depth;
        vec2 screen = gl_FragCoord.xy;
        float base_shadow = GetShadowFactorAndCenterDepth(ShadowMap, luv, lpos.z, shadow_bias, ShadowParams.zz, screen, base_map_depth);
        vec3 world_for_sprite_shadows = world.xyz + ExtraShadowParams.y * gbuffer.Normal;
        vec4 epos = ((vec4(world_for_sprite_shadows, 1)) * (ExtraShadowViewProjection));
        vec2 euv = epos.xy * vec2(0.5f, -0.5f) + vec2(0.5f);
        float extra_shadow = GetShadowFactorAndMinDepth(ExtraShadowMap, euv, epos.z, 0.0f, ExtraShadowParams.xx, screen, extra_map_depth);
        float dist = epos.z - extra_map_depth;
        float fade = clamp(dist * ExtraShadowParams.z + ExtraShadowParams.w, float(0), float(1));
        extra_shadow = mix(extra_shadow, 1.0f, fade);
        // make world shadows "block" sprite shadows
        float umbra_check = 0.1f;
        float in_base_umbra = step(base_shadow, umbra_check);
        float world_blocks_sprite = step(extra_map_depth, base_map_depth + 8.0f * ShadowParams.y);
        world_blocks_sprite = world_blocks_sprite * in_base_umbra;
        // prevent leaks at edges of world/sprite shadow overlap
        float min_shadow = min(base_shadow, extra_shadow);
        base_shadow = mix(base_shadow, min_shadow, world_blocks_sprite);
        // kill blocked sprite shadows
        extra_shadow = mix(extra_shadow, 1.0f, world_blocks_sprite);
        // prevent darker shadows showing on faces pointing away from light
        float facing_toward = smoothstep(0.0f, 0.25f, base_ndotl);
        extra_shadow = mix(1.0f, extra_shadow, facing_toward);
        shadow = min(base_shadow, extra_shadow);
        float combined_shadowness = 2.0f - base_shadow - extra_shadow;
        float select = max(0.0f, sign(base_shadow - extra_shadow));
        float intensity = mix(ShadowIntensities.x, ShadowIntensities.y, select);
        float extraness = max(0.0f, combined_shadowness - 1.0f);
        intensity = mix(intensity, ShadowIntensities.z, extraness);
        shadow_color *= vec3(intensity);
        // attenuate ambient when in extra shadow to ensure the extra shadow shows up
        float ambient_attenuate = mix(1.0f, ShadowIntensities.w, extraness);
        ambient *= vec3(ambient_attenuate);
        // remove shadow if we're past the cutoff z-plane
        float t = max(0.0f, sign(shadow_z_cutoff - world.z));
        shadow = mix(shadow, 1.0f, t);
    }
    if (has_foreground_shade)
    {
        vec3 foreground = ForegroundShading(world.xyz, gbuffer.Normal, Foreground, ForegroundAngle);
        shadow *= foreground.x;
        foreground_darkening = foreground.y;
        albedo_saturation = foreground.z;
    }
    // 	if (has_gobo)
    // 	{
    // 		// change UVs if we're using a world-based projection
    // 		const float mode = NearFar_GoboMode.z;
    // 		if (mode == GoboMode_WorldXY)
    // 		{
    // 			luv = world.xy;
    // 		}
    // 		else if (mode == GoboMode_WorldXZ)
    // 		{
    // 			luv = world.xz;
    // 		}
    // 		else if (mode == GoboMode_WorldZY)
    // 		{
    // 			luv = world.zy;
    // 		}
    //
    // 		const float2 scale = GoboUvScaleOffset.xy;
    // 		const float2 offset = GoboUvScaleOffset.zw;
    // 		float2 gobo_uv = luv * scale + offset;
    // #if defined(__NVN__) || defined(__GLSL__)
    // 		gobo_uv.y = 1.0f - gobo_uv.y;
    // #endif
    // 		float4 gobo_color = TEX2DLOD(Gobo, float4(gobo_uv, 0.0f, 0.0f));
    // 		light_color *= gobo_color.rgb;
    // 	}
    float ndotl = (base_ndotl + Direction_Wrap.w) / (1.0f + Direction_Wrap.w);
    vec4 xst_temp12 = (vec4(ndotl * 0.5f + 0.5f, 0.5f, 0, 0));
    // apply wrap lighting
    ndotl = mix(max(0.0f, ndotl), textureLod(NdotlRemap, xst_temp12.xy, xst_temp12.w).r, Color_NdotlRemap.w);
    // apply shading texture
    // float shade = lerp(1.0f, ndotl * shadow, ShadowParams.w);
    float shade = ndotl * shadow;
    light_color = mix(shadow_color, light_color, vec3(shade)) * foreground_darkening;
    if (debug_view == DebugView_None)
    {
        vec3 desat = vec3(vec3(RgbToLuminance(gbuffer.Albedo)));
        vec3 albedo = mix(desat, gbuffer.Albedo, vec3(albedo_saturation));
        color.rgb = light_color * albedo;
        color.rgb += albedo * ambient;
        color.rgb = mix(color.rgb, albedo * gbuffer.EmissiveIntensity, vec3(gbuffer.Emissivity));
        color.a = EncodeEmissive(gbuffer.EmissiveIntensity);
    }
    else if (debug_view == DebugView_Albedo)
    {
        color.rgb = gbuffer.Albedo;
    }
    else if (debug_view == DebugView_Normals)
    {
        color.rgb = gbuffer.Normal * 0.5f + vec3(0.5f);
    }
    else if (debug_view == DebugView_Emissive)
    {
        color.rgb = vec3(gbuffer.Emissivity);
    }
    else if (debug_view == DebugView_Ndotl)
    {
        color.rgb = vec3(ndotl);
    }
    else if (debug_view == DebugView_Directional)
    {
        color.rgb = light_color * gbuffer.Albedo;
    }
    else if (debug_view == DebugView_Ambient)
    {
        color.rgb = ambient;
    }
    else if (debug_view == DebugView_Ssao)
    {
        color.rgb = vec3(ao);
    }
    else if (debug_view == DebugView_Depth)
    {
        color.rgb = vec3(depth * AmbientColor_DepthScale.w);
    }
    else if (debug_view == DebugView_Shadow)
    {
        color.rgb = vec3(shadow);
    }
    else if (debug_view == DebugView_AmbientMap)
    {
        vec4 xst_temp13 = (texcoord);
        color.rgb = textureLod(AmbientMap, xst_temp13.xy, xst_temp13.w).rgb;
    }
    else if (debug_view == DebugView_LightColorMap)
    {
        vec4 xst_temp14 = (texcoord);
        color.rgb = textureLod(LightColorMap, xst_temp14.xy, xst_temp14.w).rgb;
    }
    else if (debug_view == DebugView_ShadowColorMap)
    {
        vec4 xst_temp15 = (texcoord);
        color.rgb = textureLod(ShadowColorMap, xst_temp15.xy, xst_temp15.w).rgb;
    }
    else
    {
        color.rgb = vec3(0.9f, 0.5f, 0.5f);
    }
    SV_Target0 = color;
}
