// GLSL Fragment Shader "main"
// Generated by XShaderCompiler

#version 330 core
in vec2 xsv_TEXCOORD0;
out vec4 SV_Target0;
//  computes the luminance value for an RGB color
float RgbToLuminance(vec3 color)
{
    return dot(color, vec3(0.212671, 0.715160, 0.072169));
}
uniform vec4 preg_0;
// (grayscale, vignette, overlay, overlay mode)
uniform vec4 preg_1;
// (invert, gamma, CRAZY WARP VALUE, warp amount)
uniform vec4 preg_2;
// (warp position xy, warp inverse radius xy)
// This needs to precede all texture defines for post-processing to work properly
//  TextureNames;Frame:0;Vignette:1;Overlay:2;
uniform sampler2D treg_0;
uniform sampler2D treg_1;
uniform sampler2D treg_2;


// Uniform name defines ///////////////////////////////////////
#define SettingsA preg_0
#define SettingsB preg_1
#define SettingsC preg_2
#define Frame treg_0
#define Vignette treg_1
#define Overlay treg_2
///////////////////////////////////////////////////////////////

void main()
{
    // warp
    // unintended-but-cool thing
    // const float2 warp_delta = input.TexCoord - SettingsC.xy;
    // const float2 warp_coords = SettingsC.z * warp_delta;
    // const float strength = SettingsC.w * (1.0f - min(1.0f, length(warp_coords)));
    // const float2 tex_delta = strength * (float2(0.5, -0.5) * warp_delta + 0.5f);
    vec2 tex_delta;
    if (SettingsB.z >= 0.0f)
    {
        vec2 warp_delta = SettingsC.xy - xsv_TEXCOORD0;
        vec2 warp_coords = SettingsC.zw * warp_delta;
        float strength = SettingsB.w * (1.0f - min(1.0f, length(warp_coords)));
        tex_delta = strength * warp_delta;
    }
    else
    {
        vec2 warp_delta = SettingsC.xy - xsv_TEXCOORD0;
        // vector from input coord to warp center
        // find the vector from warp center to input coord
        vec2 warpCenterToMe = -warp_delta;
        // find the vector from warp center to "correct corner" of image, which is determined by whether I'm UL UR DL or DR from the warp center itself
        vec2 correctCorner;
        correctCorner.x = float((xsv_TEXCOORD0.x < SettingsC.x) ? 0 : 1);
        correctCorner.y = float((xsv_TEXCOORD0.y < SettingsC.y) ? 0 : 1);
        vec2 warpCenterToCorner = correctCorner - xsv_TEXCOORD0;
        // dot the two, normalized
        float normalizedDot = dot(normalize(warpCenterToMe), normalize(warpCenterToCorner));
        // use as warp strength modifier
        // we want 1.0 to result in ALL POINTS being at the warp point
        // but to get there at different rates
        // so when normalizedDot is 1 or close to 1 we want it to be linear
        // and faster the farther away we get
        float warpScale = 1.0f + 3.0f * (1.0f - normalizedDot * normalizedDot);
        float lerpVal = clamp((1.0f - warpScale * SettingsB.w), float(0), float(1));
        // AH OKAY so this is "I am at a position along the line from warp point outward, where does this point stretch OUTWARD to"
        // rather than "where is this point stretching INWARD from"
        // argh
        // so if I am the INWARD point, we need to find the OUTWARD point that WOULD map to me
        // so my dist is a (lerpVal) percentage of the outward dist
        // and the total outward dist from the warp center is lerpVal + (1-lerpVal)
        // so the EXTRA OUTWARD distance is (1-lerpVal)/lerpVal
        // and we have the backward vector, so negative that
        tex_delta = -warp_delta * ((1.0f - lerpVal) / lerpVal);
    }
    vec4 t = vec4(xsv_TEXCOORD0 + tex_delta, 0, 0);
    vec4 xst_temp1 = (t);
    vec3 src = textureLod(Frame, xst_temp1.xy, xst_temp1.w).rgb;
    vec3 xsr_output = src.rgb;
    // invert
    vec3 inverse = vec3(1, 1, 1) - xsr_output;
    xsr_output = mix(xsr_output, inverse, vec3(SettingsB.x));
    // grayscale
    vec3 luminance = vec3(RgbToLuminance(xsr_output));
    xsr_output = mix(xsr_output, luminance, vec3(SettingsA.x));
    vec4 xst_temp2 = (t);
    // overlay - mode is 0 = additive; 1 = alpha
    vec4 overlay = textureLod(Overlay, xst_temp2.xy, xst_temp2.w);
    float alpha = overlay.a * SettingsA.z;
    xsr_output = overlay.rgb * alpha + xsr_output * (1.0f - SettingsA.w * alpha);
    vec4 xst_temp3 = (t);
    // vignette
    vec4 vignette = textureLod(Vignette, xst_temp3.xy, xst_temp3.w);
    xsr_output = mix(xsr_output, vignette.rgb, vec3(SettingsA.y * vignette.a));
    // gamma - SHOULD BE THE LAST THING!
    xsr_output = pow(abs(xsr_output), SettingsB.yyy);
    SV_Target0 = vec4(xsr_output, 1);
}
