// GLSL Fragment Shader "main"
// Generated by XShaderCompiler

#version 330 core

out vec4 SV_Target0;
//  This header includes functions for working with the G-buffer
struct Gbuffer
{
    vec3  Albedo;
    vec3  Normal;
    float Emissivity;
    float EmissiveIntensity;
};
vec3 UnpackNormal(vec3 sampled_normal)
{
    return sampled_normal * 2.0f - vec3(1.0f);
}
float EmissiveMax = 4.0f;
float DecodeEmissive(float e)
{
    return e * EmissiveMax;
}
Gbuffer SampleGbuffer(sampler2D gbuffer0, sampler2D gbuffer1, vec2 texcoord)
{
    Gbuffer g;
    vec4 t = vec4(texcoord, 0.0f, 0.0f);
    vec4 xst_temp0 = (t);
    vec4 sample0 = textureLod(gbuffer0, xst_temp0.xy, xst_temp0.w);
    vec4 xst_temp1 = (t);
    vec4 sample1 = textureLod(gbuffer1, xst_temp1.xy, xst_temp1.w);
    g.Albedo = sample0.rgb;
    g.Normal = UnpackNormal(sample1.rgb);
    g.Emissivity = sample0.a;
    g.EmissiveIntensity = DecodeEmissive(sample1.a);
    return g;
}
float ConvertDepthSample(float depth)
{
    return depth;
}
vec4 GetPositionFromExponentialDepth(float depth, float near, float far, vec2 screen, mat4 ivp)
{
    vec4 p = vec4(screen.x, screen.y, depth, 1.0f);
    far;
    // avoid warnings
    float z = near / min(-0.001f, depth - 1.0f);
    p *= vec4(-z);
    vec4 pw = ((p) * (ivp));
    return pw;
}
vec4 GetPositionFromDepth(float depth, float near, float far, vec2 screen, mat4 ivp)
{
    return GetPositionFromExponentialDepth(depth, near, far, screen, ivp);
}
uniform mat4 preg_0;
// camer inverse view-projection transform
uniform mat4 preg_4;
uniform vec4 preg_8;
// (bias x, bias y, UNUSED, UNUSED)
uniform vec4 preg_10;
// (use ndotl remap, use falloff texture, UNUSED, UNUSED)
uniform vec4 preg_11;
// (x, y, z, 1 / radius)
uniform vec4 preg_12;
// (r, g, b, UNUSED)
// This needs to preceed all texture defines for post-processing to work properly
// TextureNames;Gbuffer0:0;Gbuffer1:1;DepthTexture:2;NdotlRemap:3;Falloff:4;
uniform sampler2D treg_0;
uniform sampler2D treg_1;
uniform sampler2D treg_2;
uniform sampler2D treg_3;
uniform sampler2D treg_4;


// Uniform name defines ///////////////////////////////////////
#define InvProjection preg_0
#define InvView preg_4
#define Depth_InvRes preg_8
#define Settings preg_10
#define Position_Radius preg_11
#define xst_Color preg_12
#define Gbuffer0 treg_0
#define Gbuffer1 treg_1
#define DepthTexture treg_2
#define NdotlRemap treg_3
#define Falloff treg_4
///////////////////////////////////////////////////////////////

void main()
{
    vec2 tex = gl_FragCoord.xy * Depth_InvRes.zw;
    vec2 screen = tex * vec2(2, 2) + vec2(-1, -1);
    vec4 xst_temp3 = (vec4(tex, 0, 0));
    float depth = ConvertDepthSample(textureLod(DepthTexture, xst_temp3.xy, xst_temp3.w).r);
    vec4 wpos = GetPositionFromDepth(depth, Depth_InvRes.x, Depth_InvRes.y, screen, InvProjection);
    wpos = ((wpos) * (InvView));
    vec3 lpos = Position_Radius.xyz;
    float inv_radius = Position_Radius.w;
    vec3 light_vec = lpos - wpos.xyz;
    float dist2 = dot(light_vec, light_vec);
    float dist = sqrt(dist2);
    Gbuffer gbuffer = SampleGbuffer(Gbuffer0, Gbuffer1, tex);
    float ndotl = max(0.0f, dot(gbuffer.Normal, light_vec / dist));
    vec4 xst_temp4 = (vec4(ndotl * 0.5f + 0.5f, 0, 0, 0));
    ndotl = mix(max(0.0f, ndotl), textureLod(NdotlRemap, xst_temp4.xy, xst_temp4.w).r, Settings.x);
    // by default we use linear faloff
    float dist_norm = dist * inv_radius;
    vec4 xst_temp5 = (vec4(dist_norm, 0, 0, 0));
    float falloff = mix(max(0.0f, 1.0f - dist_norm), textureLod(Falloff, xst_temp5.xy, xst_temp5.w).r, Settings.y);
    // DOCUMENTING FOR THE FUTURE
    // Max wanted the toon shading circles to be affected by falloff
    // the best I found was:
    // --------
    // float falloff = max(0.0f, 1.0f - dist / radius);
    // float ndotl = max(0.0f, dot(gbuffer.Normal, light_vec / dist));
    // float tempDumb = saturate(ndotl + (falloff - 0.5f));
    // ndotl = lerp(max(0.0f, ndotl), tex1Dlod(NdotlRemap, float4(tempDumb * 0.5f + 0.5f, 0, 0, 0)).r, Settings.x);
    // --------
    // BUT this leaves everything darker than before in general.
    // @ROB HALP!
    SV_Target0.rgb = gbuffer.Albedo * xst_Color.rgb * ndotl * falloff;
    SV_Target0.rgb = mix(SV_Target0.rgb, vec3(0, 0, 0), vec3(gbuffer.Emissivity));
    SV_Target0.a = 0.0f;
    // not alpha blended, and alpha in target stores emissive intensity
}
