// GLSL Fragment Shader "main"
// Generated by XShaderCompiler

#version 330 core
void sincos(float x, out float s, out float c)
{
    s = sin(x), c = cos(x);
}
in vec2 xsv_TEXCOORD0;
out vec4 SV_Target0;
float ConvertDepthSample(float depth)
{
    return depth;
}
vec4 GetPositionFromExponentialDepth(float depth, float near, float far, vec2 screen, mat4 ivp)
{
    vec4 p = vec4(screen.x, screen.y, depth, 1.0f);
    far;
    // avoid warnings
    float z = near / min(-0.001f, depth - 1.0f);
    p *= vec4(-z);
    vec4 pw = ((p) * (ivp));
    return pw;
}
vec4 GetPositionFromDepth(float depth, float near, float far, vec2 screen, mat4 ivp)
{
    return GetPositionFromExponentialDepth(depth, near, far, screen, ivp);
}
float SHIFT_RIGHT_08 = 1.0f / 256.0f;
float MAX_24_BIT = 16777215.0f;
//  packs a float value in the [0.0, 1.0] range into three 8-bit values
vec3 unorm_to_rgb8(float value)
{
    vec3 encoded;
    float scaled = MAX_24_BIT * value;
    scaled = SHIFT_RIGHT_08 * scaled;
    encoded.z = fract(scaled);
    scaled = SHIFT_RIGHT_08 * trunc(scaled);
    encoded.y = fract(scaled);
    encoded.x = SHIFT_RIGHT_08 * trunc(scaled);
    return encoded;
}
//  converts a texture coordinate to a screen coordinate
vec2 TexToScreen(vec2 tex)
{
    return tex.xy * vec2(2.0f, -2.0f) + vec2(-1.0f, 1.0f);
}
uniform mat4 preg_0;
uniform vec4 preg_4;
uniform vec4 preg_5;
// { sample radius, footprint, use box filter, unused }
uniform vec4 preg_6;
// { sigma, beta, k, unused }
uniform sampler2D treg_0;


// Uniform name defines ///////////////////////////////////////
#define InvProjection preg_0
#define NearFar_Resolution preg_4
#define Settings preg_5
#define Settings2 preg_6
#define Depth treg_0
///////////////////////////////////////////////////////////////

float samples = 9.0f;
float inv_samples = 1.0f / samples;
float tau = 7.0f;
float two_pi = 6.283185307f;
float q = 16.0f;
// paper says use some value 2^3 <= q <= 2^5
float inv_q = 1.0f / q;
// static const float sigma = 1.0f;
// static const float beta = 0.0f;
// static const float k = 1.0f;
float epsilon = 0.0001f;
void main()
{
    // constants renamed for clarity/convenience
    float near = NearFar_Resolution.x;
    float far = NearFar_Resolution.y;
    vec2 res = NearFar_Resolution.zw;
    float radius = Settings.x;
    float footprint = Settings.y;
    float sigma = Settings2.x;
    float beta = Settings2.y;
    float k = Settings2.z;
    // center texture and screen coordinates
    vec2 screen = TexToScreen(xsv_TEXCOORD0);
    vec2 pixel = floor(xsv_TEXCOORD0.xy * res);
    vec4 xst_temp4 = (vec4(xsv_TEXCOORD0, 0.0f, 0.0f));
    // calculate camera-space position and normal
    float depth = ConvertDepthSample(textureLod(Depth, xst_temp4.xy, xst_temp4.w).r);
    vec3 c = GetPositionFromDepth(depth, NearFar_Resolution.x, NearFar_Resolution.y, screen, InvProjection).xyz;
    vec3 dcdx = dFdx(c);
    vec3 dcdy = dFdy(c);
    vec3 n = normalize(cross(dcdy, dcdx));
    float r = -radius * footprint / c.z;
    float radius2 = radius * radius;
    vec2 pix_mod2 = fract(pixel * 0.5f) * 2.0f;
    float phi = pix_mod2.x * 3.1415926535f + pix_mod2.y * 1.5707963267f;
    // const float phi = 0.0f;
    float ao = 0.0f;
    for (float i = 0.0f; i < samples; i += 1.0f)
    {
        float alpha = (i + 0.5f) * inv_samples;
        float h = r * alpha;
        float theta = two_pi * alpha * tau + phi;
        vec2 u;
        sincos(theta, u.y, u.x);
        float m = floor(log2(h * inv_q));
        vec4 t = vec4(h * u + xsv_TEXCOORD0.xy, 0.0f, m);
        vec4 xst_temp5 = (t);
        float d = ConvertDepthSample(textureLod(Depth, xst_temp5.xy, xst_temp5.w).r);
        vec3 p = GetPositionFromDepth(d, near, far, TexToScreen(t.xy), InvProjection).xyz;
        vec3 v = p - c;
        float l2 = dot(v, v);
        float x = radius2 - l2;
        // older estimator
        ao += max(0.0f, dot(v, n) + c.z * beta) / (l2 + epsilon);
    }
    // older estimator
    ao = clamp(1.0f - (2.0f * sigma * inv_samples) * ao, float(0), float(1));
    ao = pow(ao, k);
    // newer estimator
    // const float radius6 = radius2 * radius2 * radius2;
    // ao = max(0.0f, 1.0f - (5.0f / (radius6 * samples)) * ao);
    // ao = max(0.0f, 1.0f - (1.0f / (samples)) * ao);
    // box filter over 2x2 pixel quad
    if (Settings.z == 0.0f)
    {
        ao -= dFdx(ao) * (pix_mod2.x - 0.5f);
        ao -= dFdy(ao) * (pix_mod2.y - 0.5f);
    }
    vec3 packed_depth = unorm_to_rgb8(depth);
    SV_Target0 = vec4(packed_depth, ao);
}
