// GLSL Vertex Shader "main"
// Generated by XShaderCompiler

#version 330 core
in uvec4 ByteRow1;
in uvec4 ByteRow2;
out vec2 xsv_TEXCOORD0;
out vec4 xsv_TEXCOORD1;
out vec3 xsv_TEXCOORD2;
out vec4 xsv_TEXCOORD3;
out vec4 xsv_TEXCOORD4;
//  input.ByteRow1 - uint8 screenBlockX, screenBlockY, atlasBlockX, atlasBlockY;
//  input.ByteRow2 - uint8 blockBelowX, blockBelowY, blockBelowRightX, blockBelowRightY
//  Shader constant parameters
uniform mat4 vreg_0;
uniform mat4 vreg_4;
uniform mat4 vreg_8;
uniform vec4 vreg_12;
// (flipped, death anim, max row, max column)
//  pixels to add, block to divide on; a NEGATIVE pixels to add means do this VERTICALLY!
uniform vec4 vreg_13;
// (beam, beam, z bias, unused)
//  t for super shadows background, t for superflash, flashnum for superflash, depth for silhouettes
uniform vec4 vreg_14;


// Uniform name defines ///////////////////////////////////////
#define ModelToWorld vreg_0
#define WorldToView vreg_4
#define ViewToClip vreg_8
#define Flipped vreg_12
#define Beam_Zbias vreg_13
#define ScreenEffectVals vreg_14
///////////////////////////////////////////////////////////////

//  this should be extra on PS3, not on PC?
vec3 FLASH_COLORS[4] = vec3[4](vec3(255.f / 255.f, 216.f / 255.f, 0.f), vec3(-260.f / 255.f, 586.f / 255.f, 80.f / 255.f), vec3(1005.f / 255.f, 374.f / 255.f, 548.f / 255.f), vec3(-260.f / 255.f, 210.f / 255.f, 560.f / 255.f));
float Random[11] = float[11](0.591629f, 0.088310f, 0.911847f, 0.118983f, 0.699444f, 0.259557f, 0.952650f, 0.140039f, 0.450610f, 0.853216f, 0.330409f);
//  atlasBlockX and Y will be up to 127, therefore we can use the last bit as a which-edge-of-poly indicator
//  blockBelowX & 128 -> has block to left of us, blockBelowRightX & 128 -> has block to right of us
void main()
{
    uvec4 xst_ByteRow1 = ByteRow1;
    uvec4 xst_ByteRow2 = ByteRow2;
    float flashNum = round(clamp(ScreenEffectVals.z - 1.0f, 0.0f, 3.0f));
    xsv_TEXCOORD4.rgb = FLASH_COLORS[uint(flashNum)];
    bool isOnRightEdgeOfBlock = (xst_ByteRow1.z >= 128u);
    bool isOnBottomEdgeOfBlock = (xst_ByteRow2.x >= 128u);
    // and remove these flags from the calculation
    xst_ByteRow1.z -= uint((isOnRightEdgeOfBlock) ? 128 : 0);
    xst_ByteRow2.x -= uint((isOnBottomEdgeOfBlock) ? 128 : 0);
    // const bool hasBlockToLeft = (input.ByteRow2.x >= 128);
    // output.BlockInfo.x = (hasBlockToLeft) ? input.ByteRow2.x - 128 - 1 : input.ByteRow2.x - 1; // -1 means no block
    // output.BlockInfo format:
    // x = blockBelow x coord, y = blockBelow y coord, z = blockBelowRight x coord, w = blockBelowRight y coord
    xsv_TEXCOORD3.x = float(xst_ByteRow2.x) - 1.0f;
    // -1 means no block
    xsv_TEXCOORD3.y = float(xst_ByteRow2.y) - 1.0f;
    // -1 means no block
    bool hasBlockToRight = (xst_ByteRow2.z >= 128u);
    xsv_TEXCOORD3.z = (hasBlockToRight) ? float(xst_ByteRow2.z) - 128.0f - 1.0f : float(xst_ByteRow2.z) - 1.0f;
    // -1 means no block
    xsv_TEXCOORD3.w = float(xst_ByteRow2.w) - 1.0f;
    // -1 means no block
    // firstly, our world position is screenBlockX (or Y) * blocksize
    gl_Position = vec4(xst_ByteRow1.xy * 16u, 0, 1);
    // BEAMS!
    bool bDoVerticalBeam = (Beam_Zbias.x < 0.0f);
    gl_Position.x += float((!bDoVerticalBeam)) * abs(Beam_Zbias.x) * clamp(float(((2u * xst_ByteRow1.x) + uint(isOnRightEdgeOfBlock))) - (Beam_Zbias.y * 2.0f), float(0), float(1));
    gl_Position.y += float((bDoVerticalBeam)) * abs(Beam_Zbias.x) * clamp(float(((2u * xst_ByteRow1.y) + uint(!isOnBottomEdgeOfBlock))) - (Beam_Zbias.y * 2.0f), float(0), float(1));
    // death animation
    int row_shift = 3;
    int start_index = int((xst_ByteRow1.x + uint(row_shift) * xst_ByteRow1.y) % 11u);
    float column_start_offset = Random[start_index];
    float k = 3.0f;
    // number of rows that are visible disappearing at once
    float ik = 1.0f / k;
    vec2 dims = Flipped.zw;
    float scale = ik * (dims.y + k + 1.0f);
    float offset = ik * (float(xst_ByteRow1.y) - dims.y - column_start_offset);
    float t_death = clamp(scale * Flipped.y + offset, float(0), float(1));
    xsv_TEXCOORD4.a = t_death;
    float death_adjust = t_death * 8.0f;
    float block_width = (1.0f - t_death) * 16.0f;
    gl_Position.x += death_adjust;
    float death_blink_height = 70.0f;
    gl_Position.y += death_blink_height * t_death * mix(1.0f, 4.0f, t_death);
    // get farther away/faster as the time goes
    // if input.ByteRow1.x > indicated value || it is equal to indicated value and isOnRightEdgeOfBlock, add next indicated value
    gl_Position.x += (isOnRightEdgeOfBlock) ? block_width : 0;
    gl_Position.y += (isOnBottomEdgeOfBlock) ? 0 : 16.0f;
    // these were in image space
    gl_Position = ((gl_Position) * (ModelToWorld));
    gl_Position.xy -= vec2(0.0f, 0.0f);
    ;
    gl_Position = ((gl_Position) * (WorldToView));
    xsv_TEXCOORD2 = gl_Position.xyz;
    // the depth pass version outputs the camera-space position as the "world" position
    gl_Position = ((gl_Position) * (ViewToClip));
    // atlas block number to look up
    // basically, bMirrored WOULD negates if we're on the right edge or not, so we check if those are not equal instead of checking just if we're on the right
    // but we're skipping this so we can correct other problems, and we may never need to use it!
    // bool bMirrored = false;//(gNormalMirrorFlipParams.x < 0);
    // on *PC* we flip these and correct for it in the pixel shader
    // on PS3, we don't because the rasterizer is good enough to not screw the UVs up!
    // pixel offset within the block (0 = left/top, 1 = right/bottom)
    float pixelY = float((isOnBottomEdgeOfBlock ? 1 : 0));
    float pixelX;
    if (Flipped.x < 0.0f)
    {
        pixelX = float((isOnRightEdgeOfBlock ? 0 : 1));
    }
    else
    {
        pixelX = float((isOnRightEdgeOfBlock ? 1 : 0));
    }
    xsv_TEXCOORD0 = vec2(pixelX, pixelY);
    // outBlockPos:  MY block x, MY block y, BLOCK TO RIGHT x, BLOCK TO RIGHT y
    vec2 blockPosToRight = vec2(-1.0f, -1.0f);
    if (hasBlockToRight)
    {
        blockPosToRight.x = float(xst_ByteRow1.z + 1u);
        blockPosToRight.y = float(xst_ByteRow1.w);
        if (blockPosToRight.x >= float(1024 / 16))
        {
            blockPosToRight = vec2(0, blockPosToRight.y + 1.0f);
        }
    }
    xsv_TEXCOORD1 = vec4(xst_ByteRow1.zw, blockPosToRight.xy);
}
