// GLSL Fragment Shader "main"
// Generated by XShaderCompiler

#version 330 core

in vec4 xsv_TEXCOORD0;
in vec3 xsv_TEXCOORD1;
in vec4 xsv_TEXCOORD2;
out vec4 SV_Target0;
float ConvertDepthSample(float depth)
{
    return depth;
}
vec4 GetPositionFromExponentialDepth(float depth, float near, float far, vec2 screen, mat4 ivp)
{
    vec4 p = vec4(screen.x, screen.y, depth, 1.0f);
    far;
    // avoid warnings
    float z = near / min(-0.001f, depth - 1.0f);
    p *= vec4(-z);
    vec4 pw = ((p) * (ivp));
    return pw;
}
vec4 GetPositionFromDepth(float depth, float near, float far, vec2 screen, mat4 ivp)
{
    return GetPositionFromExponentialDepth(depth, near, far, screen, ivp);
}
// *
//  *  \param position  World position of the point being shaded.
//  *  \param normal    Surface normal at the point being shaded.
//  *  \param params1   Vector containing: {z scale, z offset, overdarken, saturation}
//  *  \param params2   Vector containing: {angle fade, unused, unused, unused}
//  *  \return          Vector containing: {shadow factor, darkening factor, saturation}
//
vec3 ForegroundShading(vec3 position, vec3 normal, vec4 params1, vec4 params2)
{
    float normal_scale = pow(max(1.0f - clamp(dot(vec3(0, 1, 0), normal), float(0), float(1)), 0.0001f), params2.x);
    float t = normal_scale * clamp(position.z * params1.x + params1.y, float(0), float(1));
    float shadow = (1.0f - t);
    float darken = mix(1.0f, params1.z, t);
    float saturation = mix(1.0f, params1.w, t);
    return vec3(shadow, darken, saturation);
}
// *
//  * \param base_color  Color of the surface fog will be applied to.
//  * \param distance   Surface distance.
//  * \param fog_color   Color of the fog.
//  * \param fog_max     Maximum fog strength.
//  * \param fog_bounds  Vector containing: {fog start distance, fog end distance}
//
vec3 ApplyFog(vec3 base_color, float xsr_distance, vec3 fog_color, float fog_max, vec2 fog_bounds)
{
    float fog_amount = clamp((xsr_distance - fog_bounds.x) / (fog_bounds.y - fog_bounds.x), float(0), float(1));
    return mix(base_color, fog_color, vec3(min(fog_amount, fog_max)));
}
vec3 GetPlayerPathAmbient(float zpos, vec3 base_ambient, vec3 path_ambient, vec4 path_params)
{
    float tfront = clamp(zpos * path_params.x + path_params.y, float(0), float(1));
    float tback = clamp(zpos * path_params.z + path_params.w, float(0), float(1));
    float t = max(tfront, tback);
    return mix(path_ambient, base_ambient, vec3(t));
}
vec3 DecodeLightMap(vec4 encoded)
{
    float k = 3.0f;
    float scale = k * encoded.a;
    vec3 light = scale * encoded.rgb;
    return light;
}
//  takes screen pixel coords and return a 0 or 1 checkerboard pattern
vec2 Checkerboard(vec2 screen)
{
    vec2 checkerboard = trunc(screen);
    checkerboard = 0.5f * checkerboard;
    checkerboard = fract(checkerboard);
    checkerboard = 2.0f * checkerboard;
    return checkerboard;
}
vec3 GetFlashColorLow(float index)
{
    vec3 gLowColors[4] = vec3[4](vec3(115.f / 255.f, 24.f / 255.f, 0.f / 255.f), vec3(55.f / 255.f, 120.f / 255.f, 55.f / 255.f), vec3(200.f / 255.f, 22.f / 255.f, 22.f / 255.f), vec3(5.f / 255.f, 11.f / 255.f, 17.f / 255.f));
    return gLowColors[uint(index)];
}
vec3 GetFlashColorHigh(float index)
{
    vec3 gHighColors[4] = vec3[4](vec3(255.f / 255.f, 168.f / 255.f, 100.f / 255.f), vec3(110.f / 255.f, 110.f / 255.f, 20.f / 255.f), vec3(200.f / 255.f, 168.f / 255.f, 120.f / 255.f), vec3(19.f / 255.f, 30.f / 255.f, 38.f / 255.f));
    return gHighColors[uint(index)];
}
vec3 ApplySuperShadow(vec3 xsr_input, float t)
{
    vec3 super_color = vec3(0.2f, 0.2f, 0.2f);
    vec3 luminance = vec3(1.f, 1.f, 1.f);
    float theDot = 1.f - clamp(dot(xsr_input, luminance), float(0), float(1));
    // if(theDot < 1.f / 3.f)
    // theDot = 3.f * theDot;
    // else if(theDot < 2.f / 3.f)
    // theDot = 1.f;
    // else
    // theDot = 3.f * (1.f - theDot);
    // const float eased = (3 * theDot*theDot - 2 * theDot*theDot*theDot);
    // const float twiceEased = (3 * eased*eased - 2 * eased*eased*eased);
    // return lerp(input, super_color * eased, t);
    return mix(xsr_input, super_color * theDot, vec3(t));
}
vec3 ApplySuperFlash(vec3 xsr_input, float t, vec3 colorLow, vec3 colorHigh)
{
    vec3 coeff = vec3(0.39f, 0.5f, 0.11f);
    float initialGray = dot(xsr_input, coeff);
    // -- Desaturated
    float Contrast = 1.56f;
    float Brightness = 1.5f;
    // const float Saturation = 0.25f;
    // const float Saturation = 0.15f;
    // const float Saturation = 0.10f;
    // const float Saturation = 0.05f;
    float Saturation = 0.02f;
    // -- High Contrast --
    // const float Contrast = 1.9f;
    // const float Brightness = 2.5f;;
    // const float Saturation = 1.25f;
    float gray = Brightness * (1.2f + mix(0.5f, initialGray - 0.5f, Contrast));
    // return lerp(lerp(colorLow, colorHigh, initialGray) * gray, initialGray, -Saturation);  // ALWAYS ON
    return mix(xsr_input, mix(mix(colorLow, colorHigh, vec3(initialGray)) * gray, vec3(initialGray), vec3(-Saturation)), vec3(t));
}
bool BattleFade(vec2 screen, float t, sampler2D fade_texture, vec2 tex_size)
{
    bool cut_or_not = false;
    vec4 texcoord = vec4(screen / tex_size, 0, 0);
    vec4 xst_temp3 = (texcoord);
    float c = textureLod(fade_texture, xst_temp3.xy, xst_temp3.w).r;
    cut_or_not = c <= t;
    return cut_or_not;
}
//  Shadow map functions
//  sampling kernels
vec2 Poisson9[9] = vec2[9](vec2(0.1567495f, 0.2453307f), vec2(-0.4647257f, -0.1085932f), vec2(-0.2514141f, 0.6994573f), vec2(0.5295524f, -0.3833097f), vec2(-0.4383077f, -0.7664956f), vec2(0.957235f, 0.2121297f), vec2(0.3471507f, 0.9330071f), vec2(0.1850856f, -0.8876677f), vec2(-0.8153945f, 0.4067398f));
float GetShadowFactorAndCenterDepth(sampler2D map, vec2 uv, float depth, float bias, vec2 dims, vec2 screen, out float out_depth)
{
    out_depth = 1.0f;
    // xbox one doesn't seem to support border color correctly, so have to do this manually
    // RT_TODO: benchmark it, maybe it's faster?
    vec2 tex_scale = vec2(1.0f / dims.x, 1.0f / dims.y);
    float check_depth = depth + bias;
    depth -= bias;
    float shadow = 0.0f;
    // poisson samples
    vec2 c = Checkerboard(screen);
    float cost = 1.0f - (c.x + c.y);
    float sint = c.x - c.y;
    for (int i = 0; i < 9; ++i)
    {
        vec2 p = 1.4f * Poisson9[i];
        vec2 rp = vec2(cost * p.x - sint * p.y, sint * p.x + cost * p.y);
        vec4 s = vec4(uv + rp * tex_scale, 0, 0);
        s.y = 1.0f - s.y;
        vec4 xst_temp5 = (s);
        float map_depth = textureLod(map, xst_temp5.xy, xst_temp5.w).r;
        shadow += (depth <= map_depth ? 1.0f : 0.0f);
        if (i == 0)
        {
            out_depth = map_depth;
        }
    }
    shadow /= 9.0f;
    return mix(shadow, 1.0f, max(0.0f, sign(check_depth - 1.0f)));
}
float GetShadowFactorAndMinDepth(sampler2D map, vec2 uv, float depth, float bias, vec2 dims, vec2 screen, out float out_depth)
{
    out_depth = 1.0f;
    // xbox one doesn't seem to support border color correctly, so have to do this manually
    // RT_TODO: benchmark it, maybe it's faster?
    vec2 tex_scale = vec2(1.0f / dims.x, 1.0f / dims.y);
    out_depth = 1.0f;
    float check_depth = depth + bias;
    depth -= bias;
    float shadow = 0.0f;
    // poisson samples
    vec2 c = Checkerboard(screen);
    float cost = 1.0f - (c.x + c.y);
    float sint = c.x - c.y;
    for (int i = 0; i < 9; ++i)
    {
        vec2 p = 1.4f * Poisson9[i];
        vec2 rp = vec2(cost * p.x - sint * p.y, sint * p.x + cost * p.y);
        vec4 s = vec4(uv + rp * tex_scale, 0, 0);
        s.y = 1.0f - s.y;
        vec4 xst_temp6 = (s);
        float map_depth = textureLod(map, xst_temp6.xy, xst_temp6.w).r;
        shadow += (depth <= map_depth ? 1.0f : 0.0f);
        out_depth = min(out_depth, map_depth);
    }
    shadow /= 9.0f;
    return mix(shadow, 1.0f, max(0.0f, sign(check_depth - 1.0f)));
}
//  global constants
uniform vec4 preg_0;
// (has shadows, foreground shading, vert color enabled, anim timer)
uniform vec4 preg_1;
uniform vec4 preg_2;
uniform vec4 preg_3;
uniform vec4 preg_4;
uniform vec4 preg_5;
uniform vec4 preg_6;
// (x, y, z, wrap strength)
uniform vec4 preg_7;
uniform vec4 preg_8;
// (shadow rgb, shadow cutoff z)
uniform vec4 preg_9;
uniform vec4 preg_10;
// (world bias, depth bias, size, unused)
uniform mat4 preg_11;
uniform vec4 preg_15;
// (size, world bias, fade scale, fade offset)
uniform mat4 preg_16;
uniform vec4 preg_20;
uniform vec4 preg_21;
uniform vec4 preg_22;
// (start, end, max strength)
uniform vec4 preg_23;
uniform vec4 preg_24;
uniform mat4 preg_25;
//  instance constants
uniform vec4 preg_29;
// (water, visibility, emissivity, emissive intensity)
uniform vec4 preg_30;
// (fade texture size, cutoff amount, is additive)
uniform vec4 preg_31;
// (tint rgb, unused) OR (water main color)
uniform vec4 preg_32;
uniform vec4 preg_33;
// This needs to preceed all texture defines for post-processing to work properly
// TextureNames;DiffuseTexture:0;BattleFadeTexture:1;AmbientMap:2;LightColorMap:3;ShadowColorMap:4;ShadowMap:5;ExtraShadowMap:6;DepthTexture:7;OpaquePassTexture:8;NoiseTexture:9;
uniform sampler2D treg_0;
uniform sampler2D treg_1;
uniform sampler2D treg_2;
uniform sampler2D treg_3;
uniform sampler2D treg_4;
uniform sampler2D treg_5;
uniform sampler2D treg_6;
uniform sampler2D treg_7;


// Uniform name defines ///////////////////////////////////////
#define PassSettings preg_0
#define SuperFlash preg_1
#define Ambient preg_2
#define AmbientMapTransform preg_3
#define PlayerPathAmbient preg_4
#define PlayerPathParams preg_5
#define LightDirection_Wrap preg_6
#define LightColor preg_7
#define ShadowColor_Cutoff preg_8
#define ShadowIntensities preg_9
#define ShadowParams preg_10
#define ShadowViewProjection preg_11
#define ExtraShadowParams preg_15
#define ExtraShadowViewProjection preg_16
#define Foreground1 preg_20
#define Foreground2 preg_21
#define FogBounds preg_22
#define FogColor preg_23
#define Screen_Depth preg_24
#define InvViewProjection preg_25
#define Water_Emissive preg_29
#define BattleFade_Additive preg_30
#define Tint preg_31
#define WaterHighlight preg_32
#define WaterEdge preg_33
#define DiffuseTexture treg_0
#define BattleFadeTexture treg_1
#define AmbientMap treg_2
#define LightColorMap treg_3
#define ShadowColorMap treg_4
#define ShadowMap treg_5
#define ExtraShadowMap treg_6
#define DepthTexture treg_7
///////////////////////////////////////////////////////////////

void main()
{
    vec2 FadeSize = BattleFade_Additive.xy;
    float FadeLevel = BattleFade_Additive.z;
    if (FadeLevel != 0.0f)
    {
        if (BattleFade(gl_FragCoord.xy, FadeLevel, BattleFadeTexture, FadeSize))
        {
            discard;
        }
    }
    vec4 albedo = texture(DiffuseTexture, xsv_TEXCOORD0.xy);
    albedo.a *= xsv_TEXCOORD2.a;
    if (BattleFade_Additive.w != 0.0f)
    {
        albedo.rgb *= Tint.rgb;
        SV_Target0 = albedo;
    }
    else
    {
        if (Water_Emissive.x == 1.0f)
        {
            vec4 base_color = mix(Tint, WaterHighlight, vec4(albedo.r));
            vec4 screen_tex = vec4(gl_FragCoord.xy * Screen_Depth.xy, 0.0f, 0.0f);
            vec2 screen_coords = screen_tex.xy * vec2(2, -2) + vec2(-1, 1);
            vec4 xst_temp7 = (screen_tex);
            float depth = ConvertDepthSample(textureLod(DepthTexture, xst_temp7.xy, xst_temp7.w).r);
            vec4 world = GetPositionFromDepth(depth, Screen_Depth.z, Screen_Depth.w, screen_coords, InvViewProjection);
            float foam_inv_dist2 = 1.0f / (128.0f * 128.0f);
            vec3 dist_vector = world.xyz - xsv_TEXCOORD1.xyz;
            float dist_squared = dot(dist_vector, dist_vector);
            // const float foam = smoothstep(1.0f, 0.5f, dist_squared * Water_Emissive.y);
            float foam = 1.0f - clamp(dist_squared * foam_inv_dist2, float(0), float(1));
            float visibility = clamp(sqrt(dist_squared) * Water_Emissive.y, float(0), float(1));
            albedo = mix(base_color, WaterEdge, vec4(foam));
            albedo.a = mix(albedo.a, 1.0f, visibility);
        }
        else
        {
            albedo.rgb *= Tint.rgb;
        }
        float base_ndotl = -dot(LightDirection_Wrap.xyz, xsv_TEXCOORD2.xyz);
        // RT_TODO - forward doesn't use the ndotl remap texture
        float ndotl = (base_ndotl + LightDirection_Wrap.w) / (1.0f + LightDirection_Wrap.w);
        vec3 ambient = Ambient.rgb;
        vec3 light_color = LightColor.rgb;
        vec3 shadow_color = ShadowColor_Cutoff.rgb;
        if (Ambient.x == -1.0f)
        {
            vec2 uv = xsv_TEXCOORD1.xy * AmbientMapTransform.xy + AmbientMapTransform.zw;
            vec4 t = vec4(uv, 0, 0);
            vec4 xst_temp8 = (t);
            // ambient = TEX2DLOD(AmbientMap, t).rgb;
            // light_color = TEX2DLOD(LightColorMap, t).rgb;
            // shadow_color = TEX2DLOD(ShadowColorMap, t).rgb;
            ambient = DecodeLightMap(textureLod(AmbientMap, xst_temp8.xy, xst_temp8.w));
            vec4 xst_temp9 = (t);
            light_color = DecodeLightMap(textureLod(LightColorMap, xst_temp9.xy, xst_temp9.w));
            vec4 xst_temp10 = (t);
            shadow_color = DecodeLightMap(textureLod(ShadowColorMap, xst_temp10.xy, xst_temp10.w));
        }
        ambient = GetPlayerPathAmbient(xsv_TEXCOORD1.z, ambient, PlayerPathAmbient.rgb, PlayerPathParams);
        float shaded = ndotl;
        float shadow = 1.0f;
        bool has_shadows = (PassSettings.x != 0.0f);
        if (has_shadows)
        {
            vec4 world = vec4(xsv_TEXCOORD1 + ShadowParams.x * xsv_TEXCOORD2.xyz, 1.0f);
            vec4 lpos = ((world) * (ShadowViewProjection));
            vec2 luv = lpos.xy * vec2(0.5f, -0.5f) + vec2(0.5f);
            float shadow_bias = ShadowParams.y;
            float base_map_depth;
            float extra_map_depth;
            vec2 screen = gl_FragCoord.xy;
            float base_shadow = GetShadowFactorAndCenterDepth(ShadowMap, luv, lpos.z, shadow_bias, ShadowParams.zz, screen, base_map_depth);
            vec4 world_for_sprite_shadows = vec4(xsv_TEXCOORD1 + ExtraShadowParams.y * xsv_TEXCOORD2.xyz, 1.0f);
            vec4 epos = ((world_for_sprite_shadows) * (ExtraShadowViewProjection));
            vec2 euv = epos.xy * vec2(0.5f, -0.5f) + vec2(0.5f);
            float extra_shadow = GetShadowFactorAndMinDepth(ExtraShadowMap, euv, epos.z, 0.0f, ExtraShadowParams.xx, screen, extra_map_depth);
            float dist = epos.z - extra_map_depth;
            float fade = clamp(dist * ExtraShadowParams.z + ExtraShadowParams.w, float(0), float(1));
            extra_shadow = mix(extra_shadow, 1.0f, fade);
            // make world shadows "block" sprite shadows
            float umbra_check = 0.1f;
            float in_base_umbra = step(base_shadow, umbra_check);
            float world_blocks_sprite = step(extra_map_depth, base_map_depth + 8.0f * ShadowParams.y);
            world_blocks_sprite = world_blocks_sprite * in_base_umbra;
            // prevent leaks at edges of world/sprite shadow overlap
            float min_shadow = min(base_shadow, extra_shadow);
            base_shadow = mix(base_shadow, min_shadow, world_blocks_sprite);
            // kill blocked sprite shadows
            extra_shadow = mix(extra_shadow, 1.0f, world_blocks_sprite);
            // prevent darker shadows showing on faces pointing away from light
            float facing_toward = smoothstep(0.0f, 0.25f, base_ndotl);
            extra_shadow = mix(1.0f, extra_shadow, facing_toward);
            shadow = min(base_shadow, extra_shadow);
            float combined_shadowness = 2.0f - base_shadow - extra_shadow;
            float select = max(0.0f, sign(base_shadow - extra_shadow));
            float intensity = mix(ShadowIntensities.x, ShadowIntensities.y, select);
            float extraness = max(0.0f, combined_shadowness - 1.0f);
            intensity = mix(intensity, ShadowIntensities.z, extraness);
            shadow_color *= vec3(intensity);
            // attenuate ambient when in extra shadow to ensure the extra shadow shows up
            float ambient_attenuate = mix(1.0f, ShadowIntensities.w, extraness);
            ambient *= vec3(ambient_attenuate);
            // remove shadow if we're past the cutoff z-plane
            float t = max(0.0f, sign(ShadowColor_Cutoff.w - xsv_TEXCOORD1.z));
            shadow = mix(shadow, 1.0f, t);
            shaded *= shadow;
        }
        bool foreground_shade = (PassSettings.y != 0.0f);
        if (foreground_shade)
        {
            vec3 foreground = ForegroundShading(xsv_TEXCOORD1, xsv_TEXCOORD2.xyz, Foreground1, Foreground2);
            shaded *= foreground.x;
        }
        // shaded = lerp(1.0f, shaded, ShadowParams.w);
        SV_Target0.rgb = albedo.rgb * ambient;
        SV_Target0.rgb += albedo.rgb * mix(shadow_color, light_color, vec3(max(0.0f, shaded)));
        SV_Target0.rgb = mix(SV_Target0.rgb, albedo.rgb * Water_Emissive.w, vec3(Water_Emissive.z));
        SV_Target0.a = albedo.a;
        SV_Target0.rgb = ApplyFog(SV_Target0.rgb, xsv_TEXCOORD1.z, FogColor.rgb, FogBounds.z, FogBounds.xy);
    }
    // super flash tinting
    SV_Target0.rgb = ApplySuperShadow(SV_Target0.rgb, SuperFlash.x);
    float index = max(SuperFlash.z - 1.0f, 0.0f);
    SV_Target0.rgb = ApplySuperFlash(SV_Target0.rgb, SuperFlash.y, GetFlashColorLow(index), GetFlashColorHigh(index));
}
