// GLSL Fragment Shader "main"
// Generated by XShaderCompiler

#version 330 core
in vec2 xsv_TEXCOORD0;
out vec4 SV_Target0;
float ConvertDepthSample(float depth)
{
    return depth;
}
//  computes the luminance value for an RGB color
float RgbToLuminance(vec3 color)
{
    return dot(color, vec3(0.212671, 0.715160, 0.072169));
}
vec3 GetFlashColorLow(float index)
{
    vec3 gLowColors[4] = vec3[4](vec3(115.f / 255.f, 24.f / 255.f, 0.f / 255.f), vec3(55.f / 255.f, 120.f / 255.f, 55.f / 255.f), vec3(200.f / 255.f, 22.f / 255.f, 22.f / 255.f), vec3(5.f / 255.f, 11.f / 255.f, 17.f / 255.f));
    return gLowColors[uint(index)];
}
vec3 GetFlashColorHigh(float index)
{
    vec3 gHighColors[4] = vec3[4](vec3(255.f / 255.f, 168.f / 255.f, 100.f / 255.f), vec3(110.f / 255.f, 110.f / 255.f, 20.f / 255.f), vec3(200.f / 255.f, 168.f / 255.f, 120.f / 255.f), vec3(19.f / 255.f, 30.f / 255.f, 38.f / 255.f));
    return gHighColors[uint(index)];
}
vec3 ApplySuperShadow(vec3 xsr_input, float t)
{
    vec3 super_color = vec3(0.2f, 0.2f, 0.2f);
    vec3 luminance = vec3(1.f, 1.f, 1.f);
    float theDot = 1.f - clamp(dot(xsr_input, luminance), float(0), float(1));
    // if(theDot < 1.f / 3.f)
    // theDot = 3.f * theDot;
    // else if(theDot < 2.f / 3.f)
    // theDot = 1.f;
    // else
    // theDot = 3.f * (1.f - theDot);
    // const float eased = (3 * theDot*theDot - 2 * theDot*theDot*theDot);
    // const float twiceEased = (3 * eased*eased - 2 * eased*eased*eased);
    // return lerp(input, super_color * eased, t);
    return mix(xsr_input, super_color * theDot, vec3(t));
}
vec3 ApplySuperFlash(vec3 xsr_input, float t, vec3 colorLow, vec3 colorHigh)
{
    vec3 coeff = vec3(0.39f, 0.5f, 0.11f);
    float initialGray = dot(xsr_input, coeff);
    // -- Desaturated
    float Contrast = 1.56f;
    float Brightness = 1.5f;
    // const float Saturation = 0.25f;
    // const float Saturation = 0.15f;
    // const float Saturation = 0.10f;
    // const float Saturation = 0.05f;
    float Saturation = 0.02f;
    // -- High Contrast --
    // const float Contrast = 1.9f;
    // const float Brightness = 2.5f;;
    // const float Saturation = 1.25f;
    float gray = Brightness * (1.2f + mix(0.5f, initialGray - 0.5f, Contrast));
    // return lerp(lerp(colorLow, colorHigh, initialGray) * gray, initialGray, -Saturation);  // ALWAYS ON
    return mix(xsr_input, mix(mix(colorLow, colorHigh, vec3(initialGray)) * gray, vec3(initialGray), vec3(-Saturation)), vec3(t));
}
// *
//  * \param base_color  Color of the surface fog will be applied to.
//  * \param distance   Surface distance.
//  * \param fog_color   Color of the fog.
//  * \param fog_max     Maximum fog strength.
//  * \param fog_bounds  Vector containing: {fog start distance, fog end distance}
//
vec3 ApplyFog(vec3 base_color, float xsr_distance, vec3 fog_color, float fog_max, vec2 fog_bounds)
{
    float fog_amount = clamp((xsr_distance - fog_bounds.x) / (fog_bounds.y - fog_bounds.x), float(0), float(1));
    return mix(base_color, fog_color, vec3(min(fog_amount, fog_max)));
}
uniform vec4 preg_0;
// (fxaa enabled, bypass, unused, unused)
uniform vec4 preg_1;
uniform vec4 preg_2;
uniform vec4 preg_3;
// This needs to preceed all texture defines for post-processing to work properly
// TextureNames;WorldTexture:0;DepthTexture:1;
uniform sampler2D treg_0;
uniform sampler2D treg_1;


// Uniform name defines ///////////////////////////////////////
#define Settings preg_0
#define SuperFlash preg_1
#define FogParams_NearFarPlane preg_2
#define FogColor_Max preg_3
#define WorldTexture treg_0
#define DepthTexture treg_1
///////////////////////////////////////////////////////////////

void main()
{
    vec4 texcoord = vec4(xsv_TEXCOORD0, 0.0f, 0.0f);
    vec4 xst_temp4 = (texcoord);
    vec3 lit_world = textureLod(WorldTexture, xst_temp4.xy, xst_temp4.w).rgb;
    vec3 post_color = lit_world;
    // pass through raw lit world color if we've debug toggled off the postfx
    if (Settings.y == 0.0f)
    {
        vec4 xst_temp5 = (texcoord);
        // distance fog
        float depth = ConvertDepthSample(textureLod(DepthTexture, xst_temp5.xy, xst_temp5.w).r);
        float near = FogParams_NearFarPlane.z;
        float far = FogParams_NearFarPlane.w;
        float dist = -near / min(-0.001f, depth - 1.0f);
        post_color.rgb = ApplyFog(post_color.rgb, dist, FogColor_Max.rgb, FogColor_Max.w, FogParams_NearFarPlane.xy);
        // super flash tinting
        post_color = ApplySuperShadow(post_color, SuperFlash.x);
        float index = max(SuperFlash.z - 1.0f, 0.0f);
        post_color = ApplySuperFlash(post_color, SuperFlash.y, GetFlashColorLow(index), GetFlashColorHigh(index));
    }
    // FXAA needs luma written to the alpha channel
    if (Settings.x != 0.0f)
    {
        SV_Target0 = vec4(post_color, RgbToLuminance(post_color));
    }
    else
    {
        SV_Target0 = vec4(post_color, 1.0f);
    }
}
